
function str2ab(str) {
  const buf = new ArrayBuffer(str.length);
  const bufView = new Uint8Array(buf);
  for (let i = 0, strLen = str.length; i < strLen; i++) {
    bufView[i] = str.charCodeAt(i);
  }
  return buf;
}


function importRsaKey(pem) {
  // fetch the part of the PEM string between header and footer
  const pemHeader = "-----BEGIN PUBLIC KEY-----";
  const pemFooter = "-----END PUBLIC KEY-----";
  const pemContents = pem.substring(
    pemHeader.length,
    // pem.length - pemFooter.length - 1,
    pem.length - pemFooter.length,
  );
  // base64 decode the string to get the binary data
  const binaryDerString = window.atob(pemContents);
  // convert from a binary string to an ArrayBuffer
  const binaryDer = str2ab(binaryDerString);

  return window.crypto.subtle.importKey(
    "spki",
    binaryDer,
    {
      name: "RSA-OAEP",
      hash: "SHA-1"
    },
    true,
    ["encrypt"],
  );
}

async function sha256(message) {
    // encode as UTF-8
    const msgBuffer = new TextEncoder().encode(message);                    

    // hash the message
    const hashBuffer = await crypto.subtle.digest('SHA-256', msgBuffer);

    // convert ArrayBuffer to Array
    const hashArray = Array.from(new Uint8Array(hashBuffer));

    // convert bytes to hex string                  
    const hashHex = hashArray.map(b => b.toString(16).padStart(2, '0')).join('');
    return hashHex;
}


class EncryptRequest {
  constructor(str, callback) {
    this.data = str;
    this.cb = callback;
  }
};


class Crypto {
  constructor(pemEncodedKey) {
    this.jobs = []; // Array of EncryptRequests
    this.publicKeyImportComplete = this.publicKeyImportComplete.bind(this);
    this.encrypt = this.encrypt.bind(this);

    var prom = importRsaKey(pemEncodedKey);

    prom.then((k)=>{
      this.publicKey = k;
      this.publicKeyImportComplete();
    }).catch((e)=>{
      console.log(e);
    });

  }

  publicKeyImportComplete() {
    this.jobs.forEach((i)=>{
        window.crypto.subtle.encrypt({name: "RSA-OAEP"}, this.publicKey, i.data).then((ciphertext)=>{
          i.cb(ciphertext);
      });
    })
    // this.jobs.clear();
    this.jobs = [];
  }

  // argument is an EncryptRequest
  encrypt(request) {
    if (this.publicKey == undefined) {
      this.jobs.push(request);
    }
    else {
      window.crypto.subtle.encrypt({name: "RSA-OAEP"}, this.publicKey, request.data).then((ciphertext)=>{
        request.cb(ciphertext);
      });
    }
  }
}
