"use strict";

// const this.deviceRef = 0;

class Gen5Lib {
  constructor(datastore) {
    this.electron = undefined;
    this.lib = undefined;
    this.datastore = datastore;
    this.libCallback = this.libCallback.bind(this);
    this.hasLib = this.hasLib.bind(this);
    this.hasDevice = this.hasDevice.bind(this);
    this.postMessage = this.postMessage.bind(this);
    this.getDiagnostics = this.getDiagnostics.bind(this);
    this.getDeviceInfo = this.getDeviceInfo.bind(this);
    this.getDriverVersion = this.getDriverVersion.bind(this);
    this.getCurrentState = this.getCurrentState.bind(this);
    this.requestResync = this.requestResync.bind(this);
    this.needToResyncState = this.needToResyncState.bind(this);
    this.maybeResyncState = this.maybeResyncState.bind(this);
    this.synchDSToCurrentState = this.synchDSToCurrentState.bind(this);
    this.setHostBufferSize = this.setHostBufferSize.bind(this);
    this.setExtraSafetyOffset1X = this.setExtraSafetyOffset1X.bind(this);
    this.setWaveInEnables = this.setWaveInEnables.bind(this);
    this.setWaveOutEnables = this.setWaveOutEnables.bind(this);
    this.getWaveSRCEnabled = this.getWaveSRCEnabled.bind(this);
    this.setWaveSRCEnabled = this.setWaveSRCEnabled.bind(this);

    this.deviceRef = undefined;
    this.driverParamChanged = true;

    if (typeof require !== "undefined") {
      this.electron = require("electron").ipcRenderer;
      this.lib = require(__dirname + "/../../build/Release/mseries");
      this.lib.register(this.libCallback);
      let that = this;
      this.datastore.addListener("kHostBuffer1x", function (ev, vals, setByObject) {
        if (setByObject === that) {
          return;
        }
        if (that.deviceRef == undefined) {
          console.log("NULL device ref!")
          return;
        }
        let cs = that.lib.getCurrentState(that.deviceRef);
        // Driver buffer size is for current mult, convert from 1x
        if((cs !== undefined) && (vals[0] != cs.hostBufferSizeInSamples))
          that.lib.setHostBufferSize(that.deviceRef, parseInt(vals[0]) * getSampleRateMult(cs.currentSampleRateInHertz));
      });
      this.datastore.addListener("kWaveInEnables", function (ev, vals, setByObject) {
        if (setByObject === that)
          return;
        if (that.deviceRef == undefined) {
          console.log("NULL device ref!")
          return;
        }
        var cs = that.lib.getCurrentState(that.deviceRef);
        if((cs !== undefined) && (vals[0] != cs.waveInEnables))
          that.lib.setWaveInEnables(that.deviceRef,parseInt(vals[0]));
      });
      this.datastore.addListener("kWaveOutEnables", function (ev, vals, setByObject) {
        if (setByObject === that)
          return;
        if (that.deviceRef == undefined) {
          console.log("NULL device ref!")
          return;
        }
        var cs = that.lib.getCurrentState(that.deviceRef);
        if((cs !== undefined) && (vals[0] != cs.waveOutEnables))
          that.lib.setWaveOutEnables(that.deviceRef,parseInt(vals[0]));
      });
      this.datastore.addListener("kExtraSafetyOffset1X", function (ev, vals, setByObject) {
        if (setByObject === that)
          return;
        if (that.deviceRef == undefined) {
          console.log("NULL device ref!")
          return;
        }
        var cs = that.lib.getCurrentState(that.deviceRef);
        if((cs !== undefined) && (vals[0] != cs.xso1x))
          that.lib.setExtraSafetyOffset1X(that.deviceRef, parseInt(vals[0]));
      });
      this.datastore.addListener("kWaveSRCEnabled", function(ev, vals, setByObject) {
        if (setByObject === that)
          return;
        if (that.deviceRef == undefined) {
          console.log("NULL device ref!")
          return;
        }
        var cs = that.lib.getWaveSRCEnabled(that.deviceRef);
        if ((cs !== undefined) && (vals[0] != cs.waveSRCEnabled))
          that.lib.setWaveSRCEnabled(that.deviceRef, parseInt(vals[0]));
      });
      this.datastore.addListener("kDevUID", ()=>{
        that.maybeResyncState();
      });
    }
  }

  libCallback(arg1, arg2, arg3) {
    if (arg1 == "dllNotify") {
      var n = arg2;
      var d = arg3;
      if (n == (1 << 2)) {
        // audio engine error - diagnostics
        gApp.eventSys.dispatchASync('kWinAEErr', arg3);
      }
      else {
        // parameter or device pnp
        this.driverParamChanged = true;
        this.maybeResyncState();
      }
    }
  }

  tryToAcquireDeviceRef() {
    if (this.deviceRef != undefined)
      return true;

    const val = new Uint8Array(this.datastore.getKVArrayItem('kDevUID', 0));
    const sn = new TextDecoder('utf-8').decode(val);
    if (sn.length == 10 && this.needToResyncState()) {
      const devRef = this.lib.getDevice(sn);
      if (devRef != -1) {
        this.deviceRef = devRef;
        return true;
      }
    }
    return false;
  }
  requestResync() {
    this.deviceRef = undefined;
  }
  needToResyncState() {
    return this.deviceRef == undefined || this.driverParamChanged;
  }
  maybeResyncState() {
    if (this.needToResyncState() && this.tryToAcquireDeviceRef()) {
      this.synchDSToCurrentState();
    }
  }
  synchDSToCurrentState() {
    if (this.deviceRef == undefined)
      return;
    var cs = this.lib.getCurrentState(this.deviceRef);
    if(cs !== undefined) {
      // Driver buffer size is for current mult, convert to 1x
      this.datastore.setV("kHostBuffer1x", 0, cs.hostBufferSizeInSamples / getSampleRateMult(cs.currentSampleRateInHertz), this);
      this.datastore.setV("kWaveInEnables", 0, cs.waveInEnables, this);
      this.datastore.setV("kWaveOutEnables", 0, cs.waveOutEnables, this);
      this.datastore.setV("kExtraSafetyOffset1X", 0, cs.xso1x, this); // 0 or 32
    }

    var ans = this.lib.getWaveSRCEnabled(this.deviceRef);
    if (ans !== undefined) {
      this.datastore.setV("kWaveSRCEnabled", 0, ans.waveSRCEnabled);
    }
  }
  hasLib() {
    return this.lib !== undefined;
  }
  hasDevice() {
    return this.deviceRef != undefined;
  }
  postMessage(msg) {
    // send cmd to dll updater
    if(msg.startsWith("open")) {
      if (msg.search("open 828") != -1) {
        this.electron && this.electron.send("open-828-firmware-updater");
      }
      else {
        this.electron && this.electron.send("open-firmware-updater");
      }
    }
    else if (msg == "driver")
      this.electron && this.electron.send("open-driver-download");
    else if (msg == "loadPreset_ul5")
      this.electron && this.electron.send("load-preset-ul5");
    else if (msg == "loadPreset_828")
      this.electron && this.electron.send("load-preset-828");
    else if(msg.indexOf("savePreset_ul5&") == 0) 
      this.electron && this.electron.send("save-preset-ul5", btoa(encodeURIComponent(msg.substring(15))));
    else if (msg.indexOf("savePreset_828&") == 0)
      this.electron && this.electron.send("save-preset-828", btoa(encodeURIComponent(msg.substring(15))));
  }
  getDiagnostics() {
    return this.lib.getDiagnostics(this.deviceRef);
  }
  getDeviceInfo() {
    return this.lib.getDeviceInfo(this.deviceRef);
  }
  getDriverVersion() {
    return this.lib.getDriverVersion(this.deviceRef);
  }
  getCurrentState() {
    return this.lib.getCurrentState(this.deviceRef);
  }
  setHostBufferSize(hbs) {
    this.lib.setHostBufferSize(this.deviceRef, hbs);
  }
  setExtraSafetyOffset1X(xso1x) {
    this.lib.setExtraSafetyOffset1X(this.deviceRef, xso1x);
  }
  setWaveInEnables(wie) {
    this.lib.setWaveInEnables(this.deviceRef, wie);
  }
  setWaveOutEnables(woe) {
    this.lib.setWaveOutEnables(this.deviceRef, woe);
  }
  getWaveSRCEnabled() {
    return this.lib.getWaveSRCEnabled(this.deviceRef);
  }
  setWaveSRCEnabled(enabled) {
    this.lib.setWaveSRCEnabled(this.deviceRef, enabled);
  }
}

const AutoLoginStatus = {
  none: 0,
  trying: 1,
  failed: 2
};


class Device {
  constructor(appdata) {
    this.wsConnect = this.wsConnect.bind(this);
    this.wsOpen = this.wsOpen.bind(this);
    this.wsError = this.wsError.bind(this);
    this.wsClose = this.wsClose.bind(this);
    this.wsRx = this.wsRx.bind(this);
    // this.rxUDPAnnounce = this.rxUDPAnnounce.bind(this);
    this.isConnected = this.isConnected.bind(this);
    this.isWebSocketConnected = this.isWebSocketConnected.bind(this);
    this.isAuthenticated = this.isAuthenticated.bind(this);
    this.isTest = this.isTest.bind(this);
    this.tryToReconnect = this.tryToReconnect.bind(this);
    this.checkConnection = this.checkConnection.bind(this);
    this.reconcileAnonymousDevice = this.reconcileAnonymousDevice.bind(this);
    this.hasHostMsgPipe = this.hasHostMsgPipe.bind(this);
    this.has_iOS_hostMsgPipe = this.has_iOS_hostMsgPipe.bind(this);
    this.receiveMsg = this.receiveMsg.bind(this);
    this.proxyStatusChanged = this.proxyStatusChanged.bind(this);
    this.handleProxyMessage = this.handleProxyMessage.bind(this);
    this.requestPassword = this.requestPassword.bind(this);
    this.connectToDevice = this.connectToDevice.bind(this);
    this.getConnectedDevice = this.getConnectedDevice.bind(this);
    this.sendPasswordHash = this.sendPasswordHash.bind(this);
    this.logout = this.logout.bind(this);
    this.on_midi_disconnect = this.on_midi_disconnect.bind(this);
    this.maybeReconnectViaMIDI = this.maybeReconnectViaMIDI.bind(this);
    this.submitChangePasswordRequest = this.submitChangePasswordRequest.bind(this);
    this.resetPassword = this.resetPassword.bind(this);
    this.initializeDeviceType = this.initializeDeviceType.bind(this);

    this.type = DeviceTypes.None;
    this.properties = DeviceProps_ul;

    this.appdata = appdata;
    this.datastore = appdata.datastore;
    this.eventSys = appdata.eventSys;
    this.eventSys.addListener("ProxyStatusChanged", this.proxyStatusChanged);
    this.eventSys.addListener("mididisconnect", this.on_midi_disconnect);

    this.test = false;
    this.opened = false;
    this.authenticated = false;
    this.lastMsgTime = 0;
    this.netDevice = undefined;
    this.isPasswordEnabled = false;
    this.loginStatus = AutoLoginStatus.none;
    this.initialReconnect = true;
    this.sysexMgr = new SysexPropertyMgr();
      
    this.eventSys.addListener("NetworkDeviceListChanged", ()=>{
      this.maybeReconnectViaMIDI();
      this.reconcileAnonymousDevice();
    });

    this.eventSys.addListener('kMaybeInitializeDeviceType', this.initializeDeviceType);

    // see if there is a mac or windows host msg handler
    this.jsHostMsgPipe = undefined;
    this.in_ios_wrapper = false;
    if((window.webkit != undefined) && (window.webkit.messageHandlers != undefined)){
      // this.wsIPAddr = "localhost:1281"; // Connect to the proxy server
      this.wsIPAddr = localStorage.getItem("LastWSIPAddr");
      if (!this.wsIPAddr)
        this.wsIPAddr = "127.0.0.1:1281";
      this.jsHostMsgPipe = window.webkit.messageHandlers.msgJStoObjC;
      if (!this.jsHostMsgPipe) {
        this.jsHostMsgPipe = window.webkit.messageHandlers.msgJStoObjC_iOS;
        if (this.jsHostMsgPipe != undefined)
          this.in_ios_wrapper = true;
      }
      this.iosProxy = new iOSMIDIProxy(this.jsHostMsgPipe);
    }
    else
    {
      this.webMIDI = new WebMIDIManager(this.appdata);

      this.eventSys.addListener("WebMIDIStateChanged", ()=>{
        if (this.sysexMgr.device && (this.sysexMgr.device instanceof WebMIDI_Device)) {
          if (this.webMIDI.haveDevice(this.sysexMgr.device))
            return;
          // Device is gone
          this.disconnect();
        }
        if (this.webMIDI.motuDevices.length > 0) {
          this.wsConnect("midi");
        }
      });

      this.wsIPAddr = localStorage.getItem("LastWSIPAddr");
      if (!this.wsIPAddr)
        this.wsIPAddr = "127.0.0.1:1281";
      this.timerID = setInterval(this.checkConnection, 100);
      if(appdata.gen5Lib.hasLib()) {
        this.jsHostMsgPipe = appdata.gen5Lib;
      }
    }

    setTimeout(this.tryToReconnect, 0);
  }

  modelToDevice(model) {
    if (model.startsWith(DeviceModels.UltraLite)) {
      return DeviceTypes.UltraLite;
    }
    else if (model.startsWith(DeviceModels.k828)) {
      return DeviceTypes.k828;
    }
    else {
      return DeviceTypes.None;
    }
  }

  updateType(new_type) {
    if (new_type === DeviceTypes.UltraLite) {
      this.type = new_type;
      this.properties = DeviceProps_ul;
    }
    else if (new_type === DeviceTypes.k828) {
      this.type = new_type;
      this.properties = DeviceProps_828;
    }
    else {
      this.type = DeviceTypes.None;
    }
  }

  initializeDeviceType(){
    const v = this.datastore.getKVArrayItem('kModelID', 0);
    const new_type = this.modelToDevice(v);
    this.updateType(new_type);
    this.eventSys.dispatchSync("DeviceStatusChanged");
  }
  deviceTypeChanged(model) {
    // parse the model to determine which device it is
    const new_type = this.modelToDevice(model);

    if (new_type !== this.type) {
      this.updateType(new_type);
      this.datastore.reset(this.properties);
      this.eventSys.dispatchSync("DeviceStatusChanged");
    }
  }
  hasHostMsgPipe() {
    return this.jsHostMsgPipe != undefined;
  }
  has_iOS_hostMsgPipe() {
    return this.in_ios_wrapper && this.hasHostMsgPipe();
  }
  // Sends JSON to updater running in objectiveC/C++ or electron
  sendHostMsgPipeMsg(msg) {
    if(this.jsHostMsgPipe != undefined)
      this.jsHostMsgPipe.postMessage(msg);
  }
  receiveMsg(msg) {
    var jsonMsg = JSON.parse(msg);
    if(jsonMsg.driverVersion) {
      this.eventSys.dispatchASync("driverVersion", jsonMsg.driverVersion);
    }
    else if(jsonMsg.applyPreset !== undefined) {
      this.datastore.applyPreset(jsonMsg.applyPreset);
    }
  }
  tryToReconnect() {
    this.wsConnect(this.wsIPAddr);
  }

  async handleProxyMessage(data) {
    var msgid = (data[0] << 8) | data[1];

    if (msgid == kProxyMsgJS.kPublicKeySetup.id) {
      var pubkey = data.slice(2, data.length);
      var dec = new TextDecoder();

      var pubkeyStr = dec.decode(pubkey);
      this.crypto = new Crypto(pubkeyStr);
    }
    else if (msgid == kProxyMsgJS.kChangePasswordResponse.id) {
      const result = data[2];
      if (!result) {
        this.eventSys.dispatchASync("PasswordChangeFail");
      }
      else {
        this.eventSys.dispatchASync("PasswordChangeSuccess");
      }
    }
    else if (msgid == kProxyMsgJS.kPasswordEnableStatus.id) {
      this.isPasswordEnabled = !!data[2];
      this.eventSys.dispatchASync("PasswordEnableStatusChanged");
    }
    else if (msgid == kProxyMsgJS.kAuthRequest.id) {
      var serial = new TextDecoder().decode(data.slice(2, 12));
      if (this.netDevice.serial == undefined || this.netDevice.serial == "") {
        this.netDevice.serial = serial;
      }

      if (!this.crypto) {
        console.log("Proxy requesting authentication but never sent public key!?");
        this.disconnect();
        ShowDiscoveryView();
        return;
      }

      if (this.loginStatus == AutoLoginStatus.none) {
        let pw = (new PWCache()).getPassword(this.netDevice.serial);
        if (pw != undefined) {
          this.sendPasswordHash(pw);
          this.loginStatus = AutoLoginStatus.trying;
        }
        else  {
          this.loginStatus = AutoLoginStatus.failed;

          // Don't request the password from the user if 
          // this is the initial reconnect attemp (if the app
          //  just finished launching). It's awkward.
          if (!this.initialReconnect)
            this.requestPassword();
          else
            this.disconnect();
        }
      }
      else  {
        // Apparently the stored password is incorrect
        this.loginStatus = AutoLoginStatus.failed;
        (new PWCache()).deletePassword(this.netDevice.serial);

        // Don't request the password from the user if 
        // this is the initial reconnect attemp (if the app
        //  just finished launching). It's awkward.
        if (!this.initialReconnect) {
          this.requestPassword();
        } else {
          this.disconnect();
        }
      }
    }
  }

  async sendPasswordHash(pw_hash) {
    this.crypto.encrypt(
      new EncryptRequest(encoder.encode(pw_hash), (ciphertext)=>{
        var resp = new Uint8Array(6 + ciphertext.byteLength);
        resp[0] = (kProxyMessageID & 0xFF00) >> 8;
        resp[1] = (kProxyMessageID & 0xFF);
        resp[2] = (kProxyMsgJS.kAuthResponse.id & 0xFF00) >> 8;
        resp[3] = (kProxyMsgJS.kAuthResponse.id & 0xFF);
        resp[4] = (ciphertext.byteLength & 0xFF00) >> 8;
        resp[5] = (ciphertext.byteLength & 0xFF);

        var ciphertext_array = new Uint8Array(ciphertext);
        for (let i = 0; i < ciphertext_array.byteLength; ++i) {
          resp[6+i] = ciphertext_array[i];
        }
        this.send(resp);
      })
    );
  }

  requestPassword() {

    var sendPassword = async (password) =>{
      if (this.crypto == undefined)
        return;

      var pw_hash = await sha256(password);
      // Store the password hash if we know the serial
      if (this.netDevice != undefined && this.netDevice.serial != undefined && this.netDevice.serial != "") {
        (new PWCache()).savePassword(this.netDevice.serial, pw_hash);
      }
      this.sendPasswordHash(pw_hash);
    };

    gApp.showPasswordEntry(
      PasswordEntryType.Authenticate,  // Don't require old pw
      "Device: " + this.netDevice.serial,

      // Callbacks...
      {
        canceled: ()=>{ 
          console.log("Authentication aborted");
          this.eventSys.dispatchASync("DeviceAuthenticationAborted");
          gApp.closePasswordEntry(); 
          this.disconnect();
        },
        completion: (password)=> {
          // Close the password entry in setTimeout. 
          // If we do it immediately in the click / tap handler,
          // I'm seeing an issue where the event propogates to the 
          // background canvas, which cancels the might cancel the login.
          setTimeout(()=>{gApp.closePasswordEntry()}, 0);
          sendPassword(password);
        }
      }
    );
  }

  async submitChangePasswordRequest(new_hash) {
    if (this.crypto == undefined)
      return false;

    let promise = new Promise(async (resolve)=>{
      this.crypto.encrypt(
        new EncryptRequest(encoder.encode(new_hash), (ciphertext)=>{
          var msg = new Uint8Array(6 + ciphertext.byteLength);
          msg[0] = (kProxyMessageID & 0xFF00) >> 8;
          msg[1] = (kProxyMessageID & 0xFF);
          msg[2] = (kProxyMsgJS.kChangePasswordRequest.id & 0xFF00) >> 8;
          msg[3] = (kProxyMsgJS.kChangePasswordRequest.id & 0xFF);
          msg[4] = (ciphertext.byteLength & 0xFF00) >> 8;
          msg[5] = (ciphertext.byteLength & 0xFF);

          var ciphertext_array = new Uint8Array(ciphertext);
          for (let i = 0; i < ciphertext_array.byteLength; ++i) {
            msg[6+i] = ciphertext_array[i];
          }

          let success_listener = ()=>{
            resolve(true);
            // Remove this event listener soon, but later.
            // I can't tell if it's safe to remove it now, since we'd be modifying an
            // array while iterating over it.
            setTimeout(()=>{this.eventSys.removeListener("PasswordChangeSuccess", success_listener);}, 0);
          };

          let failure_listener = ()=>{
            resolve(false);
            // Remove this event listener soon, but later.
            // I can't tell if it's safe to remove it now, since we'd be modifying an
            // array while iterating over it.
            setTimeout(()=>{this.eventSys.removeListener("PasswordChangeFail", failure_listener);}, 0);
          }

          this.eventSys.addListener("PasswordChangeSuccess", success_listener);
          this.eventSys.addListener("PasswordChangeFail", failure_listener);

          this.send(msg);
        })
      );
    });

    let result = await promise;
    return result;
  }

  resetPassword() {
    if (this.isConnected()) {
      let msg = new Uint8Array(6);
      msg[0] = (kProxyMessageID & 0xFF00) >> 8;
      msg[1] = (kProxyMessageID & 0xFF);
      msg[2] = (kProxyMsgJS.kResetPasswordRequest.id & 0xFF00) >> 8;
      msg[3] = (kProxyMsgJS.kResetPasswordRequest.id & 0xFF);
      msg[4] = 0;
      msg[5] = 0;

      this.send(msg);
    }
  }

  wsRx(d) {
    this.lastMessageTime = perfSeconds();

    // See if this is a proxy-specific message,
    // otherwise forward it to the datastore.
    var data = new Uint8Array(d);
    var b1 = data[0];
    var b2 = data[1];

    var id = (b1 << 8) | b2;

    if (id == kProxyMessageID) {
      this.handleProxyMessage(data.slice(2, data.length));
    }
    else {
      if (id != kProxyStatusID) {
        // This status doesn't require authentication
        var wasAuthenticated = this.authenticated;
        this.authenticated = true;
        if (!wasAuthenticated) {
          this.loginStatus = AutoLoginStatus.none;
          this.initialReconnect = false;
          this.appdata.eventSys.dispatchASync("DeviceAuthenticated");
          localStorage.setItem("LastWSIPAddr", this.wsIPAddr);
        }
      }
      this.datastore.recv(data);
    }
  }

  reconcileAnonymousDevice() {
    // If we manually connected to an IP address rather than through
    // the device discover list, we may not have all the details in our "netDevice"
    // property. If the IP matches the input argument's, assume it is the same device.
    if (this.netDevice != undefined && this.netDevice.hostname == "unknown_host") {
      var idx = this.appdata.networkDeviceList.deviceList.findIndex((x)=> { return x.ip == this.netDevice.ip;});
      if (idx != -1) {
        this.netDevice = this.appdata.networkDeviceList.deviceList[idx];
      }
    }
  }

  wsConnect(ip){
    const that = this;

    if (this.netDevice == undefined || this.netDevice.ip != ip) {
      // Make an anonymous NetworkDevice
      this.netDevice = new NetworkDevice({
        "ip" : ip,
        "hostname" : "unknown_host",
        "serial" : "",
        "model": ""
      });

      this.reconcileAnonymousDevice();
    }

    const open_midi = ()=>{
      if (this.sysexMgr && this.iosProxy) {
        ProxyLog("open_midi");
        this.opened = true;
        this.sysexMgr.setDevice(this.iosProxy);
        this.sysexMgr.enableSysexPropIn(this.wsRx);
      }
      else if (this.sysexMgr && this.webMIDI && this.webMIDI.motuDevices.length > 0)
      {
        console.log("Connecting to WebMIDI device");
        this.opened = true;
        this.sysexMgr.setDevice(this.webMIDI.motuDevices[0]);
        this.sysexMgr.enableSysexPropIn(this.wsRx);
      }
    };

    const close_midi = ()=>{
      if (this.sysexMgr) {
        this.sysexMgr.disableSysexPropIn();
        this.sysexMgr.setDevice(undefined);
      }
    };

    const open_test = function(){
      console.log("Opening test device");
      close_midi();
      that.test = true;
      that.opened = true;
      that.authenticated = true;
      that.wsIPAddr = ip;
      that.eventSys.dispatchASync("DeviceStatusChanged");
      that.eventSys.dispatchASync("DatastoreReset");
      that.eventSys.dispatchASync("DeviceAuthenticated");
    };


    if((ip == "Test") || (ip == "test")) {
      this.type = DeviceTypes.k828;
      this.properties = DeviceProps_828;
      open_test();
    }
    else if (ip === "Test_ul" || ip === "test_ul") {
      this.type = DeviceTypes.UltraLite;
      this.properties = DeviceProps_ul;
      open_test();
    }
    else if (ip === "midi") {
      this.test = false;
      this.wsIPAddr = ip;
      open_midi();
    }
    else if((ip != undefined) && ((this.ws == undefined) || (ip != this.wsIPAddr))) {
      //console.log("wsConnect");
      close_midi();
      this.test = false;
      this.wsIPAddr = ip;
      if (ip.indexOf(":") === -1)
        ip = ip + ":1280";
      this.lastMessageTime = perfSeconds();
      this.ws = new WebSocket("ws://"+ip);
      this.ws.binaryType = "arraybuffer";
      this.ws.onmessage = (d)=>{
        this.wsRx(d.data);
      };
      this.ws.onopen = this.wsOpen;
      this.ws.onerror = this.wsError;
      this.ws.onclose = this.wsClose;
    }
  }

  wsError(e){
    this.lastWebSocketError = e;
    if (e != undefined) {
      console.log(e);
    }
    // console.log("wsError");
    //this.wsClose();
  }
  wsOpen(){
    this.opened = true;
    // localStorage.setItem("LastWSIPAddr", this.wsIPAddr);
    this.eventSys.dispatchASync("DeviceStatusChanged");
  }
  wsClose(e){
    if (e != undefined) {
      if (e.target != this.ws)
        return;
    }
    if(this.ws != undefined) {
      console.log("wsClose");
      if(this.opened) {
        this.eventSys.dispatchASync("DeviceStatusChanged");
        this.eventSys.dispatchASync("DatastoreReset");
      }
      this.opened = false;

      try {
        // this.ws.close();
        this.ws = undefined;
        this.netDevice = undefined;
        this.authenticated = false;
        this.loginStatus = AutoLoginStatus.none;
        if (this.wsIPAddr != undefined && this.wsIPAddr != "midi" && this.wsIPAddr.search("test") == -1) {
          setTimeout(this.tryToReconnect, 1000);
        }
        this.lastWebSocketError = undefined;
        this.deviceTypeChanged(DeviceModels.UltraLite);
        this.datastore.reset();
      }
      catch (e) {
        console.log("wsClose threw an exception");
        console.log(e);
      }
    }
  }
  checkConnection() {
    if(!this.test && this.isConnected()) {
      var tnow = perfSeconds();
      if(((perfSeconds() - this.lastMessageTime) > 2.0) && (this.ws != undefined)) {
        //console.log("checkConnection closing connection");
        // this.wsClose();
        this.ws.close();
      }
    }
  }
  disconnect() {
    console.log("Device disconnect");
    ProxyLog("Device disconnect");
    this.test = false;
    if(this.opened) {
      this.eventSys.dispatchASync("DeviceStatusChanged");
      this.eventSys.dispatchASync("DatastoreReset");
    }
    if (this.ws != undefined) {
      this.ws.close();
      this.ws = undefined;
    }
    else {
      this.opened = false;
    }
    if (this.sysexMgr) {
      this.sysexMgr.disableSysexPropIn();
      this.sysexMgr.setDevice(undefined);
    }
    this.netDevice = undefined;
    this.authenticated = false;
    this.loginStatus = AutoLoginStatus.none;
    this.initialReconnect = false;

    this.deviceTypeChanged(DeviceModels.UltraLite);
    this.datastore.reset();
  }

  on_midi_disconnect() {
    if (this.isMIDIConnected()) {
      this.disconnect();
    }
  }

  maybeReconnectViaMIDI() {
    if (this.appdata) {
      const last_ip= localStorage.getItem('LastWSIPAddr');
      if (last_ip == undefined || last_ip == "midi") {
        var idx = this.appdata.networkDeviceList.deviceList.findIndex((x)=> { return x.ip == "midi"});
        if (idx != -1) {
          this.connectToDevice(this.appdata.networkDeviceList.deviceList[idx]);
          return true;
        }
      }
    }
    return false;
  }

  isConnected() {
      return (this.isWebSocketConnected() || this.isMIDIConnected()) && this.isAuthenticated();
  }
  isWebSocketConnected() {
    return this.opened && (this.ws != undefined) && (this.ws.readyState == 1); 
  }
  isMIDIConnected() {
    return this.sysexMgr && this.sysexMgr.device != undefined;
  }
  isAuthenticated() {
    return this.authenticated;
  }
  isTest() {
    return this.test;
  }
  send(d) {
    if (Array.isArray(d))
    {
      // "d" is an array of device messages.
      if (!this.test && this.sysexMgr && this.sysexMgr.device) {
        // We're connected via the sysex API. Send each property separately.
        for (const i of d) {
          const encoded = this.sysexMgr.encodeDeviceMessage(i);
          this.sysexMgr.device.sendMIDI(encoded);
        }
      }
      else if (this.isWebSocketConnected()){
        // We're connected through the websocket. Glue the properties together and send it all in one payload.
        let data = undefined;
        for (let i = 0; i < d.length; i+=2) {
          if (data == undefined) {
            data = i+1 >= d.length ? d[i] : addUint8Arrays(d[i], d[i+1]);
          }
          else {
            data = i+1 >= d.length ? addUint8Arrays(data, d[i]) : addUint8Arrays(data, addUint8Arrays(d[i], d[i+1]));
          }
        }
        if (data != undefined)
          this.ws.send(data);
      }
    }
    else
    {
      // "d" is a single device message
      if (!d || d.length == 0)
        return;

      if(!this.test && this.isWebSocketConnected()) {
        this.ws.send(d);
      }
      else if (!this.test && this.sysexMgr && this.sysexMgr.device) {
        const encoded = this.sysexMgr.encodeDeviceMessage(d);
        this.sysexMgr.device.sendMIDI(encoded);
      }
    }

  }
  proxyStatusChanged(ev, v) {
    if (v == 0)
      this.datastore.reset();
  }

  connectToDevice(netDevice) {
    if (netDevice == this.netDevice)
      return;

    if (netDevice.ip != undefined) {
      this.disconnect();
      this.netDevice = netDevice;
      this.wsConnect(netDevice.ip);
    }
  }
  getConnectedDevice() {
    return this.netDevice;
  }

  logout() {
    if (this.netDevice != undefined && this.netDevice.serial != undefined && this.netDevice.serial != "") {
      (new PWCache()).deletePassword(this.netDevice.serial);
    }
    this.disconnect();
  }
}
var decoder = new TextDecoder("utf8");
var encoder = new TextEncoder("utf8");

function CreateDeviceMessage(kv, i) {
// todo check endianess of host and always send data network (big) endian
  var x = undefined;
  switch(kv.t) {
    default:
      break;
    case kTString: {
      // Note we don't support writing strings bigger than 32 bytes
      x = new Uint8Array(6 + 32); // initialized to 0
      var str = encoder.encode(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = 0;
      x[5]  = (32 & 0x00FF);
      // always send the device a null terminated 32 byte array
      var len = Math.min(str.length, 31);
      for(var i = 0; i < len; ++i) {
        x[6+i] = str[i];
      }
    } break;
    case kTInt32: {
      var len = 4;
      x = new Uint8Array(6 + len);
      var val = Math.round(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0xFF000000) >> 24;
      x[7]  = (val & 0x00FF0000) >> 16;
      x[8]  = (val & 0x0000FF00) >> 8;
      x[9]  = (val & 0x000000FF);
    } break;
    case kTFloat: {
      var len = 4;
      x = new Uint8Array(6 + len);
      var val = To824(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0xFF000000) >> 24;
      x[7]  = (val & 0x00FF0000) >> 16;
      x[8]  = (val & 0x0000FF00) >> 8;
      x[9]  = (val & 0x000000FF);
    } break;
    case kTInt16: {
      var len = 2;
      x = new Uint8Array(6 + len);
      var val = Math.round(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0x0000FF00) >> 8;
      x[7]  = (val & 0x000000FF);
    } break;
    case kTByte: {
      var len = 1;
      x = new Uint8Array(6 + len);
      var val = Math.round(kv.v[i]);
      x[0]  = (kv.id & 0xFF00) >> 8;
      x[1]  = (kv.id & 0x00FF);
      x[2]  = (i & 0xFF00) >> 8;
      x[3]  = (i & 0x00FF);
      x[4]  = (len & 0xFF00) >> 8;
      x[5]  = (len & 0x00FF);
      x[6]  = (val & 0x000000FF);
    } break;
  }
  return x;
}
function addUint8Arrays(a, b) {
  // return new array of b appended to a
  if(a.length === 0) return b;
  var c = new Int8Array(a.length + b.length);
  c.set(a);
  c.set(b, a.length);
  return c;
}

function UnpackDeviceMessage(t, msg) {
// todo check endianess of host and always send data network (big) endian
  switch(t) {
    default:
      break;
    case kTArray: {
      if(msg.length >= 5)
        return Array.from(msg.slice(4));
    } break;
    case kTString: {
      if(msg.length >= 5)
        return decoder.decode(msg.slice(4, msg.indexOf(0,4)));
    } break;
    case kTInt32: {
      if(msg.length >= 8)
        return (msg[4] << 24) | (msg[5] << 16) | (msg[6] << 8) | msg[7];
    } break;
    case kTFloat: {
      if(msg.length >= 8)
        return From824((msg[4] << 24) | (msg[5] << 16) | (msg[6] << 8) | msg[7]);
    } break;
    case kTInt16: {
      if(msg.length >= 6)
        return (msg[4] << 8) | msg[5];
    } break;
    case kTByte: {
      if(msg.length >= 5)
        return msg[4];
    } break;
  }
  return undefined;
}

// all items in this datastore are arrays, single items are arrays of 1
// Terminology:
// k: the key to the datastore object
// kv: the datastore object (id, type, value array)
// kv.id: the id of the value array (numeric versoin of k)
// kv.t: the type of the items in the value array
// kv.v: the value array
class DataStore extends Events{
  constructor(appdata) {
    super();
    this.appdata = appdata;
    this.reset = this.reset.bind(this);
    this.onReset = this.onReset.bind(this);
    this.getKVArray = this.getKVArray.bind(this);
    this.getKVArrayItem = this.getKVArrayItem.bind(this);
    this.setV = this.setV.bind(this);
    this.sendPendingToDevice = this.sendPendingToDevice.bind(this);
    this.pendingDeviceDataLength = this.pendingDeviceDataLength.bind(this);
    this.recv = this.recv.bind(this);
    this.loadPreset = this.loadPreset.bind(this);
    this.savePreset = this.savePreset.bind(this);
    this.ds = {};
    this.map = [];
    this.devOutData = [];
    this.reset(DeviceProps_ul);
    appdata.eventSys.addListener("DatastoreReset", this.onReset);
  }
  loadPreset() {
    // tell the host dll/app to run the open dialog and load a preset, it will send it to applyPreset
    if(this.appdata.device.hasHostMsgPipe()) {
      const msgStr = this.appdata.device.type == DeviceTypes.UltraLite ? "loadPreset_ul5" : "loadPreset_828";
      this.appdata.device.sendHostMsgPipeMsg(msgStr);
    }
  }
  applyPreset(preset) {
    const json = decodeURIComponent(atob(preset));
    const devJS = JSON.parse(json);
    if (!devJS)
      return;

    const this_dev_type = DeviceTypeFromModel(this.getKVArrayItem('kModelID', 0));
    const that_dev_type = devJS.kModelID != undefined && devJS.kModelID.v != undefined && Array.isArray(devJS.kModelID.v) ? DeviceTypeFromModel(devJS.kModelID.v[0]) : DeviceTypes.None;
    const prop_is_recallable = (p)=>{
      return p.recallable == undefined || p.recallable !== false;
    };
    const prop_is_writable = (p)=>{
      return p.w == undefined || p.w;
    };

    const dev_props_prototype = (()=>{
        if (this_dev_type == DeviceTypes.k828)
          return DeviceProps_828;
        return DeviceProps_ul;
    })();

    const prop_is_loopback_first = (p)=>{
      if (dev_props_prototype && dev_props_prototype.defaultDatastore && dev_props_prototype.defaultDatastore.kLoopbackFirst) {
        return dev_props_prototype.defaultDatastore.kLoopbackFirst.id == p.id;
      }
      return false;
    }

    let deferred_loopback_first_prop = undefined;

    const send_prop_to_device = (k, kv, presetKV, deferIfNecessary = true)=>{
      if (deferIfNecessary && prop_is_loopback_first(kv)) {
        deferred_loopback_first_prop = {
          k: k,
          kv: kv,
          presetKV: presetKV
        };
      }
      else {
        // send to device
        let changed = false;
        let n = kv.v.length;
        for(var i = 0; i < n; ++i) {
          if(kv.v[i] != presetKV.v[i]) {
            changed = true;
            kv.v[i] = presetKV.v[i];
            let msg = CreateDeviceMessage(kv, i);
            if(msg != undefined) {
              if((this.pendingDeviceDataLength() + msg.length) >= 1452)
                this.sendPendingToDevice();
              this.devOutData.push(msg);
              setTimeout(this.sendPendingToDevice, 0);
            }
          }
        }
        if(changed)
          this.dispatchASync(k, kv.v);
      }
    };

    // verify the device is correct
    if((this_dev_type == that_dev_type) &&
      (devJS.kAPIVersion != undefined))
    {
      for (let [k, presetKV] of Object.entries(devJS)) {
        // Does the key exist and is it writable by cuemix?
        if(this.ds[k] != undefined) {
          var kv = this.ds[k];
          // ds keys are writable unless specifed
          if (prop_is_recallable(kv) && prop_is_writable(kv)) {
            // ID's -1000 through -1999 are from the windows dll
            if(kv.id < 0) {
              if(kv.id > -1000) {
                // send to local storage
                kv.v = presetKV.v;
                this.dispatchASync(k, kv.v);
                localStorage.setItem(k, JSON.stringify(kv.v));
              }
              // ID's -1000 through -1999 are from the windows dll and I'm not restoring them as presets
            }
            else {
              send_prop_to_device(k, kv, presetKV);
            }
          }
        }
      }

      if (deferred_loopback_first_prop) {
        send_prop_to_device(deferred_loopback_first_prop.k, deferred_loopback_first_prop.kv, deferred_loopback_first_prop.presetKV, false);
      }
    }
    else {
      // todo: put up a message
    }
  }
  pendingDeviceDataLength() {
    let n = 0;
    for (const i of this.devOutData) {
      n += i.length;
    }
    return n;
  }
  savePreset(t) {
    if(this.appdata.device.hasHostMsgPipe()) {
      // make a copy of the ds with just key names and values
      var preset = {};
      for (let [k, kv] of Object.entries(this.ds)) {
        preset[k] = { v: kv.v };
      }
      const msgStr = this.appdata.device.type == DeviceTypes.UltraLite ? "savePreset_ul5&" : "savePreset_828&";
      this.appdata.device.sendHostMsgPipeMsg(msgStr + JSON.stringify(preset));
    }
  }

  reset(device_props) {
    this.ds = {};
    this.map = [];
    // this set's the ds to a copy of the default datastore
    if (device_props === undefined)
      device_props = this.appdata.device.properties;
    this.ds = JSON.parse(JSON.stringify(device_props.defaultDatastore));
    for (let [k, kv] of Object.entries(this.ds)) {
      if(this.map[kv.id] != undefined)
        throw "PropID is not unique";
      this.map[kv.id] = k;
      // restore negative ID's from storage
      if(kv.id < 0) {
        if(kv.id > -1000) {
          var t = localStorage.getItem(k);
          if(t != undefined)
            this.ds[k].v = JSON.parse(t);
        }
        // ID's -1000 through -1999 are from the windows dll
      }
      this.dispatchASync(k, kv.v);
    }
    const gen5Lib = this.appdata.gen5Lib;
    if(gen5Lib !== undefined && gen5Lib.hasLib())
      gen5Lib.requestResync();
  }
  onReset() {
    this.reset();
  }
  getKVArray(k){ // gets the entire key value array
    return this.ds[k].v;
  }
  getKVArrayItem(k,i){ //gets the single value of the key at index
    return this.ds[k].v[i];
  }
  setV(k, i, v, setByObject){
    // update the value, notify listeners and send to the device
    //console.log("DS Set " + k + ", i: " + i + ", v " + v);
    var kv = this.ds[k];

    // Is the key writable by cuemix?
    if(kv.w != undefined && !kv.w)
      return;

    // round integral types
    switch(kv.t) {
      default:
        break;
      case kTInt32:
      case kTInt16:
      case kTByte:
        v = Math.round(v);
      break;
    }
    if(kv.v[i] != v) {
      kv.v[i] = v;
      this.dispatchASync(k, kv.v, setByObject);
      if(kv.id < 0) {
        if(kv.id > -1000) {
          // send to local storage
          localStorage.setItem(k, JSON.stringify(kv.v));
        }
        // ID's -1000 through -1999 are from the windows dll
      }
      else {
        // send to device
        var msg = CreateDeviceMessage(kv, i);
        if(msg != undefined) {
          if((this.pendingDeviceDataLength() + msg.length) >= 1452)
            this.sendPendingToDevice();
          this.devOutData.push(msg);
          setTimeout(this.sendPendingToDevice, 0);
        }
      }
    }
  }
  sendPendingToDevice() {
    this.appdata.device.send(this.devOutData);
    this.devOutData = [];
  }

  recv(d) {
    const device_type_changed = (v1, v2) => {
      const is_ul = (v)=>{
        return v && v[0] == 'U' && v[1] == 'L';
      };
      return is_ul(v1) != is_ul(v2);
    };

    var data = new Uint8Array(d);
    var id = (data[0] << 8) | data[1];
    var i = (data[2] << 8) | data[3];

    if (id == kMetersID)
      this.appdata.meterStore.set(data.slice(4));
    else if (id == kProxyStatusID)
      this.appdata.eventSys.dispatchASync("ProxyStatusChanged", data[4]);
    else {
      var k = this.map[id];
      if(k != undefined) {
        var kv = this.ds[k];
        var v = UnpackDeviceMessage(kv.t, data);

        if (k === 'kModelID' && v != undefined) {
          this.appdata.eventSys.dispatchASync('kMaybeInitializeDeviceType');
        }

        if((v != undefined) && (kv.v[i] != v)) {
          if (k === 'kModelID' && (kv.v[i].substring(0,2) != kv.v[i].substring(0,2))) {
            this.appdata.device.deviceTypeChanged(v);
          }
          kv.v[i] = v;
          this.dispatchASync(k, kv.v);
        }
      }
    }
  }
}
// Client DS helpers
function SetupDSArrayListener(datastore, that, k, i) {
  // add the key as an object to that
  that[k] = i;
  datastore.addListener(k, function (ev, vals, setByObject) {
    var v = vals[that[k]];
    that.onDSArray(ev, v, setByObject);
  });
}
function SetDSArrayIndex(that, k, i) {
    that[k] = i;
}
function GetDSArrayIndex(that, k) {
    var i = that[k];
    return i;
}
function DSArrayGet(datastore, that, k) {
  return datastore.getKVArrayItem(k, that[k]);
}
function DSArraySet(datastore, that, k, v) {
  var i = that[k];
  if(i !== undefined)
    datastore.setV(k, i, v, that);
}
// for objects with only 1 dslistener
function SetupDSArrayListenerBasic(datastore, that, k, i) {
  that.dsKey = k;
  SetupDSArrayListener(datastore, that, k, i);
}
function SetDSArrayIndexBasic(that, i) {
  SetDSArrayIndex(that, that.dsKey, i);
}
function GetDSArrayIndexBasic(that) {
  return GetDSArrayIndex(that, that.dsKey);
}
function DSArrayGetBasic(datastore, that) {
  return DSArrayGet(datastore, that, that.dsKey);
}
function DSArraySetBasic(datastore, that, v) {
  DSArraySet(datastore, that, that.dsKey, v);
}

function ProxyLog(msg) {
  if (gApp.appdata.device.hasHostMsgPipe) {
    gApp.appdata.device.sendHostMsgPipeMsg(JSON.stringify({
      "LogMessage":msg
    }));
  }
}
