
const DeviceProps_ul = function() {

  const kClockSources = [
    { v:3, l:"Internal" },
    { v:0, l:"S/PDIF" },
    { v:2, l:"Optical" }
  ];

  // PatchBase (from dylans patchbase.h)
  const PatchBase = {
    tdm:  0,  //x0
    zero: 23, //x17
    adat: 24, //x18
    i2s:  32, //x20
    spdf: 44, //x2c
    dsp:  46, //x2e

    digital: 24 // Alias for adat
  };

  // LocalInputs  (from dylans patchbase.h)
  const LocalInputs = [
    5, // Mic 1
    0, // Mic 2
    6, // Line 3
    7, // Line 4
    3, // Line 5
    4, // Line 6
    1, // Line 7
    2  // Line 8
  ];

  // LocalOutputs  (from dylans patchbase.h)
  // Also BS Mapping for DAC trims, main group and meters,
  // also meters are pre trim
  const LocalOutputs = {
    lines:[
      0x6, // Main 1
      0x7, // Main 2
      0x8, // Line 3
      0x9, // Line 4
      0x0, // Line 5
      0x1, // Line 6
      0x2, // Line 7
      0x3, // Line 8
      0x4, // Line 9
      0x5  // Line 10
    ],
    phones:[
      0xA,
      0xB
    ],
    spdif:[
      0xC,
      0xD
    ]
  };

// InputSource  (from dylans patchbase.h)
const InputSource = {
  get_tdm:function(index){ return index < 23 ? PatchBase.tdm + index : undefined; },
  zero:PatchBase.zero,
  get_adat:function(index){ return index < 8 ? PatchBase.adat + index : undefined; },
  get_line:function(index){ return index < LocalInputs.length ? PatchBase.i2s + LocalInputs[index] : undefined; },
  get_spdif:function(index) { return index < 2 ? PatchBase.spdf + index : undefined; },
  get_dsp:function(index) { return index < 32 ? PatchBase.dsp + index : undefined; }
};

// OutputDestination (from dylans patchbase.h)
const OutputDestination = {
  get_tdm:function(index){ return index < 23 ? PatchBase.tdm + index : undefined; },
  zero:PatchBase.zero,
  get_adat:function(index){ return index < 8 ? PatchBase.adat + index : undefined; },
  get_line:function(index){ return index < LocalOutputs.lines.length ? PatchBase.i2s + LocalOutputs.lines[index] : undefined; },
  get_phone:function(index) { return index < LocalOutputs.phones.length ? PatchBase.i2s + LocalOutputs.phones[index] : undefined; },
  get_spdif:function(index){ return index < 2 ? PatchBase.spdf + index : undefined; },
  get_dsp:function(index){ return index < 32 ? PatchBase.dsp + index : undefined; }
};

  const miUSBAnalOut = InputSource.get_tdm(2);
  const miUSBPhOut = InputSource.get_tdm(10);
  const miUSBSPDIFOut = InputSource.get_tdm(12);
  const miUSBADATAOut = InputSource.get_tdm(14);

  const analogIns = [
    {source:InputSource.get_line(0), destination:OutputDestination.get_tdm(0), postfx:InputSource.get_dsp(4) },
    {source:InputSource.get_line(1), destination:OutputDestination.get_tdm(1), postfx:InputSource.get_dsp(5) },
    {source:InputSource.get_line(2), destination:OutputDestination.get_tdm(2), postfx:InputSource.get_dsp(6) },
    {source:InputSource.get_line(3), destination:OutputDestination.get_tdm(3), postfx:InputSource.get_dsp(7) },
    {source:InputSource.get_line(4), destination:OutputDestination.get_tdm(4), postfx:InputSource.get_dsp(12) },
    {source:InputSource.get_line(5), destination:OutputDestination.get_tdm(5), postfx:InputSource.get_dsp(13) },
    {source:InputSource.get_line(6), destination:OutputDestination.get_tdm(6), postfx:InputSource.get_dsp(14) },
    {source:InputSource.get_line(7), destination:OutputDestination.get_tdm(7), postfx:InputSource.get_dsp(15) }
  ];

  const analogOuts = [
    {destination:OutputDestination.get_line(0),   dac:LocalOutputs.lines[0], dsp:InputSource.get_dsp(0) },
    {destination:OutputDestination.get_line(1),   dac:LocalOutputs.lines[1], dsp:InputSource.get_dsp(1) },
    {destination:OutputDestination.get_line(2),   dac:LocalOutputs.lines[2], dsp:InputSource.get_dsp(2) },
    {destination:OutputDestination.get_line(3),   dac:LocalOutputs.lines[3], dsp:InputSource.get_dsp(3) },
    {destination:OutputDestination.get_line(4),   dac:LocalOutputs.lines[4], dsp:InputSource.get_dsp(8) },
    {destination:OutputDestination.get_line(5),   dac:LocalOutputs.lines[5], dsp:InputSource.get_dsp(9) },
    {destination:OutputDestination.get_line(6),   dac:LocalOutputs.lines[6], dsp:InputSource.get_dsp(10) },
    {destination:OutputDestination.get_line(7),   dac:LocalOutputs.lines[7], dsp:InputSource.get_dsp(11) },
    {destination:OutputDestination.get_line(8),   dac:LocalOutputs.lines[8], dsp:InputSource.get_dsp(16) },
    {destination:OutputDestination.get_line(9),   dac:LocalOutputs.lines[9], dsp:InputSource.get_dsp(17) }
  ];

  const phonesOut = [
    {
      destination: { left:OutputDestination.get_phone(0), right:OutputDestination.get_phone(1) },
      line: { left:LocalOutputs.phones[0], right:LocalOutputs.phones[1] },
      dsp: InputSource.get_dsp(18)
    }
  ];

  const defaultPatch = [
    InputSource.get_line(0), // kOD_TDM1    (0x00)
    InputSource.get_line(1), // kOD_TDM2    (0x01)
    InputSource.get_line(2), // kOD_TDM3    (0x02)
    InputSource.get_line(3), // kOD_TDM4    (0x03)
    InputSource.get_line(4), // kOD_TDM5    (0x04)
    InputSource.get_line(5), // kOD_TDM6    (0x05)
    InputSource.get_line(6), // kOD_TDM7    (0x06)
    InputSource.get_line(7), // kOD_TDM8    (0x07)
    InputSource.get_dsp(18), // kOD_TDM9    (0x08)
    InputSource.get_dsp(19), // kOD_TDM10   (0x09)
    miUSBSPDIFOut,   // kOD_TDM11   (0x0A)
    miUSBSPDIFOut+1, // kOD_TDM12   (0x0B)
    InputSource.get_adat(0), // kOD_TDM13   (0x0C)
    InputSource.get_adat(1), // kOD_TDM14   (0x0D)
    InputSource.get_adat(2), // kOD_TDM15   (0x0E)
    InputSource.get_adat(3), // kOD_TDM16   (0x0F)
    InputSource.get_adat(4), // kOD_TDM17   (0x10)
    InputSource.get_adat(5), // kOD_TDM18   (0x11)
    InputSource.get_adat(6), // kOD_TDM19   (0x12)
    InputSource.get_adat(7), // kOD_TDM20   (0x13)
    InputSource.zero,    // kOD_TDM21   (0x14)
    InputSource.zero,    // kOD_TDM22   (0x15)
    InputSource.zero,    // kOD_TDM23   (0x16)
    InputSource.zero,    // kOD_Zeros   (0x17)
    miUSBADATAOut, // kOD_ADAT1   (0x18)
    miUSBADATAOut+1, // kOD_ADAT2   (0x19)
    miUSBADATAOut+2, // kOD_ADAT3   (0x1A)
    miUSBADATAOut+3, // kOD_ADAT4   (0x1B)
    miUSBADATAOut+4, // kOD_ADAT5   (0x1C)
    miUSBADATAOut+5, // kOD_ADAT6   (0x1D)
    miUSBADATAOut+6, // kOD_ADAT7   (0x1E)
    miUSBADATAOut+7, // kOD_ADAT8   (0x1F)
    analogOuts[4].dsp, // kOD_Line3   (0x20)
    analogOuts[5].dsp, // kOD_Line4   (0x21)
    analogOuts[6].dsp, // kOD_Line5   (0x22)
    analogOuts[7].dsp, // kOD_Line6   (0x23)
    analogOuts[8].dsp, // kOD_Line7   (0x24)
    analogOuts[9].dsp, // kOD_Line8   (0x25)
    analogOuts[0].dsp, // kOD_Main1   (0x26)
    analogOuts[1].dsp, // kOD_Main2   (0x27)
    analogOuts[2].dsp, // kOD_Line1   (0x28)
    analogOuts[3].dsp, // kOD_Line2   (0x29)
    phonesOut[0].dsp, // kOD_Phones1 (0x2A)
    phonesOut[0].dsp+1, // kOD_Phones2 (0x2B)
    miUSBSPDIFOut, // kOD_SPDIF1  (0x2C)
    miUSBSPDIFOut+1, // kOD_SPDIF2  (0x2D)
    InputSource.get_line(0), // kOD_DSP1    (0x2E)
    InputSource.get_line(1), // kOD_DSP2    (0x2F)
    InputSource.get_line(2), // kOD_DSP3    (0x30)
    InputSource.get_line(3), // kOD_DSP4    (0x31)
    miUSBSPDIFOut,   // kOD_DSP5    (0x32)
    miUSBSPDIFOut+1, // kOD_DSP6    (0x33)
    InputSource.get_adat(0), // kOD_DSP7    (0x34)
    InputSource.get_adat(1), // kOD_DSP8    (0x35)
    InputSource.get_line(4), // kOD_DSP9    (0x36)
    InputSource.get_line(5), // kOD_DSP10   (0x37)
    InputSource.get_line(6), // kOD_DSP11   (0x38)
    InputSource.get_line(7), // kOD_DSP12   (0x39)
    InputSource.get_adat(2), // kOD_DSP13   (0x3A)
    InputSource.get_adat(3), // kOD_DSP14   (0x3B)
    InputSource.get_adat(4), // kOD_DSP15   (0x3C)
    InputSource.get_adat(5), // kOD_DSP16   (0x3D)
    InputSource.get_tdm(0), // kOD_DSP17   (0x3E)
    InputSource.get_tdm(1), // kOD_DSP18   (0x3F)
    InputSource.get_tdm(2), // kOD_DSP19   (0x40)
    InputSource.get_tdm(3), // kOD_DSP20   (0x41)
    InputSource.get_tdm(4), // kOD_DSP21   (0x42)
    InputSource.get_adat(6), // kOD_DSP22   (0x43)
    InputSource.get_adat(7), // kOD_DSP23   (0x44)
    InputSource.zero,    // kOD_DSP24   (0x45)
    InputSource.get_tdm(5), // kOD_DSP25   (0x46)
    InputSource.get_tdm(6), // kOD_DSP26   (0x47)
    InputSource.get_tdm(7), // kOD_DSP27   (0x48)
    InputSource.get_tdm(8), // kOD_DSP28   (0x49)
    InputSource.get_tdm(9), // kOD_DSP29   (0x4A)
    InputSource.get_tdm(10), // kOD_DSP30   (0x4B)
    InputSource.get_tdm(11), // kOD_DSP31   (0x4C)
    InputSource.zero,    // kOD_DSP32   (0x4D)
    InputSource.zero,
    InputSource.zero
  ];

  const dspReverbWet = InputSource.get_dsp(22);
  const miADATAIn = InputSource.get_adat(0);
  const miSPDIFIn = InputSource.get_spdif(0);
  const miMixAnaOut = [analogOuts[2].dsp, analogOuts[3].dsp, analogOuts[4].dsp, analogOuts[5].dsp, analogOuts[6].dsp, analogOuts[7].dsp, analogOuts[8].dsp, analogOuts[9].dsp];
  const miMixPhOut = phonesOut[0].dsp;
  const miPostFXMicIn = analogIns[0].postfx;
  const miPostFXAnaIn = [analogIns[2].postfx, analogIns[3].postfx, analogIns[4].postfx, analogIns[5].postfx, analogIns[6].postfx, analogIns[7].postfx];

  // NOTE: meterstore.js is counting on the gate and comp meters being contiguous
  // if you insert anything here make sure it doesn't break meterstore.js
  // I gave up on converting these to dylans map for now
  const miPostEQMicIn = 80;
  const miPostEQAnaIn = 82;
  const miGateGRMicIn = 88;
  const miPostGateMicIn = 90;
  const miCompGRMicIn = 92;
  const miCompGRAnaIn = 94;

  const MixTypes = {
    Mic: "Mic",
    Line: "Line",
    SPDIF: "S/PDIF",
    ADATA: "ADAT",
    TOSlink: "TOSlink",
    Reverb: "Reverb"
  };

  // output: index into koChannelNames
  // input: index into kiChannelNames
  const nameIndices = {
    main: {
      output: 0
    },
    mic: {
      input: 0
    },
    line: {
      input: 2,
      output: 2
    },
    phones: {
      output: 10
    },
    spdif: {
      input: 8,
      output: 12
    },
    optical_a: {
      input: 10,
      output: 14,
    }
  };

  const numOpticalBanks = 1;

  // type: type/name of bank
  // iGain: index of first channel mix gain at sample rates
  // numCh: number of mix input channels (and bus channels) at sample rates
  // name: index of first mix input channel name
  // stereo: index of first mix input channel stereo control
  // pre: index of ffirst mix input channel pre controls (48v, pad trim)
  // eq: index of first mix input channel eq controls
  // gate: index of first mix input channel gate controls
  // comp: index of first mix input channel comp controls
  // iBus: index of first mix bus gain at sample rates
  // name: index of first mix bus name
  // stereo: index of first mix input channel stereo control
  // busEq: index of first mix bus eq controls
  const adatInfo = {
    type:MixTypes.ADATA,
    numCh:[8,8,4,4,0,0],
    iGain:10,
    stereo:10,
    name:nameIndices.optical_a.input,
    meters:[-1,-1,-1,-1,miADATAIn],
    getMeter:function(index){ return InputSource.get_adat(index); },
    getPatch:function(index){ return OutputDestination.get_adat(index); }
  };

  const tosInfo = {
    type:MixTypes.TOSlink,
    numCh:[2,2,2,2,0,0],
    iGain:10,
    stereo:10,
    name:nameIndices.optical_a.input,
    meters:[-1,-1,-1,-1,miADATAIn],
    getMeter:adatInfo.getMeter,
    getPatch:adatInfo.getPatch
  };

  // name: index into kiChannelNames
  const deviceInputs = {
    mics: {
      type:MixTypes.Mic,
      numCh:[2,2,2,2,2,2],
      iGain:0,
      stereo:0,
      pre:0,
      eq:0,
      gate:0,
      comp:0,
      name:nameIndices.mic.input,
      meters:[miPostEQMicIn,miGateGRMicIn,miPostGateMicIn,miCompGRMicIn,miPostFXMicIn]
    },
    lines: {
      type:MixTypes.Line,
      numCh:[analogIns.length-2,analogIns.length-2,analogIns.length-2,analogIns.length-2,analogIns.length-2,analogIns.length-2],
      iGain:2,
      stereo:2,
      eq:2,
      comp:2,
      name:nameIndices.line.input,
      meters:[-1,-1,miPostEQAnaIn,miCompGRAnaIn,miPostFXAnaIn],
    },
    spdif: {
      type:MixTypes.SPDIF,
      numCh:[2,2,2,2,0,0],
      iGain:8,
      stereo:8,
      name:nameIndices.spdif.input,
      meters:[-1,-1,-1,-1,miSPDIFIn],
    },
    reverb: {
      type:MixTypes.Reverb,
      numCh:[2,2,2,2,0,0],
      iGain:30,
      name:-1,
      meters:[-1,-1,-1,-1,dspReverbWet],
    },
  };

  const mixOutputs = {
    mains:{
      type:MixBusTypes.Main,
      numCh:2,
      iBus:0,
      iGain:18,
      busEq:0,
      colorIndex:0,
      name:nameIndices.main.output,
      istereo:18, // stereo index of host output to mixer
      meters:[analogOuts[0].dsp, analogIns[0].destination],
    },
    lines:{
      type:MixBusTypes.Line,
      numCh:analogOuts.length-2,
      iBus:2,
      iGain:20,
      busEq:2,
      colorIndex:2,
      name:nameIndices.line.output,
      stereo:2,
      istereo:20, // stereo index of host output to mixer
      meters:[miMixAnaOut, miUSBAnalOut],
    },
    phones:{
      type:MixBusTypes.Phones,
      numCh:2,
      iBus:10,
      iGain:28,
      busEq:10,
      colorIndex:10,
      name:nameIndices.phones.output,
      istereo:28, // stereo index of host output to mixer
      meters:[miMixPhOut, miUSBPhOut],
    },
    reverbs:{
      type:MixBusTypes.Reverb,
      numCh:2,
      iBus:12,
      busEq:11,
      colorIndex:12,
      meters:[dspReverbWet, -1],
    }
  };

  // Bus index's
  // 0: main 1
  // 1: main 2
  // 2: Line 1
  // 3: Line 2
  // 4: Line 3
  // 5: Line 4
  // 6: Line 5
  // 7: Line 6
  // 8: Line 7
  // 9: Line 8
  // 10: phones 1
  // 11: phones 2
  // 12: reverb 1
  // 13: reverb 2
  const mixBusInfo = [
    mixOutputs.mains,
    mixOutputs.phones,
    mixOutputs.lines,
    mixOutputs.reverbs
  ];

  const numInputs = 32; // the number of inputs in kiMixPan, kiMixFader, kiMixSolo, kiMixMute, kiMixStereo

  const kLoopbackOrder =  [
    { v:0, l:"USB In 9-10 (Default)" },
    { v:1, l:"USB In 1-2" }
  ];

  const kDevJS = {
    // Data

    kAPIVersion: {
      id:8,
      w:false,
      t:kTInt16,
      // version of the datastore api the device supports
      v:[kDataStoreAPIVersionUnknown]
    },
    // key value pairs defining what we need to know about the device
    kDeviceMAC: {
      id:0,
      w:false,
      // Serial number is "ULM5" + lower 3 bytes, UID is MAC with FFFE in center
      t:kTArray,
      v:[[0,0,0,0,0,0]],
    },
    kDevUID: {
      id:2,
      w:false,
      t:kTArray,
      v:[""]
    },
    kDevName: {
      id:3,
      t:kTString,
      v: [""]
    },
    kDevVersion: {
      id:4,
      w:false,
      t:kTString,
      v:["",""]
    },
    kSampleRate: {
      id:10,
      t:kTInt32,
      // kSampleRates
      v:[48000],
    },
    kClockSource: {
      id:11,
      t:kTByte,
      // kClockSources
      v:[0],
    },
    kOpticalMode: {
      id:5006,
      t:kTByte,
      v:[
        OpticalModes.ADAT, // input
        OpticalModes.ADAT  // output
      ],
    },
    kFPGALocked: {
      id:15,
      w:false,
      t:kTByte,
      // Bool 1 = locked, combine with PLL to determine if clock is locked
      v:[0],
    },
    kPLLLocked: {
      id:16,
      w:false,
      t:kTByte,
      // Bool 1 = locked, combine with FPGA to determine if clock is locked
      v:[0],
    },
    kMIDIThru: {
      id:19,
      t:kTByte,
      v:[0]
    },
    kLoopbackFirst: {
      id:20,
      t:kTByte,
      v:[0],
    },
    kABEnable: {
      id:5015,
      t:kTByte,
      v:[0],
    },
    kAEnable: {
      id:5016,
      t:kTByte,
      v:[1],
    },
    kBEnable: {
      id:5017,
      t:kTByte,
      v:[1],
    },
    kMonoEnable: {
      id:5018,
      t:kTByte,
      v:[1],
    },
    kMuteEnable: {
      id:5019,
      t:kTByte,
      v:[1],
    },
    kWordClockOut: {
      id:5024,
      t:kTByte,
      v:[0]
    },
    kOpticalExpander: {
      id:5025,
      t:kTByte,
      v:[OpticalExpanderMode.kDisabled],
    },
    kTalkbackLatch: {
      id:5026,
      t:kTByte,
      v:[1]
    },
    kTalkbackEnable: {
      id:5027,
      t:kTByte,
      v:[0]
    },
    kTalkbackLevel: {
      id:5028,
      t:kTFloat,
      v:[0]
    },
    kTalkbackDim: {
      id:5029,
      t:kTFloat,
      v:[0]
    },
    kTalkbackSource: {
      id:5030,
      t:kTByte,
      v:[0]
    },
    kTalkbackDestination: {
      id:5031,
      t:kTInt32,
      v:[0]
    },
    kModelID: {
      id:21,
      w:false,
      t:kTString,
      v:[DeviceModels.UltraLite + "M5"]
    },
    kFootSwitchEnabled: {
      id:5032,
      t:kTByte,
      v:[0]
    },
    kFootSwitchDownKey: {
      id:5033,
      t:kTString,
      v:[""]
    },
    kFootSwitchUpKey: {
      id:5034,
      t:kTString,
      v:[""]
    },

    kJackDetect: {
      id:5005,
      w:false,
      t:kTByte,
      // bool 0 = mic, 1 = line
      v:[0,0],
    },
    k48V: {
      id:5004,
      t:kTByte,
      // bool
      v:[0,0],
    },
    kPad: {
      id:5003,
      t:kTByte,
      // bool
      v:[0,0],
    },
    kiGain: {
      id:5001,
      t:kTByte,
      // in dB
      v:[
        0,0,        // mic pres range 0-74 dB currently only up to 65
        0,0,0,0,0,0 // line in's range 0 - 20 dB
      ],
    },
    kiPhase: {
      id:5002,
      t:kTByte,
      // bool, 0 = normal, 1 = invert
      v:[
        0,0,        // mic pres
        0,0,0,0,0,0 // line in's
      ],
    },
    kFPGAPatch: {
      id:5010,
      t:kTByte,
      // patch/routing table for all outputs, index defined in OutputDestination kOD_XXX
      v:defaultPatch,
    },
    kOpticalExpanderPatch: {
      id:5020,
      t:kTByte,
      // patch/routing table for all outputs, index defined in OutputDestination
      v:defaultPatch
    },
    kLoopbackSource: {
      id:5014,
      t:kTByte,
      v:[
        phonesOut[0].dsp,
        phonesOut[0].dsp+1
      ],
    },

    kiChannelNames: {
      id:6,
      t:kTString,
      // Names must be 32 bytes or less
      v:[
        "", "", // Mic
        "", "", "", "", "", "", // Line
        "", "", // spdif
        "", "", "", "", "", "", "", "", // adat a/toslink
      ],
    },

    koChannelNames: {
      id:7,
      t:kTString,
      // Names must be 32 bytes or less
      v:[
        "", "", // main
        "", "", "", "", "", "", "", "", // Line
        "", "", // phones
        "", "", // spdif
        "", "", "", "", "", "", "", "", // adat a/toslink
      ],
    },
    koTrim: {
      id:5000,
      t:kTByte,
      // range is 0-100  range is 0 to -100 dB.
      // Use LocalOutputs kLO_XXX for index
      v:[
        0, // LocalOutputs.lines[4]   = 0x0,
        0, // LocalOutputs.lines[5]   = 0x1,
        0, // LocalOutputs.lines[6]   = 0x2,
        0, // LocalOutputs.lines[7]   = 0x3,
        0, // LocalOutputs.lines[8]   = 0x4,
        0, // LocalOutputs.lines[9]  = 0x5,
        0, // LocalOutputs.lines[0]   = 0x6,
        0, // LocalOutputs.lines[1]   = 0x7,
        0, // LocalOutputs.lines[2]   = 0x8,
        0, // LocalOutputs.lines[3]   = 0x9,
        4, // LocalOutputs.phones[0] = 0xA,
        4, // LocalOutputs.phones[1] = 0xB,
      ]
    },
    kMainTrim: {
      id:5011,
      t:kTByte,
      // range is 0-100 range is 0 to -100 dB.
      v:[0],
    },
    kMainGroup: {
      id:5012,
      t:kTInt16,
      // Each bit is the enable for it's index
      // bit LocalOutputs.lines[0]
      // bit LocalOutputs.lines[1]
      // bit LocalOutputs.lines[2]
      // bit LocalOutputs.lines[3]
      // bit LocalOutputs.lines[4]
      // bit LocalOutputs.lines[5]
      // bit LocalOutputs.lines[6]
      // bit LocalOutputs.lines[7]
      // bit LocalOutputs.lines[8]
      // bit LocalOutputs.lines[9]
      v:[0xC0],
    },

    //////////////// MIX ///////////////////

    kiMixStereo: {
      id:1000,
      t:kTByte,
      // bool
      v:[
        0,0, // mic
        0,0,0,0,0,0, // Line
        1,0, // spdif
        0,0,0,0,0,0,0,0, // adat/toslink
        1,0, // host out main
        1,0, // host out phones
        1,0,1,0,1,0,1,0, // host out Line
        1,0  // dsp reverb
      ],
    },
    koMixStereo: {
      id:1001,
      t:kTByte,
      // bool
      v:[
        1,0, // main
        1,0,1,0,1,0,1,0, // Line
        1,0,1,0 // phones (unused) and reverb (unused)
      ],
    },

    // input eq 4 bands per channel (packed c1: b1 b2 b3 b4, c2: b1 b2 b3 b4 etc)
    kiEQMode: {
      id:1002,
      t:kTByte,
      // 4 state int - highpass = 3, high shelf = 2, low shelf = 1, peak = 0
      v:[
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2
      ],
    },
    kiEQBypass: {
      id:1003,
      t:kTByte,
      // bool
      v:[
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0
      ],
    },
    kiEQFreq: {
      id:1004,
      t:kTInt32,
      // float
      v:[
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000
      ],
    },
    kiEQGain: {
      id:1005,
      t:kTFloat,
      // float
      v:[
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0
      ],
    },
    kiEQBW: {
      id:1006,
      t:kTFloat,
      // float
      v:[
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1
      ],
    },

    kiGateBypass: {
      id:1007,
      t:kTByte,
      // bool
      v:[1,1],
    },
    kiGateThreshold: {
      id:1008,
      t:kTFloat,
      // float dB 0:-inf 1:0
      v:[-46,-46],
    },
    kiGateAttack: {
      id:1009,
      t:kTInt16,
      // 1-100 ms
      v:[10,10],
    },
    kiGateRelease: {
      id:1010,
      t:kTInt16,
      // 50-2000 ms
      v:[1000,1000],
    },

    kiCompBypass: {
      id:1011,
      t:kTByte,
      // bool
      v:[1,1,1,1,1,1,1,1],
    },
    kiCompThreshold: {
      id:1012,
      t:kTFloat,
      // float dB 0:-inf 1:0
      v:[-10,-10,-10,-10,-10,-10,-10,-10],
    },
    kiCompAttack: {
      id:1013,
      t:kTInt16,
      // 1-100 ms
      v:[10,10,10,10,10,10,10,10],
    },
    kiCompRelease: {
      id:1014,
      t:kTInt16,
      // 50-2000 ms
      v:[1000,1000,1000,1000,1000,1000,1000,1000],
    },
    kiCompRatio: {
      id:1015,
      t:kTFloat,
      // float 1-10?
      v:[2,2,2,2,2,2,2,2],
    },
    kiCompMakeup: {
      id:1021,
      t:kTFloat,
      // float
      v:[1,1,1,1,1,1,1,1],
    },

    // mic Line spdif atos homain holine hophones reverb
    // The number of inputs is defined in numInputs
    // The layout of index's is per bus, so channels 0-31 bus 0, then channels 0-31 bus 1 etc
    kiMixFader: {
      id:1016,
      t:kTFloat,
      // 8.24 -> float 32 in x 14 out
      v:[
        //0m 2Line        8sp  10 atos          18hm 20hLine          28hp 30r
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 1,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,1,0,0,0,0,0,0, 0,0, 0,0, // Ana 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,1,0,0,0,0,0, 0,0, 0,0, // Ana 3, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,1,0,0,0,0, 0,0, 0,0, // Ana 4, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,1,0,0,0, 0,0, 0,0, // Ana 5, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,1,0,0, 0,0, 0,0, // Ana 6, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,1,0, 0,0, 0,0, // Ana 7, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,1, 0,0, 0,0, // Ana 8, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 1,0, 0,0, // Phones 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,1, 0,0, // Phones 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 2, Ch 0-31
      ],
    },
    kiMixPan: {
      id:1017,
      t:kTFloat,
      // 8.24 -> float 32 in x 14 out
      v:[
        //0mic   2Line                  8spdif   10 atos                          18hmain  20hLine                        22hphns  30r
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Main 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Main 2, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 2, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 3, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 4, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 5, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 6, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 7, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Ana 8, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Phones 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Phones 2, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Reverb 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5, // Reverb 2, Ch 0-31
      ],
    },
    kiMixSolo: {
      id:1018,
      t:kTByte,
      // bool 32 in x 14 out
      v:[
        //0m 2Line      8sp  10 atos          18hm 20 hLine       22hp 30r
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 3, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 4, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 5, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 6, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 7, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 8, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 2, Ch 0-31
      ],
    },
    kiMixMute: {
      id:1019,
      t:kTByte,
      // bool 32 in x 14 out
      v:[
        //0m 2Line      8sp  10 atos          18hm 20 hLine       22hp 30r
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Main 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 3, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 4, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 5, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 6, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 7, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Ana 8, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Phones 2, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 1, Ch 0-31
        0,0, 0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0, // Reverb 2, Ch 0-31
      ],
    },
    kReverbDecay: {
      id:1030,
      t:kTByte,
      // decay: 0-100 as a percent of whatever 86 = default
      v:[86],
    },
    kReverbDamping: {
      id:1031,
      t:kTByte,
      // damping: 0-100 as a percent of whatever
      v:[90],
    },
    kReverbPreDelay: {
      id:1032,
      t:kTByte,
      // PreDelay samples 0-100 ms
      v:[50],
    },
    kReverbWidth: {
      id:1033,
      t:kTByte,
      // Width 0-100 percent
      v:[50],
    },
    kReverbPreset: {
      id:1034,
      t:kTByte,
      // Preset the hardware appies to the coefficients
      v:[1],
    },

    // output eq 3 bands per channel
    koEQMode: {
      id:1022,
      t:kTByte,
      // 4 state int - highpass = 3, high shelf = 2, low shelf = 1, peak = 0
      v:[
        1,0,2, // Main L-R
        1,0,2, // unused
        1,0,2, // Line 3
        1,0,2, // Line 4
        1,0,2, // Line 5
        1,0,2, // Line 6
        1,0,2, // Line 7
        1,0,2, // Line 8
        1,0,2, // Line 9
        1,0,2, // Line 10
        1,0,2, // Phones L-R
        1,0,2  // Reverb
      ],
    },
    koEQBypass: {
      id:1023,
      t:kTByte,
      // bool
      v:[
        0,0,0, // Main L-R
        0,0,0, // unused
        0,0,0, // Line 3
        0,0,0, // Line 4
        0,0,0, // Line 5
        0,0,0, // Line 6
        0,0,0, // Line 7
        0,0,0, // Line 8
        0,0,0, // Line 9
        0,0,0, // Line 10
        0,0,0, // Phones L-R
        0,0,0  // Reverb
      ],
    },
    koEQFreq: {
      id:1024,
      t:kTInt32,
      // float
      v:[
        60,800,14000, // Main L-R
        60,800,14000, // unused
        60,800,14000, // Line 3
        60,800,14000, // Line 4
        60,800,14000, // Line 5
        60,800,14000, // Line 6
        60,800,14000, // Line 7
        60,800,14000, // Line 8
        60,800,14000, // Line 9
        60,800,14000, // Line 10
        60,800,14000, // Phones L-R
        60,800,14000, // Reverb
      ],
    },
    koEQGain: {
      id:1025,
      t:kTFloat,
      // float
      v:[
        0,0,0, // Main L-R
        0,0,0, // unused
        0,0,0, // Line 3
        0,0,0, // Line 4
        0,0,0, // Line 5
        0,0,0, // Line 6
        0,0,0, // Line 7
        0,0,0, // Line 8
        0,0,0, // Line 9
        0,0,0, // Line 10
        0,0,0, // Phones L-R
        0,0,0  // Reverb
      ],
    },
    koEQBW: {
      id:1026,
      t:kTFloat,
      // float
      v:[
        1,1,1, // Main L-R
        1,1,1, // unused
        1,1,1, // Line 3
        1,1,1, // Line 4
        1,1,1, // Line 5
        1,1,1, // Line 6
        1,1,1, // Line 7
        1,1,1, // Line 8
        1,1,1, // Line 9
        1,1,1, // Line 10
        1,1,1, // Phones L-R
        1,1,1  // Reverb
      ],
    },

    // output gain 8.24fp
    koBusFader: {
      id:1027,
      t:kTFloat,
      // float: 14 out same order as mix buses
      v:[1,1, 1,1, 1,1, 1,1, 1,1, 1,1, 1,1],
    },
    koBusMute: {
      id:1028,
      t:kTByte,
      // float: 14 out same order as mix buses
      v:[0,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0],
    },

    kHostOutMixMode: {
      id:1020,
      t:kTByte,
      // bool: 0 = basic, 1 = extended
      v:[0],
    },

    kNetControlEnable: {
      id: 26,
      t:kTByte,
      v:[0]
    },

    // ID's -1 through -999 are local storage and not sent to the device
    kMixInMeterMode: {
      id:-1,
      t:kTByte,
      // bool: 0 = Pre Fader, 1 = Post Fader
      v:[0],
    },
    kUIMeterFPS: {
      id:-2,
      t:kTInt32,
      // kUIMeterFPSs
      v:[30],
    },
    kBusStripExp: {
      id:-4,
      t:kTString,
      // "on" or "off"
      v:["on"],
    },
    kTabExp: {
      id:-6,
      t:kTString,
      // "on" or "off"
      v:["on"],
    },

    kReturnDetect: {
      id:-7,  // Not implemented on ul5
      t:kTByte,
      w:false,
      v:[0, 0]
    },

    // ID's -1000 through -1999 are windows only parameters via dll
    kHostBuffer1x: {
      id:-1000,
      t:kTInt32,
      v:[128],
    },
    kExtraSafetyOffset1X: {
      id:-1001,
      t:kTInt32,
      // expect 0 (lowest latency) or 1 (default)
      v:[1],
    },
    kWaveInEnables: {
      id:-1002,
      t:kTInt32,
      v:[0xfff],
    },
    kWaveOutEnables: {
      id:-1003,
      t:kTInt32,
      v:[0xfff],
    },
    kWaveSRCEnabled: {
      id:-1004,
      t:kTInt32,
      v:[0],
    },
  };

  // These are duplicates of what's defined above in a
  // format that's more friendly for us
  return Object.freeze({
    PatchBase:PatchBase,
    kClockSources:kClockSources,
      
    MixTypes:MixTypes,
    nameIndices:nameIndices,

    // These are the meter indexes and fpga routing id's
    usbMainOut:InputSource.get_tdm(0),
    usbPhonesOut:miUSBPhOut,
    usbSPDIFOut:miUSBSPDIFOut,
    usbADATAOut:miUSBADATAOut,

    analogIns:analogIns,
    analogOuts:analogOuts,
    phonesOut:phonesOut,

    dspReverbDry:InputSource.get_dsp(20),
    dspReverbWet:dspReverbWet,

    spdifIn:miSPDIFIn,
    spdifOut:OutputDestination.get_spdif(0),

    numOpticalBanks:numOpticalBanks,

    zeroPatch:InputSource.zero,
    defaultDatastore:kDevJS,

    supportsOpticalExpander:false,
    supportsTalkback:true,
    supportsFootswitch:false,
    adatALabel:adatInfo.label,
    get_optical_info:function(mode) {
      if (mode === OpticalModes.ADAT)
        return [adatInfo];
      else
        return [tosInfo];
    },

    inputs:deviceInputs,
    channelInfo:[
      deviceInputs.mics,
      deviceInputs.lines,
      deviceInputs.spdif,
      adatInfo,
      tosInfo,
      deviceInputs.reverb
    ],
    numInputs:numInputs,

    kLoopbackOrder: kLoopbackOrder,

    hasInsertReturns: false,
    returns_igain_start_idx: 0,

    outputs:mixOutputs,
    busInfo:mixBusInfo,
    numBuses:mixBusInfo.reduce(function(accumulator, currentValue){
      return accumulator + currentValue.numCh;
    }, 0),
    // maps to the different buses
    busColors:[
      'rgb(200, 200, 200)', // main 1 white
      'rgb(200, 200, 200)', // main 2
      'rgb(255, 122, 34)', // anlg 1 orange
      'rgb(255, 122, 34)', // anlg 2
      'rgb(230, 200, 50)', // anlg 3 yellow
      'rgb(230, 200, 50)', // anlg 4
      'rgb(99, 210, 165)',// anlg 5 green
      'rgb(99, 210, 165)',// anlg 6
      'rgb(1, 146, 248)',// anlg 7 blue
      'rgb(1, 146, 248)',// anlg 8
      'rgb(214, 23, 1)', // phones 1 red
      'rgb(214, 23, 1)', // phones 2
      'rgb(154, 109, 176)', // rvb 1 purple
      'rgb(154, 109, 176)', // rvb 2
    ],

    get_mix_index:function(input, output) {
      return output * numInputs + input;
    },

    // name and bit
    waveInInfo:[
      {n:"Loopback In 1-2", b:4},
      {n:"Mic/Line/Inst In 1-2", b:0},
      {n:"Line In 3-4", b:1},
      {n:"Line In 5-6", b:2},
      {n:"Line In 7-8", b:3},
      {n:"S/PDIF In 1-2", b:5},
      {n:"Optical In 1-2", b:6},
      {n:"Optical In 3-4", b:7},
      {n:"Optical In 5-6", b:8},
      {n:"Optical In 7-8", b:9}
    ],
    waveOutInfo:[
      {n:"Main Out 1-2", b:0},
      {n:"Phones Out 1-2", b:5},
      {n:"Line Out 3-4", b:1},
      {n:"Line Out 5-6", b:2},
      {n:"Line Out 7-8", b:3},
      {n:"Line Out 9-10", b:4},
      {n:"S/PDIF Out 1-2", b:6},
      {n:"Optical Out 1-2", b:7},
      {n:"Optical Out 3-4", b:8},
      {n:"Optical Out 5-6", b:9},
      {n:"Optical Out 7-8", b:10}
    ]
  });
}();

