
const DeviceProps_828 = function() {

  const kClockSources = [
    { v:0, l:"Internal" },
    { v:5, l:"Word Clock"},
    { v:8, l:"S/PDIF" },
    { v:12, l:"Optical A" },
    { v:13, l:"Optical B"}
  ];

  // PatchBase (from 828mk5_patch.h)
  const PatchBase = function() {
    const local_base = 32;
    const digital_base = local_base + 26;
    return {
      host:   0,
      local:  local_base,
      digital:digital_base,
      dsp:    digital_base+18,
      zero:   127
    };
  }();

  // LocalInputs  (from 828mk5_patch.h)
  const LocalInputs = [
    15, // Mic 1
    2,  // Mic 2
    18, // Line 3
    5,  // Line 4
    14, // Line 5
    1,  // Line 6
    13, // Line 7
    0,  // Line 8
    17, // Line 9
    4   // Line 10
  ];

  // LocalOutputs  (from 828mk5_patch.h)
  const LocalOutputs = {
    lines:[
      5,  // Main 1
      7,  // Main 2
      21, // Line 3
      24, // Line 4
      22, // Line 5
      19, // Line 6
      6,  // Line 7
      9,  // Line 8
      11, // Line 9
      8   // Line 10
    ],
    phones:[
      15, // 1 left
      25, // 1 right
      18, // 2 left
      20  // 2 right
    ],
    nc: [
      12, 2
    ]
  };

  // DacOutputs  (from 828mk5_patch.h)
  // Also BS Mapping for DAC trims, main group and meters,
  // also meters are pre trim
  const DacOutputs = {
    lines:[
      12, // Main 1
      8,  // Main 2
      1,  // Line 3
      3,  // Line 4
      5,  // Line 5
      7,  // Line 6
      6,  // Line 7
      4,  // Line 8
      2,  // Line 9
      0   // Line 10
    ],
    phones:[
      15, // 1 left
      11, // 1 right
      13, // 2 left
      9   // 2 right
    ],
    nc: [
      10, 14
    ]
  };

  // InputSource  (from 828mk5_patch.h)
  const get_adat_bank_index = function(index){ return index >= 4 ? index + 4 : index; };
  const get_adat_index = function(index){ return index >= 8 ? get_adat_bank_index(index-8) + 4 : get_adat_bank_index(index); };
  const InputSource = {
    get_host:function(index){ return index < 32 ? PatchBase.host + index : undefined; },
    zero:PatchBase.zero,
    get_adat:function(index){ return index < 16 ? PatchBase.digital + 2 + get_adat_index(index) : undefined; },
    get_line:function(index){ return index < LocalInputs.length ? PatchBase.local + LocalInputs[index] : undefined; },
    get_spdif:function(index) { return index < 2 ? PatchBase.digital + index : undefined; },

    get_line_mix:function(index) { return index < LocalInputs.length ? PatchBase.dsp + index : undefined; },
    get_phone_mix:function(index) { return index < LocalOutputs.phones.length ? PatchBase.dsp + index + 10 : undefined; },
    get_reverb_mix:function(index) { return index < 2 ? PatchBase.dsp + 14 + index : undefined; },

    get_line_post_fx:function(index) { return index < LocalInputs.length ? PatchBase.dsp + index + 16 : undefined; }
  };

  // OutputDestination (from 828mk5_patch.h)
  const OutputDestination = {
    get_host:function(index){ return index < 32 ? PatchBase.host + index : undefined; },
    zero:PatchBase.zero,
    get_adat:function(index){ return index < 16 ? PatchBase.digital + 2 + get_adat_index(index) : undefined; },
    get_line:function(index){ return index < LocalOutputs.lines.length ? PatchBase.local + LocalOutputs.lines[index] : undefined; },
    get_phone:function(index) { return index < LocalOutputs.phones.length ? PatchBase.local + LocalOutputs.phones[index] : undefined; },
    get_spdif:function(index){ return index < 2 ? PatchBase.digital + index : undefined; }
  };

  const host_input_map = {
    main:0,
    line:2,
    phone:10,
    spdif:14,
    adat:16
  };

  const miUSBAnalOut = InputSource.get_host(host_input_map.main);
  const miUSBPhOut = InputSource.get_host(host_input_map.phone);

  const analogIns = function() {
    let ret = [];
    for (let i=0; i<LocalInputs.length; i++) {
      ret.push({ source:InputSource.get_line(i), destination:OutputDestination.get_host(host_input_map.main + i), postfx:InputSource.get_line_post_fx(i) });
    }
    return ret;
  }();

  const analogOuts = function() {
    let ret = [];
    for (let i=0; i<LocalOutputs.lines.length; i++) {
      ret.push({ destination:OutputDestination.get_line(i), source:InputSource.get_host(i), dac:DacOutputs.lines[i], dsp:InputSource.get_line_mix(i) });
    }
    return ret;
  }();

  const phonesOuts = [
    {
      destination: { left:OutputDestination.get_phone(0), right:OutputDestination.get_phone(1) },
      line: { left:DacOutputs.phones[0], right:DacOutputs.phones[1] },
      dsp: InputSource.get_phone_mix(0)
    },
    {
      destination: { left:OutputDestination.get_phone(2), right:OutputDestination.get_phone(3) },
      line: { left:DacOutputs.phones[2], right:DacOutputs.phones[3] },
      dsp: InputSource.get_phone_mix(2)
    }
  ];

  const dspReverbWet = 56;
  const dspReverbDry = InputSource.get_reverb_mix(0);
  const miSPDIFIn = InputSource.get_spdif(0);
  const miMixAnaOut = analogOuts[2].dsp;
  const miMixPhOut = phonesOuts[0].dsp;
  const miPostFXMicIn = analogIns[0].postfx;
  const miPostFXAnaIn = analogIns[2].postfx;

  // NOTE: meterstore.js is counting on the gate and comp meters being contiguous
  // if you insert anything here make sure it doesn't break meterstore.js
  // I gave up on converting these to dylans map for now
  const miPostEQMicIn = 102;
  const miPostEQAnaIn = 104;
  const miPostGateMicIn = 112;
  const miGateGRMicIn = 114;
  const miCompGRMicIn = 116;
  const miCompGRAnaIn = 118;

  const MixTypes = {
    Mic: "Mic",
    Line: "Line",
    SPDIF: "S/PDIF",
    ADATA: "ADAT A",
    ADATB: "ADAT B",
    TOSlink: "TOSlink",
    Reverb: "Reverb"
  };

  // output: index into koChannelNames
  // input: index into kiChannelNames
  const nameIndices = {
    main: {
      output: 0
    },
    mic: {
      input: 0
    },
    line: {
      input: 2,
      output: 2
    },
    phones: {
      output: 10
    },
    spdif: {
      input: 10,
      output: 14
    },
    optical_a: {
      input: 12,
      output: 16,
    },
    optical_b: {
      input: 20,
      output: 24
    }
  };

  const numOpticalBanks = 2;

  // type: type/name of bank
  // iGain: index of first channel mix gain at sample rates
  // numCh: number of mix input channels (and bus channels) at sample rates
  // name: index of first mix input channel name
  // stereo: index of first mix input channel stereo control
  // pre: index of ffirst mix input channel pre controls (48v, pad trim)
  // eq: index of first mix input channel eq controls
  // gate: index of first mix input channel gate controls
  // comp: index of first mix input channel comp controls
  // iBus: index of first mix bus gain at sample rates
  // name: index of first mix bus name
  // stereo: index of first mix input channel stereo control
  // busEq: index of first mix bus eq controls
  const adatAInfo = {
    numCh:[8,8,4,4,0,0],
    getMeter:function(index){ return InputSource.get_adat(index); },
    getPatch:function(index){ return OutputDestination.get_adat(index); },
    label: "A",
    getFx:function(index){ return 12 + index; }
  };

  const tosInfo = function() {
    let ret = Object.assign({}, adatAInfo);
    ret.type = MixTypes.TOSlink;
    ret.numCh = [2,2,2,2,0,0];
    return ret;
  }();

  const adatBInfo = {
    numCh:[8,8,4,4,0,0],
    getMeter:function(index){ return InputSource.get_adat(index + 8); },
    getPatch:function(index){ return OutputDestination.get_adat(index + 8); },
    getFx:function(index){ return 20 + index; },
    label: "B"
  };

  const deviceInputs = {
    mics: {
      type:MixTypes.Mic,
      numCh:[2,2,2,2,2,2],
      iGain:0,
      stereo:0,
      pre:0,
      eq:0,
      gate:0,
      comp:0,
      name:nameIndices.mic.input,
      meters:[miPostEQMicIn,miGateGRMicIn,miPostGateMicIn,miCompGRMicIn,miPostFXMicIn]
    },
    lines: {
      type:MixTypes.Line,
      numCh:[analogIns.length-2,analogIns.length-2,analogIns.length-2,analogIns.length-2,analogIns.length-2,analogIns.length-2],
      iGain:2,
      stereo:2,
      eq:2,
      comp:2,
      name:nameIndices.line.input,
      meters:[-1,-1,miPostEQAnaIn,miCompGRAnaIn,miPostFXAnaIn],
    },
    spdif: {
      type:MixTypes.SPDIF,
      numCh:[2,2,2,2,0,0],
      iGain:10,
      stereo:10,
      name:nameIndices.spdif.input,
      meters:[-1,-1,-1,-1,miSPDIFIn],
    },
    reverb: {
      type:MixTypes.Reverb,
      numCh:[2,2,2,2,0,0],
      iGain:42,
      name:-1,
      meters:[-1,-1,-1,-1,dspReverbWet],
    }
  };

  const mixOutputs = {
    mains:{
      type:MixBusTypes.Main,
      numCh:2,
      iBus:0,
      iGain:28,
      busEq:0,
      colorIndex:0,
      name:nameIndices.main.output,
      istereo:28, // stereo index of host output to mixer
      meters:[analogOuts[0].dsp, analogIns[0].destination],
    },
    lines:{
      type:MixBusTypes.Line,
      numCh:analogOuts.length-2,
      iBus:2,
      iGain:30,
      busEq:2,
      colorIndex:2,
      name:nameIndices.line.output,
      stereo:2,
      istereo:30, // stereo index of host output to mixer
      meters:[miMixAnaOut, miUSBAnalOut+2],
    },
    phones:{
      type:MixBusTypes.Phones,
      numCh:4,
      iBus:10,
      iGain:38,
      busEq:10,
      colorIndex:10,
      name:nameIndices.phones.output,
      istereo:38, // stereo index of host output to mixer
      meters:[miMixPhOut, miUSBPhOut],
    },
    reverbs:{
      type:MixBusTypes.Reverb,
      numCh:2,
      iBus:14,
      busEq:13,
      colorIndex:14,
      meters:[dspReverbWet, -1],
    }
  };

  // Bus index's
  // 0: main 1
  // 1: main 2
  // 2: Line 1
  // 3: Line 2
  // 4: Line 3
  // 5: Line 4
  // 6: Line 5
  // 7: Line 6
  // 8: Line 7
  // 9: Line 8
  // 10: phones 1
  // 11: phones 2
  // 12: phones 3
  // 13: phones 4
  // 14: reverb 1
  // 15: reverb 2
  const mixBusInfo = [
    mixOutputs.mains,
    mixOutputs.phones,
    mixOutputs.lines,
    mixOutputs.reverbs
  ];

  const numInputs = 44; // the number of inputs in kiMixPan, kiMixFader, kiMixSolo, kiMixMute, kiMixStereo

  const defaultPatch = function() {
    // Initialize to ZERO patch.
    let patch = new Array(256).fill(PatchBase.zero);

    // Then Set the patch we care about.
    analogIns.forEach(function(info){
      patch[info.destination] = info.source;
    });

    // loop back
    patch[OutputDestination.get_host(10)] = InputSource.get_host(0);
    patch[OutputDestination.get_host(11)] = InputSource.get_host(1);

    // spdif
    for (let i = 0; i < 2; i++) {
      patch[OutputDestination.get_host(12+i)] = InputSource.get_spdif(i);
      patch[OutputDestination.get_spdif(i)] = InputSource.get_host(14+i);
    }

    for (let i = 0; i < 16; i++) {
      patch[OutputDestination.get_host(14+i)] = InputSource.get_adat(i);
      patch[OutputDestination.get_adat(i)] = InputSource.get_host(16+i);
    }

    analogOuts.forEach(function(info){
      patch[info.destination] = info.dsp;
    });

    for (let i = 0; i < 4; i++) {
      patch[OutputDestination.get_phone(i)] = InputSource.get_host(10+i);
    }

    return patch;
  }();

  const kLoopbackOrder =  [
    { v:0, l:"USB In 11-12 (Default)" },
    { v:1, l:"USB In 1-2" }
  ];

  const kDevJS = {
    // Data

    kAPIVersion: {
      id:8,
      w:false,
      t:kTInt16,
      // version of the datastore api the device supports
      v:[kDataStoreAPIVersionUnknown]
    },
    // key value pairs defining what we need to know about the device
    kDeviceMAC: {
      id:0,
      w:false,
      // Serial number is "ULM5" + lower 3 bytes, UID is MAC with FFFE in center
      t:kTArray,
      v:[[0,0,0,0,0,0]],
    },
    kDevUID: {
      id:2,
      w:false,
      t:kTArray,
      v:[""]
    },
    kDevName: {
      id:3,
      w:true,
      t:kTString,
      v: [""]
    },
    kDevVersion: {
      id:4,
      w:false,
      t:kTString,
      v:["",""]
    },
    kSampleRate: {
      id:10,
      t:kTInt32,
      // kSampleRates
      v:[48000],
    },
    kClockSource: {
      id:11,
      t:kTByte,
      // kClockSources
      v:[0],
    },
    kOpticalMode: {
      id:5006,
      t:kTByte,
      v:[
        OpticalModes.ADAT, // input
        OpticalModes.ADAT  // output
      ],
    },
    kWordClockOut: {
      id:5024,
      t:kTByte,
      v:[0]
    },
    kOpticalExpander: {
      id:5025,
      t:kTByte,
      v:[OpticalExpanderMode.kDisabled],
    },
    kFPGALocked: {
      id:15,
      w:false,
      t:kTByte,
      // Bool 1 = locked, combine with PLL to determine if clock is locked
      v:[0],
    },
    kPLLLocked: {
      id:16,
      w:false,
      t:kTByte,
      // Bool 1 = locked, combine with FPGA to determine if clock is locked
      v:[0],
    },
    kMIDIThru: {
      id:19,
      t:kTByte,
      v:[0]
    },
    kLoopbackFirst: {
      id:20,
      t:kTByte,
      v:[0],
    },
    kABEnable: {
      id:5015,
      t:kTByte,
      v:[0],
    },
    kAEnable: {
      id:5016,
      t:kTByte,
      v:[1],
    },
    kBEnable: {
      id:5017,
      t:kTByte,
      v:[1],
    },
    kMonoEnable: {
      id:5018,
      t:kTByte,
      v:[1],
    },
    kMuteEnable: {
      id:5019,
      t:kTByte,
      v:[1],
    },
    kReturnDetect: {
      id:5021,
      t:kTByte,
      w:false,
      v:[0, 0]
    },
    kTalkbackLatch: {
      id:5026,
      t:kTByte,
      v:[1]
    },
    kTalkbackEnable: {
      id:5027,
      t:kTByte,
      v:[0]
    },
    kTalkbackLevel: {
      id:5028,
      t:kTFloat,
      v:[0]
    },
    kTalkbackDim: {
      id:5029,
      t:kTFloat,
      v:[0]
    },
    kTalkbackSource: {
      id:5030,
      t:kTByte,
      v:[0]
    },
    kTalkbackDestination: {
      id:5031,
      t:kTInt32,
      v:[0]
    },
    kModelID: {
      id:21,
      w:false,
      t:kTString,
      v:[DeviceModels.k828 + "E"]
    },
    kFootSwitchEnabled: {
      id:5032,
      recallable:false,
      t:kTByte,
      v:[0]
    },
    kFootSwitchDownKey: {
      id:5033,
      t:kTString,
      v:[""]
    },
    kFootSwitchUpKey: {
      id:5034,
      t:kTString,
      v:[""]
    },

    kJackDetect: {
      id:5005,
      w:false,
      t:kTByte,
      // bool 0 = mic, 1 = line
      v:[0,0],
    },
    k48V: {
      id:5004,
      t:kTByte,
      // bool
      v:[0,0],
    },
    kPad: {
      id:5003,
      t:kTByte,
      // bool
      v:[0,0],
    },
    kiGain: {
      id:5001,
      t:kTByte,
      // in dB
      v:[
        0,0,        // mic pres range 0-74 dB currently only up to 65
        0,0,0,0,0,0,0,0, // line in's range 0 - 20 dB
        0,0         // returns range 0 - 20 dB
      ],
    },
    kiPhase: {
      id:5002,
      t:kTByte,
      // bool, 0 = normal, 1 = invert
      v:[
        0,0,        // mic pres
        0,0,0,0,0,0 // line in's
      ],
    },
    kFPGAPatch: {
      id:5010,
      t:kTByte,
      // patch/routing table for all outputs, index defined in OutputDestination
      v:defaultPatch
    },
    kOpticalExpanderPatch: {
      id:5020,
      t:kTByte,
      // patch/routing table for all outputs, index defined in OutputDestination
      v:defaultPatch
    },
    kLoopbackSource: {
      id:5014,
      t:kTByte,
      v:[
        phonesOuts[0].dsp,
        phonesOuts[0].dsp+1
      ],
    },

    kiChannelNames: {
      id:6,
      t:kTString,
      // Names must be 32 bytes or less
      v:[
        "", "", // Mic
        "", "", "", "", "", "", "", "", // Line
        "", "", // spdif
        "", "", "", "", "", "", "", "", // adat a/toslink
        "", "", "", "", "", "", "", "", // adat b
      ],
    },

    koChannelNames: {
      id:7,
      t:kTString,
      // Names must be 32 bytes or less
      v:[
        "", "", // main
        "", "", "", "", "", "", "", "", // Line
        "", "", "", "", // phones
        "", "", // spdif
        "", "", "", "", "", "", "", "", // adat a/toslink
        "", "", "", "", "", "", "", "", // adat b
      ],
    },

    koTrim: {
      id:5000,
      t:kTByte,
      // range is 0-100  range is 0 to -100 dB.
      // Use LocalOutputs kLO_XXX for index
      v:[
        0, // Line 10
        0, // Line 3
        0, // Line 9
        0, // Line 4
        0, // Line 8
        0, // Line 5
        0, // Line 7
        0, // Line 6
        0, // Main R
        0, // Phone 2 R
        0, // NC0
        0, // Phone 1 R
        0, // Main L
        0, // Phone 2 L
        0, // NC1
        0, // Phone 1 L
      ]
    },
    kMainTrim: {
      id:5011,
      t:kTByte,
      // range is 0-100 range is 0 to -100 dB.
      v:[0],
    },
    kMainGroup: {
      id:5012,
      t:kTInt16,
      // Each bit is the enable for it's index
      // bit kLO_Main1
      // bit kLO_Main2
      // bit kLO_Line3
      // bit kLO_Line4
      // bit kLO_Line5
      // bit kLO_Line6
      // bit kLO_Line7
      // bit kLO_Line8
      // bit kLO_Line9
      // bit kLO_Line10
      v:[0xC0],
    },

    //////////////// MIX ///////////////////

    kiMixStereo: {
      id:1000,
      t:kTByte,
      // bool
      v:[
        0,0, // mic
        0,0,0,0,0,0,0,0, // Line
        1,0, // spdif
        0,0,0,0,0,0,0,0, // adat a/toslink
        0,0,0,0,0,0,0,0, // adat b
        1,0, // host out main
        1,0,1,0, // host out phones
        1,0,1,0,1,0,1,0, // host out Line
        1,0  // dsp reverb
      ],
    },
    koMixStereo: {
      id:1001,
      t:kTByte,
      // bool
      v:[
        1,0, // main
        1,0,1,0,1,0,1,0, // Line
        1,0,1,0,1,0,1,0 // phones (unused) and reverb (unused)
      ],
    },

    // input eq 4 bands per channel (packed c1: b1 b2 b3 b4, c2: b1 b2 b3 b4 etc)
    kiEQMode: {
      id:1002,
      t:kTByte,
      // 4 state int - highpass = 3, high shelf = 2, low shelf = 1, peak = 0
      v:[
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2,
        1,0,0,2
      ],
    },
    kiEQBypass: {
      id:1003,
      t:kTByte,
      // bool
      v:[
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0
      ],
    },
    kiEQFreq: {
      id:1004,
      t:kTInt32,
      // float
      v:[
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000,
        60,400,2000,14000
      ],
    },
    kiEQGain: {
      id:1005,
      t:kTFloat,
      // float
      v:[
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0,
        0,0,0,0
      ],
    },
    kiEQBW: {
      id:1006,
      t:kTFloat,
      // float
      v:[
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1,
        1,1,1,1
      ],
    },

    kiGateBypass: {
      id:1007,
      t:kTByte,
      // bool
      v:[1,1],
    },
    kiGateThreshold: {
      id:1008,
      t:kTFloat,
      // float dB 0:-inf 1:0
      v:[-46,-46],
    },
    kiGateAttack: {
      id:1009,
      t:kTInt16,
      // 1-100 ms
      v:[10,10],
    },
    kiGateRelease: {
      id:1010,
      t:kTInt16,
      // 50-2000 ms
      v:[1000,1000],
    },

    kiCompBypass: {
      id:1011,
      t:kTByte,
      // bool
      v:[1,1,1,1,1,1,1,1,1,1],
    },
    kiCompThreshold: {
      id:1012,
      t:kTFloat,
      // float dB 0:-inf 1:0
      v:[-10,-10,-10,-10,-10,-10,-10,-10,-10,-10],
    },
    kiCompAttack: {
      id:1013,
      t:kTInt16,
      // 1-100 ms
      v:[10,10,10,10,10,10,10,10,10,10],
    },
    kiCompRelease: {
      id:1014,
      t:kTInt16,
      // 50-2000 ms
      v:[1000,1000,1000,1000,1000,1000,1000,1000,1000,1000],
    },
    kiCompRatio: {
      id:1015,
      t:kTFloat,
      // float 1-10?
      v:[2,2,2,2,2,2,2,2,2,2],
    },
    kiCompMakeup: {
      id:1021,
      t:kTFloat,
      // float
      v:[1,1,1,1,1,1,1,1,1,1],
    },

    // mic Line spdif atos homain holine hophones reverb
    // The number of inputs is defined in DeviceProps.numInputs
    // The layout of index's is per bus, so channels 0-31 bus 0, then channels 0-31 bus 1 etc
    kiMixFader: {
      id:1016,
      t:kTFloat,
      // 8.24 -> float 44 in x 16 out
      v:[
        //0m 2Line            10sp 12 atos a        20 adat b        28hm 30hLine          38hp     42r
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Main 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Main 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 1,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,1,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,1,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 3, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,1,0,0,0,0, 0,0,0,0, 0,0, // Ana 4, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,1,0,0,0, 0,0,0,0, 0,0, // Ana 5, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,1,0,0, 0,0,0,0, 0,0, // Ana 6, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,1,0, 0,0,0,0, 0,0, // Ana 7, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,1, 0,0,0,0, 0,0, // Ana 8, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 1,0,0,0, 0,0, // Phones 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,1,0,0, 0,0, // Phones 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 1,0, 0,0,0,0,0,0,0,0, 0,0,1,0, 0,0, // Phones 3, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,1, 0,0,0,0,0,0,0,0, 0,0,0,1, 0,0, // Phones 4, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Reverb 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Reverb 2, Ch 0-31
      ],
    },
    kiMixPan: {
      id:1017,
      t:kTFloat,
      // 8.24 -> float 44 in x 16 out
      v:[
        //0mic   2Line                            10spdif  12 atos a                        20 adat b                        28hmain  30hLine                          38hphns          42r
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Main 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Main 2, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 2, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 3, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 4, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 5, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 6, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 7, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Ana 8, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Phones 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Phones 2, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Phones 3, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Phones 4, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Reverb 1, Ch 0-31
        0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5, 0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5, 0.5,0.5,0.5,0.5, 0.5,0.5, // Reverb 2, Ch 0-31
      ],
    },
    kiMixSolo: {
      id:1018,
      t:kTByte,
      // bool 44 in x 16 out
      v:[
      //0m   2Line            10sp 12 atos a        20 adat b        28hm 30 hLine         38hp     42r
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Main 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Main 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 3, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 4, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 5, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 6, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 7, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 8, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 3, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 4, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Reverb 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Reverb 2, Ch 0-31
      ],
    },
    kiMixMute: {
      id:1019,
      t:kTByte,
      // bool 44 in x 16 out
      v:[
        //0m 2Line            10sp 12 atos a        20 adat b        28hm 30 hLine         38hp     42r
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Main 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Main 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 3, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 4, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 5, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 6, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 7, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Ana 8, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 2, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 3, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Phones 4, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Reverb 1, Ch 0-31
        0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0, 0,0,0,0,0,0,0,0, 0,0,0,0, 0,0, // Reverb 2, Ch 0-31
      ],
    },
    kReverbDecay: {
      id:1030,
      t:kTByte,
      // decay: 0-100 as a percent of whatever 86 = default
      v:[86],
    },
    kReverbDamping: {
      id:1031,
      t:kTByte,
      // damping: 0-100 as a percent of whatever
      v:[90],
    },
    kReverbPreDelay: {
      id:1032,
      t:kTByte,
      // PreDelay samples 0-100 ms
      v:[50],
    },
    kReverbWidth: {
      id:1033,
      t:kTByte,
      // Width 0-100 percent
      v:[50],
    },
    kReverbPreset: {
      id:1034,
      t:kTByte,
      // Preset the hardware appies to the coefficients
      v:[1],
    },

    // output eq 3 bands per channel
    koEQMode: {
      id:1022,
      t:kTByte,
      // 4 state int - highpass = 3, high shelf = 2, low shelf = 1, peak = 0
      v:[
        1,0,2, // Main L-R
        1,0,2, // unused
        1,0,2, // Line 3
        1,0,2, // Line 4
        1,0,2, // Line 5
        1,0,2, // Line 6
        1,0,2, // Line 7
        1,0,2, // Line 8
        1,0,2, // Line 9
        1,0,2, // Line 10
        1,0,2, // Phones 1 L-R
        1,0,2, // unused
        1,0,2, // Phones 2 L-R
        1,0,2  // Reverb
      ],
    },
    koEQBypass: {
      id:1023,
      t:kTByte,
      // bool
      v:[
        0,0,0, // Main L-R
        0,0,0, // unused
        0,0,0, // Line 3
        0,0,0, // Line 4
        0,0,0, // Line 5
        0,0,0, // Line 6
        0,0,0, // Line 7
        0,0,0, // Line 8
        0,0,0, // Line 9
        0,0,0, // Line 10
        0,0,0, // Phones 1 L-R
        0,0,0, // unused
        0,0,0, // Phones 2 L-R
        0,0,0  // Reverb
      ],
    },
    koEQFreq: {
      id:1024,
      t:kTInt32,
      // float
      v:[
        60,800,14000, // Main L-R
        60,800,14000, // unused
        60,800,14000, // Line 3
        60,800,14000, // Line 4
        60,800,14000, // Line 5
        60,800,14000, // Line 6
        60,800,14000, // Line 7
        60,800,14000, // Line 8
        60,800,14000, // Line 9
        60,800,14000, // Line 10
        60,800,14000, // Phones 1 L-R
        60,800,14000, // unused
        60,800,14000, // Phones 2 L-R
        60,800,14000, // Reverb
      ],
    },
    koEQGain: {
      id:1025,
      t:kTFloat,
      // float
      v:[
        0,0,0, // Main L-R
        0,0,0, // unused
        0,0,0, // Line 3
        0,0,0, // Line 4
        0,0,0, // Line 5
        0,0,0, // Line 6
        0,0,0, // Line 7
        0,0,0, // Line 8
        0,0,0, // Line 9
        0,0,0, // Line 10
        0,0,0, // Phones 1 L-R
        0,0,0, // unused
        0,0,0, // Phones 2 L-R
        0,0,0  // Reverb
      ],
    },
    koEQBW: {
      id:1026,
      t:kTFloat,
      // float
      v:[
        1,1,1, // Main L-R
        1,1,1, // unused
        1,1,1, // Line 3
        1,1,1, // Line 4
        1,1,1, // Line 5
        1,1,1, // Line 6
        1,1,1, // Line 7
        1,1,1, // Line 8
        1,1,1, // Line 9
        1,1,1, // Line 10
        1,1,1, // Phones 1 L-R
        1,1,1, // unused
        1,1,1, // Phones 2 L-R
        1,1,1  // Reverb
      ],
    },

    // output gain 8.24fp
    koBusFader: {
      id:1027,
      t:kTFloat,
      // float: 14 out same order as mix buses
      v:[1,1, 1,1, 1,1, 1,1, 1,1, 1,1, 1,1],
    },
    koBusMute: {
      id:1028,
      t:kTByte,
      // float: 14 out same order as mix buses
      v:[0,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0],
    },

    kHostOutMixMode: {
      id:1020,
      t:kTByte,
      // bool: 0 = basic, 1 = extended
      v:[0],
    },

    kNetControlEnable: {
      id: 26,
      t:kTByte,
      v:[0]
    },

    // ID's -1 through -999 are local storage and not sent to the device
    kMixInMeterMode: {
      id:-1,
      t:kTByte,
      // bool: 0 = Pre Fader, 1 = Post Fader
      v:[0],
    },
    kUIMeterFPS: {
      id:-2,
      t:kTInt32,
      // kUIMeterFPSs
      v:[30],
    },
    kBusStripExp: {
      id:-4,
      t:kTString,
      // "on" or "off"
      v:["on"],
    },
    kTabExp: {
      id:-6,
      t:kTString,
      // "on" or "off"
      v:["on"],
    },


    // ID's -1000 through -1999 are windows only parameters via dll
    kHostBuffer1x: {
      id:-1000,
      t:kTInt32,
      v:[128],
    },
    kExtraSafetyOffset1X: {
      id:-1001,
      t:kTInt32,
      // expect 0 (lowest latency) or 1 (default)
      v:[1],
    },
    kWaveInEnables: {
      id:-1002,
      t:kTInt32,
      v:[0xfff],
    },
    kWaveOutEnables: {
      id:-1003,
      t:kTInt32,
      v:[0xfff],
    },
    kWaveSRCEnabled: {
      id:-1004,
      t:kTInt32,
      v:[0],
    },
  };

  // These are duplicates of what's defined above in a
  // format that's more friendly for us
  return Object.freeze({
    PatchBase:PatchBase,
    kClockSources:kClockSources,
    
    MixTypes:MixTypes,
    nameIndices:nameIndices,

    // These are the meter indexes and fpga routing id's
    usbMainOut:InputSource.get_host(host_input_map.main),
    usbPhonesOut:miUSBPhOut,
    usbSPDIFOut:InputSource.get_host(host_input_map.spdif),
    usbADATAOut:InputSource.get_host(host_input_map.adat),
    usbADATBOut:InputSource.get_host(host_input_map.adat+8),

    analogIns:analogIns,
    analogOuts:analogOuts,
    phonesOut:phonesOuts,

    dspReverbDry:dspReverbDry,
    dspReverbWet:dspReverbWet,

    spdifIn:miSPDIFIn,
    spdifOut:OutputDestination.get_spdif(0),

    numOpticalBanks:numOpticalBanks,

    zeroPatch:InputSource.zero,
    defaultDatastore:kDevJS,

    supportsOpticalExpander:true,
    supportsTalkback:true,
    supportsFootswitch:true,
    adatALabel:adatAInfo.label,
    get_optical_info:function(mode) {
      if (mode === OpticalModes.ADAT)
        return [adatAInfo, adatBInfo];
      else
        return [tosInfo, adatBInfo];
    },

    inputs:deviceInputs,
    channelInfo:[
      deviceInputs.mics,
      deviceInputs.lines,
      deviceInputs.spdif,
      {
        type:MixTypes.ADATA,
        numCh:[4,4,4,4,0,0],
        iGain:adatAInfo.getFx(0),
        stereo:adatAInfo.getFx(0),
        name:nameIndices.optical_a.input,
        meters:[-1,-1,-1,-1,adatAInfo.getMeter(0)]
      },
      {
        type:MixTypes.ADATA,
        numCh:[4,4,0,0,0,0],
        iGain:adatAInfo.getFx(4),
        stereo:adatAInfo.getFx(4),
        name:nameIndices.optical_a.input+4,
        meters:[-1,-1,-1,-1,adatAInfo.getMeter(4)]
      },
      {
        type:MixTypes.TOSlink,
        numCh:tosInfo.numCh,
        iGain:tosInfo.getFx(0),
        stereo:tosInfo.getFx(0),
        name:nameIndices.optical_a.input,
        meters:[-1,-1,-1,-1,tosInfo.getMeter(0)]
      },
      {
        type:MixTypes.ADATB,
        numCh:[4,4,4,4,0,0],
        iGain:adatBInfo.getFx(0),
        stereo:adatBInfo.getFx(0),
        name:nameIndices.optical_b.input,
        meters:[-1,-1,-1,-1,adatBInfo.getMeter(0)]
      },
      {
        type:MixTypes.ADATB,
        numCh:[4,4,0,0,0,0],
        iGain:adatBInfo.getFx(4),
        stereo:adatBInfo.getFx(4),
        name:nameIndices.optical_b.input+4,
        meters:[-1,-1,-1,-1,adatBInfo.getMeter(4)]
      },
      {
        type:MixTypes.Reverb,
        numCh:[2,2,2,2,0,0],
        iGain:42,
        name:-1,
        meters:[-1,-1,-1,-1,dspReverbWet],
      },
    ],
    numInputs:numInputs,
    kLoopbackOrder: kLoopbackOrder,
    hasInsertReturns: true,
    returns_igain_start_idx: 10,

    outputs:mixOutputs,
    busInfo:mixBusInfo,
    numBuses:mixBusInfo.reduce(function(accumulator, currentValue){
      return accumulator + currentValue.numCh;
    }, 0),
    // maps to the different buses
    busColors:[
      'rgb(200, 200, 200)', // main 1 white
      'rgb(200, 200, 200)', // main 2
      'rgb(255, 122, 34)', // anlg 1 orange
      'rgb(255, 122, 34)', // anlg 2
      'rgb(230, 200, 50)', // anlg 3 yellow
      'rgb(230, 200, 50)', // anlg 4
      'rgb(99, 210, 165)',// anlg 5 green
      'rgb(99, 210, 165)',// anlg 6
      'rgb(1, 146, 248)',// anlg 7 blue
      'rgb(1, 146, 248)',// anlg 8
      'rgb(214, 23, 1)', // phones 1 red
      'rgb(214, 23, 1)', // phones 2
      'rgb(214, 23, 1)', // phones 3 red
      'rgb(214, 23, 1)', // phones 4
      'rgb(154, 109, 176)', // rvb 1 purple
      'rgb(154, 109, 176)', // rvb 2
    ],

    get_mix_index:function(input, output) {
      return output * numInputs + input;
    },

    // name and bit (828)
    waveInInfo:[
      {n:"Loopback In 1-2", b:5},
      {n:"Mic/Line/Inst In 1-2", b:0},
      {n:"Line In 3-4", b:1},
      {n:"Line In 5-6", b:2},
      {n:"Line In 7-8", b:3},
      {n:"Line In 9-10", b:4},
      {n:"S/PDIF In 1-2", b:6},
      {n:"Optical In A 1-2", b:7},
      {n:"Optical In A 3-4", b:8},
      {n:"Optical In A 5-6", b:9},
      {n:"Optical In A 7-8", b:10},
      {n:"Optical In B 1-2", b:11},
      {n:"Optical In B 3-4", b:12},
      {n:"Optical In B 5-6", b:13},
      {n:"Optical In B 7-8", b:14},
    ],
    waveOutInfo:[
      {n:"Main Out 1-2", b:0},
      {n:"Phones 1 Out L-R", b:5},
      {n: "Phones 2 Out L-R", b:6},
      {n:"Line Out 3-4", b:1},
      {n:"Line Out 5-6", b:2},
      {n:"Line Out 7-8", b:3},
      {n:"Line Out 9-10", b:4},
      {n:"S/PDIF Out 1-2", b:7},
      {n:"Optical Out A 1-2", b:8},
      {n:"Optical Out A 3-4", b:9},
      {n:"Optical Out A 5-6", b:10},
      {n:"Optical Out A 7-8", b:11},
      {n:"Optical Out B 1-2", b:12},
      {n:"Optical Out B 3-4", b:13},
      {n:"Optical Out B 5-6", b:14},
      {n:"Optical Out B 7-8", b:15},
    ]
  });
}();

