'use strict';

const kDynamicsCtrlInfo = [
  {k: 'kiGateThreshold', l:"Threshold", d:1, u:"dB", min:-72, max:0, dft:-46, warp:false },
  {k: 'kiGateAttack', l:"Attack", d:0, u:"ms", min:1, max:100, dft:10, warp:false },
  {k: 'kiGateRelease', l:"Release", d:0, u:"ms", min:50, max:2000, dft:1000, warp:false },
  {k: 'kiCompThreshold', l:"Threshold", d:1, u:" dB", min:-72, max:0, dft:-10, warp:false },
  {k: 'kiCompRatio', l:"Ratio", d:1, u:"", min:1, max:10, dft:2, warp:false },
  {k: 'kiCompAttack', l:"Attack", d:0, u:" ms", min:1, max:100, dft:10, warp:false },
  {k: 'kiCompRelease', l:"Release", d:0, u:"ms", min:50, max:2000, dft:1000, warp:false },
  {k: 'kiCompMakeup', l:"Gain", d:1, u:"dB", min:-20, max:20, dft:0, warp:true },
];
function validateCtrl(ctrlInfo, v) {
    v = v.toFixed(ctrlInfo.d);
    if (ctrlInfo.max >= ctrlInfo.min)
    {
      v = Math.min(ctrlInfo.max, v);
      v = Math.max(ctrlInfo.min, v);
    }
    else
    {
      v = -v;
      v = Math.max(ctrlInfo.max, v);
      v = Math.min(ctrlInfo.min, v);
    }
    return v;
}
// convert ds value to range 0-1
function dynDSToV(dsv, ctrlInfo) {
  dsv = ctrlInfo.warp ? GainToDB(dsv) : dsv;
  return (dsv - ctrlInfo.min) / (ctrlInfo.max - ctrlInfo.min);
}
// convert value range 0-1 to ds
function dynVToDS(v, ctrlInfo) {
  v = v * (ctrlInfo.max - ctrlInfo.min) + ctrlInfo.min;
  return ctrlInfo.warp ? DBToGain(v) : v;
}

class DynamicsMeterBase extends View {
  // Horizontal meter with no peak hold, has l, r and gain reduction
  constructor(l,t,w,h,meterStore) {
    super(l,t,w,h);

    this.meterStore = meterStore;
    this.setup = this.setup.bind(this);
    this.enable = this.enable.bind(this);
    this.onMeters = this.onMeters.bind(this);
    this.setup = this.setup.bind(this);
    this.enable = this.enable.bind(this);

    this.enabled = this.isVis = false;
    this.isStereo = false;
    this.index = 0;
    this.setBG(kBlackBG);
  }
  setup(s, i) {
    this.isStereo = s;
    this.index = i;
  }
  enable(enable) {
    if(this.enabled != enable) {
      this.enabled = enable;
      this.setVis(enable);
      if(enable)
        this.meterStore.addListener(this.onMeters);
      else
        this.meterStore.removeListener(this.onMeters);

      this.requestRender();
    }
  }
  onMeters(v) {}
}
class DynamicsThumbMeter extends DynamicsMeterBase {
  // Horizontal meter with no peak hold, has l, r and gain reduction
  constructor(l,t,w,h,meterStore) {
    super(l,t,w,h,meterStore);
    this.lm = 0;
    this.rm = 0;
    this.grm = 0;
    this.grmIndex = 0;
  }
  setup(s, grmI, i) {
    this.isStereo = s;
    this.grmIndex = grmI;
    this.index = i;
  }
  render(c) {
    if(!this.enabled)
      return;

    if(this.isStereo){
      var y = this.t+1;

      c.fillStyle = kGRMeter;
      var w = this.grm * this.w;
      c.fillRect(this.r - w, y, w, 2);

      c.fillStyle = kLMeter;
      y += 2;
      c.fillRect(this.l, y, this.lm * this.w, 3);

      y += 4;
      c.fillRect(this.l, y, this.rm * this.w, 3);
    }
    else {
      var y = this.t + 1;

      c.fillStyle = kGRMeter;
      var w = this.grm * this.w;
      c.fillRect(this.r - w, y, w, 2);

      c.fillStyle = kLMeter;
      y += 4;
      c.fillRect(this.l, y, this.lm * this.w, 5);
    }
  }
  onMeters(v) {
    if(!this.enabled)
      return;
    if((this.index >= 0) && (this.index < v.length)) {
      var lm = CalcDBToCtrlLinear(v[this.index].m);
      var grm = 1.0 - CalcDBToCtrlLinear(v[this.grmIndex].mnd);
      if(this.isStereo) {
        var rm = CalcDBToCtrlLinear(v[this.index + 1].m);
        if((this.lm != lm) || (this.rm != rm) || (this.grm != grm)) {
          this.lm = lm;
          this.rm = rm;
          this.grm = grm;
          this.requestRender();
        }
      } else if((this.lm != lm) || (this.grm != grm)) {
          this.lm = lm;
          this.grm = grm;
          this.requestRender();
        }

    }
  }
}

class ReductionMeter extends DynamicsMeterBase {
  // Vertical meter with no peak hold
  constructor(l,t,w,h,meterStore) {
    super(l,t,w,h,meterStore);
    this.grm = 0;
  }
  render(c) {
    if(!this.enabled)
      return;

    c.fillStyle = kGRMeter;
    var h = this.grm * this.h;
    c.fillRect(this.l + 1, this.t + 1, this.w - 2, h);
  }
  onMeters(v) {
    if(!this.enabled)
      return;
    if((this.index >= 0) && (this.index < v.length)) {
      var grm = 1.0 - CalcDBToCtrlLinear(v[this.index].mnd);
      if(this.grm != grm) {
          this.grm = grm;
          this.requestRender();
        }
    }
  }
}
class HLMeter extends DynamicsMeterBase {
  // Horizontal meter with no peak hold, has l, r
  constructor(l,t,w,h,meterStore) {
    super(l,t,w,h,meterStore);
    this.lm = 0;
    this.rm = 0;
  }
  render(c) {
    if(!this.enabled)
      return;

    if(this.isStereo){
      var h = this.h/2 - 1;

      c.fillStyle = kLMeter;
      c.fillRect(this.l, this.t, this.lm * this.w, h);
      c.fillRect(this.l, this.t + h + 1, this.rm * this.w, h);
    }
    else {
      c.fillStyle = kLMeter;
      c.fillRect(this.l, this.t, this.lm * this.w, this.h);
    }
  }
  onMeters(v) {
    if(!this.enabled)
      return;
    if((this.index >= 0) && (this.index < v.length)) {
      var lm = CalcDBToCtrlLinear(v[this.index].m);
      if(this.isStereo) {
        var rm = CalcDBToCtrlLinear(v[this.index + 1].m);
        if((this.lm != lm) || (this.rm != rm)) {
          this.lm = lm;
          this.rm = rm;
          this.requestRender();
        }
      } else if(this.lm != lm) {
          this.lm = lm;
          this.requestRender();
        }

    }
  }
}
class DynamicsThumb  extends FxThumb {
  constructor(l, t, isFxChStrip, appdata, modalDialog) {
    super(l, t, kDynamicsFxType, isFxChStrip, modalDialog);

    this.onDSArray = this.onDSArray.bind(this);
    this.setup = this.setup.bind(this);

    this.gateOn = 0;
    this.compOn = 0;

    this.datastore = appdata.datastore;
    SetupDSArrayListener(this.datastore, this, 'kiGateBypass', 0);
    SetupDSArrayListener(this.datastore, this, 'kiGateThreshold', 0);
    SetupDSArrayListener(this.datastore, this, 'kiCompBypass', 0);
    SetupDSArrayListener(this.datastore, this, 'kiCompThreshold', 0);

    if(isFxChStrip) {
      this.addSub(this.gMeter = new DynamicsThumbMeter(this.l+8, this.t + 12, this.w - 16, 15, appdata.meterStore));
      this.addSub(this.cMeter = new DynamicsThumbMeter(this.l+8, this.t + 40, this.w - 16, 15, appdata.meterStore));
    }
    else {
      this.addSub(this.gMeter = new DynamicsThumbMeter(this.l+8, this.t + 6, this.w - 16, 10, appdata.meterStore));
      this.addSub(this.cMeter = new DynamicsThumbMeter(this.l+8, this.t + 28, this.w - 16, 10, appdata.meterStore));
    }
  }
  onDSArray(k, v) {
    if(this.info != undefined) {
      this.gateOn = !DSArrayGet(this.datastore, this, 'kiGateBypass');
      this.compOn = !DSArrayGet(this.datastore, this, 'kiCompBypass');
      this.gMeter.enable(this.gateOn && this.info.gate != undefined && this.info.gate != -1);
      this.cMeter.enable(this.compOn && this.info.comp != undefined && this.info.comp != -1);
      this.requestRender();
    }
  }
  setup(info) {
    this.setValid(false);
    super.setup(info);

    if(info.gate != undefined && info.gate != -1) {
      SetDSArrayIndex(this, 'kiGateBypass', info.gate);
      SetDSArrayIndex(this, 'kiGateThreshold', info.gate);
      this.gMeter.setup(info.isStereo, info.meters[kMIGateReduction], info.meters[kMIPreGate]);
      this.setValid(true);
    }
    if(info.comp != undefined && info.comp != -1) {
      SetDSArrayIndex(this, 'kiCompBypass', info.comp);
      SetDSArrayIndex(this, 'kiCompThreshold', info.comp);
      this.cMeter.setup(info.isStereo, info.meters[kMICompReduction], info.meters[kMIPreComp]);
      this.setValid(true);
    }
    this.onDSArray();
  }
  render(c) {
    super.render(c);
    if(this.isValid) {

      c.strokeStyle = "yellow";

      c.lineWidth = 1.0;
      if(this.gMeter.enabled) {
        var v = dynDSToV(DSArrayGet(this.datastore, this, 'kiGateThreshold'), kDynamicsCtrlInfo[0]);
        var x = this.gMeter.l + 5 + (v * (this.gMeter.w - 10));
        var y = this.gMeter.b + 1;
        c.beginPath();
        c.moveTo(x,y);
        c.lineTo(x-5, y+7);
        c.lineTo(x+5, y+7);
        c.closePath();
        c.stroke();
      }
      if(this.cMeter.enabled) {
        var v = dynDSToV(DSArrayGet(this.datastore, this, 'kiCompThreshold'), kDynamicsCtrlInfo[3]);
        var x = this.cMeter.l + 5 + (v * (this.cMeter.w - 10));
        var y = this.cMeter.b + 1;
        c.beginPath();
        c.moveTo(x,y);
        c.lineTo(x-5, y+7);
        c.lineTo(x+5, y+7);
        c.closePath();
        c.stroke();
      }
    }
  }
}


class DynamicsEnableBtn extends SimpleDSBtn {
  constructor(l, t, datastore, dsKey, i, label)  {
    super(l, t, 200, 40, datastore, dsKey, i);
    this.c = 'rgb(56,177,255)';
    this.setBG(kFxEditBG);
    this.label = label;
  }
  render(c) {

    c.fillStyle = this.c;
    c.strokeStyle = this.c;

    drawPwrIcon(c, 3, this.l + 19, this.t + 20, 18, (this.v == 0) ? this.c : kEQLabel);

    c.fillStyle = kTabTxt;
    c.font = "bold 18px sans-serif";
    c.textAlign = 'left';
    c.fillText(this.label, this.l + 60, 10 + this.cy);
  }
}

class DynamicsCtrlTextEdit extends EditableTextView {
  constructor(l,t,w,h, ctrlInfo, appdata) {
    super(l,t,w,h, 'eq-input-detail', '12px sans-serif', appdata.keyFilter);
    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);

    this.ctrlInfo = ctrlInfo;
    this.setBG(kFxEditBG);
    this.c = kEQValue;

    this.datastore = appdata.datastore;
    SetupDSArrayListenerBasic(this.datastore, this, ctrlInfo.k, 0);
  };
  setup(i) {
    SetDSArrayIndexBasic(this, i);
    var v = DSArrayGetBasic(this.datastore, this);
    v = this.ctrlInfo.warp ? GainToDB(v) : v;
    this.v = trimFloat(v, this.ctrlInfo.d) + " " + this.ctrlInfo.u;
    this.requestRender();
  }
  onDSArray(k, v) {
    v = this.ctrlInfo.warp ? GainToDB(v) : v;
    this.v = trimFloat(v, this.ctrlInfo.d) + " " + this.ctrlInfo.u;
    this.requestRender();
  }
  valueChanged(v) {
    v = validateCtrl(this.ctrlInfo, Number(v.replace(/[^0-9\.\-]/g, '')));
    DSArraySetBasic(this.datastore, this, this.ctrlInfo.warp ? DBToGain(v) : v);
    return v + this.ctrlInfo.u;
  }
}


class DynamicKnob extends SimpleDSKnob {
  constructor(l, t, cE, cD, ctrlInfo, appdata, i) {
    super(l, t, 60, 100, appdata.datastore, ctrlInfo.k, i);

    this.setup = this.setup.bind(this);
    this.enable = this.enable.bind(this);

    this.cr = new Rect(l + 2, t, 54, 60);
    this.cE = cE;
    this.cD = cD;
    this.setBG(kFxEditBG);
    this.ctrlInfo = ctrlInfo;
    this.enabled = false;

    this.addSub(this.edit = new DynamicsCtrlTextEdit(this.l, this.t + (this.h - 42), this.w, 18, ctrlInfo, appdata));
  }
  setup(i) {
    super.setup(i);
    this.edit.setup(i);
  }
  offset(dx, dy) {
    this.cr.offset(dx,dy);
    super.offset(dx, dy);
  }
  enable(enable) {
    if(this.enabled != enable) {
      this.enabled = enable;
      this.requestRender();
    }
  }
  handleDblClick() {
    this.setVal(this.ctrlInfo.warp ? DBToGain(this.ctrlInfo.dft) : this.ctrlInfo.dft);
  }
  dsToV(dsv) {
    return dynDSToV(dsv, this.ctrlInfo);
  }
  vToDS(v) {
    return dynVToDS(v, this.ctrlInfo);
  }
  render(c) {

    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, 2.25 * Math.PI);
    c.lineWidth = 7;
    c.strokeStyle = kDynCtrlBG;
    c.stroke();

    var v = (0.75 + this.v * 1.5) * Math.PI;
    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, v);
    c.lineWidth = 7;
    c.strokeStyle = this.enabled ? this.cE : this.cD;
    c.stroke();
    //console.log(v + ' ' + this.v);

    c.fillStyle = kTabTxt;
    c.font = "bold 18px Helvetica";
    c.textAlign = 'center';
    c.fillText(this.ctrlInfo.l, this.cx, this.t + (this.h - 2));
  }
}


class GridTracker extends SimpleDSKnob {
  constructor(r, datastore, ctrlInfo, i) {
    super(r.l,r.t,r.w,r.h, datastore, ctrlInfo.k, i);
    this.ctrlInfo = ctrlInfo;
  }
  dsToV(dsv) {
    return dynDSToV(dsv, this.ctrlInfo);
  }
  vToDS(v) {
    return dynVToDS(v, this.ctrlInfo);
  }
  render(c) { }
}


function drawGrid(r, c, min) {

  c.fillStyle = kBlackBG;
  c.fillRect(r.l, r.t, r.w, r.h);

  c.fillStyle = kTabTxt;
  c.font = "18px sans-serif";
  c.textAlign = 'right';

  const numLines = 6;
  var gs = r.h/numLines;
  var t = r.t;
  c.lineWidth = 1;
  c.strokeStyle = kGridColor;
  c.beginPath();
  for(var i = 0; i < numLines; ++i) {
    c.fillText((i*min/numLines).toFixed(0), r.l - 5, t+10);
    c.moveTo(r.l, t); c.lineTo(r.r, t);
    t += gs;
  }
  c.fillText(min.toFixed(0), r.l - 5, r.b+14);
  c.moveTo(r.l, r.b); c.lineTo(r.r, r.b);
  c.stroke();

  c.beginPath();
  t = r.l;
  c.moveTo(r.l, r.t); c.lineTo(r.l, r.b);
  for(var i = 1; i < numLines; ++i) {
    t += gs;
    c.fillText(((numLines - i)*min/numLines).toFixed(0), t+10, r.b+16);
    c.moveTo(t, r.t); c.lineTo(t, r.b);
  }
  c.fillText("0", r.r+5, r.b+16);
  c.moveTo(r.r, r.t); c.lineTo(r.r, r.b);

  c.stroke();
}
class GateView extends View {
  constructor(l,t,w,h,appdata) {
    super(l,t,w,h);

    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);
    this.datastore = appdata.datastore;

    this.addSub(this.eb = new DynamicsEnableBtn(l+36, t + 30, this.datastore, 'kiGateBypass', 0, "Gate"));
    this.addSub(this.thr = new DynamicKnob(l + 107, t + 430, kGateCtrl, kGateCtrlDisabled, kDynamicsCtrlInfo[0], appdata, 0));
    this.addSub(this.att = new DynamicKnob(this.thr.r + 34, t + 430, kGateCtrl, kGateCtrlDisabled, kDynamicsCtrlInfo[1], appdata, 0));
    this.addSub(this.rel = new DynamicKnob(this.att.r + 34, t + 430, kGateCtrl, kGateCtrlDisabled, kDynamicsCtrlInfo[2], appdata, 0));

    this.addSub(this.redMeter = new ReductionMeter(l + 373, this.t + 90, 8, 280, appdata.meterStore));
    this.addSub(this.lMeter = new HLMeter(l + 85, this.t + 395, 280, 8, appdata.meterStore));

    this.vr = new Rect(l + 85, t + 90, 280, 280);
    this.addSub(this.gtThr = new GridTracker(this.vr, this.datastore, kDynamicsCtrlInfo[0], 0));

    SetupDSArrayListener(this.datastore, this, 'kiGateBypass', 0);
    SetupDSArrayListener(this.datastore, this, 'kiGateThreshold', 0);
  }
  setup(info) {
    this.info = info;
    if(info.gate != undefined && info.gate != -1) {
      SetDSArrayIndex(this, 'kiGateBypass', info.gate);
      SetDSArrayIndex(this, 'kiGateThreshold', info.gate);
      this.eb.setup(info.gate);
      this.thr.setup(info.gate);
      this.gtThr.setup(info.comp);
      this.att.setup(info.gate);
      this.rel.setup(info.gate);
      this.redMeter.setup(false, info.meters[kMIGateReduction]);
      this.lMeter.setup(info.isStereo, info.meters[kMIPreGate]);
      this.onDSArray();
    }
  }
  onDSArray(k, v) {
      var gateOn = !DSArrayGet(this.datastore, this, 'kiGateBypass') && this.info.gate != undefined && this.info.gate != -1;
      this.thr.enable(gateOn);
      this.att.enable(gateOn);
      this.rel.enable(gateOn);
      this.redMeter.enable(gateOn);
      this.lMeter.enable(gateOn);
      this.requestRender();
  }
  render(c) {
    c.fillStyle = kFxEditBG;
    c.fillRect(this.l, this.t, this.w, this.h);

    drawGrid(this.vr, c, this.thr.ctrlInfo.min);

    c.lineWidth = 1;
    c.strokeStyle = kEQValue;
    c.beginPath();
    c.moveTo(this.vr.l, this.vr.b);
    c.lineTo(this.vr.r, this.vr.t);
    c.stroke();

    var v = (this.thr.vToDS(this.thr.v) - this.thr.ctrlInfo.min) * this.vr.w/(this.thr.ctrlInfo.max - this.thr.ctrlInfo.min);
    c.lineWidth = 3;
    c.strokeStyle = this.att.enabled ? kGateCtrl : kGateCtrlDisabled;
    c.beginPath();
    c.moveTo(this.vr.l + v, this.vr.b);
    c.lineTo(this.vr.l + v, this.vr.b - v);
    c.lineTo(this.vr.r, this.vr.t);
    c.stroke();
  }
}


class CompView extends View {
  constructor(l,t,w,h,appdata) {
    super(l,t,w,h);

    this.setL = this.setL.bind(this);
    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);

    this.datastore = appdata.datastore;
    this.addSub(this.eb = new DynamicsEnableBtn(l + 54, t + 30, this.datastore, 'kiCompBypass', 0, "Compression"));
    this.addSub(this.thr = new DynamicKnob(l + 28, t + 430, kCompCtrl, kCompCtrlDisabled, kDynamicsCtrlInfo[3], appdata, 0));
    this.addSub(this.rat = new DynamicKnob(this.thr.r + 34, t + 430, kCompCtrl, kCompCtrlDisabled, kDynamicsCtrlInfo[4], appdata, 0));
    this.addSub(this.att = new DynamicKnob(this.rat.r + 34, t + 430, kCompCtrl, kCompCtrlDisabled, kDynamicsCtrlInfo[5], appdata, 0));
    this.addSub(this.rel = new DynamicKnob(this.att.r + 34, t + 430, kCompCtrl, kCompCtrlDisabled, kDynamicsCtrlInfo[6], appdata, 0));
    this.addSub(this.mku = new DynamicKnob(this.rel.r + 34, t + 430, kCompCtrl, kCompCtrlDisabled, kDynamicsCtrlInfo[7], appdata, 0));

    this.addSub(this.redMeter = new ReductionMeter(l + 390, this.t + 90, 8, 280, appdata.meterStore));
    this.addSub(this.lMeter = new HLMeter(l + 102, this.t + 395, 280, 8, appdata.meterStore));

    SetupDSArrayListener(this.datastore, this, 'kiCompBypass', 0);
    SetupDSArrayListener(this.datastore, this, 'kiCompThreshold', 0);
    SetupDSArrayListener(this.datastore, this, 'kiCompRatio', 0);

    this.vr = new Rect(l + 102, t + 90, 280, 280);

    this.addSub(this.gtThr = new GridTracker(this.vr, this.datastore, kDynamicsCtrlInfo[3], 0));
  }
  setL(l) {
    var ol = l - this.l
    this.offset(ol, 0);

    this.vr.setL(l + 102);
    this.redMeter.move(l + 390, this.t + 90);
    this.lMeter.move(l + 102, this.t + 395);
  }
  setup(info) {
    this.info = info;
    if(info.comp != undefined && info.comp != -1) {
      SetDSArrayIndex(this, 'kiCompBypass', info.comp);
      SetDSArrayIndex(this, 'kiCompThreshold', info.comp);
      SetDSArrayIndex(this, 'kiCompRatio', info.comp);
      this.eb.setup(info.comp);
      this.thr.setup(info.comp);
      this.gtThr.setup(info.comp);
      this.att.setup(info.comp);
      this.rel.setup(info.comp);
      this.rat.setup(info.comp);
      this.mku.setup(info.comp);
      this.redMeter.setup(false, info.meters[kMICompReduction]);
      this.lMeter.setup(info.isStereo, info.meters[kMIPreComp]);
      this.onDSArray();
    }
  }
  onDSArray(k, v) {
      var compOn = !DSArrayGet(this.datastore, this, 'kiCompBypass') && this.info.comp != undefined && this.info.comp != -1;
      this.thr.enable(compOn);
      this.att.enable(compOn);
      this.rel.enable(compOn);
      this.rat.enable(compOn);
      this.mku.enable(compOn);
      this.redMeter.enable(compOn);
      this.lMeter.enable(compOn);
      this.requestRender();
  }
  render(c) {
    c.fillStyle = kFxEditBG;
    c.fillRect(this.l, this.t, this.w, this.h);

    drawGrid(this.vr, c, this.thr.ctrlInfo.min);
    c.lineWidth = 1;
    c.strokeStyle = kEQValue;
    c.beginPath();
    c.moveTo(this.vr.l, this.vr.b);
    c.lineTo(this.vr.r, this.vr.t);
    c.stroke();

    var v = (this.thr.vToDS(this.thr.v) - this.thr.ctrlInfo.min) * this.vr.w/(this.thr.ctrlInfo.max - this.thr.ctrlInfo.min);

    var s = 1/this.rat.vToDS(this.rat.v);
    var x1 = this.vr.l + v;
    var y1 = this.vr.b - v;
    var x2 = this.vr.r;
    var y2 = y1 - (s * (x2 - x1));

    c.lineWidth = 3;
    c.strokeStyle = this.att.enabled ? kCompCtrl : kCompCtrlDisabled;
    c.beginPath();
    c.moveTo(this.vr.l, this.vr.b);
    c.lineTo(x1, y1);
    c.lineTo(x2, y2);
    c.stroke();
  }
}

class DynamicsView extends View {
  constructor(l,t,w,h,appdata) {
    super(l,t,w,h);
    this.setBG(kFxEditBG);

    this.setup = this.setup.bind(this);
    this.addSub(this.gv = new GateView(this.l, this.t, this.w * 0.465, this.h, appdata));
    this.addSub(this.cv = new CompView(this.gv.r, this.t, this.w * 0.535, this.h, appdata));
  }
  setup(info) {
    this.gv.setup(info);
    this.cv.setup(info);

    if((info.gate != undefined) && (info.gate != -1))
    {
      this.gv.setVis(true);
      this.cv.setL(this.gv.r);
    }
    else
    {
      this.gv.setVis(false);
      this.cv.setL(this.cx - this.cv.w/2);
    }

    if((info.comp != undefined) && (info.comp != -1))
      this.cv.setVis(true);
    else
      this.cv.setVis(false);
    this.requestRender();
  }
}


