'use strict';

class Fader extends View {
  constructor(l,t,w,h) {
    super(l,t,w,h);
    this.setV = this.setV.bind(this);
    this.setGain = this.setGain.bind(this);
    this.setH = this.setH.bind(this);
    this.update  = this.update.bind(this);
    this.color = "white";

    this.wantsMouse = true;
    this.dragB = this.b - kFaderCapRadius;
    this.dragT = this.t + kFaderCapRadius;
    this.dragH = this.dragB - this.dragT;
    this.dragging = false;
    this.v = 0;
    this.dB = CalcCtrl24To12dB(this.v);
    this.gain = DBToGain(this.dB);
    this.capC = this.dragB;
    this.setBG(kMixBG);
  }
  update() {
    this.capC = this.dragT + ((1-this.v) * this.dragH);
    this.requestRender();
  }
  setV(v) {
    v = Math.min(1, Math.max(0, v));

    if(this.v != v) {
      this.v = v;
      this.dB =  CalcCtrl24To12dB(this.v);
      this.gain = DBToGain(this.dB);
      this.update();
    }
  }
  setGain(gain) {
    gain = Math.min(4, Math.max(0, gain));

    if(this.gain != gain) {
      this.gain = gain;
      this.dB =  GainToDB(this.gain);
      this.v = Calc12dBToCtrl24(this.dB);
      this.update();
    }
  }
  setH(h) {
      super.setH(h);
      this.dragB = this.b - (kFaderCapRadius + 2);
      this.dragT = this.t + (kFaderCapRadius + 2);
      this.dragH = this.dragB - this.dragT;
      this.update();
  }
  handleDblClick() {
    this.setV(this.v == 0 ? Calc12dBToCtrl24(0) : 0);
  }
  getMouseHandler(xy) {
    // Only allow click on the fader cap
    var mh = super.getMouseHandler(xy);
    if(mh == this) {
      if((xy.y >= (this.capC - kFaderCapRadius)) &&
        (xy.y <= (this.capC + kFaderCapRadius))) {
          return this;
      }
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y) &&
      (xy.y >= (this.capC - kFaderCapRadius)) &&
      (xy.y <= (this.capC + kFaderCapRadius))) {
      this.dragging = true;
      this.anchor = this.capC - xy.y;
    }
  }
  handleMouseMoved(xy, delta) {
    if(this.dragging) {
      var v = 1 - ((xy.y + this.anchor) - this.dragT)/this.dragH;
      this.setV(v);
    }
  }
  handleMouseUp() {
    this.dragging = false;
  }
  setMeterTH(t, h) {
    this.meterT = t;
    this.meterH = h;
  }
  render(c)
  {
    c.fillStyle = kBlackBG;
    c.fillRect(this.cx - 2, this.t + kFaderCapRadius, 3, this.h - (2 * kFaderCapRadius));

    if (this.meterT && this.meterH)
        draw12dBTicksNoText(c, kMeterTickColor, this.meterT, this.l, this.meterH);

    var x = this.l + 25;
    var y = this.capC;
    var faderValue = (this.v == 0 ? "-∞" : Math.trunc(this.dB))

    fillCircle(c, x, y, kFaderCapRadius, kMixBG);

    c.lineWidth = 3;

    c.beginPath();
    c.arc(x, y, kFaderCapRadius, 0, 2 * Math.PI);
    c.strokeStyle = kBlackBG;
    c.stroke();

    c.beginPath();
    c.arc(x, y, kFaderCapRadius - 4, 0, 2 * Math.PI);
    c.strokeStyle = this.color;
    c.stroke();

    c.fillStyle = this.color;
    c.font = "11px sans-serif";
    c.textAlign = "center";
    c.fillText(faderValue ,x, y + 4);
  }
}


class StripFader extends Fader {
  constructor(l,t,w,h, appdata, dsKey, meter) {
    super(l,t,w,h);
    this.refresh = this.refresh.bind(this);
    this.setup = this.setup.bind(this);
    this.dsKey = dsKey;
    this.gainIndex = 0;
    this.isStereo = false;
    this.meter = meter;
    this.mixManager = appdata.mixManager;
    this.datastore = appdata.datastore;

    this.datastore.addListener(this.dsKey, this.refresh);
    this.datastore.addListener('kMixInMeterMode', this.update);
  }
  update() {
    if(this.isVis && this.isPVis) {
      this.capC = this.dragT + ((1-this.v) * this.dragH);
      this.requestRender();
      var attDB = (this.datastore.getKVArrayItem('kMixInMeterMode', 0) == 1) ? this.dB : 0;
      if(this.meter)
        this.meter.setAttenuation(attDB);
    }
  }
  setV(v) {
    // called by the control when the mouse moves
    super.setV(v);
    this.datastore.setV(this.dsKey, this.gainIndex, this.gain);
    //console.log(this.v + " ctrl, " + gain + " gain, " + dB + " dB");
  }
  refresh() {
    // called when the fader gain changed through some other place
    var gain = this.datastore.getKVArrayItem(this.dsKey, this.gainIndex);
    if(gain != this.gain) {
      this.setGain(gain);
    }
  }
  setup(gainIndex, isStereo, meterIndex) {
    this.color = this.mixManager.getActiveBusInfo().color;
    this.gainIndex = gainIndex
    this.meterIndex = meterIndex;
    this.isStereo = isStereo;
    this.refresh();
  }
}


class PanFader extends View {
  constructor(l,t,w,h, appdata) {
    super(l,t,w,h);

    this.refresh = this.refresh.bind(this);
    this.setup = this.setup.bind(this);
    this.setV = this.setV.bind(this);
    this.color = kMixBG;

    this.scale = (this.w - 4);
    this.wantsMouse = true;
    this.dragging = false;
    this.v = 0.5;
    this.setBG(kMixBG);

    this.dsKey = 'kiMixPan';
    this.panIndex = 0;
    this.enabled = true;

    this.mixManager = appdata.mixManager;
    this.datastore = appdata.datastore;
    this.datastore.addListener(this.dsKey, this.refresh);
  }
  refresh() {
    // called when the fader changed through some other place
    this.setV(this.datastore.getKVArrayItem(this.dsKey, this.panIndex));
  }
  setup(panIndex, isStereo) {
    this.color = this.mixManager.getActiveBusInfo().color;
    this.panIndex = panIndex;
    this.enabled = !isStereo;
    this.refresh();
  }
  setV(v) {
    v = Math.min(1, Math.max(0, v));

    if(this.v != v) {
      this.v = v;
      this.requestRender();
      this.datastore.setV(this.dsKey, this.panIndex, this.v);
    }
  }
  handleDblClick() {
  if(this.enabled)
    this.setV(0.5);
  }
  handleMouseDown(xy) {
    if(this.enabled && this.contains(xy.x, xy.y)) {
      this.dragging = true;
    }
  }
  handleMouseMoved(xy, delta) {
    if(this.enabled && this.dragging) {
      var v = (xy.x - (this.l + 2))/this.scale
      this.setV(v);
    }
  }
  handleMouseUp() {
    this.dragging = false;
  }
  render(c)
  {
    if(this.enabled) {
      c.fillStyle = kBlackBG;
      c.fillRect(this.l, this.cy-2, this.w, 3);

      var x = this.l + 2 + (this.v * this.scale);

      c.fillStyle = "white";
      c.fillRect(x, this.cy-1, this.cx - x, 0.8);

      c.fillStyle = this.color;
      c.fillRect(x-1, this.cy-5, 3, 10);
    }
  }
}
