'use strict';

const kSpeakerImg = new OffScreenImage("speaker.svg");
const kPhonesImg = new OffScreenImage("phones.svg");
const kAlertImg = new OffScreenImage("alert.svg");

const kMicPreKnobType = Object.freeze({
  PreampType: 0,
  InsertTrimType:1
});

class MicPreHomeWithInsert extends View {
  constructor(l, t, w, h, i, returns_trim_idx, appdata) {
    super(l, t, w, h);
    this.setup = this.setup.bind(this);
    this.onJack = this.onJack.bind(this);
    this.gainTypeChanged = this.gainTypeChanged.bind(this);

    this.chIdx = i;

    const knob_title_styles = {
      fillStyle: kToggleBlue,
      font: '10px sans-serif'
    };

    this.addSub(this.preamp_gain = new IOKnobWithTitle(l, t + 115, kToggleBlue, kGateCtrlDisabled, appdata, kPreInsertEnabledPreampCtrlInfo, i, appdata, ["Preamp", "Gain"], knob_title_styles));
    this.addSub(this.insert_trim = new IOKnobWithTitle(l, t + 115, kToggleBlue, kGateCtrlDisabled, appdata, kPreInsertEnabledInsertTrimCtrlInfo, returns_trim_idx, appdata, ["Insert", "Gain"], knob_title_styles));

    const btns_top = this.t + 27;

    const row_width = this.row_width = this.preamp_gain.w + 5;
    const x_padding = 5;
    const y_padding = 5;
    const mini_btn_width = this.mini_btn_width = (row_width - x_padding) / 2;
    const row_height = 25;

    const textStyles = { font: "10px Arial "};
    
    this.addSub(this.b48V = new ToggleBtnWithTextStyle(0, btns_top, mini_btn_width, "48v", appdata.datastore, 'k48V', i, kToggleRed, kEmptyBG, textStyles));
    this.b48V.setH(row_height);

    this.addSub(this.pad = new ToggleBtnWithTextStyle(0, btns_top, mini_btn_width, "Pad", appdata.datastore, 'kPad', i, kToggleBlue, kEmptyBG, textStyles));
    this.pad.setH(row_height);

    this.gainType = new ValueProvider(kMicPreKnobType.PreampType);
    this.gainType.addListener(this.gainTypeChanged);

    this.addSub(this.preamp_switch = new Button(0, btns_top + (row_height + y_padding), row_width, row_height, {
      label: "Preamp Gain",
      bgcolor: kEmptyBG,
      color: kToggleBlue,
      textStyles: textStyles,
      onmousedown: ()=>{
        this.gainType.set(kMicPreKnobType.PreampType);
      }
    }));

    this.addSub(this.trim_switch =  new Button(0, btns_top + (2*(row_height + y_padding)), row_width, row_height, {
      label: "Insert Gain",
      bgcolor: kEmptyBG,
      color: kToggleBlue,
      textStyles: textStyles,
      onmousedown: ()=>{
        this.gainType.set(kMicPreKnobType.InsertTrimType);
      }
    }));

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.datastore.addListener('kJackDetect', this.onJack);
    
    this.gainTypeChanged(this.gainType.value);
    this.onJack('kJackDetect', this.datastore.getKVArray('kJackDetect'));
  }
  setup(preamp_idx, returns_trim_idx) {
    this.chIdx = preamp_idx;
    this.preamp_gain.setup(preamp_idx);
    this.insert_trim.setup(returns_trim_idx);
  }
  gainTypeChanged(v) {
    if (v == kMicPreKnobType.PreampType) {
      this.insert_trim.setVis(false);
      this.preamp_gain.setVis(true);
      this.preamp_switch.drawActive = true;
      this.trim_switch.drawActive = false;
    }
    else {
      this.insert_trim.setVis(true);
      this.preamp_gain.setVis(false);
      this.preamp_switch.drawActive = false;
      this.trim_switch.drawActive = true;
    }
    this.requestRender();
  }

  uRBCXCY() {
    super.uRBCXCY();

    this.preamp_gain.centerOnX(this.cx);
    this.insert_trim.centerOnX(this.cx);
    this.preamp_switch.centerOnX(this.cx);
    this.trim_switch.centerOnX(this.cx);

    this.b48V.setL(this.trim_switch.l);
    this.pad.setL(this.preamp_switch.r - this.pad.w);
  }

  offset(dx, dy) {
    super.offset(dx, dy);
    this.uRBCXCY();
  }

  onJack(k, v) {
    if(k == 'kJackDetect') {
      // bool 0 = mic, 1 = line/inst
      this.b48V.setEnabled(v[this.chIdx] == 0);
      this.pad.setEnabled(v[this.chIdx] == 0);
    }
  }
}

class MicPreHomeWithoutInsert extends View {
  constructor(l, t, w, h, i, appdata) {
    super(l, t, w, h);
    this.onJack = this.onJack.bind(this);
    this.chIdx = i;
    this.addSub(this.b48V = new ToggleBtn(l + 8, t + 27, 52, "48V", appdata.datastore, 'k48V', i, kToggleRed, kEmptyBG));
    this.addSub(this.pad = new ToggleBtn(l + 8, t + 73, 52, "Pad", appdata.datastore, 'kPad', i, kToggleBlue, kEmptyBG));
    this.addSub(this.gain = new IOKnob(l, t + 115, kToggleBlue, kGateCtrlDisabled, appdata, kPreTrimCtrlInfo, i));

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.datastore.addListener('kJackDetect', this.onJack);

    this.uRBCXCY();
  }

  uRBCXCY() {
    super.uRBCXCY();

    this.b48V.centerOnX(this.cx);
    this.pad.centerOnX(this.cx);
    this.gain.centerOnX(this.cx);
  }

  offset(dx, dy) {
    super.offset(dx, dy);
    this.uRBCXCY();
  }

  onJack(k, v) {
    if(k == 'kJackDetect') {
      // bool 0 = mic, 1 = line/inst
      this.b48V.setEnabled(v[this.chIdx] == 0);
      this.pad.setEnabled(v[this.chIdx] == 0);
    }
  }
}

class MicPreHome extends View {
  constructor(p, l, t, i, appdata) {
    super(l,t,135,200);

    this.setBG(kEmptyBG);
    p.addSub(this);
    this.onReturnsDetect = this.onReturnsDetect.bind(this);

    this.chIdx = i;
    const device_props = appdata.device.properties;
    this.addSub(this.meter = new IOMeter(this.l + 17, this.t + 25, 10, 150, device_props.analogIns[i].source, appdata.meterStore));
    this.addSub(this.standardControls = new MicPreHomeWithoutInsert(this.meter.r, this.t, this.r - this.meter.r, this.h, i, appdata));

    const returns_trim_idx = i + device_props.returns_igain_start_idx;
    this.addSub(this.insertControls = new MicPreHomeWithInsert(this.meter.r, this.t, this.r - this.meter.r, this.h, i, returns_trim_idx, appdata));

    this.standardControls.setVis(false);
    this.device = appdata.device;
    this.datastore = appdata.datastore;

    this.datastore.addListener('kReturnDetect', this.onReturnsDetect);
    this.onReturnsDetect();
  }
  uRBCXCY() {
    super.uRBCXCY();

    const returns_detected = this.datastore.getKVArrayItem('kReturnDetect', this.chIdx);

    const dx = 12;
    this.standardControls.centerOnX(this.cx + dx);
    this.insertControls.centerOnX(this.cx + dx);

    const meter_left = returns_detected ? (this.l + (((this.insertControls.preamp_switch.l + 4  - this.l) - this.meter.w) / 2.0)) : (this.l + ((this.standardControls.gain.l - this.l) / 2.0));
    this.meter.setL(meter_left);
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.uRBCXCY();
  }
  recreate() {
    const device_props = this.device.properties;
    this.meter.index = device_props.analogIns[this.chIdx].source;

    const returns_trim_idx = this.chIdx + device_props.returns_igain_start_idx;
    this.insertControls.setup(this.chIdx, returns_trim_idx);
    this.requestRender();
  }

  onReturnsDetect() {
    const returns_detected = this.datastore.getKVArrayItem('kReturnDetect', this.chIdx);
    this.standardControls.setVis(!returns_detected);
    this.insertControls.setVis(returns_detected);

    this.uRBCXCY();
    this.requestRender();
  }

  render(c) {
    drawLabelGroup(c, getLineInName(this.device, this.datastore, this.chIdx), this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}
class PhonesVolHome extends View {
  constructor(p, l, t, index, appdata) {
    super(l,t,135,200);
    p.addSub(this);
    this.index = index;
    this.phoneInfo = undefined;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.showNum = true;
    this.addSub(this.m1 = new IOMeter(l + 16, this.t + 25, 10, 150, -1, appdata.meterStore));
    this.addSub(this.m2 = new IOMeter(this.m1.l + 12, this.t + 25, 10, 150, -1, appdata.meterStore));
    this.addSub(this.trim = new IOKnob(l + 52, t + 115, kToggleBlue, kGateCtrlDisabled, appdata, kPhonesTrimCtrlInfo, -1, -1));

    this.meterWidth = this.m2.r - this.m1.l;

    this.datastore.addListener('kFPGAPatch', this.onPatchChanged.bind(this));
    this.datastore.addListener('kOpticalExpander', this.updateMeters.bind(this));
  }
  uRBCXCY() {
    super.uRBCXCY();

    const dx = 18;
    this.trim.centerOnX(this.cx + dx);

    const meterLeft = this.l + (((this.trim.l - this.l) - this.meterWidth) / 2) + 3;
    this.m1.setL(meterLeft);
    this.m2.setL(meterLeft + 12);
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.uRBCXCY();
  }
  updateMeters() {
    if (this.phoneInfo != undefined) {
      const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 
      this.m1.index = this.datastore.getKVArrayItem(patchTableKey, this.phoneInfo.destination.left);
      this.m2.index = this.datastore.getKVArrayItem(patchTableKey, this.phoneInfo.destination.right);
    }
  }

  onPatchChanged(k, v){
    if (this.phoneInfo !== undefined) {
      this.updateMeters();
    }
  }

  recreate() {
    const phonesList = this.device.properties.phonesOut;
    const exists = this.index < phonesList.length;
    this.setVis(exists);

    if (exists) {
      this.showNum = phonesList.length > 1;
      const info = phonesList[this.index];
      this.phoneInfo = info;
      this.updateMeters(info);
      this.trim.setup(info.line.left, info.line.right);
    }
    else {
      this.phoneInfo = undefined;
      this.m1.index = -1;
      this.m2.index = -1;
      this.trim.setup(-1);
    }
  }
  render(c) {
    const num_label = this.showNum ? " " + (this.index+1) : "";
    drawLabelGroup(c, "Phones" + num_label, this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}
class MainVolGroupHome extends View {
  constructor(p, l, t, appdata) {
    super(l,t,135,200);

    this.recreate = this.recreate.bind(this);
    this.deviceTypeChanged = this.deviceTypeChanged.bind(this);

    p.addSub(this);
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.addSub(this.m1 = new IOMeter(l + 16, this.t + 25, 10, 150, -1, appdata.meterStore));
    this.addSub(this.m2 = new IOMeter(this.m1.l + 12, this.t + 25, 10, 150, -1, appdata.meterStore));
    this.addSub(this.muteBtn = new ToggleBtn(l + 52, this.t + 27, 68, "MUTE", this.datastore, 'kMuteEnable', 0, kToggleRed, kEmptyBG));
    this.addSub(this.monoBtn = new ToggleBtn(l + 52, this.t + 73, 68, "MONO", this.datastore, 'kMonoEnable', 0, kToggleBlue, kEmptyBG));
    this.addSub(this.vol = new IOKnob(l + 52, t + 115, kToggleBlue, kGateCtrlDisabled, appdata, kMainVolCtrlInfo, 0));

    this.meterWidth = this.m2.r - this.m1.l;

    this.datastore.addListener('kOpticalExpander', this.recreate);
    this.datastore.addListener('kOpticalExpanderPatch', this.recreate);
    this.datastore.addListener('kModelID', this.deviceTypeChanged);
  }
  uRBCXCY() {
    super.uRBCXCY();

    const dx = 18;
    this.muteBtn.centerOnX(this.cx + dx);
    this.monoBtn.centerOnX(this.cx + dx);
    this.vol.centerOnX(this.cx + dx);

    const meterLeft = this.l + (((this.vol.l - this.l) - this.meterWidth) / 2) + 3;
    this.m1.setL(meterLeft);
    this.m2.setL(meterLeft + 12);
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.uRBCXCY();
  }
  recreate() {
    const device_props = this.device.properties;
    const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 

    this.m1.index = this.datastore.getKVArrayItem(patchTableKey, device_props.analogOuts[0].destination);
    this.m2.index = this.datastore.getKVArrayItem(patchTableKey, device_props.analogOuts[1].destination);
  }
  render(c) {
    drawLabelGroup(c, "Monitor", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
  deviceTypeChanged() {
    const supportsMuteAndMono = (()=>{
      const model_id = this.datastore.getKVArrayItem('kModelID', 0);
      if (model_id.search("828") == -1)
        return false;
      return true;
    })();

    if (this.muteBtn != undefined)
      this.muteBtn.setVis(supportsMuteAndMono);
    if (this.monoBtn != undefined)
      this.monoBtn.setVis(supportsMuteAndMono);
  }
}

class TalkbackDestinationSelect extends MultiSelect {
  constructor(d, l, t, w, h, ctrlID, appdata) {
    super(d, l, t, w, h, ctrlID, appdata.eventSys);

    this.onDs = this.onDs.bind(this);

    this.datastore = appdata.datastore;
    this.device = appdata.device;
    this.destinations = makeBitset();
    this.scrollBarV.setColor("rgb(20,100,155)");

    this.datastore.addListener('kTalkbackDestination', this.onDs);
  }

  onDs(k, v){
    const value = this.datastore.getKVArrayItem('kTalkbackDestination', 0);
    const destinations = makeBitset(value);
    this.destinations = destinations;
    this.cBoxes.forEach(function(cbox){
      cbox.checked = destinations.get_bit(cbox.v);
    });
  }

  changeCB(id, checked){
    this.destinations.set_bit(id, checked);
    this.destinations.set_bit(id+1, checked);
    this.datastore.setV('kTalkbackDestination', 0, this.destinations.get());
  }

  recreate() {
    this.clearOptions();
    const mixes = this.device.properties.outputs;

    const mains = mixes.mains;
    const num_mains = mains.numCh / 2;
    for (let i = 0; i < num_mains; i++) {
      const index = i * 2;
      this.addOption("Main Out " + (index+1) + "-" + (index+2), mains.iBus + index);
    }

    const phones = mixes.phones;
    const num_phones = phones.numCh / 2;
    for (let i = 0; i < num_phones; i++) {
      const index = i * 2;
      this.addOption("Phones Out " + (i+1), phones.iBus + index);
    }

    const lines = mixes.lines;
    const num_lines = lines.numCh / 2;
    for (let i = 0; i < num_lines; i++) {
      const index = i * 2;
      this.addOption("Line Out " + (index+3) + "-" + (index+4), lines.iBus + index);
    }
  }
};

class TackbackFxThumbs extends View {
  constructor(l,t,w,h, appdata) {
    super(l,t,w,h);

    this.setChNum = this.setChNum.bind(this);
    this.setEnabled = this.setEnabled.bind(this);

    this.mixManager = appdata.mixManager;
    this.datastore = appdata.datastore;

    // create the modal dialogs
    const fxbasecanvas = new FxModalCanvas(appdata);
    this.chFxEdit = new ChFxEditView(appdata, fxbasecanvas, false);

    const itemW = w;
    const itemL = l;
    const itemH = (h - 5) / 2;

    this.addSub(this.eq = new EQThumb(itemL, t, appdata, true, false, this.chFxEdit));
    this.eq.setH(itemH);
    this.fxSlotH = this.eq.h + 5;
    this.addSub(this.dyn = new DynamicsThumb(itemL, t + this.fxSlotH, false, appdata, this.chFxEdit));
    this.dyn.setH(itemH);

    this.setChNum(0);
    this.enableClipping(true);
    this.eq.enableClipping(true);
    this.dyn.enableClipping(true);
    this.eq.setValid(true);
    this.dyn.setValid(true);
  }

  setChNum(chNum) {
    this.chNum = chNum;
    this.chFxEdit.setActiveCh(chNum);

    var info = this.mixManager.getChInfo(this.chNum);
    this.eq.setup(info);
    this.dyn.setup(info);
  }


  activate(channel, fxType) {
    this.mixManager.setActiveCh(channel);

    if(fxType == undefined)
      fxType = this.lastFXType;

    this.lastFXType = fxType;
    this.removeSub(this.eq);
    this.removeSub(this.dyn);

    var f = this.fx[fxType];
    if(f != undefined) {

      this.addSub(f);
      this.bg = f.bg;

      this.canvas.activate(this);
      this.requestRender();
    }
  }


  getMouseHandler(xy) {
    if (!this.enabled) {
      return undefined;
    }
    return super.getMouseHandler(xy);
  }

  setEnabled(e) {
    if (this.enabled != e) {
      this.enabled = e;
      if (!e) {
        this.eq.setAlpha(kDimmedAlpha);
        this.dyn.setAlpha(kDimmedAlpha);
      }
      else {
        this.eq.setAlpha(1.0);
        this.dyn.setAlpha(1.0);
      }
      this.requestRender();
    }
  }

  render(c) {
    c.fillStyle = kEmptyBG;
    c.fillRect(this.l, this.t, this.w, this.h);
  }
}

class Talkback extends View {
  constructor(parent, left, top, appdata) {
    super(left,top,556,215);

    this.setBG(kEmptyBG);

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.latchMode = true;
    this.talkbackEnabled = false;
    this.talkHeld = false;

    parent.addSub(this);
    this.latchCBChanged = this.latchCBChanged.bind(this);
    this.latchChanged = this.latchChanged.bind(this);
    this.enabledChanged = this.enabledChanged.bind(this);
    this.sourceMenuChanged = this.sourceMenuChanged.bind(this);
    this.sourceChanged = this.sourceChanged.bind(this);
    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);
    this.wantsMouse = true;
    this.validSource = false;

    this.enableHitR = new Rect(left + 30, top + 40, 90, 30);

    const bottom_margin = top + 194;

    this.addSub(this.latchBox = new CheckBox(parent.c.div, left + 48, this.enableHitR.b + 5, "Latch", 50, this.latchCBChanged));
    this.addSub(this.fxThumbs = new TackbackFxThumbs(left + 30, this.latchBox.b + 12, kChStripW - 14, bottom_margin - (this.latchBox.b + 12), appdata));

    const knob_left = this.enableHitR.r + 45;
    this.addSub(this.sourceMenu = new SelectMenu(parent.c.div, knob_left - 10, top + 54, 172, 27, this.sourceMenuChanged));

    this.addSub(this.level = new IOKnob(knob_left, top + 100, kToggleBlue, kGateCtrlDisabled, appdata, kTalkbackLevelCtrlInfo, 0));
    this.addSub(this.dim = new IOKnob(knob_left + 90, top + 100, kToggleBlue, kGateCtrlDisabled, appdata, kTalkbackDimCtrlInfo, 0));
    this.addSub(this.destSelect = new TalkbackDestinationSelect(parent.c.div, left + 360, top + 54, 172, bottom_margin - (top + 54), 0, appdata));

    this.datastore.addListener('kTalkbackLatch', this.latchChanged);
    this.datastore.addListener('kTalkbackEnable', this.enabledChanged);
    this.datastore.addListener('kTalkbackSource', this.sourceChanged);
    this.datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);
  }

  opticalExpanderChanged(k, v){
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0;
    this.setEnabled(e);
  }

  recreate() {
    this.sourceMenu.clearOptions();
    this.sourceMenu.addOption(255, "None");
    const device_props = this.device.properties;
    const mics = device_props.inputs.mics;
    const num_mics = mics.numCh[0];
    for (let i = 0; i < num_mics; i++) {
      this.sourceMenu.addOption(mics.iGain + i, "Mic input " + (i+1));
    }

    const lines = device_props.inputs.lines;
    const num_lines = lines.numCh[0];
    for (let i = 0; i < num_lines; i++) {
      this.sourceMenu.addOption(lines.iGain + i, "Line input " + (i+3));
    }
    this.destSelect.recreate();
  }

  setEnabled(e) {
    super.setEnabled(e);
    this.latchBox.setEnabled(e);
    this.sourceMenu.setEnabled(e);
    this.destSelect.setEnabled(e);
  }

  latchCBChanged(id, checked){
    this.datastore.setV('kTalkbackLatch', 0, checked ? 1 : 0);
  }
  latchChanged(k, v) {
    this.latchMode = this.datastore.getKVArrayItem('kTalkbackLatch', 0) > 0;
    this.latchBox.check(this.latchMode);
  }
  enabledChanged(k, v) {
    this.talkbackEnabled = this.datastore.getKVArrayItem('kTalkbackEnable', 0) > 0;
    this.requestRender(this.enableHitR);
  }
  sourceMenuChanged() {
    const new_source = parseInt(this.sourceMenu.sel.value);
    this.datastore.setV('kTalkbackSource', 0, new_source);
  }
  sourceChanged(k, v) {
    let src = this.datastore.getKVArrayItem('kTalkbackSource', 0);
    this.sourceMenu.sel.value = src;
    if (src != undefined && src != 255) {
      this.fxThumbs.setEnabled(true);
      this.fxThumbs.setChNum(src);
      this.level.setEnabled(true);
      this.dim.setEnabled(true);
      this.validSource = true;
    }
    else {
      this.fxThumbs.setEnabled(false);
      this.level.setEnabled(false);
      this.dim.setEnabled(false);
      this.validSource = false;
    }
    this.requestRender();
  }

  offset(dx, dy) {
    super.offset(dx, dy);
    this.enableHitR.offset(dx, dy);
  }

  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    let mh = super.getMouseHandler(xy);
    if(mh == this) {
      if (this.enableHitR.contains(xy.x, xy.y)) {
        return this;
      }
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) {
    if (this.enableHitR.contains(xy.x, xy.y)) {
      this.datastore.setV('kTalkbackEnable', 0, this.latchMode && this.talkbackEnabled ? 0 : 1);
      this.talkHeld = true;
    }
    else {
      this.talkHeld = false;
    }
  }
  handleMouseUp(xy) {
    if (this.talkHeld && !this.latchMode) {
      this.datastore.setV('kTalkbackEnable', 0, 0);
    }
    this.talkHeld = false;
  }

  render(c) {
    drawLabelGroup(c, "Talkback", this.l, this.t+1, this.w-2, this.h, 6, kEQLabel, kEQLabel);

    c.font = "300 14px Arial";
    c.fillText("Source:", this.sourceMenu.l, this.sourceMenu.t-10);

    const knob_label_left = this.level.r + 20;
    c.save();

    if (!this.validSource) {
      c.globalAlpha = kDimmedAlpha;
    }

    c.textAlign = "center";
    c.font = "800 13px Arial";
    c.fillText("Level", this.level.cx, this.level.b + 20);
    c.fillText("Dim", this.dim.cx, this.dim.b+ 20);

    c.restore();

    c.fillText("Destinations:", this.destSelect.l + 3, this.destSelect.t - 10);

    drawButton(c, this.enableHitR, "TALK", this.talkbackEnabled, kToggleBlue);
  }
}

const MonitorState = Object.freeze({
  Off:0, On:1, Edited:2, Disabled: 3
});

function getChInputMonitorState(ch, faders, mutes, pans, stereo, tlkSrc) {
  // 0 = off, 1 = on, 2 = edited, 3 = disabled
  if (ch == tlkSrc)
    return MonitorState.Disabled;

  var roundedFader = Math.round(100*faders[ch])/100;
  if(faders[ch] == 0) {
    return MonitorState.Off;
  }
  if(!stereo && (mutes[ch] == 0) && (pans[ch] == 0.5) && roundedFader == 1) {
    return MonitorState.On;
  }
  return MonitorState.Edited;
}
function getMainsInputMonitorState(device, i, faders, stereos, mutes, pans, tlkSrc) {
  // this counts on index's for line in mixer channels (Mic1-line8) being contiguous
  const device_props = device.properties;
  const stereo = stereos[i + device_props.outputs.mains.iBus] !== 0;
  return getChInputMonitorState(device_props.get_mix_index(i, device_props.outputs.mains.iBus), faders, mutes, pans, stereo, device_props.get_mix_index(tlkSrc, device_props.outputs.mains.iBus));

}
function getPhones1InputMonitorState(device, i, faders, stereos, mutes, pans, tlkSrc) {
  // this counts on index's for line in mixer channels (Mic1-line8) being contiguous
  const device_props = device.properties;
  const stereo = stereos[i + device_props.outputs.mains.iBus] !== 0;
  return getChInputMonitorState(device_props.get_mix_index(i, device_props.outputs.phones.iBus), faders, mutes, pans, stereo, device_props.get_mix_index(tlkSrc, device_props.outputs.phones.iBus));
}
function getPhones2InputMonitorState(device, i, faders, stereos, mutes, pans, tlkSrc) {
  // this counts on index's for line in mixer channels (Mic1-line8) being contiguous
  const device_props = device.properties;
  const stereo = stereos[i + device_props.outputs.mains.iBus] !== 0;
  return getChInputMonitorState(device_props.get_mix_index(i, device_props.outputs.phones.iBus + 2), faders, mutes, pans, stereo, device_props.get_mix_index(tlkSrc, device_props.outputs.phones.iBus + 2));
}
function enableInputMonitorState(datastore, i, enable) {
  if(!enable)
    datastore.setV('kiMixFader', i, 0);
  else {
    datastore.setV('kiMixFader', i, 1);
    datastore.setV('kiMixStereo', i, 0);
    datastore.setV('kiMixMute', i, 0);
    datastore.setV('kiMixSolo', i, 0);
    datastore.setV('kiMixPan', i, 0.5);
  }
}
function drawIMButton(c, r, label, imState) {
  if(imState === MonitorState.Edited)
    drawButton(c, r, label, true, kEQLabel);
  else
    drawButton(c, r, label, imState === MonitorState.On, kToggleBlue, imState != MonitorState.Disabled);
}

class InputMonitoring extends View {
  constructor(l, t, appdata) {
    super(l, t, 605, 210);

    this.update = this.update.bind(this);
    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);
    this.hasMorePhones = false;

    this.setBG(kEmptyBG);
    this.wantsMouse = true;

    this.infoR = new Rect(this.r-35, this.t+20, 30, 30);
    this.xR = new Rect(this.r-40, this.t+30, 20, 20);
    this.showInfoMessage = false;

    const w1 = 49;
    const t1 = this.t + 64;
    const t2 = this.t + 119;
    const t3 = t2 + 55;
    const l1 = this.l + 195;

    this.mains = [];
    this.phones1 = [];
    this.phones2 = [];
    for (let i = 0; i < 10; i++) {
      const left = l1 + i * w1;
      this.mains.push({
        rect:new Rect(left, t1, 25, 25),
        state:MonitorState.Off
      });
      this.phones1.push({
        rect:new Rect(left, t2, 25, 25),
        state:MonitorState.Off
      });
      this.phones2.push({
        rect:new Rect(left, t3, 25, 25),
        state:MonitorState.Off
      });
    }

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.datastore.addListener('kiMixFader', this.update);
    this.datastore.addListener('kiMixStereo', this.update);
    this.datastore.addListener('kiMixMute', this.update);
    this.datastore.addListener('kiMixPan', this.update);
    this.datastore.addListener('kTalkbackSource', this.update);
    this.datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);
  }

  opticalExpanderChanged(k, v) {
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0;
    this.setEnabled(e);
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    var mh = super.getMouseHandler(xy);
    if(mh == this) {
      if (!this.showInfoMessage) {
        const buttons = [this.mains, this.phones1];
        if (this.hasMorePhones) {
          buttons.push(this.phones2);
        }
        const contained = buttons.some(function(button_set) {
          return button_set.some(function(button){
            return button.rect.contains(xy.x, xy.y);
          });
        });
        if (contained || this.infoR.contains(xy.x, xy.y))
          return this;
      }
      else if (this.xR.contains(xy.x, xy.y)) {
        return this;
      }
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) {
    if (! this.showInfoMessage) {
      const datastore = this.datastore;
      const device_props = this.device.properties;
      const num_buttons = device_props.analogIns.length;
      const mainItem = this.mains.findIndex(function(button){
        return button.rect.contains(xy.x, xy.y);
      });
      if (0 <= mainItem && mainItem < num_buttons && this.mains[mainItem].state != MonitorState.Disabled) {
        enableInputMonitorState(datastore, device_props.get_mix_index(mainItem, device_props.outputs.mains.iBus), this.mains[mainItem].state == 0);
      }
      const phones1Item = this.phones1.findIndex(function(button){
        return button.rect.contains(xy.x, xy.y);
      });
      if (0 <= phones1Item && phones1Item < num_buttons && this.phones1[phones1Item].state != MonitorState.Disabled) {
        enableInputMonitorState(datastore, device_props.get_mix_index(phones1Item, device_props.outputs.phones.iBus), this.phones1[phones1Item].state == 0);
      }
      const phones2Item = this.phones2.findIndex(function(button){
        return button.rect.contains(xy.x, xy.y);
      });
      if (0 <= phones2Item && phones2Item < num_buttons && this.phones2[phones2Item].state != MonitorState.Disabled) {
        enableInputMonitorState(datastore, device_props.get_mix_index(phones2Item, device_props.outputs.phones.iBus+2), this.phones2[phones2Item].state == 0);
      }
      if (this.infoR.contains(xy.x, xy.y)) {
        this.showInfoMessage = true;
        this.requestRender();
      }
    }
    else if (this.xR.contains(xy.x, xy.y)) {
      this.showInfoMessage = false;
      this.requestRender();
    }
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.infoR.offset(dx, dy);
    this.xR.offset(dx, dy);
  }
  uRBCXCY() {
    super.uRBCXCY();
    this.xR.setL(this.r-40);
  }
  update() {
    var faders = this.datastore.getKVArray('kiMixFader');
    var stereos = this.datastore.getKVArray('kiMixStereo');
    var mutes = this.datastore.getKVArray('kiMixMute');
    var pans = this.datastore.getKVArray('kiMixPan');

    let tlkSrc = this.datastore.getKVArrayItem('kTalkbackSource', 0);

    const device = this.device;
    this.mains.forEach(function(item, index) {
      item.state = getMainsInputMonitorState(device, index, faders, stereos, mutes, pans, tlkSrc);
    });
    this.phones1.forEach(function(item, index) {
      item.state = getPhones1InputMonitorState(device, index, faders, stereos, mutes, pans, tlkSrc);
    });
    this.phones2.forEach(function(item, index) {
      item.state = getPhones2InputMonitorState(device, index, faders, stereos, mutes, pans, tlkSrc);
    });
    this.requestRender();
  }
  recreate() {
    const phonesList = this.device.properties.phonesOut;
    this.hasMorePhones = phonesList.length > 1;
    const buttons = [this.mains, this.phones1, this.phones2];
    const top = this.t;
    const updateButtons = function(left) {
      const h_offset = left - buttons[0][0].rect.l;
      const v_offset = top + 64 - buttons[0][0].rect.t;
      buttons.forEach(function(button_set) {
        button_set.forEach(function(button){
          button.rect.offset(h_offset, v_offset);
        });
      });
    };
    if (this.hasMorePhones) {
      this.setWH(760, 282);
      updateButtons(this.l + 220);
    }
    else {
      this.setWH(760, 185);
      updateButtons(this.l + 195);
    }
    this.infoR.setL(this.r - 35);
    this.requestRender();
  }
  render(c) {
    c.save();
    const hasMorePhones = this.hasMorePhones;
    const mains = this.mains;
    const num_lines = this.device.properties.analogIns.length;
    for (let i = 0; i < num_lines; i++) {
      drawIMButton(c, mains[i].rect, i+1, mains[i].state);
      drawIMButton(c, this.phones1[i].rect, i+1, this.phones1[i].state);
      if (hasMorePhones) {
        drawIMButton(c, this.phones2[i].rect, i+1, this.phones2[i].state);
      }
    }
    c.font = "300 14px Arial";
    c.textAlign = 'left';
    c.strokeStyle = c.fillStyle = kEQLabel;
    const phones1_label = hasMorePhones ? " 1" : "";
    c.fillText("Main Out 1-2", this.l + 94, this.t +81);
    c.fillText("Phones" + phones1_label, this.l + 94, this.t + 137);
    if (hasMorePhones) {
      c.fillText("Phones 2", this.l + 94, this.t + 192);
    }

    const mic_center = function() {
      const left = mains[0].rect.l;
      return (mains[1].rect.r - left) / 2 + left;
    }();
    const lines_left = mains[2].rect.l;
    const lines_right = mains[num_lines - 1].rect.r;
    const lines_center = (lines_right - lines_left) / 2 + lines_left;
    c.textAlign = 'center';
    c.fillText("Mic/Line/Inst", mic_center, this.t + 50);
    c.fillText("Line In", lines_center, this.t + 50);

    drawLabelGroup(c, "Input Monitoring", this.l+2, this.t+2, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);

    c.lineWidth = 1;
    c.beginPath();
    c.moveTo(lines_left + 1, this.t + 44);
    c.lineTo(lines_center - 33, this.t + 44);
    c.moveTo(lines_center + 28, this.t + 44);
    c.lineTo(lines_right, this.t + 44);
    c.moveTo(this.l + 40, this.t + 104);
    c.lineTo(lines_right, this.t + 104);
    if (hasMorePhones) {
      c.moveTo(this.l + 40, this.t + 159);
      c.lineTo(lines_right, this.t + 159);
    }
    c.stroke();

    kSpeakerImg.render(c, this.l + 45, this.t + 48);
    kPhonesImg.render(c, this.l + 38, this.t + 112);
    if (hasMorePhones) {
      kPhonesImg.render(c, this.l + 38, this.t + 167);
    }
    
    c.restore();

    drawInfoIcon(c, this.infoR.cx, this.infoR.cy, kEQLabel);

    if (this.hasMorePhones)
    {
      const exp1 = "Click an input to hear it directly on the Main Mix or Phones Mix."
      const exp2 = "When doing so, turn off input monitoring in your host software."
      const text_base = this.b - (this.hasMorePhones ? 37 : 24);

      c.font = "600 11px Arial";
      c.textAlign = 'left';

      const alert_top = this.b -  68;
      const alert_left = this.l + 238;
      c.fillText(exp1, alert_left + 52, text_base - 17);
      c.fillText(exp2, alert_left + 52, text_base);

      kAlertImg.render(c, alert_left, alert_top);

    }
    else {
      const info_text = [
        "Click an input to hear it",
        "directly on the Main Mix",
        "or Phones Mix. When",
        "doing so, turn off input",
        "monitoring in your host",
        "software"
      ];

      const l = this.mains[num_lines - 1].rect.r + 35;
      const r = this.r - (kAlertImg.width + 3);
      const w = r - l;
      const cx = l + (w/2);

      c.font = "600 11px Arial";
      c.textAlign = 'left';
      let t = this.t + 80;
      for (const s of info_text) {
        c.fillText(s, l, t);
        t += 15;
      }

      kAlertImg.render(c, cx - kAlertImg.width/2, this.t + 25);
    }
    if(this.showInfoMessage) {
      roundRectPath(c, this.l+14, this.t+24, this.w-24, this.h-32, 6);
      c.fillStyle = 'black';
      c.fill()
      c.strokeStyle = kEQLabel;
      c.stroke()

      c.beginPath();
      c.moveTo(this.xR.l, this.xR.t);
      c.lineTo(this.xR.r, this.xR.b);
      c.moveTo(this.xR.r, this.xR.t);
      c.lineTo(this.xR.l, this.xR.b);
      c.stroke()

      const info1l1 = "Click an input to hear it on the Main Outs or Headphone Output.";
      const info1l2 = "For more control over the monitor mix, use the Main Out 1-2 Mix";
      const info1l3 = "and Phones Mix tabs."

      const info2l1 = "If you make adjustments to an input in its mix tab channel strip,"
      const info2l2 = "the button here will turn gray.";

      const getDevName = ()=>{
        const model_id = this.datastore.getKVArrayItem('kModelID', 0);
        if (model_id.search("828") != -1)
          return "828";
        return "UltraLite-mk5";
      };

      const info3l1 = "When using these input monitoring features in the " + getDevName() + ",";
      const info3l2 = "be sure to disable input monitoring in your host software.";
      const info3l3 = "Otherwise, the input signal will get doubled on the output.";

      // c.font = "12px Arial";
      c.font = "11px Arial";
      c.fillStyle = kEQLabel;
      const x = this.l + 152;
      let y = this.t + 30;
      c.fillText(info1l1, x, y += 13);
      c.fillText(info1l2, x, y += 13);
      c.fillText(info1l3, x, y += 13);

      c.fillText(info2l1, x, y += 24);
      c.fillText(info2l2, x, y += 13);

      c.fillText(info3l1, x, y += 24);
      c.fillText(info3l2, x, y += 13);
      c.fillText(info3l3, x, y += 13);

      drawIMButton(c, new Rect(this.l + 110, this.t+40, 25, 25), 1, 1);
      drawIMButton(c, new Rect(this.l + 110, this.t+84, 25, 25), 1, 2);
      kAlertImg.render(c, this.l + 102, this.t + 120);
    }
  }
}


class Home extends View{
  constructor(appdata) {
    super(0, 0, 0, 0);
    var xPos = kTabSelW;

    this.c = new CanvasRoot(appdata.animator);
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    this.scrollToTalkback = this.scrollToTalkback.bind(this);
    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);

    appdata.eventSys.addListener('wResized', this.resized);

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "homeContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.width = "calc(100% - " + xPos + "px)";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "home";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.tabDiv = this.scrollBarVH.containerDiv;

    //this.c.addSub(this.dnh = new DevNameHeader());
    this.m1 = new MicPreHome(this, 38, 350, 0, appdata);
    this.m2 = new MicPreHome(this, this.m1.r + 21, 350, 1, appdata);
    this.mv = new MainVolGroupHome(this, this.m2.r + 21, 350, appdata);
    this.pv1 = new PhonesVolHome(this, this.mv.r + 21, 350, 0, appdata);
    this.pv2 = new PhonesVolHome(this, this.pv1.r + 21, 350, 1, appdata);

    this.toprow = [this.m1, this.m2, this.mv, this.pv1, this.pv2];

    this.abmonitor = new ABMonitoring(this, this.m1.l, this.m1.b + 20, appdata.datastore, true);
    this.ab_default_width = this.abmonitor.w;

    this.talkback = new Talkback(this, this.abmonitor.r + 21, this.abmonitor.t, appdata);
    this.addSub(this.im = new InputMonitoring(this.m1.l, this.abmonitor.b + 20, appdata));
    this.setBG(kEmptyBG);

    this.appdata = appdata;

    appdata.datastore.addListener("kOpticalExpander", this.opticalExpanderChanged);
    this.opticalExpanderChanged();
  }

  opticalExpanderChanged(k, v) {
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0);
    if (e && this.optExpanderOverlay == undefined) {
      this.optExpanderOverlay = new OpticalExpanderOverlay(this.im.cx, this.talkback.t + 100, this.c.rootDiv, kEmptyBG, this.c.animator, this.appdata.datastore);
    }
    else if (!e && this.optExpanderOverlay) {
      this.optExpanderOverlay.dismiss();
      this.optExpanderOverlay = undefined;
    }
  }

  recreate() {
    const content_top = 73
    this.m1.setT(content_top);
    this.m2.setT(content_top);
    this.mv.setT(content_top);
    this.pv1.setT(content_top);
    this.pv2.setT(content_top);
    this.abmonitor.setT(this.m1.b + 20);
    this.talkback.setT(this.abmonitor.t);
    this.im.setT(this.abmonitor.b + 20);

    this.m1.recreate();
    this.m2.recreate();
    this.mv.recreate();
    this.pv1.recreate();
    this.pv2.recreate();
    this.talkback.recreate();
    this.talkback.setVis(this.device.properties.supportsTalkback);
    this.im.recreate();

    let topNumVis = 0;
    for (const v of this.toprow) {
      if (v.isVis)
        ++topNumVis;
    }
    const kPaddingX = 21;
    const topRowWidth = this.im.r - this.im.l;
    const w = (topRowWidth - (kPaddingX * (topNumVis - 1)))/ topNumVis;

    let l = this.toprow[0].l;
    for (const v of this.toprow) {
      v.setW(w);
      v.setL(l);
      l += w + kPaddingX;
    }

    if (Math.abs(w - this.ab_default_width) < 10)
      this.abmonitor.setW(w);
    else
      this.abmonitor.setW(this.ab_default_width);

    this.talkback.setL(this.abmonitor.r + kPaddingX);
    this.talkback.setW(this.l + this.im.r - this.talkback.l); // Align talkback right edge with everything else

    this.requestRender();
  }
  resized() {
    const w = 830;
    const h = this.im.b + 20;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
    this.requestRender();
  }
  render(c) {
    drawDeviceNameHeader(c, this.l + 40, this.t, 765, this.device);
  }
  scrollToTalkback() {
    let top = this.talkback.t;
    let left = this.talkback.l;
    this.c.div.scrollTop = top;
    this.c.div.scrollLeft = left;
  }
}
