'use strict';

// helper to get input and output names
const getLineInName = function(device, datastore, i) {
  const getDefault = function(i) {
    if (i < device.properties.inputs.mics.numCh[0]) {
      return "Mic/Line/Inst " + (i+1);
    }
    else {
      return "Line " + (i+1);
    }
  };

  let n = "Unknown";
  let left = i & ~1;
  if (datastore.getKVArrayItem('kiMixStereo', left)) {
    n = datastore.getKVArrayItem('kiChannelNames', left);
    if((n == undefined) || n == "")
      n = getDefault(i);
    else
      n += ((left == i) ? ' (L)' : ' (R)');
  }
  else {
    n = datastore.getKVArrayItem('kiChannelNames', i);
    if((n == undefined) || n == "")
      n = getDefault(i);
  }
  return n;
}

const getLineOutName = function(device, datastore, i) {
  const getDefault = function(i) {
    if (i < device.properties.inputs.mics.numCh[0]) {
      return "Main " + (i+1);
    }
    else {
      return "Line " + (i+1);
    }
  };

  let n = "Unknown";
  let left = i & ~1;
  if (datastore.getKVArrayItem('koMixStereo', left)) {
    n = datastore.getKVArrayItem('koChannelNames', left);
    if((n == undefined) || n == "")
      n = getDefault(left) + '-' + String(left + 2);
  }
  else {
    n = datastore.getKVArrayItem('koChannelNames', i);
    if((n == undefined) || n == "")
      n = getDefault(i);
  }
  return n + ' Mix';
};

class IOMeter extends View {
  constructor(l,t,w,h,index,meterStore) {
    super(l,t,w,h);
    this.onMeters = this.onMeters.bind(this);
    this.setBG(kBlackBG);
    this.lm = {m:0, ph:0};
    this.index = index;
    meterStore.addListener(this.onMeters);
  }
  render(c) {
    let l = this.l + 1;
    let w = this.w - 2;
    let sh = Math.floor(this.lm.m * this.h);
    let sy = this.h - sh;
    var l3 = Math.trunc(Calc0dBToCtrl24(-3) * this.h);
    var l12 = Math.trunc(Calc0dBToCtrl24(-12) * this.h);

    if(sh != 0) {
      if(sh > l3) {
        c.fillStyle = kMeterBlue;
        c.fillRect(l, this.t + sy, w, sh);
        c.fillStyle = kMeterLightBlue;
        c.fillRect(l, this.t + sy, w, sh-l12);
        c.fillStyle = kMeterOrange;
        c.fillRect(l, this.t + sy, w, sh-l3)
      }
      else if(sh > l12) {
        c.fillStyle = kMeterBlue;
        c.fillRect(l, this.t + sy, w, sh);
        c.fillStyle = kMeterLightBlue;
        c.fillRect(l, this.t + sy, w, sh-l12)
      }
      else {
        c.fillStyle = kMeterBlue;
        c.fillRect(l, this.t + sy, w, sh)
      }
    }

    let y = this.h - (this.lm.ph * this.h);
    if(y < (this.h - 2)) {
      c.fillStyle = "red";
      c.fillRect(l, this.t + y, w, 2);
    }
  }
  onMeters(v) {
    if((this.index >= 0) && (this.index < v.length)) {
      let m = Calc0dBToCtrl24(v[this.index].m);
      let ph = Calc0dBToCtrl24(v[this.index].ph);
      if((this.lm.m != m) || (this.lm.ph != ph)) {
        this.lm.m = m;
        this.lm.ph = ph;
        this.requestRender();
      }
    }
  }
}

const kPreTrimCtrlInfo = {k: 'kiGain', d:0, u:"dB", min:0, max:74, dft:0 };
const kPreInsertEnabledPreampCtrlInfo = {k: 'kiGain', d:0, u:"dB", min:0, max:54, dft:0};
const kPreInsertEnabledInsertTrimCtrlInfo = {k: 'kiGain', d:0, u:"dB", min:0, max: 20, dft:0};
const kLineInTrimCtrlInfo = {k: 'kiGain', d:0, u:"dB", min:0, max:20, dft:0 };
const kOutputTrimCtrlInfo = {k: 'koTrim', d:0, u:"dB", min:100, max:0, dft:0 };
const kPhonesTrimCtrlInfo = {k: 'koTrim', d:0, u:"dB", min:100, max:0, dft:24 };
const kMainVolCtrlInfo = {k: 'kMainTrim', d:0, u:"dB", min:100, max:0, dft:24 };
const kTalkbackLevelCtrlInfo = {k:'kTalkbackLevel', d:0, u:"dB", min:-48, max:12, dft:0, warp:true };
const kTalkbackDimCtrlInfo = {k: 'kTalkbackDim', d:0, u:"dB", min:-48, max:0, dft:0, warp:true };

class IOKnobTextEdit extends EditableTextView {
  constructor(l,t,w,h, appdata, ctrlInfo) {
    super(l,t,w,h, 'eq-input-detail', '12px sans-serif', appdata.keyFilter);
    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);

    this.ctrlInfo = ctrlInfo;
    this.setBG(kEmptyBG);
    this.c = kEQLabel;

    this.datastore = appdata.datastore;
    SetupDSArrayListenerBasic(this.datastore, this, ctrlInfo.k, 0);
  };
  setup(i) {
    SetDSArrayIndexBasic(this, i);
    let v = DSArrayGetBasic(this.datastore, this);
    v = this.ctrlInfo.warp ? GainToDB(v) : v;
    this.v = trimFloat(v, this.ctrlInfo.d) + " " + this.ctrlInfo.u;
    this.requestRender();
  }
  onDSArray(k, v) {
    v = this.ctrlInfo.warp ? GainToDB(v) : v;
    v = trimFloat(v, this.ctrlInfo.d);
    if((this.ctrlInfo.k == 'kMainTrim') || (this.ctrlInfo.k == 'koTrim')) {
      if (v == this.ctrlInfo.min)
        v = '-∞';
      else if (v != this.ctrlInfo.max)
        v = '-' + v;
    }
    else if ((this.ctrlInfo.k === 'kTalkbackLevel') || (this.ctrlInfo.k === 'kTalkbackDim')) {
      if (v <= this.ctrlInfo.min) {
        v = '-∞';
      }
    }
    this.v = v + ' ' + this.ctrlInfo.u;
    this.requestRender();
  }
  valueChanged(v) {
    v = v.replace(/[∞]|inf/g, this.ctrlInfo.min);
    v = v.replace(/[^0-9\.\-]/g, '');
    if ((v.length == 0) || isNaN(v))
      return this.v;
    v = validateCtrl(this.ctrlInfo, Number(v));
    DSArraySetBasic(this.datastore, this, this.ctrlInfo.warp ? DBToGain(v) : v);
    if((this.ctrlInfo.k == 'kMainTrim') || (this.ctrlInfo.k == 'koTrim')) {
      if (v == this.ctrlInfo.min)
        v = '-∞';
      else if (v != this.ctrlInfo.max)
        v = '-' + v;
    }
    return v + ' ' + this.ctrlInfo.u;
  }
}

class IOKnob extends SimpleDSKnob {
  constructor(l, t, cE, cD, appdata, ctrlInfo, iL, iR) {
    super(l, t, 68, 75, appdata.datastore, ctrlInfo.k, iL);

    this.setup = this.setup.bind(this);
    this.enable = this.enable.bind(this);

    this.cr = new Rect(l + 2, t, 64, 70);
    this.cE = cE;
    this.cD = cD;
    this.setBG(kEmptyBG);
    this.ctrlInfo = ctrlInfo;
    this.enabled = true;

    this.addSub(this.edit = new IOKnobTextEdit(this.l, this.t + (this.h - 18), this.w, 18, appdata, ctrlInfo));

    this.setup(iL, iR);
  }
  setup(left, right) {
    this.iR = right;
    super.setup(left);
    this.edit.setup(left);
  }
  offset(dx, dy) {
    this.cr.offset(dx,dy);
    super.offset(dx, dy);
  }
  enable(enable) {
    if(this.enabled != enable) {
      this.enabled = enable;
      this.requestRender();
    }
  }
  handleDblClick() {
    this.setVal(this.ctrlInfo.warp ? DBToGain(this.ctrlInfo.dft) : (dynVToDS(this.v, this.ctrlInfo) == this.ctrlInfo.min ? this.ctrlInfo.dft : this.ctrlInfo.min));
  }
  dsToV(dsv) {
    return dynDSToV(dsv, this.ctrlInfo);
  }
  vToDS(v) {
    return dynVToDS(v, this.ctrlInfo);
  }
  setVal(v) {
    // overridden so we can change both trims when the knob is stereo moved
    let i = this[this.ctrlInfo.k];
    if(i != undefined) {
      this.datastore.setV(this.ctrlInfo.k, i, v);
      if(this.iR)
        this.datastore.setV(this.ctrlInfo.k, this.iR, v);
    }
  }
  onDSArray(k, v, setByObject) {
    super.onDSArray(k,v, setByObject);
    
    // If the text edit widget changed the value, and this is a stereo-linked knob, set the right channel as well.
    if (setByObject != undefined && setByObject == this.edit && this.iR != undefined) {
      this.datastore.setV(this.ctrlInfo.k, this.iR, v, this);
    }
  }
  render(c) {
    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, 2.25 * Math.PI);
    c.lineWidth = 7;
    c.strokeStyle = kDynCtrlBG;
    c.stroke();

    let v = (0.75 + this.v * 1.5) * Math.PI;
    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, v);
    c.lineWidth = 7;
    c.strokeStyle = this.enabled ? this.cE : this.cD;
    c.stroke();
  }
}

class IOKnobWithTitle extends IOKnob {
  constructor(l, t, cE, cD, appdata, ctrlInfo, iL, iR, title_lines, titleStyles) {
    super(l, t, cE, cD, appdata, ctrlInfo, iL, iR);

    this.title_lines = title_lines != undefined ? title_lines : [""];
    this.titleStyles = titleStyles;
  }

  render(c) {
    c.save();
    super.render(c);
    c.restore();

    if (this.title_lines != undefined) {
      if (this.titleStyles != undefined) {
        for (const [k,v] of Object.entries(this.titleStyles)) {
          c[k] = v;
        }
      }

      const line_height = stringDimensions(kStrDimRefText, c, this.root).height;

      const line_padding_y = 3;
      const title_height = (line_height * this.title_lines.length) + (line_padding_y * (Math.max(0, this.title_lines.length - 1))); 

      c.textAlign = "center";

      let t = this.cy - (title_height / 2);

      for (let i = 0; i < this.title_lines.length; ++i) {
        t += ((line_height + line_padding_y) * i);
        c.fillText(this.title_lines[i], this.cx, t + (line_height / 2));
      }
    }

  }
}


class InPrePostFXMenu extends View {
  constructor(p, l, t, w, datastore) {
    super(l,t,w,40);

    this.change = this.change.bind(this);
    this.fpgaPatchChanged = this.fpgaPatchChanged.bind(this);
    this.loopbackFirstChanged = this.loopbackFirstChanged.bind(this);

    this.datastore = datastore;
    this.loopbackFirst = datastore.getKVArrayItem('kLoopbackFirst', 0);
    this.setBG(kEmptyBG);

    p.addSub(this);
    datastore.addListener('kFPGAPatch', this.fpgaPatchChanged);
    datastore.addListener('kLoopbackFirst', this.loopbackFirstChanged);
    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l, t+12, this.w, 28, this.change));
  }
  setup(info) {
    this.info = info;
    this.recreateMenu();
  }
  recreateMenu() {
    this.menu.clearOptions();
    if (this.info !== undefined) {
      this.menu.addOption(this.info.source, "Pre Fx");
      this.menu.addOption(this.info.postfx, "Post Fx");
      this.menu.sel.value = this.datastore.getKVArrayItem('kFPGAPatch', this.info.destination + (this.loopbackFirst * 2));
    }
  }
  fpgaPatchChanged(k, v) {
    if (this.info !== undefined) {
      let vl = v[this.info.destination + (this.loopbackFirst * 2)];
      if(this.menu.sel.value != vl) {
        this.menu.sel.value = vl;
        this.requestRender();
      }
    }
  }
  loopbackFirstChanged(k, v) {
    this.loopbackFirst = v[0];
  }
  change() {
    let vl = parseInt(this.menu.sel.value);
    this.datastore.setV('kFPGAPatch', this.info.destination + (this.loopbackFirst * 2), vl);
  }
  render(c) {
    c.fillStyle = kEQLabel;
    c.font = "300 14px Arial";
    c.textAlign = 'left';
    c.fillText("To USB host:", this.l+2, this.t+8);
  }
}

class MicPreGainKnob extends View {
  constructor(l, t, w, h, i, appdata) {
    super(l, t, w, h);
    this.addSub(this.gain = new IOKnob(l, t, kToggleBlue, kGateCtrlDisabled, appdata, kPreTrimCtrlInfo, -1));
  }
}

class MicPreGainKnobsWithReturns extends View {
  constructor(l, t, w, h, i, returns_igain_idx, appdata) {
    super(l,t,w,h);

    this.chIdx = i;
    const knob_title_styles = {
      fillStyle: kToggleBlue,
      font: '10px sans-serif'
    };

    this.addSub(this.gain = new IOKnobWithTitle(l, t, kToggleBlue, kGateCtrlDisabled, appdata, kPreInsertEnabledPreampCtrlInfo, i, appdata, ["Preamp", "Gain"], knob_title_styles));
    this.addSub(this.insert_trim = new IOKnobWithTitle(l, this.gain.b + 5, kToggleBlue, kGateCtrlDisabled, appdata, kPreInsertEnabledInsertTrimCtrlInfo, returns_igain_idx, appdata, ["Insert", "Gain"], knob_title_styles));
  }

  setup(chIdx, returns_idx) {
    this.chIdx = chIdx;
    this.gain.setup(chIdx);
    this.insert_trim.setup(returns_idx);
  }
}

class MicPre extends View {
  constructor(p, l, t, i, appdata) {
    super(l,t,180,230);

    this.setBG(kEmptyBG);
    p.addSub(this);
    this.onJack = this.onJack.bind(this);
    this.onReturnsDetect = this.onReturnsDetect.bind(this);
    this.chIdx = i;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.addSub(this.meter = new IOMeter(this.l + 12, this.t + 25, 10, 126, -1, appdata.meterStore));
    this.addSub(this.b48V = new ToggleBtn(l + 35, t + 27, 52, "48V", this.datastore, 'k48V', i, kToggleRed, kEmptyBG));
    this.addSub(this.pad = new ToggleBtn(l + 35, t + 73, 52, "Pad", this.datastore, 'kPad', i, kToggleBlue, kEmptyBG));
    this.addSub(new ToggleBtn(l + 35, t + 119, 52, "\u00D8", this.datastore, 'kiPhase', i, kToggleBlue, kEmptyBG));

    const device_props = this.device.properties;
    this.addSub(this.gain = new MicPreGainKnob(l + 100, t + 27, this.r - l + 100, this.b - t + 27, i, appdata));
    this.addSub(this.gainWithReturns = new MicPreGainKnobsWithReturns(l + 100, t + 15, this.r - l + 100, this.b - t + 27, i, i + device_props.returns_igain_start_idx, appdata));

    this.fxmenu = new InPrePostFXMenu(this, this.l+12, t+166, this.w-24, this.datastore);
    this.recreate();

    this.datastore.addListener('kJackDetect', this.onJack);
    this.datastore.addListener('kReturnDetect', this.onReturnsDetect);
  }
  recreate() {
    const device_props = this.device.properties;
    this.meter.index = device_props.analogIns[this.chIdx].source;
    this.gain.gain.setup(this.chIdx);
    this.gainWithReturns.setup(this.chIdx, this.chIdx + device_props.returns_igain_start_idx);
    this.fxmenu.setup(device_props.analogIns[this.chIdx]);
  }
  onJack(k, v) {
    if(k == 'kJackDetect') {
      // bool 0 = mic, 1 = line/inst
      this.b48V.setEnabled(v[this.chIdx] == 0);
      this.pad.setEnabled(v[this.chIdx] == 0);
    }
  }
  onReturnsDetect() {
    const returns_detected = this.datastore.getKVArrayItem('kReturnDetect', this.chIdx);
    this.gain.setVis(!returns_detected);
    this.gainWithReturns.setVis(returns_detected);

    this.requestRender();
  }
  render(c) {
    drawDoubleLabelGroup(c, [getLineInName(this.device, this.datastore, this.chIdx), this.chIdx + 1], this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}


class LineIn extends View {
  constructor(p, l, t, i, appdata) {
    super(l,t,113,230);
    p.addSub(this);
    this.chIdx = i;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.addSub(this.meter = new IOMeter(this.l + 12, this.t + 25, 10, 126, -1, appdata.meterStore));
    this.addSub(this.gain = new IOKnob(l + 33, t + 27, kToggleBlue, kGateCtrlDisabled, appdata, kLineInTrimCtrlInfo, -1));
    this.addSub(new ToggleBtn(l + 42, t + 119, 52, "\u00D8", this.datastore, 'kiPhase', i, kToggleBlue, kEmptyBG));
    this.fxmenu = new InPrePostFXMenu(this, this.l+12, t+166, this.w-24, this.datastore);
    this.recreate();
  }
  recreate() {
    const device_props = this.device.properties;
    if (this.chIdx < device_props.analogIns.length) {
      this.meter.index = device_props.analogIns[this.chIdx].source;
      this.gain.setup(this.chIdx);
      this.fxmenu.setup(device_props.analogIns[this.chIdx]);
    } else {
      this.meter.index = -1;
      this.gain.setup(-1);
      this.fxmenu.setup(undefined);
    }
  }
  render(c) {
    drawDoubleLabelGroup(c, [getLineInName(this.device, this.datastore, this.chIdx), this.chIdx + 1], this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}

function make_optical_bank_label(label) {
  if (label === undefined) {
    return "";
  }
  else {
    return " " + label;
  }
}


class OpticalIn extends View {
  constructor(parent, left, top, width, appdata) {
    super(left,top,width,120);
    this.setBG(kEmptyBG);
    this.opticalChanged = this.opticalChanged.bind(this);
    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);
    this.getNumOpticalBanks = this.getNumOpticalBanks.bind(this);

    parent.addSub(this);
    this.wantsMouse = true;

    this.opticalMode = OpticalModes.ADAT;
    this.meterStore = appdata.meterStore;
    this.device = appdata.device;
    this.datastore = appdata.datastore;

    this.meter_banks = [];
    let bank_left = this.l + 225;
    const bank_top = this.t + 25;

    for (let bank = 0; bank < 2; bank++){
      let meter_left = bank_left;

      const meters = [];
      for (let i = 0; i < 4; i++) {
        const left_meter = new IOMeter(meter_left, bank_top, 10, 80, -1, this.meterStore);
        this.addSub(left_meter);
        meters.push(left_meter);
        meter_left = left_meter.r + 2;

        const right_meter = new IOMeter(meter_left, bank_top, 10, 80, -1, this.meterStore);
        this.addSub(right_meter);
        meters.push(right_meter);
        meter_left = right_meter.r + 6;
      }
      this.meter_banks.push(meters);
      bank_left = bank_left + 142;
    }

    this.datastore.addListener('kOpticalMode', this.opticalChanged);
    this.datastore.addListener('kSampleRate', this.opticalChanged);
    this.datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);

    this.adatHitR = new Rect(left + 32, top + 58, 48, 26);
    this.tosHitR = new Rect(this.adatHitR.r + 17, top + 58, 48, 26);

    this.opticalExpanderChanged();
  }
  opticalExpanderChanged(k, v){
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0;
    this.setEnabled(e);
  }
  opticalChanged() {
    this.opticalMode = this.datastore.getKVArrayItem('kOpticalMode', 0)
    this.recreateMeters();
  }

  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    let mh = super.getMouseHandler(xy);
    if(mh == this) {
      if(this.adatHitR.contains(xy.x, xy.y) || this.tosHitR.contains(xy.x, xy.y))
        return this;
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) {
    const device_props = this.device.properties;
    if(this.adatHitR.contains(xy.x, xy.y)) {
      if(this.opticalMode != OpticalModes.ADAT)
        this.datastore.setV('kOpticalMode', 0, OpticalModes.ADAT);
    }
    else if(this.tosHitR.contains(xy.x, xy.y)) {
      this.datastore.setV('kOpticalMode', 0, OpticalModes.TOSlink);
    }
  }
  render(c) {
    drawLabelGroup(c, "Optical In", this.l+1, this.t+1, this.w, this.h-2, 6, kEQLabel, kEQLabel);

    const sr = this.datastore.getKVArrayItem('kSampleRate', 0);

    if (sr <= 96000) {
      c.font = "300 14px Arial";
      const bank_a_label = make_optical_bank_label(this.device.properties.adatALabel);
      c.fillText("Bank" + bank_a_label + " format:", this.adatHitR.l + 1, this.adatHitR.t - 12);

      drawButton(c, this.adatHitR, "ADAT", this.opticalMode == OpticalModes.ADAT, kToggleBlue);
      drawButton(c, this.tosHitR, "TOS", this.opticalMode == OpticalModes.TOSlink, kToggleBlue);
    }
    else {
      c.font = "300 14px Arial";
      c.textAlign = "center";
      c.fillText("Optical is disabled at 4x sample rates", this.cx, this.cy + 10);
    }
  }

  getNumOpticalBanks() {
    return this.device.properties.get_optical_info(this.opticalMode).length;
  }

  createMeters() {
    const sr = this.datastore.getKVArrayItem('kSampleRate', 0);
    const sr_index = getSampleRateIndex(sr);
    const opticalInfo = this.device.properties.get_optical_info(this.opticalMode);

    const bank_width = 142;
    const meter_width = 10;
    const pair_width = (2*meter_width) + 8;

    this.meter_banks.forEach((bank, bankIndex)=>{
      let num_meters = 0;
      let meterGetter = function(index){ return -1; };
      if (bankIndex < opticalInfo.length) {
        const info = opticalInfo[bankIndex];
        num_meters = info.numCh[sr_index];
        meterGetter = info.getMeter;
      }

      bank.forEach((meter, meterIndex)=>{
        meter.setVis(meterIndex < num_meters);
        meter.index = meterGetter(meterIndex);
        const meters_offset_x = opticalInfo.length == 2 ? 225 : 175;
        const bank_offset = this.l + meters_offset_x + (bankIndex * bank_width);
        const stereoIndex = Math.floor(meterIndex / 2);
        const is_left = (meterIndex % 2) == 0;
        const offset = is_left ? (stereoIndex * pair_width) : (stereoIndex * pair_width) + 12;
        meter.setL(bank_offset + offset);
      });
    });
  }

  recreateMeters() {
    this.createMeters();
    this.requestRender();
  }
}

class SPDIFIn extends View {
  constructor(p, l, t, m, appdata) {
    super(l,t,120,120);

    this.srChanged = this.srChanged.bind(this);

    p.addSub(this);
    this.addSub(this.meter1 = new IOMeter(l + 44, this.t + 25, 10, 80, m, appdata.meterStore));
    this.addSub(this.meter2 = new IOMeter(this.meter1.r + 2, this.t + 25, 10, 80, m+1, appdata.meterStore));

    this.datastore = appdata.datastore;
    this.datastore.addListener('kSampleRate', this.srChanged);

    this.uRBCXCY();
  }

  uRBCXCY() {
    super.uRBCXCY();
    const meter_width = this.meter2.r - this.meter1.l;

    this.meter1.setL(this.l + (this.w - meter_width) / 2.0);
    this.meter2.setL(this.meter1.r + 2);
  }

  setup(meterIndex) {
    this.meter1.index = meterIndex;
    this.meter2.index = meterIndex+1;
  }
  srChanged() {
    let isVis = this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000;

    for (const v of this.subViews.items) {
      v.setVis(isVis);
    }
  }
  render(c) {
    drawLabelGroup(c, "S/PDIF", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);

    let isVis = this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000;
    if (! isVis) {
      c.textAlign = "left";
      c.font = "12px Arial";

      const stringHeight = (str)=>{
        const measurement = c.measureText(str);
        return (measurement.actualBoundingBoxAscent + measurement.actualBoundingBoxDescent) / this.root.divScale;
      };
      const stringWidth = (str)=>{
        const measurement = c.measureText(str);
        return measurement.width / this.root.divScale;
      };

      let y = this.t + 50;
      c.fillText("S/PDIF is", this.l + 10, y);

      y += stringHeight("S/PDIF is") + 5;
      c.fillText("disabled at", this.l + 10, y);

      y += stringHeight("disabled at") + 5;
      c.fillText("4x sample", this.l + 10, y);

      y += stringHeight("4x sample") + 3;
      c.fillText("rates", this.l + 10, y);
    }

  }
}

class FootSwitchView extends View {
  constructor(parent, left, top, appdata) {
    super(left, top, 762, 120);

    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);

    parent.addSub(this);
    this.wantsMouse = true;
    this.setBG(kEmptyBG);

    this.datastore = appdata.datastore;
    this.switchEnabled = false;

    this.enableHitR = new Rect(left + 23, top + 54, 58, 26);

    const menu_top = top + 60;
    this.addSub(this.downEntry = new KeyEntryView(this.enableHitR.r + 282, menu_top, 145, 28, this.downKeyCallback.bind(this)));
    this.addSub(this.upEntry = new KeyEntryView(this.downEntry.r + 57, menu_top, 145, 28, this.upKeyCallback.bind(this)));

    this.datastore.addListener('kFootSwitchEnabled', this.enabledChanged.bind(this));
    this.datastore.addListener('kFootSwitchDownKey', this.downCommandChanged.bind(this));
    this.datastore.addListener('kFootSwitchUpKey', this.upCommandChanged.bind(this));
    this.datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);

    this.opticalExpanderChanged();
  }
  opticalExpanderChanged(k, v){
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0;
    this.setEnabled(e);
  }
  downCommandChanged(k, v) {
    this.downEntry.setCommand(this.datastore.getKVArrayItem('kFootSwitchDownKey', 0));
  }

  downKeyCallback(cmd) {
    this.datastore.setV('kFootSwitchDownKey', 0, cmd);
  }

  upCommandChanged(cmd) {
    this.upEntry.setCommand(this.datastore.getKVArrayItem('kFootSwitchUpKey', 0));
  }

  upKeyCallback(cmd) {
    this.datastore.setV('kFootSwitchUpKey', 0, cmd);
  }

  enabledChanged(k, v) {
    this.switchEnabled = this.datastore.getKVArrayItem('kFootSwitchEnabled', 0) !== 0;
    this.requestRender(this.enableHitR);
  }

  getMouseHandler(xy) {
    const mh = super.getMouseHandler(xy);
    if (mh === this && !this.enableHitR.contains(xy.x, xy.y)) {
      return undefined;
    }
    return mh;
  }

  handleMouseDown(xy) {
    if (this.enableHitR.contains(xy.x, xy.y)) {
      this.datastore.setV('kFootSwitchEnabled', 0, this.switchEnabled ? 0 : 1);
    }
  }

  render(c) {
    drawLabelGroup(c, "Foot Switch", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);

    kAlertImg.render(c, this.l + 96, this.t + 48);

    const info_message = [
      "Enabling the foot switch will cause",
      "the USB connection to disconnect",
      "briefly. Your PC will connect the",
      'foot switch as a "USB keyboard"',
      "(so it can trigger keystrokes)."
    ];

    let message_top = this.t + 37;
    const message_left = this.l + 145;
    c.font = "600 10px Arial";
    c.textAlign = 'left';
    c.strokeStyle = c.fillStyle = kEQLabel;
    const that = this;
    info_message.forEach(function(msg){
      c.fillText(msg, message_left, message_top);
      message_top += 15;
    });

    c.font = "300 14px Arial";
    c.fillText("Down keystroke:", this.downEntry.l + 3, this.downEntry.t - 10);
    c.fillText("Up keystroke:", this.upEntry.l + 3, this.upEntry.t - 10);

    drawButton(c, this.enableHitR, "Enable", this.switchEnabled, kToggleBlue);
  }
};


class Inputs extends View {
  constructor(root, l, t, appdata) {
    super(l, t, 0, 0);

    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);

    root.addSub(this);

    this.setBG(kEmptyBG);
    let left = l + 38;
    const m1 = new MicPre(this, left, t + 44, 0, appdata);
    const m2 = new MicPre(this, left, 10 + m1.b, 1, appdata);
    this.ins = [m1, m2];

    left = m1.r + 20;
    this.device = appdata.device;
    const device_props = appdata.device.properties;
    for (let i=1; i < 5; i++) {
      const base_index = i * 2;
      const i1 = new LineIn(this, left, t + 44, base_index, appdata);
      const i2 = new LineIn(this, left, i1.b + 10, base_index+1, appdata);
      left = i1.r + 20;
      this.ins.push(i1, i2);
    }
    const gain_right = left - 10;
    this.gain_width = gain_right - l;
    this.gain_height = m2.b - m1.t + 50;

    this.spd = new SPDIFIn(this, l + 20, m2.b + 35, device_props.spdifIn, appdata);

    const optical_left = 20 + this.spd.r;
    this.optical = new OpticalIn(this, optical_left, m2.b + 35, gain_right - optical_left + 20, appdata);

    this.footswitch = new FootSwitchView(this, l + 20, this.optical.b + 25, appdata);

    appdata.datastore.addListener("kOpticalExpander", this.opticalExpanderChanged);

    this.datastore = appdata.datastore;

    this.recreate();
    this.opticalExpanderChanged();
  }
  displayWidth() {
    return this.gain_width + this.l - 20;
  }
  recreate() {
    const device_props = this.device.properties;
    const num_ins = device_props.analogIns.length;
    this.ins.forEach(function(input, index){
      input.setVis(index < num_ins);
      input.recreate();
    });
    const gain_right = this.ins[num_ins-1].r + 10;
    this.gain_width = gain_right - this.l;

    const opticalWidth = this.optical.getNumOpticalBanks() == 2 ?  gain_right - this.optical.l + 20 : (gain_right - 170) - this.optical.l + 20;
    this.optical.setW(opticalWidth);

    if (this.optical.getNumOpticalBanks() == 2) {
      this.setW(800);
      this.setH(815);
    }
    else {
      this.setW(640);
      this.setH(655);
    }

    this.spd.setup(device_props.spdifIn);
    this.footswitch.setVis(device_props.supportsFootswitch);

    this.requestRender();
  }
  opticalExpanderChanged(k, v) {
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0);
    if (e && this.optExpanderOverlay == undefined) {
      this.optExpanderOverlay = new OpticalExpanderOverlay(this.cx, this.optical.b - 25, this.root.rootDiv, kEmptyBG, this.root.animator, this.datastore);
    }
    else if (!e && this.optExpanderOverlay) {
      this.optExpanderOverlay.dismiss();
      this.optExpanderOverlay = undefined;
    }
  }
  render(c) {
    drawLabelGroup(c, "Input Gain", this.l+20, this.t+14, this.gain_width, this.gain_height, 6, kEQLabel, kEQLabel);
  }
}

function get_optical_index_getter(device, patch) {
  const device_props = device.properties;
  const start_index = device.properties.analogIns.findIndex(function(input){ return input.source === patch; });
  if (start_index !== -1) {
    return function(index){ // inputs are not contiguous
      if ((start_index + index) >= device_props.analogIns.length) {
        return device_props.zeroPatch;
      }
      else {
        return device_props.analogIns[start_index + index].source;
      }
    };
  }
  else {
    return function(index){
      return patch + index;
    };
  }
}

// Handle mouse clicks.
class MouseView extends View {
  constructor(mousee, l, t, w, h) {
    super(l,t,w,h);
    this.mousee = mousee;
    this.wantsMouse = true;
  }
  handleMouseDown(xy) {
    this.mousee.handleMouseDown(xy);
  }
}
class OpticalOut extends View {
  constructor(p, l, t, appdata) {
    super(l,t,686,120);

    this.changeBankA = this.changeBankA.bind(this);
    this.changeBankB = this.changeBankB.bind(this);
    this.onDS = this.onDS.bind(this);
    this.rebuildUI = this.rebuildUI.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    this.setPatches = this.setPatches.bind(this);

    this.opticalMode = OpticalModes.ADAT;

    const device_props = appdata.device.properties;

    p.addSub(this);
    this.setBG(kEmptyBG);

    const aligned_top = t + 53;

    // The optical menu won't get mouse clicks unless I do this
    this.addSub(this.mouser = new MouseView(this, l + 23, aligned_top, 109, 30));
    this.adatHitR = new Rect(l + 20, aligned_top, 49, 30);
    this.tosHitR = new Rect(this.adatHitR.r + 8, aligned_top, 49, 30);

    const menu_a = new SelectMenu(p.root.getRootDiv(), this.tosHitR.r + 21, aligned_top, 141, 30, this.changeBankA);
    this.addSub(menu_a);
    this.menus = [menu_a];

    this.meter_banks = [];

    let left = menu_a.r + 23;
    const meter_top = this.t + 25;
    for (let bank = 0; bank < 2; bank++)
    {
      let meter_set = [];
      for (let pair = 0; pair < 4; pair++) {
        for (let i = 0; i < 2; i++) {
          const meter = new IOMeter(left, meter_top, 8, 80, device_props.zeroPatch, appdata.meterStore);
          left = meter.r + 2;
          meter_set.push(meter);
          this.addSub(meter);
        }
        left = left + 3;
      }
      this.meter_banks.push(meter_set);
      left = left + 9;
    }

    const menu_b = new SelectMenu(p.root.getRootDiv(), left + 8 , aligned_top, 141, 30, this.changeBankB);
    this.addSub(menu_b);
    this.menus.push(menu_b);

    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.datastore.addListener('kOpticalMode', this.onDS);
    this.datastore.addListener('kFPGAPatch', this.onDS);
    this.datastore.addListener('kSampleRate', this.onDS);
    this.datastore.addListener('koChannelNames', this.labelsChanged);
    this.datastore.addListener('koMixStereo', this.labelsChanged);

    this.rebuildUI();
  }
  setEnabled(e) {
    super.setEnabled(e);
    for (const m of this.menus) {
      m.setEnabled(e);
    }
  }
  offset(dx, dy) {
    super.offset(dx, dy);

    if (this.adatHitR != undefined) {
      this.adatHitR.offset(dx, dy);
    }
    if (this.tosHitR != undefined) {
      this.tosHitR.offset(dx, dy);
    }
  }
  setPatches(info, patch, mode) {
    // set the datastore patches given the first patch
    const get_index = get_optical_index_getter(this.device, patch);
    const count = mode === OpticalModes.ADAT ? 8 : 2;

    for(let i = 0; i < count; ++i) {
      this.datastore.setV('kFPGAPatch', info.getPatch(i), get_index(i));
    }
  }
  rebuildUI() {
    const sr_index = getSampleRateIndex(this.datastore.getKVArrayItem('kSampleRate', 0));
    this.updateLabels(sr_index);
    const that = this;

    let isVis = this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000;
    const banks = this.device.properties.get_optical_info(this.opticalMode);
    this.meter_banks.forEach(function(bank, bank_index){
      // setup the meters and menus
      if (bank_index < banks.length){
        const info = banks[bank_index];

        const menu = that.menus[bank_index];
        menu.clearOptions();
        const menu_items = that.menu_items[bank_index];
        menu_items.forEach(function(item){
          menu.addOption(item.patch, item.label);
        });
        menu.setVis(isVis);
        let v = that.datastore.getKVArrayItem('kFPGAPatch', info.getPatch(0));
        menu.sel.value = v;

        const num_meters = info.numCh[sr_index];

        const v_test = menu_items.find(function(item){
          return item.patch === v;
        });
        if (v_test === undefined) {
          v = menu_items[0].patch;
        }
        const get_meter_index = get_optical_index_getter(that.device, v);

        bank.forEach(function(meter, index){
          const visible = index < num_meters;
          meter.setVis(visible && isVis);
          if (visible) {
            meter.index = get_meter_index(index);
          }
        });

      } else {
        that.menus[bank_index].setVis(false);
        bank.forEach(function(meter){
          meter.setVis(false);
        });
      }
    });

    this.requestRender();
  }
  updateLabels(sr_index) {
    const device_props = this.device.properties;
    const banks = device_props.get_optical_info(this.opticalMode);
    const that = this;
    const makeLabels = function(mode, bank_label, usb_source, num_opticals) {
      const optical_label = make_optical_bank_label(bank_label);
      let labels = [
        { // first item is adat from host device_props.usbADATAOut
          patch: usb_source,
          label: "USB Out Optical" + optical_label + " (Default)"
        }
      ];
      if (mode === OpticalModes.ADAT) {
        const make_analog_range = function(start){
          const real_start = start-1;
          const end = Math.min(real_start + num_opticals, device_props.analogIns.length);
          return start + "-" + end;
        };

        labels.push(
          {
            patch: device_props.analogIns[0].source,
            label: "Analog Inputs " + make_analog_range(1) + " (Optical Expansion)"
          }
        );

        for (let i = 2; i < device_props.analogIns.length; i += 2) {
          labels.push({
            patch: device_props.analogIns[i].source,
            label: "Analog Inputs " + make_analog_range(i+1)
          });
        }
      }
      else {
        labels.push(
          {
            patch: device_props.analogOuts[0].dsp,
            label: getMainOutputPairName(that.device, that.datastore)
          },
          {
            patch: device_props.phonesOut[0].dsp,
            label: getPhones1OutputPairName(that.device, that.datastore)
          },
          {
            patch: device_props.analogOuts[2].dsp,
            label: getLine3OutputPairName(that.device, that.datastore)
          },
          {
            patch: device_props.analogOuts[4].dsp,
            label: getLine5OutputPairName(that.device, that.datastore)
          },
          {
            patch: device_props.analogOuts[6].dsp,
            label: getLine7OutputPairName(that.device, that.datastore)
          },
          {
            patch: device_props.analogOuts[8].dsp,
            label: getLine9OutputPairName(that.device, that.datastore)
          },
          {
            patch: device_props.dspReverbDry,
            label: "Reverb Dry Mix (L-R)"
          }
        );

        for (let i = 0; i < device_props.analogIns.length; i += 2) {
          labels.push({
            patch: device_props.analogIns[i].source,
            label: "Analog Inputs " + (i+1) + "-" + (i+2)
          });
        }
      }
      return labels;
    };

    this.menu_items = [makeLabels(this.opticalMode, banks[0].label, device_props.usbADATAOut, banks[0].numCh[sr_index])];
    for (let i = 1; i < banks.length; i++){
      this.menu_items.push(makeLabels(OpticalModes.ADAT, banks[i].label, device_props.usbADATBOut, banks[i].numCh[sr_index]));
    }
  }
  labelsChanged() {
    if (this.opticalMode == OpticalModes.TOSlink) {
      this.updateLabels(getSampleRateIndex(this.datastore.getKVArrayItem('kSampleRate', 0)));
      this.menus[0].renameOptions(this.menu_items[0].map(function(item){
        return item.label;
      }));
    }
  }
  onDS() {
    // datastore patches changed or
    // datastore optical mode changed, rebuild the menu and meters
    let isVis = this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000;
    this.menus.forEach(function(menu){
      menu.setVis(isVis);
      menu.sel.style.display = isVis ? 'block' : 'none';
    });
    this.opticalMode = this.datastore.getKVArrayItem('kOpticalMode', 1);
    const banks = this.device.properties.get_optical_info(this.opticalMode);
    this.setPatches(banks[0], this.datastore.getKVArrayItem('kFPGAPatch', banks[0].getPatch(0)), this.opticalMode);
    for (let i = 1; i < banks.length; i++) {
      this.setPatches(banks[i], this.datastore.getKVArrayItem('kFPGAPatch', banks[i].getPatch(0)), OpticalModes.ADAT);
    }
    for (const v of this.subViews.items) {
      v.setVis(isVis);
    }
    this.rebuildUI();
  }
  changeBankA() {
    // menu changed
    const bank = this.device.properties.get_optical_info(this.opticalMode)[0];
    this.setPatches(bank, parseInt(this.menus[0].sel.value), this.opticalMode);
  }
  changeBankB() {
    // menu changed
    const bank = this.device.properties.get_optical_info(this.opticalMode)[1];
    this.setPatches(bank, parseInt(this.menus[1].sel.value), OpticalModes.ADAT);
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    let mh = super.getMouseHandler(xy);
    if(mh == this) {
      if(this.adatHitR.contains(xy.x, xy.y) || this.tosHitR.contains(xy.x, xy.y))
        return this;
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) {
    if(this.opticalMode != OpticalModes.ADAT) {
      if(this.adatHitR.contains(xy.x, xy.y))
        this.datastore.setV('kOpticalMode', 1, OpticalModes.ADAT);
    } else {
      if(this.tosHitR.contains(xy.x, xy.y))
        this.datastore.setV('kOpticalMode', 1, OpticalModes.TOSlink);
    }
  }
  render(c) {
    drawLabelGroup(c, "Optical Out", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);

    let isVis = this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000;
    if (isVis) {
      const opticals = this.device.properties.get_optical_info(this.opticalMode);
      c.font = "300 14px Arial";
      const bank_a_label = make_optical_bank_label(opticals[0].label);
      c.fillText("Bank" + bank_a_label + " Source:", this.menus[0].l, this.menus[0].t-10);
      if (this.menus[1].isVis) {
        c.fillText("Bank" + make_optical_bank_label(opticals[1].label) + " Source:", this.menus[1].l, this.menus[1].t-10);
      }
      c.fillText("Bank" + bank_a_label + " format:", this.adatHitR.l + 2, this.adatHitR.t - 10);
      drawButton(c, this.adatHitR, "ADAT", this.opticalMode == OpticalModes.ADAT, kToggleBlue);
      drawButton(c, this.tosHitR, "TOS", this.opticalMode == OpticalModes.TOSlink, kToggleBlue);
    }
    else {
      c.textAlign = "center";
      c.fillText("Optical is disabled at 4x sample rates", this.cx, this.cy + 10);
    }
  }
}

class SPDIFOut extends View {
  constructor(p, l, t, appdata) {
    super(l,t,334,120);

    this.defualtWidth = 334;
    this.defaultPadding = 37;
    this.padding = this.defaultPadding;
    this.change = this.change.bind(this);
    this.fpgaPatchChanged = this.fpgaPatchChanged.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    this.srChanged = this.srChanged.bind(this);
    this.repositionElements = this.repositionElements.bind(this);

    p.addSub(this);
    this.addSub(this.meter1 = new IOMeter(l + this.padding, this.t + 25, 10, 80, -1, appdata.meterStore));
    this.addSub(this.meter2 = new IOMeter(this.meter1.r + 2, this.t + 25, 10, 80, -1, appdata.meterStore));

    this.device = appdata.device;
    this.datastore = appdata.datastore;

    this.datastore.addListener('kFPGAPatch', this.fpgaPatchChanged);
    this.datastore.addListener('koChannelNames', this.labelsChanged);
    this.datastore.addListener('koMixStereo', this.labelsChanged);
    this.datastore.addListener('kOpticalExpander', this.fpgaPatchChanged);
    this.datastore.addListener('kOpticalExpanderPatch', this.fpgaPatchChanged);

    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.meter2.r + (this.padding * 1.24), this.t + 60, 160, 30, this.change));
    this.recreate();
    this.datastore.addListener('kSampleRate', this.srChanged);

    this.srChanged();
  }
  repositionElements() {
    this.meter1.setL(this.l + this.padding);
    this.meter2.setL(this.meter1.r + 2);
    this.menu.setL(this.meter2.r + (this.padding * 1.24));
  }
  setEnabled(e) {
    super.setEnabled(e);
    this.menu.setEnabled(e);
  }
  recreate() {
    const device_props = this.device.properties;
    const index = device_props.spdifOut;
    this.updateLabels();
    this.menu.clearOptions();
    const that = this;
    this.options.forEach(function(option){
      that.menu.addOption(option.patch, option.label);
    });
    this.menu.sel.value = this.datastore.getKVArrayItem('kFPGAPatch', index);

    this.meter1.index = index;
    this.meter2.index = index + 1;
  }
  updateLabels() {
    const device_props = this.device.properties;
    this.options = [
      {patch:device_props.usbSPDIFOut, label:"USB Out SPDIF 1-2"},
      {patch:device_props.analogOuts[0].dsp, label:getMainOutputPairName(this.device, this.datastore)},
      {patch:device_props.phonesOut[0].dsp, label:getPhones1OutputPairName(this.device, this.datastore)},
      {patch:device_props.analogOuts[2].dsp, label:getLine3OutputPairName(this.device, this.datastore)},
      {patch:device_props.analogOuts[4].dsp, label:getLine5OutputPairName(this.device, this.datastore)},
      {patch:device_props.analogOuts[6].dsp, label:getLine7OutputPairName(this.device, this.datastore)},
      {patch:device_props.analogOuts[8].dsp, label:getLine9OutputPairName(this.device, this.datastore)},
      {patch:device_props.dspReverbDry, label:"Reverb Dry Mix (L-R)"}
    ];
  }
  fpgaPatchChanged() {
    const device_props = this.device.properties;

    const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 

    // Reference the main patch table for the menu selection state.
    let vl = this.datastore.getKVArrayItem('kFPGAPatch', device_props.spdifOut);
    let vr = this.datastore.getKVArrayItem('kFPGAPatch', device_props.spdifOut+1);
    if(this.menu.sel.value != vl) {
      this.menu.sel.value = vl;
    }

    // Reference the active patch table for meters.
    vl = this.datastore.getKVArrayItem(patchTableKey, device_props.spdifOut);
    vr = this.datastore.getKVArrayItem(patchTableKey, device_props.spdifOut+1);

    if (this.meter1.index != vl || this.meter2.index != vr) {
      this.meter1.index = vl;
      this.meter2.index = vr;
      this.requestRender();
    }
  }
  labelsChanged(k, v) {
    this.updateLabels();
    const labels = this.options.map(function(option){ return option.label; });
    this.menu.renameOptions(labels);
  }
  srChanged() {
    let isVis = this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000;
    for (const v of this.subViews.items) {
      v.setVis(isVis);
    }
    this.menu.sel.style.display = isVis ? 'block' : 'none';
  }
  change() {
    let vl = parseInt(this.menu.sel.value);
    let vr =  vl + 1;
    const device_props = this.device.properties;
    this.datastore.setV('kFPGAPatch', device_props.spdifOut, vl);
    this.datastore.setV('kFPGAPatch', device_props.spdifOut+1, vr);
  }
  render(c) {
    drawLabelGroup(c, "S/PDIF Out", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";

    if (this.datastore.getKVArrayItem('kSampleRate', 0) <= 96000) {
      c.fillText("Source:", this.menu.l, this.menu.t-6);
    }
    else {
      c.textAlign = "center";
      c.fillText("S/PDIF is disabled at 4x sample rates", this.cx, this.cy + 10);
    }
  }
}

class LineOut extends View {
  constructor(p, l, t, i, appdata) {
    super(l,t,113,120);
    this.recreate = this.recreate.bind(this);

    p.addSub(this);
    this.chIdx = i;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    const device_props = appdata.device.properties;
    this.addSub(this.meter = new IOMeter(this.l + 12, this.t + 25, 10, 80, device_props.zeroPatch, appdata.meterStore));
    this.addSub(this.trim = new IOKnob(l + 35, t + 27, kToggleBlue, kGateCtrlDisabled, appdata, kOutputTrimCtrlInfo, device_props.analogOuts[i].dac));
    this.recreate();
    this.datastore.addListener('kOpticalExpander', this.recreate);
    this.datastore.addListener('kOpticalExpanderPatch', this.recreate);
  }
  recreate() {
    const device_props = this.device.properties;
    const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 

    this.meter.index = this.datastore.getKVArrayItem(patchTableKey, device_props.analogOuts[this.chIdx].destination);
    this.trim.setup(device_props.analogOuts[this.chIdx].dac);
  }
  render(c) {
    drawDoubleLabelGroup(c, [getLineOutName(this.device, this.datastore, this.chIdx), this.chIdx + 1], this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}

function make_output_patch_menu(device, datastore, sampleRate, includeOpticalIn = false) {
  const is4x = (sampleRate > 96000);
  const is2x4x = (sampleRate > 48000);
  const unavailable4x = is4x ? " (Unavailable)" : "";
  const unavailable2x4x = is2x4x ? " (Unavailable)" : "";
  const device_props = device.properties;
  let options = [
    {patch:device_props.usbMainOut, label:"USB Out Main 1-2", disabled:false },
  ];

  device_props.phonesOut.forEach(function(info, index){
    const base = index * 2;
    options.push({patch:device_props.usbPhonesOut+base, label:"USB Out Phones " + (base+1) + "-" + (base+2) + unavailable4x, disabled:is4x });
  });

  options.push(
    {patch:device_props.usbMainOut+2, label:"USB Out Line 3-4", disabled:false },
    {patch:device_props.usbMainOut+4, label:"USB Out Line 5-6", disabled:false },
    {patch:device_props.usbMainOut+6, label:"USB Out Line 7-8", disabled:false },
    {patch:device_props.usbMainOut+8, label:"USB Out Line 9-10", disabled:false },
    {patch:device_props.usbSPDIFOut, label:"USB Out SPDIF 1-2" + unavailable4x, disabled:is4x}
  );

  if (device.properties.numOpticalBanks == 2) {
    options.push(
      {patch:device_props.usbADATAOut, label:"USB Out Optical A 1-2" + unavailable4x, disabled:is4x},
      {patch:device_props.usbADATAOut+2, label:"USB Out Optical A 3-4" + unavailable4x, disabled:is4x},
      {patch:device_props.usbADATAOut+4, label:"USB Out Optical A 5-6" + unavailable2x4x, disabled:is2x4x},
      {patch:device_props.usbADATAOut+6, label:"USB Out Optical A 7-8" + unavailable2x4x, disabled:is2x4x},
      {patch:device_props.usbADATBOut, label:"USB Out Optical B 1-2" + unavailable4x, disabled:is4x},
      {patch:device_props.usbADATBOut+2, label:"USB Out Optical B 3-4" + unavailable4x, disabled:is4x},
      {patch:device_props.usbADATBOut+4, label:"USB Out Optical B 5-6" + unavailable2x4x, disabled:is2x4x},
      {patch:device_props.usbADATBOut+6, label:"USB Out Optical B 7-8" + unavailable2x4x, disabled:is2x4x}
    );

    if (includeOpticalIn) {
      options.push(
        {patch:device_props.PatchBase.digital, label:"Optical In A 1-2" + unavailable4x, disabled:is4x},
        {patch:device_props.PatchBase.digital+2, label:"Optical In A 3-4" + unavailable4x, disabled:is4x},
        {patch:device_props.PatchBase.digital+4, label:"Optical In A 5-6" + unavailable2x4x, disabled:is2x4x},
        {patch:device_props.PatchBase.digital+6, label:"Optical In A 7-8" + unavailable2x4x, disabled:is2x4x},
        {patch:device_props.PatchBase.digital+8, label:"Optical In B 1-2" + unavailable4x, disabled:is4x},
        {patch:device_props.PatchBase.digital+10, label:"Optical In B 3-4" + unavailable4x, disabled:is4x},
        {patch:device_props.PatchBase.digital+12, label:"Optical In B 5-6" + unavailable2x4x, disabled:is2x4x},
        {patch:device_props.PatchBase.digital+14, label:"Optical In B 7-8" + unavailable2x4x, disabled:is2x4x}
      );
    }
  }
  else {
    options.push(
      {patch:device_props.usbADATAOut, label:"USB Out Optical 1-2" + unavailable4x, disabled:is4x},
      {patch:device_props.usbADATAOut+2, label:"USB Out Optical 3-4" + unavailable4x, disabled:is4x},
      {patch:device_props.usbADATAOut+4, label:"USB Out Optical 5-6" + unavailable2x4x, disabled:is2x4x},
      {patch:device_props.usbADATAOut+6, label:"USB Out Optical 7-8" + unavailable2x4x, disabled:is2x4x}
    );

    if (includeOpticalIn) {
      options.push(
        {patch:device_props.PatchBase.digital, label:"Optical In 1-2" + unavailable4x, disabled:is4x},
        {patch:device_props.PatchBase.digital+2, label:"Optical In 3-4" + unavailable4x, disabled:is4x},
        {patch:device_props.PatchBase.digital+4, label:"Optical In 5-6" + unavailable2x4x, disabled:is2x4x},
        {patch:device_props.PatchBase.digital+6, label:"Optical In 7-8" + unavailable2x4x, disabled:is2x4x}
      );
    }
  }

  options.push({patch:device_props.analogOuts[0].dsp, label:getMainOutputPairName(device, datastore), disabled:false});

  device_props.phonesOut.forEach(function(info, index){
    const base = index * 2;
    options.push({patch:info.dsp, label:getPhonesOutputPairName(device, datastore, index), disabled:false })
  });

  options.push(
    {patch:device_props.analogOuts[2].dsp, label:getLine3OutputPairName(device, datastore), disabled:false},
    {patch:device_props.analogOuts[4].dsp, label:getLine5OutputPairName(device, datastore), disabled:false},
    {patch:device_props.analogOuts[6].dsp, label:getLine7OutputPairName(device, datastore), disabled:false},
    {patch:device_props.analogOuts[8].dsp, label:getLine9OutputPairName(device, datastore), disabled:false},
    {patch:device_props.dspReverbDry, label:"Reverb Dry Mix (L-R)" + unavailable4x, disabled:is4x},
    {patch:device_props.dspReverbWet, label:"Reverb Wet Mix (L-R)" + unavailable4x, disabled:is4x}
  );
  return options;
}



class PhonesVol extends View {
  constructor(p, l, t, index, appdata) {
    super(l,t,334,120);
    this.index = index;
    this.device = appdata.device;
    this.info = this.device.properties.phonesOut[index];
    this.change = this.change.bind(this);
    this.phonesSourceChanged = this.phonesSourceChanged.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    this.updateMeters = this.updateMeters.bind(this);
    p.addSub(this);
    this.chIdx = 10;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    const meterIndex = this.device.zeroPatch;
    this.addSub(this.m1 = new IOMeter(this.l + 17, this.t + 25, 10, 80, meterIndex, appdata.meterStore));
    this.addSub(this.m2 = new IOMeter(this.m1.r + 2, this.t + 25, 10, 80, meterIndex, appdata.meterStore));
    this.addSub(this.trim = new IOKnob(l + 53, t + 27, kToggleBlue, kGateCtrlDisabled, appdata, kPhonesTrimCtrlInfo, -1, -1));
    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l + 140, this.t + 60, 180, 30, this.change));
    this.recreate();
    this.datastore.addListener('kFPGAPatch', this.phonesSourceChanged);
    this.datastore.addListener('koChannelNames', this.labelsChanged);
    this.datastore.addListener('koMixStereo', this.labelsChanged);
    this.datastore.addListener('kSampleRate', this.labelsChanged);
    this.datastore.addListener('kOpticalExpander', this.updateMeters);
    this.datastore.addListener('kOpticalExpanderPatch', this.updateMeters);
  }
  render(c) {
    const num_text = this.device.properties.phonesOut.length > 1 ? " " + (this.index + 1) : "";
    drawLabelGroup(c, "Phones" + num_text, this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";
    c.fillText("Source:", this.menu.l, this.menu.t-6);
  }
  updateLabels() {
    const sr = this.datastore.getKVArrayItem('kSampleRate', 0);
    this.options = make_output_patch_menu(this.device, this.datastore, sr, true);

    if (this.info != undefined) {
      for (const o of this.options) {
        if (o.patch == this.info.dsp) {
          o.label += " (Default)";
        }
      }
    }
  }
  recreate() {
    this.info = this.device.properties.phonesOut[this.index];
    this.updateLabels();
    this.menu.clearOptions();
    const that = this;
    this.options.forEach(function(option){
      that.menu.addOption(option.patch, option.label, option.disabled);
    });
    if (this.info !== undefined) {
      this.menu.sel.value = this.datastore.getKVArrayItem('kFPGAPatch', this.info.destination.left);

      const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 

      this.trim.setup(this.info.line.left, this.info.line.right);
      this.updateMeters();
    } else {
      this.m1.index = this.device.properties.zeroPatch;
      this.m2.index = this.device.properties.zeroPatch;
      this.trim.setup(-1, -1);
    }
  }
  phonesSourceChanged(k, v) {
    if (this.info !== undefined) {
      const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 
      let vl = this.datastore.getKVArrayItem(patchTableKey, this.info.destination.left);
      let vr = this.datastore.getKVArrayItem(patchTableKey, this.info.destination.right);
      if(this.menu.sel.value != vl) {
        this.menu.sel.value = vl;
        this.requestRender();
      }
      this.updateMeters();
    }
  }
  labelsChanged(k, v) {
    this.updateLabels();
    const labels = this.options.map(function(option){ return option.label; });
    const disabled = this.options.map(function(option){ return option.disabled; });
    this.menu.renameOptions(labels, disabled);
  }
  updateMeters() {
    if (this.info != undefined) {
      const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 
      let vl = this.datastore.getKVArrayItem(patchTableKey, this.info.destination.left);
      let vr = this.datastore.getKVArrayItem(patchTableKey, this.info.destination.right);
      this.m1.index = vl;
      this.m2.index = vr;
    }
  }
  change() {
    let vl = parseInt(this.menu.sel.value);
    let vr =  vl + 1;
    this.datastore.setV('kFPGAPatch', this.info.destination.left, vl);
    this.datastore.setV('kFPGAPatch', this.info.destination.right, vr);

    this.updateMeters();
  }
}

class Loopback extends View {
  constructor(p, l, t, appdata) {
    super(l,t,334,120);
    this.change = this.change.bind(this);
    this.loopbackSourceChanged = this.loopbackSourceChanged.bind(this);
    this.labelsChanged = this.labelsChanged.bind(this);
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    p.addSub(this);
    this.updateLabels();
    this.addSub(this.m1 = new IOMeter(this.l + 28, this.t + 25, 10, 80, -1, appdata.meterStore));
    this.addSub(this.m2 = new IOMeter(this.m1.r + 2, this.t + 25, 10, 80, -1, appdata.meterStore));
    this.addSub(this.menu = new SelectMenu(p.root.getRootDiv(), this.l + 78, this.t + 60, 210, 30, this.change));
    const that = this;
    this.options.forEach(function(option){
      that.menu.addOption(option.patch, option.label, option.disabled);
    });
    this.recreate();
    this.datastore.addListener('kLoopbackSource', this.loopbackSourceChanged);
    this.datastore.addListener('koChannelNames', this.labelsChanged);
    this.datastore.addListener('koMixStereo', this.labelsChanged);
    this.datastore.addListener('kSampleRate', this.labelsChanged);
  }
  setEnabled(e) {
    super.setEnabled(e);
    this.menu.setEnabled(e);
  }
  recreate() {
    this.m1.index = this.datastore.getKVArrayItem('kLoopbackSource', 0);
    this.m2.index = this.datastore.getKVArrayItem('kLoopbackSource', 1);

    this.updateLabels();
    this.menu.clearOptions();
    const that = this;
    this.options.forEach(function(option){
      that.menu.addOption(option.patch, option.label, option.disabled);
    });
    this.menu.sel.value = this.datastore.getKVArrayItem('kLoopbackSource', 0);
  }
  render(c) {
    drawLabelGroup(c, "Loopback", this.l+1, this.t+1, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
    c.font = "300 14px Arial";
    c.fillText("Source:", this.menu.l, this.menu.t-6);
  }
  updateLabels() {
    let sr = this.datastore.getKVArrayItem('kSampleRate', 0);
    this.options = make_output_patch_menu(this.device, this.datastore, sr);
  }
  loopbackSourceChanged(k, v) {
    let vl = v[0];
    let vr = v[1];
    if(this.menu.sel.value != vl) {
      this.menu.sel.value = vl;
      this.requestRender();
    }
    this.m1.index = vl;
    this.m2.index = vr;
  }
  labelsChanged(k, v) {
    this.updateLabels();
    const labels = this.options.map(function(option){ return option.label; });
    const disabled = this.options.map(function(option){ return option.disabled; });
    this.menu.renameOptions(labels, disabled);
  }
  change() {
    let vl = parseInt(this.menu.sel.value);
    let vr =  vl + 1;
    this.datastore.setV('kLoopbackSource', 0, vl);
    this.datastore.setV('kLoopbackSource', 1, vr);
    this.m1.index = vl;
    this.m2.index = vr;
  }
}

const ABMonitorType = Object.freeze({
  AOnly:1,
  BOnly:2,
  Both:3,
  None:4
});

class ABMonitoring extends View {
  constructor(parent, left, top, datastore, vertical, drawLabel) {
    const is_horizontal = vertical === undefined || !vertical;
    super(left,top,
      function() { return is_horizontal ? 260 : 184; }(),
      function() { return is_horizontal ? 138 : 215 }()
    );

    this.drawLabel = drawLabel == undefined ? true : drawLabel;
    this.whichChanged = this.whichChanged.bind(this);
    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);
    this.updatePositions = this.updatePositions.bind(this);
    this.is_ab_enabled = this.is_ab_enabled.bind(this);

    parent.addSub(this);
    this.wantsMouse = true;

    this.abEnabled = false;
    this.which = ABMonitorType.Both;
    this.muted = true;
    this.mono = true;

    this.is_horizontal = is_horizontal;

    this.updatePositions();

    this.datastore = datastore;
    datastore.addListener('kAEnable', this.whichChanged);
    datastore.addListener('kBEnable', this.whichChanged);
    datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);
  }

  updatePositions() {
    const innerRect = new Rect(this.l, this.t + 15, this.w - 2, this.h - 15);

    const btnWidth = 60;
    const btnHeight = 30;
    const btnRectFromCenter = (x,y) =>{
      return new Rect(x - btnWidth/2.0, y - btnHeight/2.0, btnWidth, btnHeight);
    };

    if (this.is_horizontal) {
      const first_row_cy = innerRect.t + 40;
      const second_row_cy = innerRect.t + innerRect.h/2.0;
      const third_row_cy = innerRect.b - 40;

      const second_row_offset_x = 25 + btnWidth/2.0;

      this.aHitR = btnRectFromCenter(innerRect.cx, first_row_cy);
      this.onHitR = btnRectFromCenter(innerRect.l + second_row_offset_x, second_row_cy);
      this.bothHitR = btnRectFromCenter(innerRect.r - second_row_offset_x, second_row_cy);
      this.bHitR = btnRectFromCenter(innerRect.cx, third_row_cy);
    }
    else {
      const first_row_cy = innerRect.t + 40;
      const second_row_cy = innerRect.t + innerRect.h/2.0;
      const third_row_cy = innerRect.b - 40;

      const second_row_offset_x = 25 + btnWidth/2.0;

      this.onHitR = btnRectFromCenter(innerRect.cx, first_row_cy);
      this.aHitR = btnRectFromCenter(innerRect.l + second_row_offset_x, second_row_cy);
      this.bHitR = btnRectFromCenter(innerRect.r - second_row_offset_x, second_row_cy);
      this.bothHitR = btnRectFromCenter(innerRect.cx, third_row_cy);
    }
  }

  opticalExpanderChanged(k, v){
    this.setEnabled(this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0);
  }
  offset(dx, dy) {
    super.offset(dx, dy);
  }
  uRBCXCY() {
    super.uRBCXCY();

    this.updatePositions();
  }

  is_ab_enabled() {
    const a_state = this.datastore.getKVArrayItem('kAEnable', 0) !== 0;
    const b_state = this.datastore.getKVArrayItem('kBEnable', 0) !== 0;
    return a_state || b_state;
  }
  whichChanged(k, v) {
    const a_state = this.datastore.getKVArrayItem('kAEnable', 0) !== 0;
    const b_state = this.datastore.getKVArrayItem('kBEnable', 0) !== 0;
    if (a_state) {
      this.which = b_state ? ABMonitorType.Both : ABMonitorType.AOnly;
    }
    else {
      this.which = b_state ? ABMonitorType.BOnly : ABMonitorType.None;
    }

    this.abEnabled = this.is_ab_enabled();
    this.requestRender();
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    let mh = super.getMouseHandler(xy);
    if(mh == this) {
      const list = [this.onHitR, this.aHitR, this.bothHitR, this.bHitR];

      if (list.some(function(button){ return button.contains(xy.x, xy.y); })) {
        return this;
      }
      return undefined;
    }
    return mh;
  }

  handleMouseDown(xy) {
    if (this.onHitR.contains(xy.x, xy.y)) {
      const wasEnabled = this.abEnabled;
      this.datastore.setV('kABEnable', 0, this.abEnabled ? 0 : 1);
      if (wasEnabled) {
        // Disable both A and B
        this.datastore.setV('kAEnable', 0, 0);
        this.datastore.setV('kBEnable', 0, 0);
      }
      else {
        // Enable A
        this.datastore.setV('kAEnable', 0, 1);
        this.datastore.setV('kBEnable', 0, 0);
      }
    }
    else {
      if (this.aHitR.contains(xy.x, xy.y)) {
        this.datastore.setV('kBEnable', 0, 0);
        this.datastore.setV('kAEnable', 0, this.which !== ABMonitorType.AOnly ? 1 : 0);
      }
      else if (this.bHitR.contains(xy.x, xy.y)) {
        this.datastore.setV('kAEnable', 0, 0);
        this.datastore.setV('kBEnable', 0, this.which !== ABMonitorType.BOnly ? 1 : 0);
      }
      else if (this.bothHitR.contains(xy.x, xy.y)) {
        const new_state = this.which === ABMonitorType.Both ? 0 : 1;
        this.datastore.setV('kAEnable', 0, new_state);
        this.datastore.setV('kBEnable', 0, new_state);
      }
      this.datastore.setV('kABEnable', 0, this.is_ab_enabled());
    }
  }
  render(c) {
    c.fillStyle = kEmptyBG;

    const eraseRect = (r) => {
      c.fillRect(r.l, r.t, r.w, r.h);
    };

    eraseRect(this.onHitR);
    eraseRect(this.aHitR);
    eraseRect(this.bHitR);
    eraseRect(this.bothHitR);

    if (this.drawLabel) {
      eraseRect(this);
      drawLabelGroup(c, "A/B Monitoring", this.l, this.t+1, this.w-2, this.h, 6, kEQLabel, kEQLabel);
    }
    drawButton(c, this.onHitR, "AB ON", this.abEnabled, kToggleBlue);

    drawButton(c, this.aHitR, "A", this.which === ABMonitorType.AOnly, kToggleBlue);
    drawButton(c, this.bothHitR, "BOTH", this.which === ABMonitorType.Both, kToggleBlue);
    drawButton(c, this.bHitR, "B", this.which === ABMonitorType.BOnly, kToggleBlue);
  }
}


class MainVolGroupEnableMultiSelect extends SimpleDSView {
  constructor(l, t, appdata) {
    super(l, t, 685, 140, appdata.datastore, 'kMainGroup', 0);

    this.setBG(kEmptyBG);
    this.wantsMouse = true;
    this.items = [];
    this.abEnabled = true;
    this.abEnabledChanged = this.abEnabledChanged.bind(this);
    this.deviceTypeChanged = this.deviceTypeChanged.bind(this);
    this.recreate = this.recreate.bind(this);

    this.addSub(this.m1 = new IOMeter(l + 16, this.t + 25, 10, this.h - 40, -1, appdata.meterStore));
    this.addSub(this.m2 = new IOMeter(this.m1.l + 12, this.t + 25, 10, this.h - 40, -1, appdata.meterStore));
    this.addSub(this.vol = new IOKnob(l + 53, t + 39, kToggleBlue, kGateCtrlDisabled, appdata, kMainVolCtrlInfo, 0));
    this.addSub(this.muteBtn = new ToggleBtn(this.vol.r + 25, this.vol.t, 68, "MUTE", this.datastore, 'kMuteEnable', 0, kToggleRed, kEmptyBG));
    this.addSub(this.monoBtn = new ToggleBtn(this.vol.r + 25, this.vol.b - 30, 68, "MONO", this.datastore, 'kMonoEnable', 0, kToggleBlue, kEmptyBG));

    this.device = appdata.device;
    const device_props = appdata.device.properties;
    const num_button_cols = (device_props.analogOuts.length - 2) / 2;
    const top_button_t = this.t + 58;
    const bottom_button_t = this.t + 88;

    this.label_center = this.monoBtn.r + 45;
    let button_left = this.label_center + 23;

    for (let i = 0; i < num_button_cols; i++)
    {
      const index = i * 2 + 2;
      this.items.push(new Rect(button_left, top_button_t, 25, 25));
      this.items.push(new Rect(button_left, bottom_button_t, 25, 25));
      button_left += 33;
    }

    this.abmonitor = new ABMonitoring(this, button_left, this.t, appdata.datastore, false, false);
    appdata.datastore.addListener('kABEnable', this.abEnabledChanged);
    appdata.datastore.addListener('kOpticalExpander', this.recreate);
    appdata.datastore.addListener('kModelID', this.deviceTypeChanged);

    this.datastore = appdata.datastore;
  }
  recreate() {
    const device_props = this.device.properties;
    const patchTableKey = this.datastore.getKVArrayItem('kOpticalExpander', 0) == 0 ? "kFPGAPatch" : "kOpticalExpanderPatch"; 

    this.m1.index = this.datastore.getKVArrayItem(patchTableKey, device_props.analogOuts[0].destination);
    this.m2.index = this.datastore.getKVArrayItem(patchTableKey, device_props.analogOuts[1].destination);
  }
  abEnabledChanged(k, v) {
    this.abEnabled = this.datastore.getKVArrayItem('kABEnable', 0) !== 0;
    this.requestRender();
  }
  deviceTypeChanged() {
    const supportsMuteAndMono = (()=>{
      const model_id = this.datastore.getKVArrayItem('kModelID', 0);
      if (model_id.search("828") == -1)
        return false;
      return true;
    })();

    if (this.muteBtn != undefined)
      this.muteBtn.setVis(supportsMuteAndMono);
    if (this.monoBtn != undefined)
      this.monoBtn.setVis(supportsMuteAndMono);
  }
  getMouseHandler(xy) {
    // Only allow click in the buttons so ios scrolls
    let mh = super.getMouseHandler(xy);
    if(mh == this) {
      if (this.abEnabled)
        return undefined;
      if (this.items.some(function(r){ return r.contains(xy.x, xy.y); })) {
        return this;
      }
      return undefined;
    }
    return mh;
  }
  handleMouseDown(xy) {
    let bits = DSArrayGetBasic(this.datastore, this);
    const index = this.items.findIndex(function(r){ return r.contains(xy.x, xy.y); });
    const min_index = this.abEnabled ? 2 : 0;
    if (index >= min_index) {
      let bit = (1 << this.device.properties.analogOuts[index+2].dac);
      bits = (bits & bit) ? (bits & ~bit) : (bits | bit);
      DSArraySetBasic(this.datastore, this, bits);
      this.requestRender();
    }
  }
  render(c) {
    let bits = DSArrayGetBasic(this.datastore, this);
    const that = this;

    c.save();

    c.globalAlpha *= this.abEnabled ? kDimmedAlpha : 1.0; // Draw dimmed if AB enabled.

    this.items.forEach(function(item, index){
      if (!that.abEnabled || index >= 2) {
        const bit = (1 << that.device.properties.analogOuts[index+2].dac);
        const isOn = (bits & bit) ? 1 : 0;
        drawButton(c, item, index+3, isOn, kToggleBlue);
      }
    });

    c.restore();

    c.font = "300 14px Arial";
    c.textAlign = 'center';
    c.fillStyle = c.strokeStyle = kEQLabel;

    const buttons_left = this.items[0].l;
    const buttons_right = this.items[this.items.length - 1].r;

    const static_assigned_max = this.abEnabled ? 4 : 2;

    const stringHeight = (str)=>{
      const measurement = c.measureText(str);
      return (measurement.actualBoundingBoxAscent + measurement.actualBoundingBoxDescent) / this.root.divScale;
    };
    const stringWidth = (str)=>{
      const measurement = c.measureText(str);
      return measurement.width / this.root.divScale;
    };

    const outputLabelTop = this.t + 75;
    const outputLabelStr1 = this.abEnabled ? "A/B" : "Main";
    const outputLabelStr2 = this.abEnabled ? "Outputs" : "Out"
    const outputLabelHeight = stringHeight(outputLabelStr1);

    c.save();

    c.fillStyle = kToggleBlue;
    c.fillText(outputLabelStr1, this.label_center, outputLabelTop);
    c.fillText(outputLabelStr2, this.label_center, outputLabelTop + outputLabelHeight + 5);
    c.fillText("1 - " + static_assigned_max, this.label_center, outputLabelTop + outputLabelHeight + stringHeight("Outputs") + 10);

    c.restore();

    const maxStringWidth = Math.max(stringWidth("A/B"), stringWidth("Main"), stringWidth("Outputs"), stringWidth("Out"), stringWidth("1 - " + static_assigned_max));

    const line_left = this.label_center - maxStringWidth/2.0;

    const monitorGroupStr = "Monitor Group";
    const monitorGroupStrWidth = stringWidth(monitorGroupStr);

    const lines = [
      {
        start: line_left,
        end: (line_left + ((buttons_right - line_left) - monitorGroupStrWidth)/2.0) - 6
      },
      {
        start: (buttons_right - monitorGroupStrWidth/2.0) + 5,
        end: buttons_right
      }
    ];

    c.beginPath();

    for (const l of lines) {
      c.moveTo(l.start, this.t + 40);
      c.lineTo(l.end, this.t + 40);
    }

    c.fillText("Monitor Group", line_left + (buttons_right - line_left)/2.0 , this.t + 44);

    c.lineWidth = 1;
    c.stroke();

    drawLabelGroup(c, "Monitoring", this.l+2, this.t+2, this.w-2, this.h-2, 6, kEQLabel, kEQLabel);
  }
}

//todo listen to kFPGAPatch and uptate meters

class Outputs extends View {
  constructor(root, l, t, appdata) {
    super(l, t, 705, 0);

    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);

    root.addSub(this);
    this.setBG(kEmptyBG);

    this.device = appdata.device;
    const device_props = this.device.properties;
    this.addSub(this.mvs = new MainVolGroupEnableMultiSelect(l + 20, t + 12, appdata));
    const num_phones = device_props.outputs.phones.numCh / 2;
    this.pv1 = new PhonesVol(this, 20 + l, this.mvs.b + 20, 0, appdata);
    this.pv2 = new PhonesVol(this, this.pv1.r + 18, this.mvs.b + 20, 1, appdata);
    this.o1 = new LineOut(this, 38 + l, this.pv1.b + 50, 0, appdata);
    this.o2 = new LineOut(this, 38 + l, this.o1.b + 10, 1, appdata);
    this.o3 = new LineOut(this, 20 + this.o1.r, this.pv1.b + 50, 2, appdata);
    this.o4 = new LineOut(this, 20 + this.o1.r, this.o1.b + 10, 3, appdata);
    this.o5 = new LineOut(this, 20 + this.o3.r, this.pv1.b + 50, 4, appdata);
    this.o6 = new LineOut(this, 20 + this.o3.r, this.o1.b + 10, 5, appdata);
    this.o7 = new LineOut(this, 20 + this.o5.r, this.pv1.b + 50, 6, appdata);
    this.o8 = new LineOut(this, 20 + this.o5.r, this.o1.b + 10, 7, appdata);
    this.o9 = new LineOut(this, 20 + this.o7.r, this.pv1.b + 50, 8, appdata);
    this.o10 = new LineOut(this, 20 + this.o7.r, this.o1.b + 10, 9, appdata);

    this.opt = new OpticalOut(this, l + 20, this.o2.b + 35, appdata);
    this.spd = new SPDIFOut(this, l + 20, this.opt.b + 15, appdata);
    const element_next_to_loop = num_phones > 1 ? this.spd : this.pv1;
    this.loopback = new Loopback(this, 0, 0, appdata);

    this.repositionElements();

    this.datastore = appdata.datastore;

    this.datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);

    this.opticalExpanderChanged();
  }
  opticalExpanderChanged(k, v) {
    const e = this.datastore.getKVArrayItem('kOpticalExpander', 0);

    this.mvs.setEnabled(!e);
    this.loopback.setEnabled(!e);
    this.opt.setEnabled(!e);

    if (e && this.optExpanderOverlay == undefined) {
      this.optExpanderOverlay = new OpticalExpanderOverlay(this.cx, this.mvs.cy - (OptExpOvlySize.height/2.0), this.root.rootDiv, kEmptyBG, this.root.animator, this.datastore);
    }
    else if (!e && this.optExpanderOverlay) {
      this.optExpanderOverlay.dismiss();
      this.optExpanderOverlay = undefined;
    }
  }
  repositionElements() {
    const device_props = this.device.properties;
    const num_phones = device_props.outputs.phones.numCh / 2;

    this.pv2.setVis(num_phones > 1);

    if (num_phones > 1) {
      this.spd.setT(this.o2.b + 35);
      this.opt.move(this.l + 20, this.spd.b + 15);
      this.opt.setW(this.r - this.opt.l);
      this.loopback.move(this.l+20, this.spd.t);
      this.spd.setL(this.loopback.r + 15);
      this.spd.setW(this.spd.defualtWidth);
      this.spd.padding = this.spd.defaultPadding;
      this.spd.repositionElements();
      this.pv1.move(this.l + 20, this.mvs.b + 20);
      this.setH(890);
    }
    else {
      this.opt.move(this.l + 20, this.o2.b + 35);
      this.opt.setW(425);
      this.spd.move(this.opt.r + 15, this.opt.t);
      this.spd.setW(this.r - this.spd.l);
      this.spd.padding = 20;
      this.spd.repositionElements();
      this.loopback.move(this.l + 20, this.pv1.t);
      this.pv1.move(this.loopback.r + 20, this.pv1.t);
      this.setH(735);
    }

    this.requestRender();
  }
  recreate() {
    this.mvs.recreate();
    this.pv1.recreate();
    this.pv2.recreate();
    this.o1.recreate();
    this.o2.recreate();
    this.o3.recreate();
    this.o4.recreate();
    this.o5.recreate();
    this.o6.recreate();
    this.o7.recreate();
    this.o8.recreate();
    this.o9.recreate();
    this.o10.recreate();
    this.loopback.recreate();
    this.spd.recreate();
    this.opt.rebuildUI();
    this.repositionElements();
  }
  render(c)
  {
    drawLabelGroup(c, "Output Trim", this.o1.l-18, this.o1.t-30, this.w-20, (this.o10.b - this.o1.t) + 50, 6, kEQLabel, kEQLabel);
  }
}

// phase, trim, 48v amd pad controls for mic pres
// phones and loopback assignment menus
// Separate volume controls for phones and main outs?

class ISetup extends View{
  constructor(appdata) {
    super(0, 0, 0, 0);

    this.c = new CanvasRoot(appdata.animator);
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    appdata.eventSys.addListener('wResized', this.resized);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "inSetupContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "inSetup";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.tabDiv = this.scrollBarVH.containerDiv;

    this.ins = new Inputs(this, 30, 60, appdata);

    this.device = appdata.device;

    this.setBG(kEmptyBG);
  }
  recreate() {
    this.ins.recreate();
    this.resized();
  }
  render(c){
    drawDeviceNameHeader(c, this.l + 50, this.t, this.ins.displayWidth(), this.device);

  }
  resized() {
    let w = this.ins.r + 30;
    let h = this.ins.b + 30;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
  }
}

class OSetup extends View{
  constructor(appdata) {
    super(0, 0, 0, 0);

    this.c = new CanvasRoot(appdata.animator);
    this.c.addSub(this);

    this.resized = this.resized.bind(this);
    appdata.eventSys.addListener('wResized', this.resized);

    this.scrollBarVH = new VHScrollBar(this.c.div);
    this.scrollBarVH.containerDiv.id = "outSetupContainer";
    this.scrollBarVH.containerDiv.style.position = "relative";
    this.scrollBarVH.containerDiv.style.height = "100vh";
    this.scrollBarVH.containerDiv.style.float = "left";

    this.c.div.id = "outSetup";
    this.c.div.style.overflow = "auto";
    this.c.div.style.position = "relative";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kEmptyBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.tabDiv = this.scrollBarVH.containerDiv;
    this.outs = new Outputs(this, 30, 60, appdata);

    this.device = appdata.device;

    this.setBG(kEmptyBG);
  }
  recreate() {
    this.outs.recreate();
    this.resized();
  }
  render(c){
    drawDeviceNameHeader(c, this.l + 50, this.t, 690, this.device);
  }
  resized() {
    let w = this.outs.r + 30;
    let h = this.outs.b + 30;

    this.c.setVisibleRect(0, 0, w, h);

    this.setR(this.c.r);

    this.scrollBarVH.vs.updateScroll(h);
    this.scrollBarVH.hs.updateScroll(w);
  }
}
