'use strict';


// preload canvas images
const kGearImage = new OffScreenImage("settings.svg");
const kLogoImage = new OffScreenImage("logo.svg");

class MixerScrollableCanvas extends CanvasRoot {
    constructor(appdata, w, scrollH) {
    super(appdata.animator, true);

    this.mixManager = appdata.mixManager;
    this.eventSys = appdata.eventSys;
    this.scrolled = this.scrolled.bind(this);
    this.resized = this.resized.bind(this);
    this.mixCfgChanged = this.mixCfgChanged.bind(this);

    this.div.id = "mixInCh";
    this.div.style.width = "100%";
    this.div.style.height = "calc(100% - " + scrollH + "px)";
    this.div.style.overflow = "scroll";
    this.div.style.backgroundColor = kMixBG;
    this.div.style.WebkitOverflowScrolling = "touch";

    this.div.onscroll = this.scrolled;

    this.scroller = document.createElement('div');
    this.scroller.id = "scroller";
    this.scroller.style.width = w.toFixed() + "px";
    this.scroller.style.height = "100%";
    this.scroller.style.backgroundColor = kMixBG;

    // Attach the scroller to the canvas and the canvas to the scroller
    this.rootDiv = this.scroller;
    this.div.appendChild(this.scroller);
    this.rootDiv.appendChild(this.canvas);

    this.eventSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  mixCfgChanged() {
    var w = this.mixManager.getNumCh() * kChStripW;
    this.scroller.style.width = w.toFixed() + "px";
    this.eventSys.dispatchASync('wResized');
  }
  resized() {
    super.resized();
    this.eventSys.dispatchSync('updateMixScroll', this);
  }
  scrolled() {
    this.setVisibleRect(this.div.scrollLeft, this.div.scrollTop, this.r.w, this.r.h);
    this.eventSys.dispatchSync('updateMixScroll', this);
  }
}


class MixerBusStripCanvas extends CanvasRoot {
    constructor(animator) {
    super(animator);

    this.div.id = "mix out";
    this.div.style.width = "100%";
    this.div.style.height = "100%";
    this.div.style.backgroundColor = kBusStripBG;
    this.div.style.float = "left";
    this.div.style.boxShadow = "-3px 0 6px 0 rgba(0,0,0,0.5)";
    this.resized = this.resized.bind(this);
  }

  resized() {
    super.resized();
   }
}


class StripMeter extends View {
  constructor(l,t,w,h,appdata) {
    super(l,t,w,h);
    this.onMeters = this.onMeters.bind(this);
    this.setup = this.setup.bind(this);
    this.setAttenuation = this.setAttenuation.bind(this);
    this.updateZeroPosition = this.updateZeroPosition.bind(this);
    this.mixManager = appdata.mixManager;

    this.updateZeroPosition();

    this.lm = {m:0, ph:0};
    this.rm = {m:0, ph:0};
    this.index = 0;
    this.attDB = 0;
    this.hasClip = 0;
    this.c = kMeterBlue;
    appdata.meterStore.addListener(this.onMeters);
  }
  updateZeroPosition() {
    this.zeroH = Math.floor(Calc12dBToCtrl24(0) * this.h);
    this.zeroT = this.t + (this.h - this.zeroH);
  }
  setup(info) {
    this.isStereo = info.isStereo;
    this.index = info.meter ? info.meter : info.meters[kMIPostFx];
    this.c = this.mixManager.getActiveBusInfo().color;
    this.hasClip = (info.stripType == "input") && ((info.type == "Mic") || (info.type == "Line"));
  }
  setAttenuation(attDB){
    // attenuation in dB set by fader in post fader mode
    this.attDB = attDB;
  }
  renderMeter(c, m, l, w) {
    var sh = Math.floor(Math.min(Calc12dBToCtrl24(m.m + this.attDB) * this.h, this.zeroH));
    var sy = this.h - sh;
    if(sh && (m.m <= 0)) {
      c.fillStyle = this.c;
      c.fillRect(l, this.t + sy, w, sh)
    }

    var y = this.h - Math.floor(Math.min(Calc12dBToCtrl24(m.ph + this.attDB) * this.h, this.zeroH));
    if(y < (this.h - 2)) {
      c.fillStyle = "red";
      c.fillRect(l, this.t + y, w, 2);
      if(this.hasClip && (Math.floor(m.ph + this.attDB) >= 0)) {
        c.fillRect(l, this.zeroT - 19 - 1, w, 9);
      }
    }
  }
  render(c) {
    c.fillStyle = kBlackBG;
    c.fillRect(this.l, this.zeroT, this.w, this.zeroH);
    if(this.hasClip)
      c.fillRect(this.l, this.zeroT - 19 - 6, this.w, 19);
    if(this.isStereo){
      var w = (this.w/2)-6;
      this.renderMeter(c, this.lm, this.l + 5, w);
      this.renderMeter(c, this.rm, this.l + 7 + w, w);
    }
    else {
      this.renderMeter(c, this.lm, this.l + 5, this.w - 10);
    }

  }
  setMeter(v, i, dest) {
    var m = v[i].m;
    var ph = v[i].ph;
    if((dest.m != m) || (dest.ph != ph)) {
      dest.m = m;
      dest.ph = ph;
      return true;
    }
    return false;
  }
  onMeters(v) {
    if((this.index >= 0) && (this.index < v.length)) {
      if(this.setMeter(v, this.index, this.lm))
        this.requestRender();
      if(this.isStereo) {
        if (this.setMeter(v, this.index + 1, this.rm))
          this.requestRender();
      }
    }
  }
}


class ChName extends View {
  constructor(l,t,w,h, appdata, dsKey, modalDialog) {
    super(l,t,w,h);

    this.refresh = this.refresh.bind(this);
    this.setup = this.setup.bind(this);

    this.modalDialog = modalDialog;
    this.dsKey = dsKey;
    this.info = 0;
    this.setBG(kMixBG);
    this.wantsMouse = true;

    this.mixManager = appdata.mixManager;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.datastore.addListener(dsKey, this.refresh);
  }
  handleMouseDown(xy) {
    this.modalDialog.activate(this.info.chNum);
  }
  refresh() {
    this.v = getChannelName(this.datastore, this.dsKey, this.info);
    this.requestRender();
  }
  setup(info) {
    this.info = info;
    this.refresh();
  }
  render(c) {
    c.font = '16px Helvetica';
    c.textAlign = 'center';

    // fill it black then stroke if it's not an input with the bus color
    roundRectPath(c, this.l, this.t+2, this.w, this.h, 2);
    c.fillStyle = "black";
    c.fill();

    const device_props = this.device.properties;
    const outputs = device_props.outputs;
    if(this.info.stripType !== "input") {
      switch (this.info.type) {
        default:
          c.strokeStyle = this.mixManager.getActiveBusInfo().color;
          break;
        case MixBusTypes.Main:
          c.strokeStyle = device_props.busColors[outputs.mains.colorIndex];
          break;
        case MixBusTypes.Line:
          c.strokeStyle = device_props.busColors[outputs.lines.colorIndex + this.info.bankChNum];
          break;
        case MixBusTypes.Phones:
          c.strokeStyle = device_props.busColors[outputs.phones.colorIndex + this.info.bankChNum];
          break;
        case MixBusTypes.Reverb:
          c.strokeStyle = device_props.busColors[outputs.reverbs.colorIndex];
          break;
      }
      c.lineWidth = 1.1;
      c.stroke();
    }
    c.fillStyle = "white";

    var t = fitText(c, this.v, this.w);
    c.fillText(t, this.l + (this.r - this.l)/2, 7 + this.cy);

    if(this.dsKey == 'kiChannelNames') {
      var w = c.measureText(t).width;
      c.fillRect(this.cx - w/2, 9 + this.cy, w, 1);
    }
  }
}


class RudeSoloListener {
  constructor(soloChangedCB, appdata) {
    this.soloChanged = this.soloChanged.bind(this);
    this.soloChangedCB = soloChangedCB
    this.busNum = -1;
    this.rs = false;
    this.mixManager = appdata.mixManager;
    appdata.eventSys.addListener('soloChanged', this.soloChanged);
  }
  setBusNum(bn) { // -1 means disable
    this.busNum = bn;
    this.soloChanged();
  }
  soloChanged() {
    let rs = this.rs;

    if(this.busNum < 0)
      rs = false;
    else {
      let busInfo = this.mixManager.getBusInfo(this.busNum);
      if (busInfo) {
        rs = this.mixManager.getBusRudeSolo(busInfo.gainOCh);
      }
    }

    if(this.rs != rs) {
      this.rs = rs;
      this.soloChangedCB();
    }
  }
}


class XBox extends View {
  constructor(l,t,w,h,canvas) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.canvas = canvas;
    this.setBG(kFxEditBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y))
      this.canvas.deactivate();
  }
  render(c) {
    c.strokeStyle = kTabTxt;
    c.lineWidth = 3;
    c.beginPath();
    const t = 8
    c.moveTo(this.l + t, this.t + t);
    c.lineTo(this.r -t, this.b -t);
    c.moveTo(this.r -t, this.t + t);
    c.lineTo(this.l + t, this.b -t);
    c.stroke();
  }
}


class RBox extends View {
  constructor(l,t,w,h, mixManager) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.mixManager = mixManager;
    this.setBG(kFxEditBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      this.mixManager.incActiveCh(false);
    }
  }
  render(c) {
    c.strokeStyle = kTabTxt;
    c.lineWidth = 3;
    c.beginPath();
    const t = 8
    c.moveTo(this.l + t, this.t + t);
    c.lineTo(this.r - t*2, this.cy);
    c.lineTo(this.l + t, this.b - t);
    c.stroke();
  }
}


class LBox extends View {
  constructor(l,t,w,h, mixManager) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.mixManager = mixManager;
    this.setBG(kFxEditBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      this.mixManager.incActiveCh(true);
    }
  }
  render(c) {
    c.strokeStyle = kTabTxt;
    c.lineWidth = 3;
    c.beginPath();
    const t = 8
    c.moveTo(this.r - t, this.t + t);
    c.lineTo(this.l + t*2, this.cy);
    c.lineTo(this.r - t, this.b - t);
    c.stroke();
  }
}


class ChNameHeader extends EditableTextView {
  constructor(l,t,w,h,appdata,dsKey) {
    super(l,t,w,h, 'user-input-input', 'Bold 30px Helvetica', appdata.keyFilter);

    this.refresh = this.refresh.bind(this);
    this.setup = this.setup.bind(this);

    this.fx = "";
    this.n = "";
    this.setBG(kFxEditBG);

    this.datastore = appdata.datastore;
    this.info = 0;
    this.dsKey = dsKey;
    this.datastore.addListener(this.dsKey, this.refresh);
  }
  handleMouseDown() {
    if ((this.info.type != "Reverb") && (this.info.stripType != "host"))
      super.handleMouseDown();
  }
  valueChanged(v) {
    this.datastore.setV(this.dsKey, this.info.name, v);
    return v;
  }
  refresh() {
    this.v = getChannelName(this.datastore, this.dsKey, this.info);
    this.requestRender();
  }
  setup(info) {
    this.info = info;
    this.refresh();
  }
}



class ChFxEditView extends View {
  constructor(appdata, canvas, syncToMixMgr) {
    super(0,0,kFxWidth,kFxHeight);

    const ki = 4;
    const ki2 = 2 * ki;
    const t = ki + 50;
    const h = kFxHeight - (ki2 + 50);

    this.activate = this.activate.bind(this);
    this.activeChChanged = this.activeChChanged.bind(this);
    this.setActiveCh = this.setActiveCh.bind(this);

    this.syncToMixMgr = (syncToMixMgr != undefined) ? syncToMixMgr : true;

    if (this.syncToMixMgr) {
      this.addSub(this.lBox = new LBox(kFxChStripW + 25, this.t + 10, 40, 40, appdata.mixManager));
      this.addSub(this.rBox = new RBox(kFxChStripW + 95, this.t + 10, 40, 40, appdata.mixManager));
    }

    let headerL = this.rBox ? this.rBox.r : kFxChStripW + 135; 
    this.addSub(this.xBox = new XBox(this.r - 75, this.t + 10, 40, 40, canvas));
    this.addSub(this.chNav = new ChNameHeader(headerL, this.t + 10, this.w - (headerL + 135), 40, appdata, 'kiChannelNames'));

    this.addSub(this.cs = new ChannelStrip(ki, ki, kFxChStripW, kFxHeight - ki2, 0, true, appdata, this));
    this.cs.resize(kFxHeight - ki2, kFxHeight - 80, kFxHeight - 400);
    this.eq = new EQFxView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h, appdata, true);
    this.dyn = new DynamicsView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h, appdata);

    this.fx  = { [kEQFxType]:this.eq, [kDynamicsFxType]:this.dyn };
    this.setBG(kMixBG);
    this.lastFXType = kEQFxType;

    this.canvas = canvas;
    this.mixManager = appdata.mixManager;

    if (this.syncToMixMgr) {
      appdata.eventSys.addListener('activeChChanged', this.activeChChanged);
    }
  }
  activeChChanged() {
    if (this.syncToMixMgr) {
      var ch = this.mixManager.getActiveCh();
      this.setActiveCh(ch);
    }
  }

  setActiveCh(ch) {
    let info = this.mixManager.getChInfo(ch);
    this.cs.setChNum(ch);
    this.chNav.setup(info);
    this.eq.setup(info);
    this.dyn.setup(info);
  }
  activate(channel, fxType) {
    if (this.syncToMixMgr)
      this.mixManager.setActiveCh(channel);

    if(fxType == undefined)
      fxType = this.lastFXType;

    this.lastFXType = fxType;
    this.removeSub(this.eq);
    this.removeSub(this.dyn);

    var f = this.fx[fxType];
    if(f != undefined) {

      this.addSub(f);
      this.bg = f.bg;

      this.canvas.activate(this);
      this.requestRender();
    }
  }
  render(c){
    c.strokeStyle = kNavBtnGrey;
    c.lineWidth = 3;
    c.strokeRect(this.l + 1, this.t + 1, this.w - 2, this.h - 2);
  }
}


class TalkbackBtn extends View {
  constructor(l,t,w,h, appdata) {
    super(l,t,w,h);

    this.latchChanged = this.latchChanged.bind(this);
    this.enabledChanged = this.enabledChanged.bind(this);

    this.datastore = appdata.datastore;
    this.wantsMouse = true;
    this.talkHeld = false;

    this.datastore.addListener('kTalkbackLatch', this.latchChanged);
    this.datastore.addListener('kTalkbackEnable', this.enabledChanged);

    this.latchChanged();
    this.enabledChanged();

    this.setBG(kMixBG);
  }

  latchChanged(k, v) {
    this.latchMode = this.datastore.getKVArrayItem('kTalkbackLatch', 0) > 0;
  }
  enabledChanged(k, v) {
    this.talkbackEnabled = this.datastore.getKVArrayItem('kTalkbackEnable', 0) > 0;
    this.requestRender();
  }
  handleMouseDown(xy) {
    this.datastore.setV('kTalkbackEnable', 0, this.latchMode && this.talkbackEnabled ? 0 : 1);
    this.talkHeld = true;
  }
  handleMouseUp(xy) {
    if (this.talkHeld && !this.latchMode) {
      this.datastore.setV('kTalkbackEnable', 0, 0);
    }
    this.talkHeld = false;
  }
  render(c) {
    drawButton(c, this, "TALK", this.talkbackEnabled, kToggleBlue);
  }
}


class TalkbackSettingsBtn extends View {
  constructor(l,t,w,h, appdata) {
    super(l,t,w,h);
    this.wantsMouse = true;
    this.appdata = appdata;
  }

  handleMouseUp(xy) {
    this.appdata.eventSys.dispatchASync(kTabEv, kHome);
    setTimeout(()=>{
      gApp.home.scrollToTalkback();
    }, 0);
  }

  render(c) {
    drawButton(c, this, "SETTINGS", this.talkbackEnabled, kToggleBlue);
  }
}


class ChannelStrip extends View {
  constructor(l,t,w,h, chNum, isFxChStrip, appdata, modalDialog) {
    super(l,t,w,h);

    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    this.resize = this.resize.bind(this);
    this.setChNum = this.setChNum.bind(this);
    this.isFxChStrip = isFxChStrip;

    const itemW = (isFxChStrip ?  kFxChStripW : kChStripW) - 14;
    const itemL = l + 7;
    var fxSlotT = 10;
    if(isFxChStrip) {
      this.addSub(this.stereo = new ToggleBtn(itemL, t + 10, itemW, "Stereo", appdata.datastore, 'kiMixStereo', 0, kToggleBlue));
      fxSlotT = this.stereo.b + 5;
    }

    this.mixManager = appdata.mixManager;
    this.datastore = appdata.datastore;
    this.eq = new EQThumb(itemL, fxSlotT, appdata, true, isFxChStrip, modalDialog);
    this.fxSlotH = this.eq.h + 5;
    this.dyn = new DynamicsThumb(itemL, fxSlotT + this.fxSlotH, isFxChStrip, appdata, modalDialog);
    if(isFxChStrip)
    {
      this.addSub(this.pan = new PanFader(itemL, this.dyn.b + 5, itemW, 30, appdata));
      this.pan.setH(60);
    }
    else {
      this.addSub(this.name = new ChName(l+7, this.dyn.b + 5, itemW, 35, appdata, 'kiChannelNames', modalDialog));
      this.addSub(this.pan = new PanFader(itemL, this.name.b + 5, itemW, 30, appdata));
    }

    this.addSub(this.mute = new ToggleBtn(itemL, this.pan.b + 10, itemW, "MUTE", this.datastore, 'kiMixMute', 0, kToggleRed));
    this.addSub(this.meter = new StripMeter(l+(isFxChStrip ? 52 : 28), this.mute.b + 20, isFxChStrip ? 25 : 19, 357, appdata));
    this.addSub(this.fader = new StripFader(l+(isFxChStrip ? kFaderL + 35 : kFaderL), this.mute.b + 20, kFaderL, 367, appdata, 'kiMixFader', this.meter));
    this.addSub(this.solo = new ToggleBtn(itemL, this.fader.b + 10, itemW, "SOLO", this.datastore, 'kiMixSolo', 0, kToggleYellow));
    this.setBG(kMixBG);

    this.chNum = chNum;
    this.stripType = "input";

    this.addSub(this.talkButton = new TalkbackBtn(this.solo.l, this.solo.t, this.solo.w, this.solo.h, appdata));
    this.talkButton.setVis(false);

    this.addSub(this.talkSettingsButton = new TalkbackSettingsBtn(this.mute.l, this.mute.t, this.mute.w, this.mute.h, appdata));
    this.talkSettingsButton.setVis(false);

    appdata.eventSys.addListener('mixCfgChanged', this.mixCfgChanged);
    this.datastore.addListener('kTalkbackSource', this.mixCfgChanged);

    this.isTalkbackSrc = false;
  }
  setChNum(chNum) {
    this.chNum = chNum;
    this.mixCfgChanged();
  }
  showTalkback() {

  }
  mixCfgChanged() {
    this.setVis(this.mixManager.getNumCh() > this.chNum);
    if(!this.isVis) {
      this.eq.setup();
      return;
    }

    let tlkSrc = this.datastore.getKVArrayItem('kTalkbackSource', 0);
    var info = this.mixManager.getChInfo(this.chNum);
    var busInfo = this.mixManager.getActiveBusInfo();
    var gainIndex = this.mixManager.getChGainIndex(this.chNum);

    if(this.stereo != undefined) {
      this.stereo.setup(info.stereoIndex);
      this.stereo.setVis(info.stereoIndex != undefined);
    }

    this.stripType = info.stripType;
    this.eq.setup(info);
    this.dyn.setup(info);
    if(this.name)
      this.name.setup(info);
    this.pan.setup(gainIndex, info.isStereo);
    this.pan.setVis(busInfo.isStereo);
    this.meter.setup(info);
    this.fader.setup(gainIndex, info.isStereo, this.meter.index);
    this.solo.setup(gainIndex);
    this.mute.setup(gainIndex);

    if (tlkSrc == info.gainICh) {
      this.isTalkbackSrc = true;
      this.pan.setVis(false);
      this.fader.setVis(false);
      this.solo.setVis(false);
      this.mute.setVis(false);

      this.talkButton.setVis(true);
      this.talkSettingsButton.setVis(true);
    }
    else {
      this.isTalkbackSrc = false;
      this.pan.setVis(true);
      this.fader.setVis(true);
      this.solo.setVis(true);
      this.mute.setVis(true);

      this.talkButton.setVis(false);
      this.talkSettingsButton.setVis(false);
    }
    this.resize(this.h);
  }

  resize(chH) {
    this.setH(chH);

    if(this.stereo != undefined)
      this.removeSub(this.stereo);
    this.removeSub(this.eq);
    this.removeSub(this.dyn);

    var t = this.t + 10;
    var minH = 340;
    var seq = 0;
    var sd = 0;

    if(chH > (minH + this.fxSlotH))
      seq = 1;
    if(chH > (minH +  2 * this.fxSlotH))
      sd = 1;

    if((this.stereo != undefined) && (chH > (minH + this.stereo.h + 5 + 2 * this.fxSlotH))){
      this.addSub(this.stereo);
      t += this.stereo.h + 5;
    }

    if(getSampleRateMult(this.datastore.getKVArrayItem('kSampleRate', 0)) == 4) {
      // leave an empty slot because the bus strips will have eq at 4x
      t += this.fxSlotH;
    }
    else {
      if(seq) {
        if(this.eq.isValid) {
          this.eq.move(this.eq.l, t);
          this.addSub(this.eq);
        }
        t += this.fxSlotH;
      }
      if(sd) {
        if(this.dyn.isValid) {
          this.dyn.move(this.dyn.l, t);
          this.addSub(this.dyn);
        }
        t += this.fxSlotH;
      }
    }

    if(this.name) {
      this.name.setT(t);
      t = this.name.b + 5;
    }
    if(this.pan.isVis || this.isTalkbackSrc) {
      this.pan.setT(t);
      t = this.pan.b + 10;
    }
    this.mute.setT(t);
    this.talkSettingsButton.setT(t);
    this.solo.setT(chH - (this.solo.h + 10));
    this.talkButton.setT(this.solo.t);

    var faderH = (this.solo.t - this.mute.b) - 20;
    var meterT = this.mute.b + 12 + kFaderCapRadius;
    var meterH = faderH - (2 * (kFaderCapRadius + 2));
    this.meter.setT(meterT);
    this.meter.setH(meterH + 3);
    this.meter.updateZeroPosition();
    this.fader.setMeterTH(meterT, meterH);
    this.fader.setT(this.mute.b + 10);
    this.fader.setH(faderH);

    this.requestRender();
  }
  render(c) {

    if(this.name && (this.stripType == "host")) {
      c.fillStyle = kMeterTickColor;
      c.textAlign = 'center';
      c.font = "16px Helvetica";
      c.fillText("Computer", this.cx, this.name.t - 35);
      c.fillText("USB", this.cx, this.name.t - 13);
    }

    draw12dBTicks(c, kMeterTickColor, this.fader.meterT, this.l+(this.isFxChStrip ? 20 : 2), this.fader.meterH, this.fader.l);

    c.lineWidth = 1.5;
    if (this.chNum && !this.isFxChStrip) {
      c.strokeStyle = "rgb(67,67,67)";
      c.beginPath();
      c.moveTo(this.l, this.t);
      c.lineTo(this.l, this.b-1);
      c.stroke();
    }
    c.strokeStyle = "rgb(60,60,60)";
    c.beginPath();
    c.moveTo(this.r, this.t);
    c.lineTo(this.r, this.b-1);
    c.stroke();
    if (!this.isFxChStrip) {
      c.strokeStyle = "black";
      c.beginPath();
      c.moveTo(this.l, this.b);
      c.lineTo(this.r, this.b);
      c.stroke();
    }
  }
}


class BusFxEditView extends View {
  constructor(appdata, canvas) {
    super(0,0,kFxWidth,kFxHeight);

    const ki = 4;
    const ki2 = 2 * ki;
    const t = ki + 50;
    const h = kFxHeight - (ki2 + 50);

    this.activate = this.activate.bind(this)
    this.mixCfgChanged = this.mixCfgChanged.bind(this)

    this.addSub(this.xBox = new XBox(this.r - 75, this.t + 10, 40, 40, canvas));
    this.addSub(this.chNav = new ChNameHeader(kFxChStripW + 95, this.t + 10, this.xBox.l - (kFxChStripW + 95), 40, appdata, 'koChannelNames'));
    this.addSub(this.bs = new BusStrip(ki, ki, kFxChStripW, kFxHeight - ki2, true, appdata, this));
    this.bs.resize(kFxHeight - ki2, 235, kFxHeight - 255);
    this.eq = new EQFxView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h, appdata, false);
    this.rvb = new ReverbFxView(kFxChStripW + ki, t, kFxWidth - (kFxChStripW + ki2), h, appdata.datastore);
    this.fx = { [kBusEQFxType]:this.eq, [kReverbFxType]:this.rvb };
    this.setBG(this.eq.bg);

    this.canvas = canvas;
    this.mixManager = appdata.mixManager;
    appdata.eventSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  mixCfgChanged() {
    var info = this.mixManager.getActiveBusInfo();
    this.chNav.setup(info);
    this.eq.setup(info);
    this.rvb.setup(info);
  }
  activate(channel, fxType) {
    if(fxType == undefined)
      fxType = kBusEQFxType;

    this.removeSub(this.eq);
    this.removeSub(this.rvb);

    var f = this.fx[fxType];
    if (f != undefined) {
      this.addSub(f);
      this.bg = f.bg;

      this.canvas.activate(this);
      this.requestRender();
    }
  }
  render(c){
    c.strokeStyle = kNavBtnGrey;
    c.lineWidth = 3;
    c.strokeRect(this.l + 1, this.t + 1, this.w - 2, this.h - 2);
  }
}


class CollapseBtn extends View {
  constructor(l, t, datastore) {
    super(l, t, 30, 30)
    this.wantsMouse = true;
    this.datastore = datastore;
    this.setBG(kBusStripBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      this.datastore.setV('kBusStripExp', 0, this.datastore.getKVArrayItem('kBusStripExp', 0) == "on" ? "off" : "on");
      this.requestRender();
    }
  }
  render(c) {
    var isExp = this.datastore.getKVArrayItem('kBusStripExp', 0) != "off";
    if(isExp) {
      drawRArrow(c, 1, this.r-21, this.b-9, kToggleDefault);
      drawLArrow(c, 1, this.r-9, this.b-21, kToggleDefault);
    }
    else {
      drawLArrow(c, 1, this.r-21, this.b-9, kToggleDefault);
      drawRArrow(c, 1, this.r-9, this.b-21, kToggleDefault);
    }
  }
}


class ClearSoloBtn extends View {
  constructor(l, t, appdata) {
    super(l, t, 30, 30)
    this.soloChanged = this.soloChanged.bind(this);
    this.wantsMouse = true;
    this.rs = new RudeSoloListener(this.soloChanged, appdata);
    this.mixManager = appdata.mixManager;
    this.setBG(kBusStripBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      this.mixManager.clearBusSolos(this.mixManager.getActiveBusInfo().gainOCh);
    }
  }
  soloChanged() {
    this.requestRender();
  }
  render(c) {
    if(this.rs.rs)
      drawButton(c, this, "CLEAR SOLO", false, this.mixManager.getActiveBusInfo().color);
  }
}


class BusStrip extends View {
  constructor(l,t,w,h, isFxBusStrip, appdata, modalDialog) {
    super(l,t,w,h);

    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    this.resize = this.resize.bind(this);

    this.modalDialog = modalDialog;
    this.isFxBusStrip = isFxBusStrip;

    const itemW = (isFxBusStrip ?  kFxChStripW : kBusStripW) - 14;
    const itemL = l + 7;

    var fxSlotT = t + 10;
    if(isFxBusStrip) {
      this.addSub(this.stereo = new ToggleBtn(itemL, t + 10, itemW, "Stereo", appdata.datastore, 'koMixStereo', 0, kToggleBlue, kBusStripBG));
      fxSlotT = this.stereo.b + 5;
    }
    var tt = t + 0;

    // note that the eq thumb is the same size as the chstrip because I wan't it's hieght to be the same
    const left_offset = isFxBusStrip ? 0 : 13;
    this.appdata = appdata;
    this.addSub(this.eq = new EQThumb(itemL + left_offset, fxSlotT, appdata, false, isFxBusStrip, modalDialog));
    this.fxSlotH = this.eq.h + 5;
    this.addSub(this.rvb = new ReverbThumb(itemL + left_offset, fxSlotT + this.fxSlotH, appdata, isFxBusStrip, modalDialog));

    this.addSub(this.mute = new ToggleBtn(itemL, tt, itemW, "MUTE", appdata.datastore, 'koBusMute', 0, kToggleRed));
    this.mute.setBG(kBusStripBG);

    this.addSub(this.meter = new StripMeter(l+26, tt, 30, 357, appdata));
    this.addSub(this.fader = new StripFader(l+kFaderL+17, tt, kFaderL, 357, appdata, 'koBusFader'));

    this.setBG(kBusStripBG);
    this.fader.setBG(kBusStripBG);

    if(!isFxBusStrip) {
      this.addSub(this.name = new ChName(l+12, this.mute.t - 80, kBusStripW - 24, 35, appdata, 'koChannelNames', modalDialog));
      this.name.setBG(kBusStripBG);

      this.addSub(this.collapseBtn = new CollapseBtn(l+26, this.b-35, appdata.datastore));
      this.addSub(this.csb = new ClearSoloBtn(30,30, appdata));
    }

    this.isStereo = true;

    appdata.eventSys.addListener('mixCfgChanged', this.mixCfgChanged);
  }
  updateNamePosition() {
    if (this.name)
      this.name.setT(this.mute.t - (40 + (40 * this.isStereo)));
  }
  mixCfgChanged() {
    var info = this.appdata.mixManager.getActiveBusInfo();
    if(this.stereo != undefined) {
      this.stereo.setup(info.stereoIndex);
      this.stereo.setVis(info.stereoIndex != undefined);
    }

    var fxSlotT = this.t + 10;
    if(this.isFxBusStrip && (info.type != "Reverb")) {
      fxSlotT = this.stereo.b + 5;
    }

    if(this.eq.t != fxSlotT) {
      // EQ and Reverb views need to be constructed at their postion
      // they do not gracefully move around by changing their view rect
      var l = this.eq.l;
      this.removeSub(this.eq);
      this.removeSub(this.rvb);
      this.addSub(this.eq = new EQThumb(l, fxSlotT, this.appdata, false, this.isFxBusStrip, this.modalDialog));
      this.addSub(this.rvb = new ReverbThumb(l, fxSlotT + this.fxSlotH, this.appdata, this.isFxBusStrip, this.modalDialog));
    }

    if(this.name)
      this.name.setup(info);
    this.meter.setup(info);
    this.fader.setup(info.gainOCh, info.isStereo);

    this.mute.setup(info.gainOCh);
    this.eq.setup(info);
    this.rvb.setup(info);

    if(this.csb)
      this.csb.rs.setBusNum(this.appdata.mixManager.getActiveBus());

    if(this.isStereo != info.isStereo) {
      this.isStereo = info.isStereo;
      this.updateNamePosition();
    }
    this.requestRender();
  }
  resize(chH, faderT, faderH) {
    var expBus = (this.appdata.datastore.getKVArrayItem('kBusStripExp', 0) != "off") || this.isFxBusStrip;
    this.setWH(expBus ? (this.isFxBusStrip ?  kFxChStripW : kBusStripW) : kBusStripWSmall, chH);

    this.removeSub(this.eq);
    this.removeSub(this.rvb);
    if((chH > 434) && expBus)
    {
      this.addSub(this.eq);
      if (chH > (434 + this.fxSlotH))
        this.addSub(this.rvb);
    }
    this.mute.setVis(expBus);
    this.fader.setVis(expBus);

    this.mute.setT(faderT - 40);
    if(this.isFxBusStrip) {
      this.meter.setL(55);
      this.fader.setL(this.l+kFaderL+37);
    }
    else {
      this.meter.setL(expBus ? 35 : 11);
      this.fader.setL(this.l+kFaderL+22);
    }
    var meterT = faderT + 2 + kFaderCapRadius;
    var meterH = faderH - (2 * (kFaderCapRadius + 2));
    this.meter.setT(meterT);
    this.meter.setH(meterH + 3);
    this.meter.updateZeroPosition();
    this.fader.setT(faderT);
    this.fader.setH(faderH);
    this.fader.setMeterTH(meterT, meterH);
    if(this.csb) {
      this.csb.setVis(expBus);
      this.csb.setLTWH(this.l+14,  this.fader.b + 10, this.w - 28, 30);
    }

    if(this.collapseBtn) {
      this.collapseBtn.setT(this.b - 35);
      this.collapseBtn.setL(this.cx - this.collapseBtn.w/2);
    }

    if(this.name) {
      this.updateNamePosition();
      this.name.setVis(expBus);
    }
  }
  render(c){
    if(this.fader.isVis){
      draw12dBTicks(c, kMeterTickColor, this.fader.meterT, this.l+(this.isFxBusStrip ? 20 : 5), this.fader.meterH, this.fader.l);
    }
  }
}


class Mixer {
  constructor(appdata) {

    var xPos = kTabSelW;
    var mixOutW = kBusStripW;

    // create the modal dialogs
    const fxbasecanvas = new FxModalCanvas(appdata);
    const chFxEdit = new ChFxEditView(appdata, fxbasecanvas);
    const busFxEdit = new BusFxEditView(appdata, fxbasecanvas);

    // Container for all mixer
    this.div = document.createElement('div');
    this.div.id = "mixer";
    this.div.style.position = "relative";
    this.div.style.width = "calc(100% - " + xPos + "px)";
    this.div.style.height = "100%";
    this.div.style.backgroundColor = kMixBG;
    this.div.style.float = "left";

    // Container for mixer ins and scroller
    this.mixIn = document.createElement('div');
    this.mixIn.id = "mixIn";
    this.mixIn.style.width = "calc(100% - " + mixOutW + "px)";
    this.mixIn.style.height = "100%";
    this.mixIn.style.backgroundColor = kMixBG;
    this.mixIn.style.float = "left";

    // Container for mixer outs
    this.mixOut = document.createElement('div');
    this.mixOut.id = "mixOut";
    this.mixOut.style.width = mixOutW + "px";
    this.mixOut.style.height = "100%";
    this.mixOut.style.backgroundColor = kBusStripBG;
    this.mixOut.style.float = "left";
    this.mixOut.style.zIndex = kZIndexMixer;
    this.mixOut.style.position = "absolute";
    this.mixOut.style.right = 0;
    this.mixOut.style.top = 0;

    this.appdata = appdata;
    this.setVis = this.setVis.bind(this);
    this.syncOverlayPosition = this.syncOverlayPosition.bind(this);
    this.updateOverlayVisibility = this.updateOverlayVisibility.bind(this);

    // The mix canvas that scrolls
    this.eventSys = appdata.eventSys;
    this.mixInCanvas = new MixerScrollableCanvas(appdata, kChStripW * kNumChStrips, kScrollBarSize);

    this.chStrips = [];
    for (var i = 0; i < kNumChStrips; i++) {
      var cs = new ChannelStrip(i * kChStripW, 0, kChStripW, 868, i, false, appdata, chFxEdit);
      this.mixInCanvas.addSub(cs);
      this.chStrips.push(cs);
    }

    // The bus fader
    this.mixOutCanvas = new MixerBusStripCanvas(appdata.animator);
    this.mixOutCanvas.addSub(this.busStrip = new BusStrip(0, 0, mixOutW, 868, false, appdata, busFxEdit));
    this.mixOutCanvas.subViews.add(this.busStrip);
    // The scroll bar
    this.scrollBarView = new ScrollBar(this.mixInCanvas, 'updateMixScroll', appdata);

    this.mixIn.appendChild(this.mixInCanvas.div);
    this.mixIn.appendChild(this.scrollBarView.c.div);
    this.mixOut.appendChild(this.mixOutCanvas.div);
    this.resized = this.resized.bind(this);
    appdata.eventSys.addListener('wResized', this.resized);
    appdata.eventSys.addListener('activeBusChanged', this.updateOverlayVisibility);

    this.div.appendChild(this.mixIn);
    this.div.appendChild(this.mixOut);
    this.tabDiv = this.div;

    this.onDS = this.onDS.bind(this);
    this.opticalExpanderChanged = this.opticalExpanderChanged.bind(this);

    this.datastore = appdata.datastore;
    appdata.datastore.addListener('kBusStripExp', this.onDS);
    appdata.datastore.addListener('kOpticalExpander', this.opticalExpanderChanged);

    this.onDS('kBusStripExp', appdata.datastore.getKVArray('kBusStripExp'));
    this.opticalExpanderChanged('kOpticalExpander', appdata.datastore.getKVArray('kOpticalExpander'));
  }
  opticalExpanderChanged(k,v) {
    const e = v[0];

    if (e && this.optExpanderOverlay == undefined) {
      const bgcolor_alpha = "rgba(0.14, 0.15, 0.17, 0.8)";
      const bgcolor = "rgb(5, 6, 6)";

      this.solidColorOverlay = new SolidColorOverlay(this.div, bgcolor_alpha, kZIndexMixer + 1, "100%", "calc(100% - " + kScrollBarSize + "px)", true);
      this.optExpanderOverlay = new OpticalExpanderOverlay(250, 400 - (OptExpOvlySize.height/2.0), this.solidColorOverlay.div, bgcolor, this.mixInCanvas.animator, this.datastore);

      this.mutationObserver = new MutationObserver(this.syncOverlayPosition);
      this.mutationObserver.observe(document.body, { attributes: true, childList: true, subtree: true, attributeFilter: ["style"] });
      this.syncOverlayPosition();
    }
    else if (!e && this.optExpanderOverlay) {
      this.optExpanderOverlay.dismiss();
      this.optExpanderOverlay = undefined;
      this.solidColorOverlay.dismiss();
      this.solidColorOverlay = undefined;
      if (this.mutationObserver)
        this.mutationObserver.disconnect();
      this.mutationObserver = undefined;
    }
  }
  syncOverlayPosition(arg1, arg2) {
    if (this.optExpanderOverlay && this.solidColorOverlay) {
      this.mutationObserver.disconnect();

      const style = getComputedStyle(this.solidColorOverlay.div);
      const w = parseInt(style.width);
      const h = parseInt(style.height);
      const l = Math.max(0, (w - OptExpOvlySize.width)/2.0);
      const t = Math.max(0, (h - OptExpOvlySize.height)/2.0);

      // let scale = 1.0;
      let wRatio = (w / (l+OptExpOvlySize.width));
      let hRatio = (h / (t+OptExpOvlySize.height));

      const scale = Math.min(1.0, wRatio, hRatio);
      const invScale = 1.0/scale;
      this.optExpanderOverlay.div.style.left = l + "px";
      this.optExpanderOverlay.div.style.top = t + "px";
      this.optExpanderOverlay.div.style.transform = "scale(" + String(scale) + ")"; 
      // this.optExpanderOverlay.div.style.webkitTransform = "scale(" + String(scale) + ", " + String(scale) + ")";
      this.optExpanderOverlay.div.style.transformOrigin = "top left";

      this.mutationObserver.observe(document.body, { attributes: true, childList: true, subtree: true, attributeFilter: ["style"] });
    }
  }
  updateOverlayVisibility() {
    if (this.optExpanderOverlay && this.solidColorOverlay) {
      const activeBus = this.appdata.mixManager.getActiveBusInfo();
      const vis = (activeBus.type == "Phones" || activeBus.type == "Reverb") ? "none" : "block";
      this.optExpanderOverlay.div.style.display = vis;
      this.solidColorOverlay.div.style.display = vis;      
    }
  }
  setVis(isVis) {
    this.mixInCanvas.setVis(isVis);
    this.mixOutCanvas.setVis(isVis);
    this.scrollBarView.setVis(isVis);
  }
  onDS(k, v) {
    if(k == 'kBusStripExp') {
      var expBus = v[0];

      var mixOutW = (expBus == "on") ? kBusStripW : kBusStripWSmall;

      this.mixIn.style.width = "calc(100% - " + mixOutW + "px)";
      this.mixOut.style.width = mixOutW + "px";
      this.eventSys.dispatchASync('wResized');
    }
  }
  resized() {
    // tell the ui views the window resized
    this.mixInCanvas.resized();
    this.scrollBarView.resized();
    this.mixOutCanvas.resized();

    var h = this.mixInCanvas.r.h;
    var c = this.chStrips[0];

    var n = this.chStrips.length;
    for (var i = 0; i < n; i++) {
      this.chStrips[i].resize(h);
    }
    this.busStrip.resize(h+kScrollBarSize, c.fader.t, c.fader.h);
  }
}

const kTabItemH = 50;
class TabItem extends View {
  constructor(label, v, t, p, eventSys) {
    super(0, t, kTabSelW-2, kTabItemH);
    p.addSub(this);

    this.onEv = this.onEv.bind(this);
    eventSys.addListener(kTabEv, this.onEv);

    this.setBG(kTabBG);

    this.label = label;
    this.v = v;
    this.on = false;
  }

  onEv(k,v) {
    var isItMe = (v == this.v);
    if(isItMe != this.on) {
      this.on = isItMe;
      this.setBG(this.on ? kTabSel : kTabBG);
      this.requestRender();
    }
    else if (isItMe)
      this.requestRender();
  }
  render(c) {
    if(this.on) {
      c.fillStyle = kTabTxtSel;
    } else {
      c.fillStyle = kTabTxt;
    }
    c.textAlign = "left";
    c.font = "16px Helvetica";
    c.fillText(this.label, this.l + 50, 7 + this.cy);
  }
}

class HomeTabItem extends TabItem {
  constructor(label, v, t, p, eventSys) {
    super(label, v, t, p, eventSys);
  }
  render(c) {
    super.render(c);

    var l = this.l+ 7;
    var t = this.t + 13;
    c.save();
    c.strokeStyle = 'RGB(192,194,195)';
    c.lineWidth = 1.6;
    c.lineJoin = "miter";
    c.miterLimit="0.6";
    c.beginPath();
    c.moveTo(l+30, t+12);
    c.lineTo(l+24, t+12);
    c.lineTo(l+24, t+23);
    c.lineTo(l+6,  t+23);
    c.lineTo(l+6,  t+12);
    c.lineTo(l+0,  t+12);
    c.lineTo(l+15, t+0);
    c.lineTo(l+20, t+4);
    c.lineTo(l+20, t+1);
    c.lineTo(l+24, t+1);
    c.lineTo(l+24, t+7);
    c.closePath();
    c.stroke();
    c.restore();
  }
}
class DevTabItem extends TabItem {
  constructor(label, v, t, p, eventSys) {
    super(label, v, t, p, eventSys);
  }
  render(c) {
    super.render(c);
    kGearImage.render(c, this.l+9, this.t+13);
  }
}
class IOTabItem extends TabItem {
  constructor(l1, l2, v, t, p, eventSys) {
    super(l2, v, t, p, eventSys);
    this.l1 = l1;
  }
  render(c) {
    super.render(c);

    var x = this.l + 10;
    var y = this.t + 42;

    c.fillStyle = 'RGB(192,194,195)';
    c.textAlign = "center";
    c.font = "16px Helvetica";
    c.fillText(this.l1, this.l + 22, 7 + this.cy);
  }
}


class MixTabItem extends TabItem {
  constructor(busNum, p, appdata) {
    super("", busNum,0, p, appdata.eventSys);
    this.soloChanged = this.soloChanged.bind(this);
    this.abChanged = this.abChanged.bind(this);
    this.c = "black";
    this.mixManager = appdata.mixManager;
    this.datastore = appdata.datastore;
    this.setVis(false);
    this.rs = new RudeSoloListener(this.soloChanged, appdata);
    if (busNum === 0) {
      this.abEnabled = true;
      this.datastore.addListener('kABEnable', this.abChanged);
    }
  }
  update(t) {
    var n = this.mixManager.getNumBuses();
    this.setVis(this.v < n);
    if(this.isVis) {
      var bi = this.mixManager.getBusInfo(this.v);
      this.label = getChannelName(this.datastore, 'koChannelNames', bi).toUpperCase();
      if(bi.type != "Reverb")
         this.label += " MIX";
      this.c = bi.color;
      this.setT(t);
      t += kTabItemH;
    }

    this.rs.setBusNum(this.isVis ? this.v : -1);
    return t;
  }
  soloChanged() {
    this.requestRender();
  }
  abChanged(k,v) {
    this.abEnabled = this.datastore.getKVArrayItem('kABEnable', 0) !== 0;
    this.requestRender();
  }
  render(c) {
    super.render(c);
    c.beginPath();
    c.strokeStyle = this.c;
    c.lineWidth = this.on ? 8 : 1.5;
    c.moveTo(this.l+10,this.cy);
    c.lineTo(this.l+35,this.cy);
    c.stroke();
    if(this.rs.rs) {
      c.font = "12px Arial";
      c.fillStyle = this.c;
      c.textAlign = "center";
      c.fillText("solo", this.l + 22, this.t+18);
    }
    if (this.abEnabled === true) {
      c.font = "12px Arial";
      c.fillStyle = this.c;
      c.textAlign = "center";
      c.fillText("A/B", this.l + 22, this.b-9);
    }
  }
}


const kTabSectionDividerHeight = 7;

class TabSectionDivider extends View {
  constructor(l, t, w) {
    super(l, t, w, kTabSectionDividerHeight);
    this.render = this.render.bind(this);
  }

  render(c) {
    c.save();

    var lw = 2;
    c.lineWidth = lw;
    c.strokeStyle = "rgb(55, 59, 62)";

    c.beginPath();
    c.moveTo(this.l, this.t + (lw/2.0));
    c.lineTo(this.r, this.t + (lw/2.0));
    c.moveTo(this.l, this.b - (lw/2.0));
    c.lineTo(this.r, this.b - (lw/2.0));

    c.stroke();
    c.restore();
  }

};

const kDiscoveryImg = new OffScreenImage("discovery.svg");
const kLogoutImage = new OffScreenImage("logout.svg");

class DiscoveryTabItem extends View {
  constructor(t, appdata) {
    super(0, t, kTabSelW-2, kTabItemH);

    this.render = this.render.bind(this);
    this.handleMouseDown = this.handleMouseDown.bind(this);

    this.appdata = appdata;
    this.wantsMouse = true;
  }

  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y)) {
      // TODO: show discovery view
      ShowDiscoveryView(true, true);
    }
  }

  render(c) {
    c.save();

    kDiscoveryImg.renderInRect(c, this.l + 8, this.t + 11, 28, 28);
    c.textAlign = "left";
    c.font = "16px Helvetica";
    c.fillStyle = kTabTxt;

    c.fillText("DISCOVERY", this.l + 50, this.cy + 7);
    
    c.restore();
  }
};

class LogoutTabItem extends View {
  constructor(t, appdata) {
    super(0, t, kTabSelW-2, kTabItemH);

    this.handleMouseDown = this.handleMouseDown.bind(this);
    this.handleMouseMoved = this.handleMouseMoved.bind(this);
    this.handleMouseUp = this.handleMouseUp.bind(this);
    this.render = this.render.bind(this);

    this.appdata = appdata;
    this.wantsMouse = true;
  }

  render(c) {
    c.save();

    kLogoutImage.renderInRect(c, this.l + 6, this.t + 5, 28, 28);

    c.textAlign = "left";
    c.font = "16px Helvetica";
    c.fillStyle = kTabTxt;

    c.fillText("LOG OUT OF DEVICE", this.l + 50, this.cy + 7);

    c.restore();
  }

  handleMouseDown(xy, e) {
    this.dragTask = new DragTask(xy, ()=>{});

  }
  handleMouseMoved(xy, delta, e) {
    this.dragTask.updateDrag(xy);
  }
  handleMouseUp(xy) {
    let didDrag = this.dragTask.isDragging;
    this.dragTask = undefined;

    if (didDrag)
      return;

    if (this.contains(xy.x, xy.y)) {
      this.appdata.device.logout();
    }
  }

};

class DiscoveryGestureListener {
  constructor() {
    this.touchesStarted = this.touchesStarted.bind(this);
    this.touchesMoved = this.touchesMoved.bind(this);
    this.touchesEnded = this.touchesEnded.bind(this);
    this.doAnimation = this.doAnimation.bind(this);

    this.div = document.createElement("div");
    this.div.id = "DiscoveryGestureListener";
    this.div.style.position = "absolute";
    this.div.style.width = "100%";
    this.div.style.height = "100%";
    this.div.style.zIndex = kZIndexDiscoveryGestureListener;
    this.div.style.backgroundColor = "rgba(0,0,0,0)";
    this.div.style.transformOrigin = "top";
    this.div.style.pointerEvents = "none";

    document.body.addEventListener("touchstart", this.touchesStarted);
    document.body.addEventListener("touchmove", this.touchesMoved);
    document.body.addEventListener("touchend", this.touchesEnded);
    document.body.addEventListener("touchcancel", this.touchesEnded);

    document.body.appendChild(this.div);

    this.isGestureInProgress = false;
  }

  touchesStarted(e) {
    let x = e.changedTouches[0].clientX;
    let y = e.changedTouches[0].clientY;
    let xy = {
      x : x,
      y : y
    };

    if (xy.x < kDiscoveryGestureLeftEdgeInset) {
      this.discoveryView = ShowDiscoveryView(false, true);
      if (this.discoveryView) {
        this.dragTask = new DragTask(xy, ()=>{
          let delta = this.dragTask.getDelta();
          let initial = -document.body.clientWidth;
          let pos = Math.min(initial + delta.x, 0);
          if (this.discoveryView != undefined && this.discoveryView.mainDiv != undefined)
            this.discoveryView.mainDiv.style.left = String(initial + delta.x) + "px";
        });

        this.swipeDetector = new SwipeDetector(xy);
      }
    }
  }

  touchesMoved(e) {
    let x = e.changedTouches[0].clientX;
    let y = e.changedTouches[0].clientY;
    let xy = {
      x : x,
      y : y
    };

    if (this.dragTask != undefined) {
      this.dragTask.updateDrag(xy);
      if (this.dragTask.isDragging)
        this.isGestureInProgress = true;
    }
    if (this.swipeDetector != undefined) {
      this.swipeDetector.updateDrag(xy);
    }
  }

  touchesEnded(e) {
    let x = e.changedTouches[0].clientX;
    let y = e.changedTouches[0].clientY;
    let xy = {
      x : x,
      y : y
    };

    if (this.swipeDetector != undefined) {
      this.swipeDetector.touchEnded(xy);

      let swipe = this.swipeDetector.getSwipe();
      if (swipe == Swipe.Right) {
        this.doAnimation(true);
        this.swipeDetector = undefined;
        this.dragTask = undefined;
        this.isGestureInProgress = false;
        return;
      }
      else {
        this.swipeDetector = undefined;
      }
    }

    if (this.dragTask != undefined) {
      let delta = this.dragTask.getDelta();
      if (delta.x > document.body.clientWidth/2) {
        // The user meant to slide the discovery view in.
        this.doAnimation(true);
      }
      else {
        this.doAnimation(false);
      }
    }

    this.swipeDetector = undefined;
    this.dragTask = undefined;
    this.isGestureInProgress = false;
  }

  doAnimation(animateIn) {
    if (this.discoveryView == undefined)
      return;

    var targetLeft = 0;
    if (!animateIn) {
      targetLeft = -document.body.clientWidth;
    }

    let animation = this.discoveryView.mainDiv.animate(
      [
        {
          // from
        },
        {
          // to
          left: String(targetLeft) + "px"
        }
      ],
      {
        duration: kDiscoverySwipeDuration,
        easing: 'cubic-bezier(.25,.1,.25,1)',
      }
    );

    // Once the animation is finished, permanently place the div on the screen
    // (or else it will return to it's starting offscreen location)
    animation.onfinish = ()=>{
      if (animateIn)
        this.discoveryView.mainDiv.style.left = String(targetLeft) + "px";
      else
        document.body.removeChild(this.discoveryView.mainDiv);
    };
  }
};

class TabView extends View {
  constructor(appdata) {
    super(0, 0, 0, 0);

    this.resized = this.resized.bind(this);
    this.mixCfgChanged = this.mixCfgChanged.bind(this);
    this.getTab = this.getTab.bind(this);
    this.wantsMouse = true;
    this.mouseDown = false;

    this.appdata = appdata;
    appdata.eventSys.addListener('wResized', this.resized);
    appdata.eventSys.addListener('mixCfgChanged', this.mixCfgChanged);
    this.c = new CanvasRoot(appdata.animator);
    this.c.addSub(this);

    this.scrollBar = new VScrollBar(this.c.div);
    this.div = this.scrollBar.containerDiv;

    this.div.id = "tabViewContainer";
    this.div.style.position = "relative";
    this.div.style.height = "100vh";
    this.div.style.float = "left";
    this.div.style.width = kTabSelW + "px";
    this.div.style.boxShadow = "rgba(0,0,0,0.5) 3px 0px 6px 0px";
    this.div.style.zIndex = kZIndexMixer;

    this.c.div.id = "tabView";
    this.c.div.style.overflowX = "hidden";
    this.c.div.style.overflowY = "auto";
    this.c.div.style.position = "relative";
    this.c.canvas.style.height = "1200px";
    this.c.div.style.height = "100vh";
    this.c.div.style.backgroundColor = kTabBG;
    this.c.canvas.style.position = "static";
    this.c.div.onscroll = this.resized;

    this.setBG(kTabBG);

    this.mixManager = appdata.mixManager;
    this.eventSys = appdata.eventSys;
    this.tabs = [];
    this.homeTab = new HomeTabItem("HOME", kHome, kTabItemH, this, appdata.eventSys);
    this.devTab = new DevTabItem("DEVICE", kDev, this.homeTab.b, this, appdata.eventSys);
    this.iTab = new IOTabItem("I", "INPUT", kISetup, this.devTab.b, this, appdata.eventSys);
    this.oTab = new IOTabItem("O", "OUTPUT", kOSetup, this.iTab.b, this, appdata.eventSys);
    this.addSub(this.homeTab); this.tabs.push(this.homeTab);
    this.addSub(this.devTab); this.tabs.push(this.devTab);
    this.addSub(this.iTab); this.tabs.push(this.iTab);
    this.addSub(this.oTab); this.tabs.push(this.oTab);

    this.mixTabs = [];
    const device_props = appdata.device.properties;
    for(var i = 0; i < device_props.numBuses; ++i) {
      var tab = new MixTabItem(i, this, appdata);
      this.mixTabs.push(tab); this.tabs.push(tab);
    }
    this.downTab = undefined;

    this.datastore = appdata.datastore;
    this.datastore.addListener('koChannelNames', this.mixCfgChanged);

    this.addSub(this.divider = new TabSectionDivider(0, 0, kTabSelW-2));
    this.addSub(this.discoveryTab = new DiscoveryTabItem(0, appdata));
    this.addSub(this.logoutTab = new LogoutTabItem(0, appdata));


  }
  mixCfgChanged() {
    let t = this.oTab.b;
    this.mixTabs.forEach(function(tab){ t = tab.update(t); });
    if((!this.homeTab.on) && (!this.devTab.on) && (!this.iTab.on) && (!this.oTab.on))
      this.eventSys.dispatchASync(kTabEv, this.mixManager.getActiveBus());
  }
  getTab(xy) {
    var numTabs = this.tabs.length;
    for(var i = 0; i < numTabs; ++i) {
      var tab = this.tabs[i];
      if(tab.isVis && (tab.l <= xy.x) && ((tab.r -20) > xy.x) && (tab.t <= xy.y) && (tab.b > xy.y))
        return tab;
    }
    return undefined;
  }
  handleMouseDown(xy, e) {
    this.mouseDown = true;
    if(xy.y < kTabItemH) {
      this.datastore.setV('kTabExp', 0, this.datastore.getKVArrayItem('kTabExp', 0) == "on" ? "off" : "on");
      this.eventSys.dispatchASync("wResized");
      this.requestRender();
    }
    else {
      var tab = this.getTab(xy);
      if(tab != undefined) {
        this.lastY = e.touches ? e.touches[0].clientY : e.y;
        this.downTab = tab;
      }
    }
  }
  handleMouseMoved(xy, delta, e) {
    this.gestureInProgress = this.appdata.discoveryGestureRecognizer.isGestureInProgress;
    if (this.gestureInProgress)
      e.stopImmediatePropagation();

    var tab = this.getTab(xy);
    if((this.downTab != undefined) &&
      ((delta.y < -5) || (delta.y > 5) || (tab != this.downTab)))
    {
      this.downTab = undefined;
    }
    else if (!this.gestureInProgress) {
      var y = e.touches ? e.touches[0].clientY : e.y;
      var st = this.scrollBar.contentDiv.scrollTop - (y - this.lastY);
      this.lastY = y;
      this.scrollBar.setScrollTop(st);
    }

  }
  handleMouseUp(xy) {
    this.mouseDown = false;

    if(!this.gestureInProgress && this.downTab != undefined) {
      this.eventSys.dispatchASync(kTabEv, this.downTab.v);
    }
    this.downTab = undefined;
    this.gestureInProgress = false;
  }

  render(c)
  {
    kLogoImage.render(c, 60, 15);
    if(this.datastore.getKVArrayItem('kTabExp', 0) == "off") {
      drawRArrow(c, 1, 28, 19, "white");
      drawLArrow(c, 1, 16, 31, "white");
    }
    else {
      drawLArrow(c, 1, 28, 19, "white");
      drawRArrow(c, 1, 16, 31, "white");
    }
  }
  resized() {
    if(this.mouseDown)
      return;

    var w = kTabSelW;
    var h = this.mixTabs[this.mixManager.getNumBuses() - 1].b + 10;

    this.divider.setT(h);
    this.discoveryTab.setT(h + this.divider.h + 10);
    this.logoutTab.setT(this.discoveryTab.b);

    h = this.logoutTab.b;

    this.c.setVisibleRect(0, 0, w, h);
    this.setR(this.c.r);

    this.scrollBar.updateScroll(h);
    this.scrollBar.setEnable(this.datastore.getKVArrayItem('kTabExp', 0) == "on");
  }
}


class FxModalCanvas extends ModalCanvas {
  constructor(appdata) {
    super(appdata, true);
    this.mixManager = this.mixManager;
  }
  filterKey(keycode) {
    if(this.mv != null) {
      if(keycode == 'escape') {
        this.deactivate();
      }
      else if(keycode == 'larrow') {
        this.mixManager.incActiveCh(true);
      }
      else if(keycode == 'rarrow') {
        this.mixManager.incActiveCh(false);
      }
      else {
        return false;
      }
      return true;
    }
    return false;
  }
}

function MMAddListener(mqString, cb) {
  // safari didn't like addeventlistener so I changed to this but I'm not sure if it's correct
  var mm = matchMedia(mqString);
  if(mm.addEventListener != undefined)
    mm.addEventListener("change", cb);
  else
    mm.addListener(cb);
}


class App {
  constructor(appdata) {

    appdata.mixManager = new MixMgr(appdata);

    this.resized = this.resized.bind(this);
    this.onTab = this.onTab.bind(this);
    this.onDS = this.onDS.bind(this);
    this.deviceStatusChanged = this.deviceStatusChanged.bind(this);
    this.showPasswordEntry = this.showPasswordEntry.bind(this);
    this.setActiveDevice = this.setActiveDevice.bind(this);
    this.getActiveDevice = this.getActiveDevice.bind(this);
    this.eventSys = appdata.eventSys;
    this.mixManager = appdata.mixManager;
    this.device = appdata.device;
    this.deviceType = appdata.device.type;

    document.body.onresize = this.resized;

    this.tab = new TabView(appdata);
    appdata.eventSys.addListener(kTabEv, this.onTab);

    // create the mixer ui in a scrollable canvas
    this.home = new Home(appdata);
    this.setup = new DevSetup(appdata);
    this.isetup = new ISetup(appdata);
    this.osetup = new OSetup(appdata);
    this.mix = new Mixer(appdata);
    this.visible = false;

    document.body.appendChild(this.tab.div);
    this.tab.setVis(false);
    this.tabChild = null;

    let mqString = `(resolution: ${window.devicePixelRatio}dppx)`;
    MMAddListener(mqString, this.resized);

    appdata.networkDeviceList = new NetworkDeviceList(appdata);

    appdata.datastore.addListener('kTabExp', this.onDS);
    this.onDS('kTabExp', appdata.datastore.getKVArray('kTabExp'));

    appdata.eventSys.addListener('DeviceStatusChanged', this.deviceStatusChanged);

    var t = localStorage.getItem("tab");
    appdata.eventSys.dispatchSync(kTabEv, t ? t : kHome);

    this.appdata = appdata;

    this.passwordEntry = undefined;

    this.settingActiveDevice = false;
    this.fallbackDevice = undefined;
    this.fallbackInProgress = false;

    appdata.eventSys.addListener('DeviceAuthenticated', ()=>{ 
      console.log("Receiving notification for device authenticated");
      this.settingActiveDevice = false;
      this.fallbackDevice = undefined;

      // Don't hide the discovery view if the user canceled a password prompt
      // and we quietly reconnected to the previous device.
      if (!this.fallbackInProgress)
        HideDiscoveryView();
      else {
        this.fallbackInProgress = false;
      }

      // Kludge to workaround certain elements not being in the correct
      // position after connecting to a different device.
      this.appdata.eventSys.dispatchASync("wResized");
    });

    appdata.eventSys.addListener('DeviceAuthenticationAborted', ()=>{ 

      if (this.settingActiveDevice) {
        console.log("Receiving notification for authentication aborted");

        this.settingActiveDevice = false;

        if (this.fallbackDevice != undefined) {
          // Try to reconnect to the previously connected device
          this.fallbackInProgress = true;
          this.device.connectToDevice(this.fallbackDevice);
        }
        this.fallbackDevice = undefined;
      }
    });
  }

  setActiveDevice(netDevice) {
    // if (this.settingActiveDevice) {
    //   console.log("Error: setting active device before previous set complete");
    //   return;
    // }
    let fallbackDevice = this.device.getConnectedDevice();
    if (fallbackDevice != netDevice) {
      console.log("Setting the active device...");
      this.settingActiveDevice = true;
      this.fallbackDevice = fallbackDevice;
      this.device.connectToDevice(netDevice);
    }
  }

  getActiveDevice() { 
    return this.device.getConnectedDevice();
  }

  onDS(k, v) {
    if(k == 'kTabExp') {
      var expTab = v[0];
      var tabW = (expTab == "on") ? kTabSelW : 42;

      this.mix.div.style.width = "calc(100% - " + tabW + "px)";
      this.home.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.setup.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.isetup.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.osetup.tabDiv.style.width = "calc(100% - " + tabW + "px)";
      this.tab.div.style.width = tabW + "px";
      this.eventSys.dispatchASync('wResized');
    }
  }
  onTab(k,v) {
    var tc = this.home;
    switch(v) {
      case kHome: {
        tc = this.home;
      } break;
      case kDev: {
        tc = this.setup;
      } break;
      case kISetup: {
        tc = this.isetup;
      } break;
      case kOSetup: {
        tc = this.osetup;
      } break;
      default:{
        this.mixManager.setActiveBus(v);
        tc = this.mix;
      } break;
    }
    localStorage.setItem("tab", v);

    if((tc != null) && (tc != this.tabChild)) {
      if(this.tabChild != null) {
        document.body.removeChild(this.tabChild.tabDiv);
        this.tabChild.setVis(false);
      }
      this.tabChild = tc;
      if (this.visible) {
        this.tabChild.setVis(true);
        document.body.appendChild(tc.tabDiv);
      }
      tc.resized();
    }
  }
  resized() {
    kCanvasScale = window.devicePixelRatio || 2;
    this.eventSys.dispatchSync('wResized');
  }

  deviceStatusChanged() {
    const curType = this.device.type;
    if (this.deviceType !== curType) {
      this.deviceType = curType;
      this.mixManager.update();
      this.mixManager.setActiveBus(0);
      this.osetup.recreate();
      this.isetup.recreate();
      this.home.recreate();
      this.setup.recreate();
    }
  }

  show() {
    if (!this.visible) {
      this.visible = true;
      this.tab.setVis(true);
      if (this.tabChild !== null) {
        this.tabChild.setVis(this.visible);
        document.body.appendChild(this.tabChild.tabDiv);
      }
      this.tabChild.resized();
      this.eventSys.dispatchASync("AppVisibilityChanged");
    }
  }

  hide() {
    if (this.visible) {
      this.visible = false;
      this.tab.setVis(false);
      if (this.tabChild !== null) {
        document.body.removeChild(this.tabChild.tabDiv);
        this.tabChild.setVis(false);
      }
      this.eventSys.dispatchASync("AppVisibilityChanged");
    }
  }

  showPasswordEntry(pwEntryType, message, callbacks) {
    if (pwEntryType != PasswordEntryType.Authenticate)
    {
      let w = 400;
      let h = 345;

      let cx = document.body.clientWidth / 2.0;
      let cy = document.body.clientHeight / 2.0;

      let x = cx - (w/2.0);
      let y = cy - (h/2.0);
      this.passwordEntry = new ChangePassword(x,y, w, h, this.appdata, callbacks);
    }
    else
    {
      let w = 400;
      let h = 225;

      let cx = document.body.clientWidth / 2.0;
      let cy = document.body.clientHeight / 2.0;

      let x = cx - (w/2.0);
      let y = cy - (h/2.0);
      this.passwordEntry = new PasswordEntry(x,y, w, h, this.appdata, message, callbacks);
    }
  }

  closePasswordEntry() {
    if (this.passwordEntry != undefined) {
      this.passwordEntry.dismiss();
      this.passwordEntry = undefined;
    }
  }
}

window.onload = function()
{
  isIOS = (/iPad|iPhone|iPod/.test(navigator.platform) ||
    (navigator.platform === 'MacIntel' && navigator.maxTouchPoints > 1)) &&
    !window.MSStream;

  // this disables zooming/scaling by the broswer, it stops mobile browsers from scaling the page
  // it's not needed because I set it in the html meta. I may want this later ...
  // document.querySelector("meta[name=viewport]").setAttribute('content', 'width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0');

  isMobile = isIOS; // I consider iOS mobile
  // device detection, i consider it mobile if the user agent says so or if the screen is small
  if(/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|ipad|iris|kindle|Android|Silk|lge |maemo|midp|mmp|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows (ce|phone)|xda|xiino/i.test(navigator.userAgent)
    || /1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(navigator.userAgent.substr(0,4))) {
    isMobile = true;
  }

  // init globals here
  const afd = new AnimationFrameDispatcher;
  const appdata = {
    animator:afd,
    meterStore:new MeterStore(afd),
    eventSys:new Events,
    keyFilter:new KeyDownFilter
  };
  appdata.datastore = new DataStore(appdata);
  appdata.gen5Lib = new Gen5Lib(appdata.datastore);
  appdata.device = new Device(appdata);
  appdata.updater = new Updater(appdata);
  appdata.discoveryGestureRecognizer = new DiscoveryGestureListener();

  const app = new App(appdata);
  
  gApp = app;
  gApp.welcome = new WelcomeView(app, appdata);

  // enable the updater once we've fully connected to a device
  appdata.eventSys.addListener("DeviceAuthenticated", ()=>{
    appdata.updater.start();
  });

  if(appdata.gen5Lib.hasLib()) {
    appdata.gen5Lib.synchDSToCurrentState();
    appdata.updater.updateDriverVersion(appdata.gen5Lib.getDriverVersion());
  }

  if (appdata.device.hasHostMsgPipe()) {
    appdata.device.sendHostMsgPipeMsg("CM5Init");
  }

  ShowDiscoveryView(false, false);
}
