'use strict';
const kDecayM = 2.0; // time to decay full scale meter
const kDecayPH = 2.0; // time to decay peak
const kMeterHold = 1.0/30;
const kPeakHold = 1.0;
const kMinMeter = -72;//-255/2;

class MeterStore {
  constructor(animator) {
    this.addListener = this.addListener.bind(this);
    this.removeListener = this.removeListener.bind(this);
    this.dispatch = this.dispatch.bind(this);
    this.meterTest = this.meterTest.bind(this);
    this.set = this.set.bind(this);
    this.clear = this.clear.bind(this);

    this.meterInterval = kMaxMeterInterval;
    this.meterUpdateTS = 0;

    this.listeners = [];
    this.lastDecay = perfSeconds(); // time last decay was calculated

    this.testPeaks = [];
    this.meterTestTS = 0;

    // allocate an array big enough for all meters
    // meter is { meter, peakHold, holdTime },
    //   meter is the max of incoming peak or last meter
    //   peakHold is the current peak hold meter
    //   holdTime is the time peakHold is held before being decayed
    //   mnd = meter no decay;
    this.meters = [];
    for(var i = 0; i < 128; i++)  {
      this.testPeaks.push(0);
      this.meters.push( { m: kMinMeter, mnd:kMinMeter, mht:0, ph:kMinMeter, pht:0 } );
    }

    animator.queueMeters(this.dispatch);

    this.testInterval = 1000/30;
    this.testV = 0;
    // if(this.testInterval != 0)
    //    setTimeout(this.meterTest, this.testInterval);
  }
  meterTest(){
    var tnow = perfSeconds();
    if(this.meterTestTS < tnow) {
      this.meterTestTS = tnow + 4.0;
      var n = this.meters.length;
      for(var i = 0; i < n; ++i)
        this.testPeaks[i] = 12 * 2;
      this.set(this.testPeaks);
    }
    // var n = this.meters.length;
    // for(var i = 0; i < n; ++i)
    //   this.testPeaks[i] = (i & 1) ? this.testV : 1 - this.testV;
    // this.testV += 1.0/30;
    // if(this.testV > 1)
    //   this.testV -= 1;
    // this.set(this.testPeaks);

    setTimeout(this.meterTest, this.testInterval);
}
  addListener(l) {
    var o = this.listeners;
    if(o.indexOf(l) == -1)
      o.push(l);
  }
  removeListener(l) {
    var o = this.listeners;
    if(o != undefined) {
      var i = o.indexOf(l);
      if(i > -1)
        o.splice(i, 1);
    }
  }
  clear() {
    var n = this.meters.length;
    for(var i = 0; i < n; i++) {
      var m = this.meters[i];
      m.m = m.mnd = m.ph = kMinMeter;
      m.mht = m.pht = 0;
    }
  }
  set(peaks) {
    var now = perfSeconds();
    var pht = now + kPeakHold;
    var mht = now + kMeterHold;

    var n = Math.min(peaks.length, this.meters.length);
    for(var i = 0; i < n; i++) {
      var m = this.meters[i];
      // peaks from device are in a range 0-255 (.5 dB incrments).
      var p = m.mnd = -peaks[i]/2;
      if(p >= m.m) {
        m.m = p;
        m.mht = mht;
        if(p >= m.ph) {
          m.ph =  p;
          m.pht = pht;
        }
      }
    }
  }
  dispatch(ts) {
    // do the decay
    if(this.meterUpdateTS <= ts) {
      var dM = kMinMeter * (this.meterInterval / kDecayM);
      var dPH = kMinMeter * (this.meterInterval / kDecayPH);
      this.lastDecay = ts;

      var n = this.meters.length;
      for(var i = 0; i < n; i++) {
        var m = this.meters[i];
        // decay the meter
        if((m.m > kMinMeter) && (m.mht < ts)) {
          m.m = Math.max(kMinMeter, m.m + dM);
        }
        if((m.ph > kMinMeter) && (m.pht < ts)) {
         m.ph = Math.max(m.m, m.ph + dPH);
        }
      }

      // call back the clients
      var that = this;
      if(this.listeners != undefined)
        this.listeners.forEach(function(cb) { cb(that.meters) });

      this.meterUpdateTS = ts + this.meterInterval;
    }
  }
}

