
class InvalidNetworkDevice {
  constructor() {}
};

class NetworkDevice {
  constructor(dict) {

    this.isEqualTo = this.isEqualTo.bind(this);
    this.getFriendlyName = this.getFriendlyName.bind(this);
    this.isLocalDevice = this.isLocalDevice.bind(this);
    
    this.ip = dict["ip"];
    this.serial = dict["serial"];
    this.hostname = dict["hostname"];
    this.model = dict["model"];
    this.deviceName = dict["deviceName"];

    if (this.ip == undefined || this.serial == undefined || this.hostname == undefined || this.model == undefined)
      throw new InvalidNetworkDevice();
  }

  isEqualTo(rhs) {
    return this.ip == rhs.ip && this.serial == rhs.serial && this.hostname == rhs.hostname && this.model == rhs.model;
  }

  getFriendlyName() {
    if (this.deviceName == undefined || this.deviceName == "")
    {
      if (this.model.search("828") != -1) {
        return "828";
      }
      return "Ultralite-mk5";
    }
    return this.deviceName;
  }
  isLocalDevice() {
    return this.ip != undefined && this.ip.startsWith("127.0.0.1") || this.ip.startsWith("localhost") || this.ip == "midi";
  }
};

const Test828Dict = {
  "ip" : "test",
  "serial" : "Virtual Device",
  "hostname" : "Local",
  "model" : "828"
};
const TestUL5Dict = {
  "ip" : "test_ul",
  "serial" : "Virtual Device",
  "hostname" : "Local",
  "model" : "Ultralite-mk5"
 };

var gTest828 = new NetworkDevice(Test828Dict);
var gTestUL5 = new NetworkDevice(TestUL5Dict);

const kEnableTestDevices = true;

class NetworkDeviceList {
  constructor(appdata) {

    this.deviceList = [];
    this.appdata = appdata;
    this.eventSys = appdata.eventSys;

    this.discoveryAdd = this.discoveryAdd.bind(this);
    this.discoveryRemove = this.discoveryRemove.bind(this);
    this.deviceForIP = this.deviceForIP.bind(this);
    this.webMIDIStateChanged = this.webMIDIStateChanged.bind(this);

    if (appdata != undefined && appdata.gen5Lib != undefined && appdata.gen5Lib.hasLib()) {
      appdata.gen5Lib.lib.registerNetworkCallbacks(this.discoveryAdd, this.discoveryRemove);
    }

    this.eventSys.addListener("WebMIDIStateChanged", this.webMIDIStateChanged);
  }

  webMIDIStateChanged() {
    // Kludge to make the discovery view useful when used with WebMIDI and no app wrapper
    const idx = this.deviceList.findIndex((x)=>{
      return x.ip == "midi" && x.hostname == "WebMIDI";
    });
    if (this.appdata.device.webMIDI.motuDevices.length) {
      // There is a CM5 compatible MIDI device connected. Do we know about it?
      const wmd = this.appdata.device.webMIDI.motuDevices[0];
      const dict = {
        "ip": "midi",
        "serial": wmd.serial,
        "hostname": "WebMIDI",
        "model": (wmd.modelID.search("828") != -1) ? "828" : "Ultralite-mk5",
        "deviceName": wmd.deviceName
      };

      const idx2 = this.deviceList.findIndex((x)=>{
        return x.ip == "midi" && x.serial == dict["serial"] && x.hostname == "WebMIDI";
      });

      if (idx2 != -1) {
        // Already have it
        return;
      }

      if (idx != -1) {
        // Have a different one. Delete it.
        this.deviceList.splice(idx, 1);
      }
      const nd = new NetworkDevice(dict);
      this.deviceList.push(nd);
      this.deviceList.sort();
      this.eventSys.dispatchASync("NetworkDeviceListChanged");
    }
    else {
      if (idx != -1) {
        this.deviceList.splice(idx, 1);
        this.eventSys.dispatchASync("NetworkDeviceListChanged");
      }
    }
  }

  // Called from ObjC // Electron
  discoveryAdd(b64) {
    try {
      let dev = new TextDecoder('utf-8').decode(new Uint8Array(atob(b64).split('').map(function(char) {
        return char.charCodeAt(0);
      })));

      let newDev = new NetworkDevice(JSON.parse(dev));
      if (this.deviceList.find((x)=>{ return x.isEqualTo(newDev);}) == undefined) {
        this.deviceList.push(new NetworkDevice(JSON.parse(dev)));
        this.deviceList.sort();
        this.eventSys.dispatchASync("NetworkDeviceListChanged");
      }
    } catch (e) {
      console.log(e);
    }
  }

  // Called from ObjC // Electron
  discoveryRemove(b64) {
    try {
      let dev = new TextDecoder('utf-8').decode(new Uint8Array(atob(b64).split('').map(function(char) {
        return char.charCodeAt(0);
      })));
      var temp = new NetworkDevice(JSON.parse(dev));
      var idx = this.deviceList.findIndex((x)=> { return x.isEqualTo(temp);});
      if (idx != -1) {
        this.deviceList.splice(idx, 1);
        this.eventSys.dispatchASync("NetworkDeviceListChanged");
      }
    } catch(e) {
      console.log(e);
    }
  }

  deviceForIP(ipaddress) {
    var idx = this.deviceList.findIndex((x)=> { 
      return x.ip == ipaddress;
    });

    if (idx != -1) {
      return this.deviceList[idx];
    }
    return undefined;
  }

};