
class EscListenerCanvas extends ModalCanvas {
  constructor(appdata, dismissCallback) {
    super(appdata, true);

    this.filterKey = this.filterKey.bind(this);
    this.mouseDeactivate = this.mouseDeactivate.bind(this);
    this.callback = dismissCallback;
  }
  filterKey(keycode) {
    if(keycode == 'escape') {
      // this.invokeCallback;
      this.callback();
      return true;
    }
    return false;
  }
  mouseDeactivate(e) {
    // This is identical to superclass version, but we also call the "esc/cancel" callback
    // look for mouse clicks that are not in the content div and deactivate
    var xy = this.mouseGetXY(e);
    if(this.r.contains(xy.x, xy.y))
      return;

    this.callback();
    this.deactivate();
    e.preventDefault();
    e.stopImmediatePropagation();
  }
};

class TextEntryView extends EditableTextView {
  constructor(l, t, w, h, font, keyfilter, callbacks, div, isPassword) {
    super(l, t, w, h, "user-input-input", font, keyfilter, false);

    this.getTypeAttributeValue = this.getTypeAttributeValue.bind(this);
    this.render = this.render.bind(this);
    this.valueChanged = this.valueChanged.bind(this);
    this.resized = this.resized.bind(this);
    this.callbacks = callbacks;
    this.containerDiv = div;
    this.attributeValue = isPassword ? "password" : "text";
  }

  resized() {
    if (this.tb != undefined) {
      var s = this.tb.style;
      s.font = this.font;
      s.left = this.l + "px";
      s.top = this.t + "px";
      s.width = this.w + "px";
      s.height = this.h + "px";
      s.lineHeight = s.height;
      s.boxSizing = "border-box"; // Without this, iOS adds padding to the input element that makes it look off-center
    }
  }
  getTypeAttributeValue() {
    return this.attributeValue ;
  }

  render(c) {
    // Don't expose the password!
  }

  filterKey(keycode) {
    if((keycode == 'enter') && (this.tb != undefined)) {
      this.v = this.tb.value;
      this.tb.blur();
      this.callbacks.completion(this.v, this);
      return true;
    }
    if((keycode == 'escape') && (this.tb != undefined)) {
      this.tb.blur();
      this.callbacks.canceled(this);
      return true;
    }
    return false;
  }
};

class PasswordTextEdit extends TextEntryView {
  constructor(l, t, w, h, font, keyfilter, callbacks, div) {
    super(l, t, w, h, font, keyfilter, callbacks, div, true);
  }
};

class TextEntryDialog extends View {
  constructor(l,t,w,h, appdata, prompt, message, callbacks, isPassword, validateFunc = undefined) {
    super(l, t, w, h);

    this.prompt = prompt;
    this.message = message;
    this.callbacks = callbacks;
    this.resized = this.resized.bind(this);
    this.render = this.render.bind(this);
    this.dismiss = this.dismiss.bind(this);
    this.handleMouseDown = this.handleMouseDown.bind(this);

    appdata.eventSys.addListener('wResized', this.resized);

    this.c = new EscListenerCanvas(appdata, ()=>{
      this.c.deactivate(this);
      this.callbacks.canceled(this);
    });

    this.c.mdiv.style.zIndex = kZIndexPassword;

    var myCallbacks = {
      completion: (v)=>{callbacks.completion(v, this);},
      canceled: (v)=>{callbacks.canceled(v, this);}
    };
    this.textEntry = new TextEntryView(0,0,0,0, "Bold 30px Helvetica", appdata.keyFilter, myCallbacks, this.c.div, isPassword);

    this.c.addSub(this);
    this.addSub(this.textEntry);
    this.c.setBG("rgba(0,0,0,0)");

    this.c.resized();

    new ResizeObserver(this.resized).observe(document.body);

    this.c.activate(this);
    this.textEntry.activate();

    this.wantsMouse = true;
    this.validateFunc = validateFunc ? validateFunc : this.defaultValidate;
  }

  defaultValidate(str) {
    return {
      ok: true,
      errMsg: undefined
    };
  }

  getMouseHandler(xy) {
    if (this.cancelBox != undefined && this.cancelBox.contains(xy.x, xy.y)) {
      return this;
    }
    else if (this.okBox != undefined && this.okBox.contains(xy.x, xy.y)) {
      return this;
    }
    return undefined;
  }
  handleMouseDown(xy) {
    if (this.cancelBox != undefined && this.cancelBox.contains(xy.x, xy.y)) {
      this.callbacks.canceled(this);
    }
    else if (this.okBox != undefined && this.okBox.contains(xy.x, xy.y)) {
      const validationResult = this.validateFunc(this.textEntry.tb.value);
      if (validationResult.ok)
        this.callbacks.completion(this.textEntry.tb.value, this);
      else {
        this.errMsg = validationResult.errMsg;
        this.requestRender();
      }
    }
  }
  resized() {
    this.c.resized();
    this.setLTWH(this.c.r.l, this.c.r.t, this.c.r.w, this.c.r.h);
    const kPadding = 40;
    const kWidth = this.w - (2*kPadding);
    const kHeight = 50;

    this.textEntry.setLTWH(this.cx - (kWidth / 2.0), this.cy - (kHeight / 2.0), kWidth, kHeight);
    this.textEntry.resized();

    var cancelBoxLeft = this.textEntry.l;
    var okBoxRight = this.textEntry.r;

    const kBtnWidth = 125;
    const kBtnHeight = 40;
    this.cancelBox = new Rect(cancelBoxLeft, this.textEntry.b + 25, kBtnWidth, kBtnHeight);
    this.okBox = new Rect(okBoxRight - kBtnWidth, this.cancelBox.t, kBtnWidth, kBtnHeight);

    this.c.div.style.left = this.l;
    this.c.div.style.top = this.t;
    this.c.div.style.width = this.w;
    this.c.div.style.height = this.h;
    
    this.requestRender();
  }

  render(c) {
    c.save();
    c.fillStyle = "rgb(100, 100, 100)";
    c.fillRect(this.l, this.t, this.w, this.h);

    c.fillStyle = "rgb(255, 255, 255)";
    c.font = '20px Helvetica';
    c.textAlign = 'center';

    c.fillText(this.prompt, this.l + (this.r - this.l)/2, 40 + this.t);

    let msg = this.msg;

    c.fillStyle = kToggleDefault;
    c.font = '14px Helvetica';

    if (this.errMsg) {
      c.fillStyle = "red";
      msg = this.errMsg;
    } 
    if (msg)
      c.fillText(msg, this.l + (this.r - this.l)/2, 70 + this.t);

    c.restore();

    if (this.cancelBox != undefined) {
      drawButton(c, this.cancelBox, "Cancel", false, "rgba(0,0,0,0)", true);
    }
    if (this.okBox != undefined) {
      drawButton(c, this.okBox, "Ok", false, "rgba(0,0,0,0)", true);
    }
  }

  dismiss() {
    this.c.deactivate(this);
  }
}


class PasswordEntry extends TextEntryDialog {
  constructor(l,t,w,h, appdata, message, callbacks) {
    super(l,t,w,h, appdata, "Enter Password", message, callbacks, true);
    this.PasswordEntry = this.textEntry;
    this.c.mdiv.id = "PasswordEntry";
  }
}


class ChangePassword extends View {
  constructor(l,t,w,h, appdata, callbacks) {
    super(l, t, w, h);

    this.callbacks = callbacks;
    this.resized = this.resized.bind(this);
    this.render = this.render.bind(this);
    this.dismiss = this.dismiss.bind(this);
    this.handleMouseDown = this.handleMouseDown.bind(this);
    this.validate = this.validate.bind(this);

    appdata.eventSys.addListener('wResized', this.resized);

    this.c = new EscListenerCanvas(appdata, ()=>{
      this.c.deactivate(this);
      // this.callbacks.canceled();
      ()=> { this.callbacks.canceled(this); }
    });

    this.c.mdiv.id = "PasswordEntry";
    this.c.mdiv.style.zIndex = kZIndexPassword;

    let cb = {
      completion: this.validate,
      canceled: ()=>{ callbacks.canceled(this); }
    };

    this.newPasswordEntry1 = new PasswordTextEdit(0,0,0,0, "Bold 30px Helvetica", appdata.keyFilter, cb, this.c.div);
    this.newPasswordEntry2 = new PasswordTextEdit(0,0,0,0, "Bold 30px Helvetica", appdata.keyFilter, cb, this.c.div);

    this.c.addSub(this);

    this.addSub(this.newPasswordEntry1);
    this.addSub(this.newPasswordEntry2);

    this.c.setBG("rgba(0,0,0,0)");

    this.c.resized();

    new ResizeObserver(this.resized).observe(document.body);

    this.c.activate(this);
    this.newPasswordEntry1.activate();
    this.wantsMouse = true;
    this.errorMsg = "";
  }


  async validate() {
    let noValue = (pwEntry)=>{
      return pwEntry.tb.value == undefined || pwEntry.tb.value == "";
    };

    if (noValue(this.newPasswordEntry1)) {
      this.errorMsg = "New password cannot be empty";
      this.requestRender();
      return;
    }
    if (noValue(this.newPasswordEntry2)) {
      this.errorMsg = "New password cannot be empty";
      this.requestRender();
      return;
    }

    let new_pw_hash1 = await sha256(this.newPasswordEntry1.tb.value);
    let new_pw_hash2 = await sha256(this.newPasswordEntry2.tb.value);

    if (new_pw_hash1 != new_pw_hash2) {
      this.errorMsg = "Passwords do not match";
      this.requestRender();
      return;
    } 

    this.callbacks.completion(new_pw_hash1);
  }

  getMouseHandler(xy) {
    if (this.cancelBox != undefined && this.cancelBox.contains(xy.x, xy.y)) {
      return this;
    }
    else if (this.okBox != undefined && this.okBox.contains(xy.x, xy.y)) {
      return this;
    }
    return undefined;
  }
  handleMouseDown(xy) {
    if (this.cancelBox != undefined && this.cancelBox.contains(xy.x, xy.y)) {
      this.callbacks.canceled(this);
    }
    else if (this.okBox != undefined && this.okBox.contains(xy.x, xy.y)) {
      this.validate();
    }
  }
  resized() {
    this.c.resized();
    this.setLTWH(this.c.r.l, this.c.r.t, this.c.r.w, this.c.r.h);
    const kPadding = 40;
    const kWidth = this.w - (2*kPadding);
    const kHeight = 50;
    const kPaddingY = 30;
    const kHeightWithText = kHeight + kPaddingY;

    const numTextEntryBoxes = this.requireOldPassword ? 3.0 : 2.0;
    var top = this.cy + kPaddingY - ((kHeightWithText * numTextEntryBoxes) / 2.0);

    this.newPasswordEntry1.setLTWH(this.cx - (kWidth / 2.0), top, kWidth, kHeight);
    this.newPasswordEntry1.resized();

    top += kHeightWithText
    this.newPasswordEntry2.setLTWH(this.cx - (kWidth / 2.0), top, kWidth, kHeight);
    this.newPasswordEntry2.resized();

    top += kHeightWithText;

    var cancelBoxLeft = this.newPasswordEntry2.l;
    var okBoxRight = this.newPasswordEntry2.r;

    const kBtnWidth = 125;
    const kBtnHeight = 40;
    this.cancelBox = new Rect(cancelBoxLeft, this.newPasswordEntry2.b + 25, kBtnWidth, kBtnHeight);
    this.okBox = new Rect(okBoxRight - kBtnWidth, this.cancelBox.t, kBtnWidth, kBtnHeight);

    this.c.div.style.left = this.l;
    this.c.div.style.top = this.t;
    this.c.div.style.width = this.w;
    this.c.div.style.height = this.h;
    
    this.requestRender();
  }

  render(c) {
    c.save();
    c.fillStyle = "rgb(100, 100, 100)";
    c.fillRect(this.l, this.t, this.w, this.h);

    c.fillStyle = "rgb(255, 255, 255)";
    c.font = '20px Helvetica';
    c.textAlign = 'center';

    c.fillText("Set Password", this.l + (this.r - this.l)/2, 40 + this.t);

    c.fillStyle = kToggleDefault;
    c.font = '14px Helvetica';

    if (this.errorMsg != undefined && this.errorMsg.length> 0) {
      let oldFillStyle = c.fillStyle;
      c.fillStyle = "rgb(255, 0, 0)";
      c.fillText(this.errorMsg, this.l + (this.r - this.l)/2, 70 + this.t);
      c.fillStyle = oldFillStyle;
    }

    const kPaddingY = 30;
    var topInputBox = this.newPasswordEntry1;
    var headingBox = new Rect(this.l + (this.r - this.l)/2, topInputBox.t - (kPaddingY / 2.0) +4, topInputBox.w, kPaddingY);

    headingBox.setT(this.newPasswordEntry1.t - (kPaddingY / 2.0) +4);
    c.fillText("Enter new password", headingBox.l, headingBox.t);

    headingBox.setT(this.newPasswordEntry2.t - (kPaddingY / 2.0) +4);
    c.fillText("Re-enter new password", headingBox.l, headingBox.t);

    c.restore();

    if (this.cancelBox != undefined) {
      drawButton(c, this.cancelBox, "Cancel", false, "rgba(0,0,0,0)", true);
    }
    if (this.okBox != undefined) {
      drawButton(c, this.okBox, "Ok", false, "rgba(0,0,0,0)", true);
    }
  }

  dismiss() {
    this.c.deactivate(this);
  }
}


const PasswordEntryType = {
  Authenticate: 0,
  SetPassword: 1
};
