'use strict';

const kMaxPreDelay = 100; // 100 ms
const kReverbPreDelayColor = 'RGB(171,91,246)';
const kReverbDampingColor = 'RGB(209,131,39)';
const kReverbDecayColor = 'RGB(45,130,194)';
const kReverbWidthColor = 'RGB(230, 200, 50)';
const kThumbHitSize = 30;

const kRvbSizeColors = [
  'rgb(167, 44, 88)',
  'rgb(1, 114, 241)',
  'rgb(0, 255, 251)',
];

const kRvbSizeCurveColors = [
  ['rgba(167, 44, 88, 0.7)','rgba(167, 44, 88, 0.1)'],
  ['rgba(1, 114, 241, 0.7)','rgba(1, 114, 241, 0.1)'],
  ['rgba(0, 255, 251, 0.7)','rgba(0, 255, 251, 0.1)'],
];


// convert ds value to range 0-1
function rvbDSToV(dsv, ctrlInfo) {
  return (dsv - ctrlInfo.min) / (ctrlInfo.max - ctrlInfo.min);
}
// convert value range 0-1 to ds
function rvbVToDS(v, ctrlInfo) {
  return v * (ctrlInfo.max - ctrlInfo.min) + ctrlInfo.min;
}

class RvbPlotter {
  constructor(r, isThumb) {
    this.Plot = this.Plot.bind(this);
    this.PreDelayMs_ToXPos = this.PreDelayMs_ToXPos.bind(this);
    this.XPosTo_PreDelayMs = this.XPosTo_PreDelayMs.bind(this);
    this.XPosTo_Decay = this.XPosTo_Decay.bind(this);
    this.XPosTo_Damping = this.XPosTo_Damping.bind(this);
    this.DrawGrid = this.DrawGrid.bind(this);
    this.DrawGridLabels = this.DrawGridLabels.bind(this);
    this.Update = this.Update.bind(this);

    this.isThumb = isThumb;
    this.r = r;

    this.preDelayTicks = [0,25,50,75,100];
    this.edgeW = isThumb ? 2 : 20; // border
    this.maxDecayW = (this.r.w - 2*this.edgeW) * 0.7; // max width decay can be
    this.maxPreDelayW =((this.r.w - 2*this.edgeW) * 0.3); // width of p[redelay section

    this.xGrid = this.PreDelayMs_ToXPos(this.preDelayTicks);
    for(var x = (this.xGrid[this.xGrid.length - 1] + 102); x < this.r.r; x += 102)
      this.xGrid.push(x);

    this.yGrid = [];
    for(var y = this.r.b - 20; y > this.r.t; y -= 60)
      this.yGrid.push(y);
  }

  PreDelayMs_ToXPos(ms) {
    var l = this.edgeW + this.r.l;
    var w = this.maxPreDelayW;
    var ms2x = function(ms) {
      return l + (w * ms / kMaxPreDelay);
    };

    if (ms.constructor === Array) {
      var x = [];
      for (var i = 0; i < ms.length; i++) {
        x.push(ms2x(ms[i]));
      }
      return x;
    }

    return ms2x(ms);
  };
  XPosTo_PreDelayMs(x) {
    var ctrlX = x - (this.r.l + this.edgeW);
    return Math.min(kMaxPreDelay, Math.max(0, Math.round((ctrlX*kMaxPreDelay/this.maxPreDelayW))));
  }
  XPosTo_Decay(x) {
    return Math.min(100, Math.max(kDecayMin, 100*(x - this.predelayX)/this.maxDecayW));
  }
  XPosTo_Damping(x) {
    var ctrlX = Math.min(this.decayW-80,  Math.max(0, x - (this.predelayX + 40)));
    return 100*ctrlX/(this.decayW-80);
  }

  DrawGrid(c) {
    if(this.isThumb)
      return;

    c.lineWidth = 1;
    c.strokeStyle = kGridColor;

    c.beginPath();
    c.strokeRect(this.r.l+1, this.r.t+1, this.r.w-2, this.r.h-2);

    for(var x = 0; x < this.xGrid.length; x++) {
      c.moveTo(this.xGrid[x], this.r.t);
      c.lineTo(this.xGrid[x], this.r.b);
    }

    for(var y = 0; y < this.yGrid.length; y++) {
      c.moveTo(this.r.l, this.yGrid[y]);
      c.lineTo(this.r.r, this.yGrid[y]);
    }
    c.stroke();
  };

  DrawGridLabels(c) {
    if(this.isThumb)
      return;

    c.fillStyle = kFxEditBG;
    c.rect(this.r.l, this.r.b, this.r.w, 30);
    c.fill();

    c.font = "bold 14px Arial";
    c.textAlign = 'center';
    c.lineWidth = 1;

    c.strokeStyle = c.fillStyle = kToggleDefault;
    var lpd = this.xGrid[0];
    var rpd = this.xGrid[this.preDelayTicks.length-1];
    var ld = rpd + 20;
    var rd = this.xGrid[this.xGrid.length-1];
    var t = this.r.b + 20;
    var b = this.r.b + 34;

    c.beginPath();
    c.moveTo(ld, b);
    c.lineTo(ld+250, b);
    c.moveTo(rd, t);
    c.lineTo(rd, b);
    c.lineTo(rd-250, b);
    c.stroke();
    c.fillText("Decay", ld + (rd-ld)/2, b+3);

    c.strokeStyle = c.fillStyle = kReverbPreDelayColor;

    // draw the predelay labels
    for(var x = 0; x < this.preDelayTicks.length; x++)
      c.fillText(String(this.preDelayTicks[x]), this.xGrid[x], this.r.b + 14);

    c.beginPath();
    c.moveTo(lpd, t);
    c.lineTo(lpd, b);
    c.lineTo(lpd+60, b);
    c.moveTo(rpd, t);
    c.lineTo(rpd, b);
    c.lineTo(rpd-60, b);
    c.stroke();
    c.fillText("Pre-delay (ms)", lpd + (rpd-lpd)/2, b+3);
  }
  XYToRvbControl(x, y) {
    var b = this.yGrid[0];

    if((y > (b-kThumbHitSize)) && (y < (b+kThumbHitSize))) {
      // damping is always on top
      if ((x > (this.dampingX-kThumbHitSize)) && (x < (this.dampingX+kThumbHitSize)))
        return 'kReverbDamping';
      if ((x > (this.decayX-kThumbHitSize)) && (x < (this.decayX+kThumbHitSize)))
        return 'kReverbDecay';
      if ((x > (this.predelayX-kThumbHitSize)) && (x < (this.predelayX+kThumbHitSize)))
        return 'kReverbPreDelay';
    }
    return undefined;
  }

  Update(decay, damping, predelay, preset) {
    this.predelayX = this.PreDelayMs_ToXPos(predelay);
    this.decayW = (decay * this.maxDecayW)/100;
    this.decayX = this.predelayX + this.decayW;
    this.dampingStop = damping/100;
    this.dampingX = this.predelayX+40 + (damping * (this.decayW-80))/100;
    this.curveColor = kRvbSizeCurveColors[preset];
  }
  Plot(c) {
    // decay: 0-100%
    // daping: 0-100%
    // predelay:0-100ms
    var l = this.predelayX;
    var t = this.isThumb ? this.r.t+5 : this.yGrid[5];
    var r = this.decayX;
    var b = this.isThumb ? this.r.b-5 : this.yGrid[0];
    var curveB = this.isThumb ? this.r.b-5 : this.yGrid[0];//-4;
    var curveH = curveB - t;

    var grd = c.createLinearGradient(l, curveB, this.decayX, curveB);
    grd.addColorStop(0, this.curveColor[0]);
    grd.addColorStop(this.dampingStop, this.curveColor[0]);
    grd.addColorStop(1,  this.curveColor[1]);
    c.beginPath();
    c.moveTo(l, b);
    c.lineTo(l, t);
    c.quadraticCurveTo(this.predelayX + this.decayW/4, curveB - curveH/4, this.decayX, curveB);
    c.lineTo(r, b);
    c.closePath();
    c.fillStyle = grd;
    c.fill();

    if(!this.isThumb) {
      fillCircle(c, this.predelayX, b, 20, kEQControlGray);
      fillCircle(c, this.predelayX, b, 8, kReverbPreDelayColor);

      fillCircle(c, this.decayX, b, 20, kEQControlGray);
      fillCircle(c, this.decayX, b, 8, kReverbDecayColor);

      fillCircle(c, this.dampingX, b, 20, kEQControlGray);
      fillCircle(c, this.dampingX, b, 8, kReverbDampingColor);
    }
  }

}

const kDecayMin = 30;
const kReverbCtrlInfo = [
  {k: 'kReverbPreDelay', l:"Pre-delay", d:0, u:"ms", min:0, max:100 },
  {k: 'kReverbDamping', l:"Damping", d:0, u:"", min:0, max:100 },
  {k: 'kReverbDecay', l:"Decay", d:0, u:"", min:kDecayMin, max:100 },
  {k: 'kReverbWidth', l:"Width", d:0, u:"", min:0, max:100 },
];

class ReverbKnob extends SimpleDSKnob {
  constructor(l, t, cE, cD, datastore, ctrlInfo) {
    super(l, t, 100, 100, datastore, ctrlInfo.k, 0);

    this.cr = new Rect(this.cx - 27, t, 54, 60);
    this.cE = cE;
    this.cD = cD;
    this.setBG(kFxEditBG);
    this.ctrlInfo = ctrlInfo;
    this.enabled = true;

    super.setup(0);
  }
  offset(dx, dy) {
    this.cr.offset(dx,dy);
    super.offset(dx, dy);
  }
  dsToV(dsv) {
    return rvbDSToV(dsv, this.ctrlInfo);
  }
  vToDS(v) {
    return rvbVToDS(v, this.ctrlInfo);
  }
  render(c) {

    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, 2.25 * Math.PI);
    c.lineWidth = 7;
    c.strokeStyle = kDynCtrlBG;
    c.stroke();

    var v = (0.75 + this.v * 1.5) * Math.PI;
    c.beginPath();
    c.arc(this.cr.cx, this.cr.cy, this.cr.w/2 - 5, 0.75 * Math.PI, v);
    c.lineWidth = 7;
    c.strokeStyle = this.enabled ? this.cE : this.cD;
    c.stroke();
    //console.log(v + ' ' + this.v);

    c.fillStyle = kTabTxt;
    c.textAlign = 'center';

    c.font = "14px Arial";
    switch(this.ctrlInfo.k) {
      case 'kReverbDecay':
        c.fillText("short", this.cx-30, this.b-34);
        c.fillText("long", this.cx+30, this.b-34);
        break
      case 'kReverbDamping':
        c.fillText("more", this.cx-30, this.b-34);
        c.fillText("less", this.cx+30, this.b-34);
        break
      case 'kReverbWidth':
        c.fillText("narrow", this.cx-30, this.b-34);
        c.fillText("wide", this.cx+30, this.b-34);
        break
      case 'kReverbPreDelay':
        c.fillText((100 * this.v).toFixed() + " ms", this.cx, this.b-34);
        break;
    }

    c.font = "bold 18px Helvetica";
    c.fillText(this.ctrlInfo.l, this.cx, this.t + (this.h - 2));
  }
}


class ReverbPresetSelect extends View {
  constructor(l, t, datastore) {
    super(l,t,245,100);

    this.onDS = this.onDS.bind(this);

    this.wantsMouse = true;
    this.datastore = datastore;
    datastore.addListener('kReverbPreset', this.onDS);

    this.smallR = new Rect(l + 5, t + 5, 70, 30);
    this.mediumR = new Rect(l + 85, t + 5, 70, 30);
    this.largeR = new Rect(l + 165, t + 5, 70, 30);
    this.preset = 1;
  }
  onDS(k, v) {
    if(k != 'kReverbPreset')
      return;
    this.preset = v[0];
    this.requestRender();
  }
  handleMouseDown(xy) {
    if(this.smallR.contains(xy.x, xy.y))
      this.datastore.setV('kReverbPreset', 0, 0);
    else if(this.mediumR.contains(xy.x, xy.y))
      this.datastore.setV('kReverbPreset', 0, 1);
    else if(this.largeR.contains(xy.x, xy.y))
      this.datastore.setV('kReverbPreset', 0, 2);
  }
  render(c) {
    drawButton(c, this.smallR, "Small", this.preset == 0, kRvbSizeColors[0]);
    drawButton(c, this.mediumR, "Medium", this.preset == 1, kRvbSizeColors[1]);
    drawButton(c, this.largeR, "Large", this.preset == 2, kRvbSizeColors[2]);

    c.fillStyle = c.strokeStyle = kToggleDefault;
    c.lineWidth = 2;
    c.beginPath();
    c.moveTo(this.l + 5, this.b - 20);
    c.lineTo(this.cx - 30, this.b - 20);
    c.moveTo(this.cx + 30, this.b - 20);
    c.lineTo(this.r - 5, this.b - 20);
    c.stroke();
    c.font = "bold 18px Helvetica";
    c.textAlign = 'center';
    c.fillText("Size", this.cx, this.b - 15);
  }
}


class ReverbView extends View {
  constructor(r, datastore, isThumb) {
    super(r.l,r.t,r.w,r.h);

    this.onDS = this.onDS.bind(this);

    this.isThumb = isThumb;
    this.wantsMouse = !isThumb;
    this.dragging = undefined;
    this.setBG(kFxEditBG);

    if(isThumb)
      this.gr = new Rect(this.l, this.t, this.w, this.h);
    else
      this.gr = new Rect(this.l, this.t, this.w, this.h - 200);
    this.p = new RvbPlotter(this.gr, isThumb);

    this.decay = 86;
    this.damping = 20;
    this.preDelay = 50;
    this.preset = 1;
    this.p.Update(this.decay, this.damping, this.preDelay, this.preset);
    this.datastore = datastore;
    datastore.addListener('kReverbDecay', this.onDS); // damping
    datastore.addListener('kReverbDamping', this.onDS); // damping
    datastore.addListener('kReverbPreDelay', this.onDS); // predelay
    datastore.addListener('kReverbPreset', this.onDS); // reverb preset

    if(!this.isThumb) {
      var kt = this.t + 410;
      this.addSub(this.presetSelect = new ReverbPresetSelect(this.l + 29, kt + 10, datastore));
      this.addSub(this.preDelayKnob = new ReverbKnob(this.presetSelect.r + 29, kt, kReverbPreDelayColor, kGateCtrlDisabled, datastore, kReverbCtrlInfo[0]));
      this.addSub(this.dampKnob = new ReverbKnob(this.preDelayKnob.r + 29, kt, kReverbDampingColor, kGateCtrlDisabled, datastore, kReverbCtrlInfo[1]));
      this.addSub(this.decayKnob = new ReverbKnob(this.dampKnob.r + 29, kt, kReverbDecayColor, kGateCtrlDisabled, datastore, kReverbCtrlInfo[2]));
      this.addSub(this.widthKnob = new ReverbKnob(this.decayKnob.r + 29, kt, kReverbWidthColor, kGateCtrlDisabled, datastore, kReverbCtrlInfo[3]));
      this.resetButtonR = new Rect(this.widthKnob.r + 29, kt + 15, 70, 30);
      this.resetButtonOn = false;
    }
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.gr.offset(dx,dy); // shared by plotter
  }
  onDS(k, v) {
    switch(k) {
      case 'kReverbDecay':
        this.decay = v[0];
        break
      case 'kReverbDamping':
        this.damping = v[0];
        break
      case 'kReverbPreDelay':
        this.preDelay = v[0];
        break;
      case 'kReverbWidth':
        this.preDelay = v[0];
        break;
      case 'kReverbPreset':
        this.preset = v[0];
        break;
    }
    this.p.Update(this.decay, this.damping, this.preDelay, this.preset);
    this.requestRender();
  }
  handleMouseDown(xy) {
    if(this.gr.contains(xy.x, xy.y)) {
      this.dragging = this.p.XYToRvbControl(xy.x, xy.y);
    }
    else if (!this.isThumb && this.resetButtonR.contains(xy.x, xy.y)) {
      this.resetButtonOn = true;
      this.requestRender();
    }
  }
  handleMouseMoved(xy, delta) {
    if(this.dragging == undefined)
      return;

    switch(this.dragging) {
    case 'kReverbDecay':
      this.datastore.setV('kReverbDecay', 0, this.p.XPosTo_Decay(xy.x));
      break
    case 'kReverbDamping':
      this.datastore.setV('kReverbDamping', 0, this.p.XPosTo_Damping(xy.x));
      break
    case 'kReverbPreDelay':
      this.datastore.setV('kReverbPreDelay', 0, this.p.XPosTo_PreDelayMs(xy.x));
      break;
    }
  }
  handleMouseUp(xy) {
    this.dragging = undefined;
    if (!this.isThumb && this.resetButtonOn) {
      this.resetButtonOn = false;
      if (this.resetButtonR.contains(xy.x, xy.y)) {
        this.datastore.setV('kReverbDecay', 0, 86);
        this.datastore.setV('kReverbDamping', 0, 90);
        this.datastore.setV('kReverbPreDelay', 0, 50);
        this.datastore.setV('kReverbWidth', 0, 50);
        this.datastore.setV('kReverbPreset', 0, 1);
      }
      this.requestRender();
    }
  }
  render(c) {
    c.save();
    c.beginPath();

    c.fillStyle = kBlackBG;
    c.fillRect(this.gr.l, this.gr.t, this.gr.w, this.gr.h);

    var p = this.p;

    if(!this.isThumb) {
      p.DrawGrid(c);
      p.DrawGridLabels(c);
      drawButton(c, this.resetButtonR, "Defaults", this.resetButtonOn, "rgb(154, 109, 176)");
    }

    c.beginPath();
    c.rect(this.gr.l, this.gr.t, this.gr.w, this.gr.h);
    c.clip();
    c.beginPath();
    p.Plot(c);
    c.restore();
  }
}


class ReverbFxView  extends View {
  constructor(l,t,w,h,datastore) {
    super(l,t,w,h);
    this.setup = this.setup.bind(this);
    this.rvbV = new ReverbView(this, datastore, false);
    this.setBG(kFxEditBG);
  }
  setup(info) {
    this.removeSub(this.rvbV);
    this.setVis(false);
    if(info.type == "Reverb") {
     this.addSub(this.rvbV);
     this.setVis(true);
    }
  }
}


class ReverbThumb  extends FxThumb {
  constructor(l, t, appdata, isFxChStrip, modalDialog) {
    super(l, t, kReverbFxType, isFxChStrip, modalDialog);
    this.addSub(this.rvbV = new ReverbView(this, appdata.datastore, true));
  }
  setup(info) {
    super.setup(info);
    this.setValid(info.type == "Reverb");
    this.setVis(this.isValid);
  }
}
