'use strict';


// Div Scroller base
class DivScrollBase{
  constructor(contentDiv) {

    this.updateScroll = this.updateScroll.bind(this);
    this.setEnable = this.setEnable.bind(this);
    this.setVis = this.setVis.bind(this);
    this.mouseBegin = this.mouseBegin.bind(this);
    this.mousePerform = this.mousePerform.bind(this);
    this.mouseEnd = this.mouseEnd.bind(this);
    this.mouseEnter = this.mouseEnter.bind(this);
    this.mouseLeave = this.mouseLeave.bind(this);
    this.setAutoHide = this.setAutoHide.bind(this);
    this.setColor = this.setColor.bind(this);

    this.contentDiv = contentDiv;
    this.thumbRatio = 1;
    this.thumb = 0;
    this.isVis = false;
    this.enable = true;
    this.autoHide = true;

    this.track = document.createElement('div');
    this.track.id = "scrollTrack";
    this.track.style.position = 'absolute';
    this.track.style.right = '0px';
    this.track.style.top =  '0px';
    this.track.style.width = '0px';
    this.track.style.height = '0px';
    this.track.style.backgroundColor = 'rgba(0,0,0,0.0)';
    this.track.style.display = 'none';

    this.thumb = document.createElement('div');
    this.thumb.id = "scrollThumb";
    this.thumb.style.borderRadius = '8px';
    this.thumb.style.position = 'absolute';
    this.thumb.style.right = '0px';
    this.thumb.style.top =  '0px';
    this.thumb.style.width = '0px';
    this.thumb.style.height = '0px';
    this.thumb.style.boxSizing = "border-box";
    this.thumb.style.backgroundColor = 'rgb(89,92,94)';
    this.thumb.style.zIndex = kZIndexScrollDiv;
    this.track.appendChild(this.thumb);

    this.thumb.addEventListener('mousedown', this.mouseBegin);
    this.thumb.addEventListener('touchstart', this.mouseBegin);
  }
  setAutoHide(autoHide) {
    // I aded this so some controls could always show the scroll bars
    this.autoHide = autoHide;
    if(!this.autoHide)
      this.track.style.display = 'block';
  }

  setEnable(t) {
    this.enable = t;
    if(!this.dragging && this.autoHide)
      this.track.style.display = (this.scrollIsValid && this.isVis && this.enable && !isIOS) ? 'block' : 'none';
  }
  setVis(isVis) {
    this.isVis = isVis;
    if(!this.dragging && this.autoHide)
      this.track.style.display = (this.scrollIsValid && this.isVis && this.enable && !isIOS) ? 'block' : 'none';
  }
  updateScroll(w_or_h) {}
  mouseBegin(e) {
    e.preventDefault();
    document.addEventListener('mousemove', this.mousePerform);
    document.addEventListener('touchmove', this.mousePerform);
    document.addEventListener('mouseup', this.mouseEnd);
    document.addEventListener('touchend', this.mouseEnd);
  }
  mousePerform(e) {
    e.preventDefault();
  }
  mouseEnd() {
    document.removeEventListener('mousemove', this.mousePerform);
    document.removeEventListener('touchmove', this.mousePerform);
    document.removeEventListener('mouseup', this.mouseEnd);
    document.removeEventListener('touchend', this.mouseEnd);
    this.dragging = false;
    if(!this.isVis && this.autoHide)
      this.track.style.display = 'none';
  }
  mouseEnter(e) {
    this.setVis(true);
  }
  mouseLeave(e) {
    this.setVis(false);
  }

  setColor(col) {
    this.thumb.style.backgroundColor = col;
  }
}


// Scrolls div, owner calls updateScroll and setVis
class VScrollBarImpl extends DivScrollBase {
  constructor(contentDiv) {
    super(contentDiv);

    this.setScrollTop = this.setScrollTop.bind(this);

    this.thumbT = 0;
    this.track.style.width = '10px';
    this.track.style.height = '100%';
    this.thumb.style.width = '10px';
  }
  updateScroll(h) { // h is the size of the div being scrolled within its aperture
    if(this.dragging)
      return;

    var bcr = this.contentDiv.getBoundingClientRect();

    this.thumbRatio = bcr.height / h;
    this.thumbT = this.thumbRatio*this.contentDiv.scrollTop;
    this.thumb.style.top =   this.thumbT + 'px';
    this.thumb.style.height = (this.thumbRatio*98) +'%';
    this.scrollIsValid = bcr.height <= h;
    if(this.autoHide)
      this.track.style.display = (this.scrollIsValid && this.isVis && !isIOS) ? 'block' : 'none';
  }
  mouseBegin(e) {
    if((e.button == 0) || e.touches) {
      super.mouseBegin(e);

      var r = this.track.getBoundingClientRect();
      var y = (e.touches ? e.touches[0].clientY : e.y) - r.top;

      this.dragging = true;
      this.anchor = y-this.thumbT; // anchor is delta from y to thumb top
    }
  }
  mousePerform(e) {
    if(this.dragging) {
      super.mousePerform(e);
      var rThumb = this.thumb.getBoundingClientRect();
      var rTrack = this.track.getBoundingClientRect();
      var y = ((e.touches ? e.touches[0].clientY : e.y) - rTrack.top) - this.anchor;
      y = Math.max(y, 0);
      y = Math.min(y, rTrack.height - rThumb.height);
      this.thumbT = y;
      this.thumb.style.top = this.thumbT + 'px';
      this.contentDiv.scrollTop = this.thumbT / this.thumbRatio;
    }
  }
  setScrollTop(y) {
    this.contentDiv.scrollTop = y;
    this.thumbT = this.thumbRatio*this.contentDiv.scrollTop;
    this.thumb.style.top = this.thumbT + 'px';
    //this.thumb.style.height = (this.thumbRatio*98) +'%';
  }
}

class VScrollBar extends VScrollBarImpl {
  constructor(contentDiv) {
    super(contentDiv);

    this.containerDiv = document.createElement('div');
    this.containerDiv.appendChild(contentDiv);
    this.containerDiv.appendChild(this.track);
    this.containerDiv.addEventListener('mouseenter', this.mouseEnter);
    this.containerDiv.addEventListener('mouseleave', this.mouseLeave);
  }
}


//Scrolls div, owner calls updateScroll and setVis
class HScrollBarImpl extends DivScrollBase {
  constructor(contentDiv) {
    super(contentDiv);
    this.thumbL = 0;
    this.track.style.height = '10px';
    this.thumb.style.height = '10px';
  }
  updateScroll(w) { // w is the size of the div being scrolled within its aperture
    if(this.dragging)
      return;

    var bcr = this.contentDiv.getBoundingClientRect();

    this.thumbRatio = bcr.width / w;
    this.thumbL = this.thumbRatio*this.contentDiv.scrollLeft;
    this.track.style.top = (bcr.bottom - 10) + 'px';
    this.track.style.width = bcr.width + 'px';
    this.thumb.style.left = this.thumbL + 'px';
    this.thumb.style.width = (this.thumbRatio*98) +'%';
    this.scrollIsValid = bcr.width <= w;
    if(this.autoHide)
      this.track.style.display = (this.scrollIsValid && this.isVis && !isIOS) ? 'block' : 'none';
  }
  mouseBegin(e) {
    if((e.button == 0) || e.touches) {
      super.mouseBegin(e);

      var r = this.track.getBoundingClientRect();
      var x = (e.touches ? e.touches[0].clientX : e.x) - r.left;

      this.dragging = true;
      this.anchor = x-this.thumbL; // anchor is delta from y to thumb top
    }
  }
  mousePerform(e) {
    if(this.dragging) {
      super.mousePerform(e);
      var rThumb = this.thumb.getBoundingClientRect();
      var rTrack = this.track.getBoundingClientRect();
      var x = ((e.touches ? e.touches[0].clientX : e.x) - rTrack.left) - this.anchor;
      x = Math.max(x, 0);
      x = Math.min(x, rTrack.width - rThumb.width);
      this.thumbL = x;
      this.thumb.style.left =   this.thumbL + 'px';
      this.contentDiv.scrollLeft = this.thumbL / this.thumbRatio;
    }
  }
}

class VHScrollBar {
  constructor(contentDiv) {

    this.mouseEnter = this.mouseEnter.bind(this);
    this.mouseLeave = this.mouseLeave.bind(this);

    this.vs = new VScrollBarImpl(contentDiv);
    this.hs = new HScrollBarImpl(contentDiv);

    this.containerDiv = document.createElement('div');
    this.containerDiv.appendChild(contentDiv);
    this.containerDiv.appendChild(this.vs.track);
    this.containerDiv.appendChild(this.hs.track);

    this.containerDiv.addEventListener('mouseenter', this.mouseEnter);
    this.containerDiv.addEventListener('mouseleave', this.mouseLeave);
  }
  mouseEnter(e) {
    this.vs.mouseEnter(e);
    this.hs.mouseEnter(e);
  }
  mouseLeave(e) {
    this.vs.mouseLeave(e);
    this.hs.mouseLeave(e);
  }
}

// Scroll bar for mixer input section
class ScrollBar extends View {
  constructor(scrollCanvas, updateEvent, appdata) {
    super(0,0,0, kScrollBarSize);

    this.setV = this.setV.bind(this);
    this.resized = this.resized.bind(this);
    this.updateScroll = this.updateScroll.bind(this);

    this.scrollCanvas = scrollCanvas;

    this.c = new CanvasRoot(appdata.animator);
    this.c.addSub(this);

    this.c.div.id = "mixScroll";
    this.c.div.style.width = "100%";
    this.c.div.style.height = this.h + "px";
    this.c.div.style.backgroundColor = kEmptyBG;

    this.wantsMouse = true;
    this.sRect = new Rect(4, 5, 40, this.h-14);
    this.dragging = false;
    this.setBG("rgb(35,38,44)");
    appdata.eventSys.addListener(updateEvent, this.updateScroll);
  }
  updateScroll() {
    if(!this.dragging) {
      var cw = this.scrollCanvas.scroller.clientWidth;
      var v = this.scrollCanvas.r.l / (cw - this.scrollCanvas.r.w);
      var w = Math.max(this.root.r.w * this.root.r.w / cw, this.sRect.h + 20.5);
      var l = v * (this.root.r.w - w);
      this.sRect.setL(l);
      this.sRect.setW(w);
      this.requestRender();
    }
  }
  resized() {
    this.c.resized();
    this.setWH(this.c.r.w, this.c.r.h);
    this.updateScroll();
  }
  setV(v) {
      v = Math.max(v, this.l);
      v = Math.min(v, this.r - this.sRect.w);
      if(v != this.sRect.l) {
        this.sRect.setL(v);
        this.requestRender();
        this.scrollCanvas.div.scrollLeft = (this.scrollCanvas.scroller.clientWidth - this.scrollCanvas.r.w) * v / (this.root.r.w - this.sRect.w);
      }
  }
  handleMouseDown(xy) {
    if(this.sRect.contains(xy.x, xy.y)) {
      this.dragging = true;
      this.anchor = xy.x - this.sRect.l;
    }
    else {
      this.dragging = true;
      this.anchor = this.sRect.w/2;
      this.handleMouseMoved(xy);
    }
  }
  handleMouseMoved(xy) {
    if(this.dragging)
      this.setV(xy.x - this.anchor);
  }
  handleMouseUp() {
    this.dragging = false;
  }
  render(c) {
    if(this.sRect.w >= this.w)
      return;
    var rad = this.sRect.h/2;
    c.lineWidth = this.h - 5;
    c.strokeStyle = 'rgb(89,92,94)';
    c.beginPath();
    var y = rad + this.sRect.t + 2;
    c.moveTo(this.sRect.l + 10 + rad, y);
    c.lineTo(this.sRect.r - (10 + rad), y)
    c.lineCap = "round";
    c.stroke();
    c.lineCap = "butt";
  }
}

