'use strict';


function drawRArrow(c, lw, cx, cy, color) {
    // draw arrow pointing up and right centered on cx, cy
    const k1 = 0.8;
    const k2 = 4;
    const k3 = 8;
    const k4 = 3;

    var sx = cx - k3/2;
    var sy = cy + k3/2;
    c.lineWidth = lw;
    c.strokeStyle = c.fillStyle =  color;
    c.beginPath();
    c.moveTo(sx, sy);
    c.lineTo(sx+k1, sy+k1);
    c.lineTo(sx+k2+k1, (sy-k2)+k1);
    c.lineTo(sx+k2+k4, (sy-k2)+k4);
    c.lineTo(sx+k3, sy-k3);
    c.lineTo((sx+k2)-k4, (sy-k2)-k4);
    c.lineTo((sx+k2)-k1, (sy-k2)-k1);
    c.lineTo(sx-k1, sy-k1);
    c.closePath();
    c.lineJoin = "miter";
    c.fill();
}

function drawLArrow(c, lw, cx, cy, color) {
    // draw arrow pointing down and left centered on cx, cy
    const k1 = 0.8;
    const k2 = 4;
    const k3 = 8;
    const k4 = 3;

    var sx = cx + k3/2;
    var sy = cy - k3/2;

    c.strokeStyle = c.fillStyle =  color;
    c.lineWidth = lw;
    c.beginPath();
    c.moveTo(sx, sy);
    c.lineTo(sx-k1, sy-k1);
    c.lineTo((sx-k2)-k1, (sy+k2)-k1);
    c.lineTo((sx-k2)-k4, (sy+k2)-k4);
    c.lineTo(sx-k3, sy+k3);
    c.lineTo((sx-k2)+k4, (sy+k2)+k4);
    c.lineTo((sx-k2)+k1, (sy+k2)+k1);
    c.lineTo(sx+k1, sy+k1);
    c.closePath();
    c.lineJoin = "miter";
    c.fill();
}


function drawPwrIcon(c, lw, cx, cy, r, color) {

    c.strokeStyle =  color;
    c.beginPath();
    c.arc(cx, cy, r, 1.75 * Math.PI, 1.25 * Math.PI);
    c.lineWidth = lw;
    c.lineCap = "round";
    c.moveTo(cx, cy);
    c.lineTo(cx, cy - r);
    c.stroke();
    c.lineCap = "butt";
}

function drawInfoIcon(c, cx, cy, color) {

    c.fillStyle = c.strokeStyle =  color;
    c.beginPath();
    c.arc(cx, cy, 10, 0, 2 * Math.PI);
    c.lineWidth = 2.2;
    c.lineCap = "round";
    c.stroke();
    c.lineCap = "butt";
    c.font = "bolder 15px Arial";
    c.textAlign = 'center';
    c.fillText("?", cx, cy+5);
}

function drawDiagnosticsIcon(c, cx, cy, color) {

    c.strokeStyle =  color;
    c.lineWidth = 2.2;
    c.lineCap = "round";
    c.beginPath();
    c.moveTo(cx-8, cy);
    c.lineTo(cx-4, cy);
    c.lineTo(cx-2, cy-6);
    c.lineTo(cx+2, cy+6);
    c.lineTo(cx+4, cy);
    c.lineTo(cx+8, cy);
    c.stroke();
    c.lineCap = "butt";
}

function roundRectPath(c, x, y, w, h, rad) {
  c.beginPath();
  c.moveTo(x + rad, y);
  c.lineTo(x + w - rad, y);
  c.quadraticCurveTo(x + w, y, x + w, y + rad);
  c.lineTo(x + w, y + h - rad);
  c.quadraticCurveTo(x + w, y + h, x + w - rad, y + h);
  c.lineTo(x + rad, y + h);
  c.quadraticCurveTo(x, y + h, x, y + h - rad);
  c.lineTo(x, y + rad);
  c.quadraticCurveTo(x, y, x + rad, y);
  c.closePath();
}

function drawLabelGroup(c, label, x, y, w, h, rad, rClr, tClr) {

  const kLineWidth = 2;
  var to = 12;
  c.font = "300 16px Arial";
  c.textAlign = 'left';

  x += (kLineWidth/2.0);
  w -= (kLineWidth);
  y += (kLineWidth/2.0);
  h -= (kLineWidth/2.0);

  var fitLabel = fitText(c, label, w - (2*rad + 6));
  var tw = c.measureText(fitLabel).width;

  c.beginPath();
  c.moveTo(x + rad + tw + 5, y + to);
  c.lineTo(x + w - rad, y + to);
  c.quadraticCurveTo(x + w, y + to, x + w, y + to + rad);
  c.lineTo(x + w, y + h - rad);
  c.quadraticCurveTo(x + w, y + h, x + w - rad, y + h);
  c.lineTo(x + rad, y + h);
  c.quadraticCurveTo(x, y + h, x, y + h - rad);
  c.lineTo(x, y + to + rad);
  c.quadraticCurveTo(x, y + to, x + rad, y + to);

  c.lineWidth = kLineWidth;
  c.strokeStyle = rClr;
  c.stroke();

  c.fillStyle = tClr;
  c.fillText(fitLabel, x + 8, y + 15);
}

function drawDoubleLabelGroup(c, labels, x, y, w, h, rad, rClr, tClr) {

  var to = 10;
  c.font = "300 14px Arial";
  c.textAlign = 'left';

  var fitLabel = fitText(c, labels[0], w - (2*rad));
  var tw = c.measureText(fitLabel).width;

  c.beginPath();
  c.moveTo(x + rad + tw + 5, y + to);
  c.lineTo(x + w - rad, y + to);
  c.quadraticCurveTo(x + w, y + to, x + w, y + to + rad);
  c.lineTo(x + w, y + h - rad);
  c.quadraticCurveTo(x + w, y + h, x + w - rad, y + h);
  c.lineTo(x + rad, y + h);
  c.quadraticCurveTo(x, y + h, x, y + h - rad);
  c.lineTo(x, y + to + rad);
  c.quadraticCurveTo(x, y + to, x + rad, y + to);

  c.lineWidth = 1;
  c.strokeStyle = rClr;
  c.stroke();

  var w2 = 16;
  var h2 = 16;
  var x2 = x + w - w2;
  var y2 = y + h - h2;
  c.beginPath();
  c.moveTo(x2 + rad, y2);
  c.lineTo(x2 + w2, y2);
  c.lineTo(x2 + w2, y2 + h2 - rad);
  c.quadraticCurveTo(x2 + w2, y2 + h2, x2 + w2 - rad, y2 + h2);
  c.lineTo(x2, y2 + h2);
  c.lineTo(x2, y2 + rad);
  c.quadraticCurveTo(x2, y2, x2 + rad, y2);
  c.stroke();
  c.fillStyle = rClr;
  c.fill();

  c.fillStyle = tClr;
  c.fillText(fitLabel, x + 8, y + 15);

  c.font = "11px Arial";
  c.textAlign = 'center';
  c.fillStyle = kEmptyBG;
  c.fillText(labels[1], x2 + w2/2, y + h - 4, w2);
}


function drawButton(c, r, label, on, onColor, enabled = true) {
  roundRectPath(c, r.l+2, r.t+2, r.w-4, r.h-4, 6);
  if(!enabled) {
    c.strokeStyle = c.fillStyle = "#303030";
    c.lineWidth = 1;
    c.stroke();
  }
  else if(on) {
    c.fillStyle = onColor;
    c.fill();
    c.fillStyle = kBlackBG;
    // c.strokeStyle = onColor;
    // c.fillStyle = onColor;//kToggleDefault;
    // c.lineWidth = 2;
    //c.stroke();
  } else {
    c.strokeStyle = c.fillStyle = kToggleDefault;
    c.lineWidth = 1;
    c.stroke();
  }

  c.font = "14px Arial";
  c.textAlign = 'center';
  c.fillText(label, r.cx, 5 + r.cy);
}

function drawButtonWithTextStyle(c, r, label, on, onColor, textStyles, root, enabled = true) {
  roundRectPath(c, r.l+2, r.t+2, r.w-4, r.h-4, 6);
  if(!enabled) {
    c.strokeStyle = c.fillStyle = "#303030";
    c.lineWidth = 1;
    c.stroke();
  }
  else if(on) {
    c.fillStyle = onColor;
    c.fill();
    c.fillStyle = kBlackBG;
  } else {
    c.strokeStyle = c.fillStyle = kToggleDefault;
    c.lineWidth = 1;
    c.stroke();
  }

  c.font = "14px Arial";
  c.textAlign = 'center';

  if (textStyles != undefined) {
    for (const [k,v] of Object.entries(textStyles)) {
      c[k] = v;
    }
  }

  const strdim = stringDimensions(kStrDimRefText, c, root);
  c.fillText(label, r.cx, r.cy + (strdim.height / 2) - 1);
}

function fillCircle(c, x, y, radius, fillColor) {
  c.beginPath();
  c.arc(x,y,radius,0,2*Math.PI);
  c.fillStyle = fillColor;
  c.fill();
};

function fillTextWrapped(c, t, x, y, maxWidth, lineHeight) {
  var words = t.split(' ');
  var line = '';

  for(var n = 0; n < words.length; n++) {
    var testLine = line + words[n] + ' ';
    var metrics = c.measureText(testLine);
    var testWidth = metrics.width;
    if (testWidth > maxWidth && n > 0) {
      c.fillText(line, x, y);
      line = words[n] + ' ';
      y += lineHeight;
    }
    else {
      line = testLine;
    }
  }
  c.fillText(line, x, y);
  return y;
}

const kULm5LogoBG = {
  icon:new ScaledSVG("Utralitemk5Logo.svg"),
  x_offset: 10,
  y_offset: -3,
}
const k828m5LogoBG = {
  icon:new ScaledSVG("828mk5Logo.svg"),
  x_offset: 10,
  y_offset: -43,
};

function drawDeviceNameHeader(c, l, t, w, device) {
  const logo = device.type === DeviceTypes.UltraLite ? kULm5LogoBG : k828m5LogoBG;
  c.fillStyle = kTabTxt;
  c.fillRect(l, t + 40, (w - logo.icon.width) + logo.x_offset, 2);

  c.fillStyle = kTabTxt;
  c.fillRect(l, t + 40, (w - logo.icon.width) + logo.x_offset, 2);
  logo.icon.render(c, l + (w - logo.icon.width) + logo.x_offset, t + logo.y_offset);
}

function fitText(c, str, maxWidth) {
    var width = c.measureText(str).width;
    var ellipsis = '...';
    var ellipsisWidth = c.measureText(ellipsis).width;
    if (width<=maxWidth || width<=ellipsisWidth) {
        return str;
    } else {
        var len = str.length;
        while (width>=maxWidth-ellipsisWidth && len-->0) {
            str = str.substring(0, len);
            width = c.measureText(str).width;
        }
        return str+ellipsis;
    }
}


class EditableTextView extends View {
  constructor(l,t,w,h, cssName, font, keyFilterer, removeOnDeactivate = true) {
    super(l,t,w,h);
    this.activate = this.activate.bind(this);
    this.change = this.change.bind(this);
    this.remove = this.remove.bind(this);
    this.filterKey = this.filterKey.bind(this);
    this.keyFilterer = keyFilterer;
    this.valueChanged = this.valueChanged.bind(this);
    this.getTypeAttributeValue = this.getTypeAttributeValue.bind(this);
    this.didBlur = this.didBlur.bind(this);

    this.enabled = true;

    this.v = "";
    this.wantsMouse = true;
    this.cssName = cssName;
    this.font = font;
    this.setBG(kBlackBG);
    this.prefix = "";
    this.c = "gray";

    this.removeOnDeactivate = removeOnDeactivate;
  }

  setPrefix(p) {
    this.prefix = p;
  }
  render(c) {
    c.fillStyle = this.enabled ? this.c : kEQLabel;
    c.font = this.font;
    c.textAlign = 'center';
    c.fillText(fitText(c, this.prefix + this.v, this.w), this.l + (this.r - this.l)/2, 4 + this.cy);
  }
  handleMouseDown(){
    if(this.enabled)
      this.activate();
  }
  enable(e) {
    this.remove();
    this.enabled = e;
    this.requestRender();
  }
  setRoot(root) {
    this.remove();
    super.setRoot(root);

    if (root != undefined)
    {
      if (!this.removeOnDeactivate) {
        this.createDomElement();
      }
    }
  }
  // Override this i.e. for password fields
  getTypeAttributeValue() {
    return "text";
  }
  createDomElement() {
    if (this.tb == undefined) {
      this.tb = document.createElement('input');

      this.tb.setAttribute('type', this.getTypeAttributeValue());
      this.tb.value = this.prefix + this.v;

      this.tb.className = this.cssName;
      this.tb.tabIndex = 0;
      var s = this.tb.style;
      s.position = "absolute";
      s.font = this.font;
      s.left = this.l + "px";
      s.top = this.t + "px";
      s.width = this.w + "px";
      s.height = this.h + "px";
      s.lineHeight = s.height;

      this.tb.addEventListener('change', this.change);
      this.tb.addEventListener('blur', this.didBlur);
      this.root.getRootDiv().appendChild(this.tb);
      this.keyFilterer.add(this);
    }
  }
  activate() {
    if(this.enabled && (this.tb == undefined)) {
      this.wantsMouse = false;
      this.createDomElement();
    }
    this.tb.select();
    this.tb.focus();
  }
  valueChanged(v) { return v; } // extend this to filter and apply the change
  change() {
    if(this.tb != undefined) {
      this.v = this.valueChanged(this.tb.value);
      this.tb.blur();
    }
  }
  didBlur() {
    if (this.removeOnDeactivate) {
      this.remove();
    }
  }
  remove() {
    if(this.tb != undefined) {
      this.keyFilterer.remove();
      var tb = this.tb;
      this.tb = undefined;
      this.root.getRootDiv().removeChild(tb);
      this.wantsMouse = true;
      this.requestRender();
    }
  }
  filterKey(keycode) {
    if((keycode == 'enter') && (this.tb != undefined)) {
      this.v = this.valueChanged(this.tb.value);
      this.tb.blur();
      return true;
    }
    if((keycode == 'escape') && (this.tb != undefined)) {
      this.tb.blur();
      return true;
    }
    return false;
  }
}


class KeyEntryView extends View {
  constructor(left, top, width, height, callback) {
    super(left, top, width, height);

    this.setBG(kEmptyBG);
    this.wantsMouse = true;

    this.mouseDownListener = this.mouseDownListener.bind(this);
    this.handleKeyDown = this.handleKeyDown.bind(this);
    this.callback = callback;

    this.displayCommand = "";
    this.active = false;
  }

  activate() {
    if (!this.active) {
      this.active = true;
      document.addEventListener("mousedown", this.mouseDownListener, true);
      document.addEventListener("keydown", this.handleKeyDown, true);
      this.requestRender();
    }
  }

  deactivate() {
    if (this.active) {
      this.active = false;
      this.requestRender();
      document.removeEventListener("mousedown", this.mouseDownListener, true);
      document.removeEventListener("keydown", this.handleKeyDown, true);
    }
  }

  getMouseHandler(xy) {
    return super.getMouseHandler(xy);
  }

  handleMouseDown(){
    this.activate();
  }

  setCommand(cmd) {
    this.displayCommand = cmd.replace("/:/g", "+");
    this.requestRender();
  }

  updateCommand(keyEvent) {
    let newCommand = "";
    const addToCommand = function(mod) {
      if (newCommand !== "") {
        newCommand += ":";
      }
      newCommand += mod;
    };

    if (keyEvent.shiftKey) {
      addToCommand("Shift");
    }
    if (keyEvent.metaKey) {
      addToCommand("Meta");
    }
    if (keyEvent.ctrlKey) {
      addToCommand("Ctrl");
    }
    if (keyEvent.altKey) {
      addToCommand("Alt");
    }

    addToCommand(keyEvent.code.replace(/(Digit)|(Key)/, ""));
    this.callback(newCommand);
    this.setCommand(newCommand);
  }

  handleKeyDown(key) {
    key.stopImmediatePropagation();

    const cancelKeys = ["Cancel", "Escape"];
    const deleteKeys = ["Backspace", "Clear", "Delete", "EraseEof"];
    const modKeys = ["Alt", "AltGraph", "CapsLock", "Control", "Fn", "FnLock", "Hyper", "Meta",
      "NumLock", "ScrollLock", "Shift", "Super", "Symbol", "SymbolLock"];

    if (cancelKeys.includes(key.key)) {
      this.deactivate();
    }
    else if (deleteKeys.includes(key.key)) {
      this.displayCommand = "";
      this.callback("");
      this.deactivate();
    }
    else if (!modKeys.includes(key.key)) {
      this.updateCommand(key);
      this.deactivate();
    }
  }

  mouseDownListener() {
    // listen in to all the mouse events and deactivate ourself if a mousedown happens
    this.deactivate();
  }

  render(c) {
    c.lineWidth = 1;
    c.strokeStyle = this.active ? kKeyEntryActive : kKeyEntryInactive;
    roundRectPath(c, this.l, this.t, this.w, this.h, 2);
    c.stroke();

    const text = this.displayCommand === "" ? "None" : this.displayCommand;
    c.fillStyle = this.active ? "white" : kEQLabel;
    c.font = "Bold 11px Arial,sans-serif";
    c.textAlign = 'center';
    c.fillText(fitText(c, text, this.w), this.l + (this.r - this.l)/2, 4 + this.cy);
  }
};


class CheckBox extends View {
  constructor(d, l, t, name, id, changeCB) {
    super(l, t, 50, 13);

    this.d = d;

    const cbox = document.createElement('input');
    this.cbox = cbox;
    cbox.type = 'checkbox';
    cbox.id = id;
    cbox.v = id;
    cbox.checked = false;
    cbox.addEventListener('change', e => { changeCB(cbox.v, cbox.checked); });

    this.div = document.createElement('div');
    this.div.className = "checkbox_container";
    this.div.style.left = this.l + "px";
    this.div.style.top = this.t + "px";
    this.div.style.position = "absolute";

    const span = document.createElement('span');
    span.className = "checkmark";

    const label = document.createElement('label');
    label.setAttribute('for', id);
    label.innerHTML = name;
    label.className = "checkbox_text";

    this.div.appendChild(cbox);
    this.div.appendChild(span);
    this.div.appendChild(label);

    d.appendChild(this.div);
    this.setBG(kEmptyBG);
  }
  check(val) {
    this.cbox.checked = val;
  }
  visChanged(){
    super.visChanged();
    const visible = this.isVis && this.isPVis;
    this.div.style.visibility = visible ? "visible" : "hidden";
  }
  setEnabled(e) {
    super.setEnabled(e);
    this.div.style.opacity = this.enabled ? 1.0 : (this.parentView ? this.parentView.getCumulativeEffectiveAlpha() : 1.0);
    this.div.style.pointerEvents = this.enabled ? "auto" : "none";
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    this.div.style.left = this.l + "px";
    this.div.style.top = this.t + "px";
  }
}


class MultiSelect extends View {
  constructor(d, l, t, w, h, ctrlID, eventSys) {
    super(l, t, w, h);
    this.addOption = this.addOption.bind(this);
    this.changeCB = this.changeCB.bind(this);

    this.ctrlID = ctrlID;
    this.resized = this.resized.bind(this);
    eventSys.addListener('wResized', this.resized);

    this.div = document.createElement('div');
    this.div.style.overflowY = 'scroll';
    // this.div.style.display = "flex";
    // this.div.style.flexDirection = "column";

    this.scrollBarV = new VScrollBar(this.div);

    this.scrollBarV.containerDiv.className = 'user-multi-select';

    this.cBoxes = [];
    this.children = [];

    var s = this.scrollBarV.containerDiv.style;
    s.left = this.l + "px";
    s.top = (1 + this.t) + "px";
    this.div.style.width = s.width = this.w + "px";
    this.div.style.height = s.height = this.h + "px";

    this.div.onscroll = this.resized;

    d.appendChild(this.scrollBarV.containerDiv);
    this.setBG(kEmptyBG);

    this.scrollBarV.updateScroll(h);
  }
  resized() {
    this.scrollBarV.updateScroll(this.div.scrollHeight);
  }
  visChanged(){
    super.visChanged();
    const visible = this.isVis && this.isPVis;
    this.scrollBarV.containerDiv.style.visibility = visible ? "visible" : "hidden";
  }
  offset(dx, dy) {
    super.offset(dx, dy);
    let s = this.scrollBarV.containerDiv.style;
    s.left = this.l + "px";
    s.top = (1 + this.t) + "px";
  }
  render(c) {}
  addOption(t, id) {
    const div = document.createElement('div');
    div.className = "checkbox_container";
    div.style.marginTop = "15px";
    div.style.marginBottom = "15px";

    const cbox = document.createElement('input');
    cbox.type = 'checkbox';
    cbox.id = this.ctrlID + id;
    cbox.v = id;
    cbox.checked = false;
    const that = this;

    cbox.addEventListener('change', e => { that.changeCB(cbox.v, cbox.checked); });

    const span = document.createElement('span');
    span.className = "checkmark";

    const label = document.createElement('label');
    label.setAttribute('for', this.ctrlID + id);
    label.innerHTML = t;
    label.className = "checkbox_text";

    div.appendChild(cbox);
    div.appendChild(span);
    div.appendChild(label);

    this.children.push(div);

    this.div.appendChild(div);
    this.cBoxes.push(cbox);
    this.scrollBarV.updateScroll(this.div.scrollHeight);
  }
  clearOptions() {
    const that = this;
    this.children.forEach(function(child){
      that.div.removeChild(child);
    });
    this.children = [];
    this.cBoxes = [];
    this.scrollBarV.updateScroll(this.div.scrollHeight);
  }
  setEnabled(e) {
    super.setEnabled(e);
    this.scrollBarV.containerDiv.style.opacity = this.enabled ? 1.0 : (this.parentView ? this.parentView.getCumulativeEffectiveAlpha() : 1.0);
    this.scrollBarV.containerDiv.style.pointerEvents = this.enabled ? "auto" : "none";
  }
  changeCB(id, checked){}
}


class SelectMenu extends View {
  constructor(d, l, t, w, h, changeCB) {
    super(l, t, w, h);
    this.addOption = this.addOption.bind(this);

    this.d = d;
    this.sel = document.createElement('select');
    this.sel.className = 'user-input-select';

    this.sel.style.left = this.l + "px";
    this.sel.style.top = (1 + this.t) + "px";
    this.sel.style.width = this.w + "px";

    this.sel.addEventListener('change', changeCB);
    d.appendChild(this.sel);
    this.setBG(kEmptyBG);
  }
  visChanged(){
    super.visChanged();
    const visible = this.isVis && this.isPVis;
    this.sel.style.visibility = visible ? "visible" : "hidden";
  }
  uRBCXCY() {
    // this is called when the rect of the view changes
    // hook it so we can update the select
    super.uRBCXCY();
    this.sel.style.left = this.l + "px";
    this.sel.style.top = (1 + this.t) + "px";
    this.sel.style.width = this.w + "px";
  }
  addOption(v, t, d) {
    var option = document.createElement('option');
    option.value = v;
    option.text = t;
    option.style.backgroundColor = kEmptyBG;
    if (d != undefined)
      option.disabled = d;
    this.sel.appendChild(option);
  }
  renameOptions(t,d) {
    var opts = this.sel.children;
    for (var i = 0; i < Math.min(opts.length, t.length); ++i) {
      opts[i].text = t[i];
    }
    if (d != undefined) {
      for (var i = 0; i < Math.min(opts.length, d.length); ++i) {
        opts[i].disabled = d[i];
      }
    }
  }
  clearOptions() {
    const opts = this.sel.children;
    while(opts.length > 0) {
      this.sel.removeChild(opts[opts.length-1]);
    }
  }
  clear() {
    this.sel.innerHTML = "";
  }
  setEnabled(e) {
    super.setEnabled(e);
    this.sel.style.opacity = this.enabled ? 1.0 : (this.parentView ? this.parentView.getCumulativeEffectiveAlpha() : 1.0);
    this.sel.style.pointerEvents = this.enabled ? "auto" : "none";
  }
}


class DSSelectMenu extends View {
  constructor(d, l, t, w, h, datastore, dsKey, vals, i) {
    super(l, t, w, h);

    this.change = this.change.bind(this);
    this.onDS = this.onDS.bind(this);

    this.addSub(this.menu = new SelectMenu(d, this.l, this.t, this.w, this.h, this.change));

    this.datastore = datastore;
    this.dsI = i; // index of the datastore array to set
    this.dsKey = dsKey;

    datastore.addListener(dsKey, this.onDS);

    var n = vals.length;
    for(var i = 0; i < n; i++)
      this.menu.addOption(vals[i].v, vals[i].l);

    this.menu.sel.value = datastore.getKVArrayItem(dsKey, this.dsI);
  }
  onDS(k, v) {
    var vi = v[this.dsI];
    if(this.menu.sel.value != vi) {
      this.menu.sel.value = vi;
      this.requestRender();
    }
  }
  change() {
    this.datastore.setV(this.dsKey, this.dsI, this.menu.sel.value);
  }
  setEnabled(e) {
    super.setEnabled(e);
    this.menu.setEnabled(e);
  }
}


class DSLabelSelectMenu extends View {
  constructor(d, l, t, label, datastore, dsKey, vals, i) {
    super(l, t, 320, 28);

    this.datastore = datastore;
    this.addSub(this.menu = new DSSelectMenu(d, this.cx, t, this.w/2, this.h, datastore, dsKey, vals, i));
    this.label = label;
  }
  render(c) {

    c.fillStyle = "white";
    c.font = "300 16px Arial";
    c.textAlign = 'left';
    c.fillText(this.label, this.l + 2, 7 + this.cy);
  }

  setMenuVals(d, dsKey, vals, i) {
    this.removeSub(this.menu);
    this.addSub(this.menu = new DSSelectMenu(d, this.cx, this.t, this.w/2, this.h, this.datastore, dsKey, vals, i));
  }

  setEnabled(e) {
    super.setEnabled(e);
    this.menu.setEnabled(e);
  }
}


class FxThumb  extends View {
  constructor(l, t, fxType, isFxChStrip, modalDialog) {
    super(l, t, (isFxChStrip ? kFxChStripW : kChStripW) - 14, (isFxChStrip ? kFxThumbH : kThumbH) - 6);
    this.wantsMouse = true;
    this.fxType = fxType;
    this.isValid = false;
    this.modalDialog = modalDialog;
    this.setBG(kMixBG);
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y) && this.isValid) {
      this.modalDialog.activate(this.info.chNum, this.fxType);
    }
  }
  setup(info) {
    this.info = info;
    this.requestRender();
  }
  setValid(iv){
    this.isValid = iv;
    this.setBG(iv ? kBlackBG : kMixBG);
    this.setVis(iv);
  }
}


// base class for simple ds views
class SimpleDSView extends View {
  constructor(l,t,w,h, datastore, dsKey, i) {
    super(l,t,w,h);

    this.setup = this.setup.bind(this);
    this.onDSArray = this.onDSArray.bind(this);
    this.dsToV = this.dsToV.bind(this);
    this.vToDS = this.vToDS.bind(this);

    this.datastore = datastore;
    SetupDSArrayListenerBasic(datastore, this, dsKey, i);
    this.v = 0;//this.dsToV(DSArrayGetBasic(datastore, this));
  }
  setup(i) {
    SetDSArrayIndexBasic(this, i);
    if (i >= 0) {
      this.v = this.dsToV(DSArrayGetBasic(this.datastore, this));
    }
    else {
      this.v = 0;
    }
    this.requestRender();
  }
  onDSArray(k, v) {
    this.v = this.dsToV(DSArrayGetBasic(this.datastore, this));
    this.requestRender();
  }
  // override these to have a scaled UI representation of the value
  dsToV(dsv) { return dsv; }
  vToDS(v) { return v; }
}


class SimpleDSBtn extends SimpleDSView {
  constructor(l,t,w,h, datastore, dsKey, i)  {
    super(l,t,w,h, datastore, dsKey, i);

    this.wantsMouse = true;
    this.dragging = false;
    this.mIn = true;
    this.onV = 1;
    this.offV = 0;
    this.v = this.offV;
  }
  handleMouseDown(xy) {
    if(this.contains(xy.x, xy.y))
    {
      this.v = this.v == this.offV ? this.onV : this.offV;
      this.dragging = true;
      this.mIn = true;
      this.requestRender();
    }
  }
  handleMouseMoved(xy, delta) {
    if(this.dragging) {
      var mIn = this.contains(xy.x, xy.y);
      if(mIn != this.mIn) {
        this.mIn = mIn;
        this.v = this.v == 0 ? this.onV : this.offV;
        this.requestRender();
      }
    }
  }
  handleMouseUp() {
    this.dragging = false;
    if(this.mIn) {
      DSArraySetBasic(this.datastore, this, this.vToDS(this.v));
    }
    this.mIn = false;
  }
}


class ToggleBtn extends SimpleDSBtn {
  constructor(l, t, w, label, datastore, dsKey, i, c, bg = undefined)  {
    super(l, t, w, 30, datastore, dsKey, i);
    this.label = label;
    this.setBG(bg == undefined ? kMixBG : bg);
    this.c = c;
  }
  setEnabled(e) {
    this.enabled = e;
    this.requestRender();
  }
  getEffectiveAlpha() {
    // Base class (View) incorporates "enabled" into the effective alpha,
    // but it looks too dim when combined with the "disabled" color used in "drawButton".
    return this.alpha;
  }
  render(c) {
    drawButton(c, this, this.label, this.v == this.onV, this.c, this.enabled);
  }
}

class ToggleBtnWithTextStyle extends ToggleBtn {
  constructor(l, t, w, label, datastore, dsKey, i, c, bg, textStyles) {
    super(l, t, w, label, datastore, dsKey, i, c, bg);

    this.textStyles = textStyles;
  }

  render(c) {
    if (this.textStyles != undefined) {
      drawButtonWithTextStyle(c, this, this.label, this.v == this.onV, this.c, this.textStyles, this.root, this.enabled);
    }
  }
}


class SimpleDSKnob extends SimpleDSView {
  constructor(l,t,w,h, datastore, dsKey, i) {
    super(l,t,w,h, datastore, dsKey, i);

    this.setVal = this.setVal.bind(this);
    this.wantsMouse = true;
    this.dragging = false;
    this.enabled = true;
  }

  handleMouseDown(xy) {
    if(this.enabled && this.contains(xy.x, xy.y))
    {
      this.dragging = true;
      this.mouse = xy;
    }
  }
  setVal(v) {
    // derived classes can override this to hook when the control is changed by the user
    DSArraySetBasic(this.datastore, this, v);
  }
  handleMouseMoved(xy, delta) {
    if(this.dragging) {
      var delta = (xy.x - this.mouse.x) - (xy.y - this.mouse.y);
      this.mouse = xy;
      if(delta != 0) {
        var v  = (delta * 0.002) + this.v;
        v = v > 1 ? 1 : v;
        v = v < 0 ? 0 : v;
        this.v = v;
        //console.log("k " + v);
        this.setVal(this.vToDS(this.v));
        this.requestRender();
      }
    }
  }
  handleMouseUp() {
    this.dragging = false;
  }
  render(c) { }
}

class Button extends View {
  constructor(l,t,w,h, opts) {
    super(l,t,w,h);
    
    this.label = opts.label;
    this.onmousedown = opts.onmousedown;
    this.onmouseup = opts.onmouseup;
    this.drawActive = !!opts.drawActive;
    this.color = opts.color;
    if (opts.bgcolor) {
      this.setBG(opts.bgcolor);
    }

    this.textStyles = opts.textStyles;

    this.mouseDownInView = false;
    this.wantsMouse = true;
  }

  handleMouseDown(xy) {
    this.mouseDownInView = true;
    if (this.onmousedown) {
      this.onmousedown(xy);
    }
    // console.log("handleMouseDown - mouseDownInView = " + String(this.mouseDownInView));
  }

  handleMouseMoved(xy, delta) {
    this.mouseDownInView = this.contains(xy.x, xy.y);
    // console.log("handleMouseMoved - mouseDownInView = " + String(this.mouseDownInView));
  }

  handleMouseUp() {
    // console.log("handleMouseUp - mouseDownInView = " + String(this.mouseDownInView));

    if (this.mouseDownInView && this.onmouseup) {
      this.onmouseup();
    }
    this.mouseDownInView = false;
  }

  render(c) {
    // drawButton(c, this, this.label, this.drawActive, this.color, this.enabled);
    drawButtonWithTextStyle(c, this, this.label, this.drawActive, this.color, this.textStyles, this.root, this.enabled);
  }
}

const OptExpOvlySize = Object.freeze({
  width: 400,
  height: 100
});

class SolidColorOverlay{
  constructor(parentDiv, color, zIndex, widthStyle, heightStyle, interceptClicks) {
    this.div = document.createElement("div");
    this.div.id = "SolidColorOverlay";
    this.div.style.left = "0px";
    this.div.style.top = "0px";
    this.div.style.width = widthStyle;
    this.div.style.height = heightStyle;
    this.div.style.position = "absolute";
    this.div.style.zIndex = zIndex;
    this.div.style.backgroundColor = color;

    interceptClicks = !!interceptClicks;

    if (interceptClicks) {
      const interceptor = (e)=>{ 
        e.preventDefault();
        e.stopImmediatePropagation();
      };

      this.div.addEventListener("click", interceptor);
      this.div.addEventListener("mousedown", interceptor);
      this.div.addEventListener("touchstart", interceptor);
    }
    else {
      this.div.style.pointerEvents = "none";
    }

    this.parentDiv = parentDiv;
    parentDiv.appendChild(this.div);
  }

  dismiss() {
    if (this.div) {
      this.parentDiv.removeChild(this.div);
      this.div = undefined;
      this.parentDiv = undefined;
    }
  }
}

class OpticalExpanderOverlay extends CanvasRoot {
  constructor(cx, t, parentDiv, bgcolor, animator, datastore) {
    super(animator);
    this.canvas.style.position = "static";
    this.exitButtonClicked = this.exitButtonClicked.bind(this);
    this.dismiss = this.dismiss.bind(this);

    this.parentDiv = parentDiv;

    const w = OptExpOvlySize.width;
    const h = OptExpOvlySize.height;

    const absoluteRect = new Rect(cx - OptExpOvlySize.width / 2.0,t,w,h);
    const relativeRect = new Rect(0, 0, w, h);
    this.setVisibleRect(relativeRect.l, relativeRect.t, relativeRect.w, relativeRect.h);

    const bw = 250;
    const bh = 40;
    const bl = this.r.cx - bw/2.0;

    this.exitBtn = new Button(bl, this.r.t + 50, bw, bh, {
      label: "EXIT OPTICAL EXPANDER MODE",
      onmouseup: this.exitButtonClicked,
      drawActive: true,
      color: kToggleBlue
    });

    this.addSub(this.exitBtn);
    this.exitBtn.requestRender();
    this.exitBtn.enableClipping(true);

    this.parentDiv.appendChild(this.rootDiv);
    this.rootDiv.id = "OpticalExpanderOverlay";
    this.rootDiv.style.position = "absolute";
    this.rootDiv.style.left = absoluteRect.l + "px";
    this.rootDiv.style.top = absoluteRect.t + "px";
    this.rootDiv.style.width = absoluteRect.w + "px";
    this.rootDiv.style.height = absoluteRect.h + "px";
    this.rootDiv.style.backgroundColor = "rgba(0,0,0,0)";

    this.datastore = datastore;
    this.setBG(bgcolor);
  }

  exitButtonClicked() {
    this.datastore.setV('kOpticalExpander', 0, 0);
  }

  dismiss(){
    this.parentDiv.removeChild(this.rootDiv);
  }

  derivedRender(c) {
    c.fillStyle = c.strokeStyle = "rgb(118, 116, 134)";
    c.textAlign = "center";
    c.font = "18px Arial";
    c.fillText("The 828 is currently in optical expander mode.", this.r.cx, this.r.t + 25);
  }
}

