'use strict';


const kWelcomeBG = new OffScreenImage("welcomeBG.png");

class WelcomeView extends View {
  constructor(app, appdata) {
    super(0,0,kFxWidth,kFxHeight);

    this.app = app;
    this.appdata = appdata;
    this.device = appdata.device;
    this.datastore = appdata.datastore;
    this.modal = new ModalCanvas(appdata, false);
    this.activate = this.activate.bind(this);
    this.deviceStatusChanged = this.deviceStatusChanged.bind(this);
    this.proxyStatusChanged = this.proxyStatusChanged.bind(this);
    this.wantsMouse = true;
    this.setBG(kBlackBG);

    // mac or pc is connected, maybe show updater
    if(appdata.gen5Lib.hasLib() || appdata.device.hasHostMsgPipe()) {
        this.addSub(this.upd = new UpdateView(620, 355, appdata, 7));
        this.upd.setVis(false);
        //this.upd.setBG('rgba(0,0,0,1)')
    }

    appdata.eventSys.addListener('ProxyStatusChanged', this.proxyStatusChanged);
    appdata.eventSys.addListener('DeviceStatusChanged', this.deviceStatusChanged);
    appdata.eventSys.addListener('DeviceAuthenticated', this.deviceStatusChanged);

    this.datastore.addListener('kAPIVersion', this.deviceStatusChanged);

    this.connected = false;
    this.apiVersion = kDataStoreAPIVersionUnknown;
    this.proxyStatus = 2;
    this.isActivated = false;

    this.activate();
  }
  proxyStatusChanged(ev, v) {
    this.proxyStatus = v;
    switch (v) {
      case 0:
        this.activate();
        break;
      case 1:
        this.startConnecting();
        break;
      case 2:
        break;
    }
  }
  deviceStatusChanged() {
    this.connected = this.device.isConnected() || this.device.isTest();
    ProxyLog("this.connected = " + this.connected);
    const careAboutProxyStatus = !this.device.isMIDIConnected();
    this.apiVersion = this.datastore.getKVArrayItem('kAPIVersion', 0);
    if(this.upd) {
      if (this.connected && (this.apiVersion != kDataStoreAPIVersion)) {
        if (careAboutProxyStatus) {
          if (this.proxyStatus == 2) {
            this.upd.setVis(true);
            this.activate();
          }
        }
        else {
          this.upd.setVis(true);
          this.activate();
        }
      }
      else if (this.connected && this.apiVersion == kDataStoreAPIVersion){
        this.upd.setVis(false);
        this.deactivate();
      }
    }
    if(this.connected && (this.device.isTest() || (this.apiVersion == kDataStoreAPIVersion))) {
      this.app.show();
      let discoveryDiv = document.getElementById("DeviceListContainer");
      if (discoveryDiv && !discoveryDiv.hasAttribute("data-userInitiated")) {
        HideDiscoveryView();
      }
      if (this.isActivated)
        this.deactivate();
    }
    else {
      this.app.hide();
      // this.activate();
      ShowDiscoveryView();
    }
  }
  deactivate() {
    if (this.isActivated) {
      this.modal.deactivate();
      this.isActivated = false;
    }
  }
  activate() {
    if (! this.isActivated) {
      this.modal.activate(this);
      this.requestRender();
      this.isActivated = true;
    }
  }
  handleMouseDown(xy) {
      if (this.root == undefined)
        return; // How did we get here?

      if (this.root.context == undefined)
        return;

      // Get the "Open Discovery" string render dimensions
      var c = this.root.context;
      c.save();

      c.font = "Lighter 18px Helvetica";
      c.textAlign = 'center';
      const measurement = c.measureText("Open Discovery");

      c.restore();

      const cx = this.l + (this.w/2.0);
      const boxHeight = (measurement.actualBoundingBoxAscent + measurement.actualBoundingBoxDescent) / this.root.divScale;
      const boxWidth = measurement.width / this.root.divScale;
      var discoverBox = new Rect(cx - boxWidth/ 2.0, this.b - (boxHeight + 40), boxWidth, boxHeight + 40);

      if (discoverBox.contains(xy.x, xy.y)) {
        ShowDiscoveryView(true, true);
      }
  }
  startConnecting() {
    this.requestRender();
  }

  render(c) {
    c.drawImage(kWelcomeBG.canvas, this.l, this.t, this.w, this.h);
    c.strokeStyle = kNavBtnGrey;
    c.lineWidth = 3;
    c.strokeRect(this.l + 1, this.t + 1, this.w - 2, this.h - 2);

    c.fillStyle = "rgb(200,202,204)";
    c.textAlign = 'left';
    if((this.apiVersion != kDataStoreAPIVersion) && this.device.isConnected() && (this.proxyStatus == 2) && (this.device.authenticated)) {
      c.font = "Lighter 20px Helvetica";
      var t = "";
      if((this.apiVersion == 0) || (kDataStoreAPIVersion > this.apiVersion)) {
        t = "CueMix requires a newer firmware version. Please update the firmware on your device.";
      }
      else if(kDataStoreAPIVersion < this.apiVersion) {
        t = "The Device requires a newer version of CueMix. Please update CueMix.";
      }
      else {
        console.log("this.apiVersion = " + this.apiVersion);
      }
      fillTextWrapped(c, t, this.cx + 60, this.cy-30, 460, 24);
    }
    else if(this.proxyStatus != 1) {
      c.font = "Lighter 30px Helvetica";
      c.fillText("Waiting for device to connect.", this.cx + 60, this.cy-15);
      c.fillText("Check power and cables...", this.cx + 60, this.cy+25);
    }
    else {
      console.log("proxy status = " + this.proxyStatus);
    }

    c.font = "Lighter 18px Helvetica";
    c.textAlign = 'center';

    c.fillText("Open Discovery", this.cx, this.b - 40);

    var w = c.measureText("Open Discovery").width;
    c.fillRect(this.cx - w/2, this.b - 35, w, 1);
  }
}

